#========================================================================================
# package      DE_EPAGES::Amazon::API::AmazonTransaction
# state        public
#----------------------------------------------------------------------------------------
# description  This is the module for inserting Amazon transactions into the db.
#========================================================================================
package DE_EPAGES::Amazon::API::AmazonTransaction;

use strict;
use DE_EPAGES::Amazon::API::Tools qw ( ConvertFromAmazonDate SplitFullname );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError HandleError );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::PerlTools qw ( GenerateGUID GetMD5Digest );
use DE_EPAGES::Core::API::TestParameter qw ( TestString );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::Customer::API::Shop qw ( GetNextCustomerNo );
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use DE_EPAGES::Object::API::Template qw ( ProcessFileTemplate );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias Transaction );
use DE_EPAGES::Object::API::Language qw ( GetCodeByLanguageID );
use DE_EPAGES::Object::API::Table::Country qw ( GetPKeyCountryByCode2 );
use DE_EPAGES::Order::API::Order qw ( CreateOrderWithoutContainer );
use DE_EPAGES::Order::API::Shop qw ( GetNextOrderNo );
use DE_EPAGES::Product::API::Shop qw ( GetProduct );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Amazon::API::Constants qw ( %MAP_LOCALE );

sub ErrorDefinition {
    return {
        ReturnedEmptyValues                =>
            'Please report error! Import of orders failed because of empty return value from Amazon for mandatory field BuyerName. ' .
                'Report number:#ReportNumber, PortalConfig:#PortalConfigAlias, Shop:#ShopAlias, Store:#Store.',
    };
}

#========================================================================================
sub new {
    my $class = shift;
    my %args  = @_;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     insert
# state        public
#----------------------------------------------------------------------------------------
# syntax       @Translist = $Object->insert( $PortalConfig, $aOrders );
#----------------------------------------------------------------------------------------
# description  Inserts Amazon transactions into the database, skips existing transactions
#----------------------------------------------------------------------------------------
# input        $PortalConfig | used AmazonConfig | DE_EPAGES::Amazon::API::Object::AmazonConfig
# input        $aOrders      | Amazon order list | array.ref.DE_EPAGES::Amazon::API::DTO::OrderReport
# return       @Translist    | array of Amazon order numbers | array.string
#========================================================================================
sub insert ($$) {
    my $self = shift;
    my ( $PortalConfig, $aOrders ) = @_;
    my $Shop        = $PortalConfig->getSite;
    my $PortalSite  = $PortalConfig->get('PortalSite');
    my $PortalLocaleID = $PortalSite->get('LocaleID');

    my $TransFolder = $PortalConfig->folder('AmazonTransactions');

    my @completedTransactions = ();
    foreach my $AmazonOrder ( @{$aOrders} ) {

        # workaround for EPC-1189
        if (! defined ($AmazonOrder->{'BuyerName'})) {
            Error(
                'ReturnedEmptyValues',
                {
                    'ReportNumber'      => $AmazonOrder->{'ReportNumber'},
                    'PortalConfigAlias' => $PortalSite->alias,
                    'ShopAlias'         => $Shop->alias,
                    'Store'             => GetCurrentDBHandle()->cacheKey,
                });
        }

        my ( $EmailLocaleID, $EmailLanguageID, $EmailTLD ) = _GetLocaleInfoFromEmailAddress( $Shop, $AmazonOrder );
        my $LocaleID = $EmailLocaleID // $PortalLocaleID;
        my $LanguageID = $EmailLanguageID // $Shop->siteLocale( $PortalLocaleID )->get('Language')->{'LanguageID'};
        GetLog->debug("LanguageID=$LanguageID, LocaleID=$LocaleID, EmailTLD=$EmailTLD");

        if ( $self->_checkExistingAmazonTransaction( $AmazonOrder->AmazonOrderID ) ) {
            push @completedTransactions, $AmazonOrder->AmazonOrderID;
            next;
        }

        my ( $Order, $Container );
        Transaction(
            sub {

                my $Customer = $self->_createCustomerForTransaction( $Shop, $AmazonOrder, $EmailTLD );
                $Order = CreateOrderWithoutContainer(
                    $Customer,
                    {
                        'Alias'        => GetNextOrderNo($Shop,$Customer),
                    }
                );
                $Container = $Order->addLineItemContainer( {
                        'Class'      => 'LineItemContainerAmazon',
                        'CurrencyID' => $AmazonOrder->Currency,
                        'LocaleID'   => $LocaleID,
                    }
                );
                $self->_setBillingAndShippingAddress( $Customer, $AmazonOrder, $Order, $EmailTLD );

                my $Clone = $Shop->get('Address')->clone({ 'Parent'=>$Order, 'Alias'=> 'ShopAddress' });
                $Order->set({'ShopAddress' => $Clone});
            }
        );
        $Container->updateTaxModel();

        my ( $itemShipping, $shippingPromotionSum ) = ( 0, 0 );
        foreach my $Item ( @{ $AmazonOrder->ItemArray } ) {
            $self->_addItemToContainer(
                $Item,
                $Container,
                {
                    'Shop'                  => $Shop,
                    'TransFolder'           => $TransFolder,
                    'rShippingPromotionSum' => \$shippingPromotionSum,
                    'rItemShipping'         => \$itemShipping,
                    'PortalConfig'          => $PortalConfig,
                    'AmazonOrder'           => $AmazonOrder,
                    'LanguageID'            => $LanguageID,
                }
            );
        }

        $self->_addGiftOptions( $AmazonOrder, $Container );
        $self->_addShippingAndPayment( $AmazonOrder, $Container,
            {
                'itemShipping'         => $itemShipping,
                'shippingPromotionSum' => $shippingPromotionSum,
                'LanguageID'           => $LanguageID,
            }
        );

        # save AmazonConfig at order and set paid status
        # this can potentially trigger other events
        # Mail events will be blocked by hook listener OnMailTypeSend
        $Order->set({
                'AmazonConfig' => $PortalConfig,
                'PaidOn'       => ConvertFromAmazonDate( $AmazonOrder->OrderDate ),
                'CreationDate' => ConvertFromAmazonDate( $AmazonOrder->OrderDate ),
        });

        $self->_finishImportOrder($Order);
        push @completedTransactions, $AmazonOrder->AmazonOrderID;
    }
    return @completedTransactions;
}

#========================================================================================
# function     _checkExistingAmazonTransaction
# state        private
#----------------------------------------------------------------------------------------
# syntax       _checkExistingAmazonTransaction( $AmazonOrderID );
#----------------------------------------------------------------------------------------
# description  Checks the database if the amazon order has already been imported
#----------------------------------------------------------------------------------------
# input        $AmazonOrderID | Amazon order code | string
# return       $IsImported | true if it is already existing | boolean
#========================================================================================
sub _checkExistingAmazonTransaction {
    my $self = shift;
    my ($AmazonOrderID) = @_;

    my $Class = LoadClassByAlias('AmazonTransaction');

    # checking if transaction already exists.
    my $aTransactions = $Class->getObjects( 'ByExternalOrderID', [$AmazonOrderID] );

    if ( scalar @{$aTransactions} > 0 ) {
        GetLog->debug( "already imported: $AmazonOrderID" );
        return 1;
    }
    return 0;
}


#========================================================================================
# function     _addItemToContainer
# state        private
#----------------------------------------------------------------------------------------
# syntax       _addItemToContainer( $Item, $Container, $hOptions );
#----------------------------------------------------------------------------------------
# description  Adds a new AmazonTransaction in the transactions folder and a
#               LineItemAmazonTransaction into the order.
#----------------------------------------------------------------------------------------
# input        $Item | amazon transaction item | DE_EPAGES::Amazon::API::DTO::Item
# input        $Container | lineitemcontainer | DE_EPAGES::Order::API::Object::LineItemContainer
# input        $hOptions | options. used keys: <ul>
#                 <li>TransFolder - transaction folder - object</li>
#                 <li>Shop - shop - object</li>
#                 <li>AmazonOrder - amazon order - object</li>
#                 <li>PortalConfig - portal config - object</li>
#                 <li>rShippingPromotionSum - shipping discount sum - ref.float</li>
#                 <li>rItemShipping - shipping amount - ref.float</li>
#                 <li>LanguageID - language id - int</li></ul> | ref.hash
#========================================================================================
sub _addItemToContainer {
    my $self = shift;
    my ( $Item, $Container, $hOptions ) = @_;

    my $TransFolder           = $hOptions->{'TransFolder'};
    my $Shop                  = $hOptions->{'Shop'};
    my $rShippingPromotionSum = $hOptions->{'rShippingPromotionSum'};
    my $rItemShipping         = $hOptions->{'rItemShipping'};
    my $AmazonOrder           = $hOptions->{'AmazonOrder'};
    my $PortalConfig          = $hOptions->{'PortalConfig'};
    my $LanguageID            = $hOptions->{'LanguageID'};

    my $Product;
    eval { $Product = $self->_getProductForSKU( $Shop, $Item->SKU ); };

    # dont log unknown product.
    if ( ExistsError() ) {
        my $Error = GetError();
        Warning($Error) unless ( $Error->code eq 'NoProductForSKU' );
    }

    my ( $itemFees, $itemPriceSum, $promotionSum ) = ( 0, 0, 0 );
    my @promoStrings = ();

    foreach my $itemPrice ( @{ $Item->ItemPriceArray } ) {
        if ( $itemPrice->IsPromotion ) {
            push @promoStrings, $itemPrice->ItemPromotionCode
              if ( defined $itemPrice->ItemPromotionCode
                and not( scalar grep { $_ eq $itemPrice->ItemPromotionCode } @promoStrings ) );
            $promotionSum += $itemPrice->Amount if ( $itemPrice->Type eq 'Principal' );
            $$rShippingPromotionSum += $itemPrice->Amount
              if ( $itemPrice->Type eq 'Shipping' );
        }
        else {
            $itemPriceSum   += $itemPrice->Amount if ( $itemPrice->Type eq 'Principal' );
            $$rItemShipping += $itemPrice->Amount if ( $itemPrice->Type eq 'Shipping' );
        }
    }

    my $Transaction = $TransFolder->insertChild(
        'AmazonTransaction',
        {
            'Alias'                   => $TransFolder->nextAlias('AmazonTransaction %s'),
            'PortalConfig'            => $PortalConfig,
            'ExternalOrderID'         => $AmazonOrder->AmazonOrderID,
            'ExternalDocumentID'      => $AmazonOrder->ReportNumber,
            'EMail'                   => $AmazonOrder->BuyerEmailAddress,
            'Product'                 => $Product,
            'PromotionClaimCode'      => join( ',', @promoStrings ),
            'FulfillmentServiceLevel' => $AmazonOrder->FulfillmentServiceLevel,

            # Amazon uebermittelt promotion amount negativ, in epages
            # werden diese Werte positiv abgelegt.
            'PromotionAmount'         => -$promotionSum,
            'PromotionShippingAmount' => -$$rShippingPromotionSum,
            'ExternalOrderItemCode'   => $Item->AmazonOrderItemCode,
            'SKU'                     => $Item->SKU,
            'Title'                   => substr($Item->Title, 0, 255),
            'Quantity'                => $Item->Quantity,
            'Price'                   => $itemPriceSum,
            'Fees'                    => $itemFees,
            'Shipping'                => $$rItemShipping,
            'OrderTimestamp'          => ConvertFromAmazonDate( $AmazonOrder->OrderDate ),
        }
    );
    $self->_addImportedLineItem(
        $Transaction->addToContainer( $Container, { 'LangID' => $LanguageID } ) );
    return;
}

#========================================================================================
# function     _addShippingAndPayment
# state        private
#----------------------------------------------------------------------------------------
# syntax       _addShippingAndPayment( $AmazonOrder, $Container, $hOptions );
#----------------------------------------------------------------------------------------
# description  Adds lineitems for shipping and payment.
# input        $AmazonOrder | amazon order | DE_EPAGES::Amazon::API::DTO::OrderReport
# input        $Container | LineItemContainer | DE_EPAGES::Order::API::Object::LineItemContainer
# input        $hOptions | options. used keys: <ul>
#                 <li>LanguageID - language id - int</li>
#                 <li>itemShipping - shipping cost - float</li>
#                 <li>shippingPromotionSum - shipping discount - float</li></ul> | ref.hash
#========================================================================================
sub _addShippingAndPayment {
    my $self = shift;
    my ( $AmazonOrder, $Container, $hOptions ) = @_;
    my $LanguageID           = $hOptions->{'LanguageID'};
    my $itemShipping         = $hOptions->{'itemShipping'};
    my $shippingPromotionSum = $hOptions->{'shippingPromotionSum'};
    my $TaxClass             = $Container->get('TaxMatrix')->defaultTaxClass;
    my $TaxRate              = $Container->taxRate($TaxClass);

    my $Shipping = $Container->insertLineItem(
        LoadClassByAlias('LineItemShipping'),
        {
            'Alias'          => 'LineItemShipping_' . GetMD5Digest(\GenerateGUID()),
            'ShippingMethod' => undef,
            'TaxClass'      => $TaxClass,
            'TaxRate'       => $TaxRate,
            'LineItemPrice' => $itemShipping + $shippingPromotionSum,
            'Discount'      => $shippingPromotionSum,
        }
    );

    $Shipping->set({
        'Name' => $self->_getShippingName( $AmazonOrder->FulfillmentServiceLevel, { 'LangCode' => GetCodeByLanguageID($LanguageID) } ),
    }, $LanguageID);

    my $Payment = $Container->insertLineItem(
        LoadClassByAlias('LineItemPayment'),
        {
            'Alias'         => 'LineItemPayment_' . GetMD5Digest(\GenerateGUID()),
            'PaymentMethod' => undef,
            'TaxClass'      => $TaxClass,
            'TaxRate'       => $TaxRate,
            'LineItemPrice' => 0,
        }
    );

    $Payment->set({
        'Name' => $self->_getPaymentName( { 'LangCode' => GetCodeByLanguageID($LanguageID) } ),
    }, $LanguageID);

    return;
}

#========================================================================================
# function     _finishImportOrder
# state        private
#----------------------------------------------------------------------------------------
# syntax       _finishImportOrder( $Order );
#----------------------------------------------------------------------------------------
# description  Recalculates and calls HookAPI_OrderCreated to finalize the creation
#               of the order.
#----------------------------------------------------------------------------------------
# input        $Order | order | DE_EPAGES::Order::API::Object::CustomerOrder
#========================================================================================
sub _finishImportOrder {
    my $self = shift;
    my ($Order) = @_;
    $Order->recalculate;

    eval {
        TriggerHook(
            'API_OrderCreated',
            {
                'Order'   => $Order,
                'Options' => { 'SendConfirmationMail' => 0 }
            }
        );
    };
    if ( ExistsError() ) {
        Warning( GetError() );
    }
    return;
}

#========================================================================================
# function     getImportedLineItems
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $Object->getImportedLineItems( $Input );
#----------------------------------------------------------------------------------------
# description  Returns all generated line items
#----------------------------------------------------------------------------------------
# return       $aLineItems | LineItemAmazonTransaction array | array.ref.DE_EPAGES::Amazon::API::Object::LineItemAmazonTransaction
#========================================================================================
sub getImportedLineItems {
    my $self = shift;
    return $self->{GeneratedLineItems};
}

#========================================================================================
# function     _addGiftOptions
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object->_addGiftOptions( $AmazonOrder, $LineItemContainer, $hOptions );
#----------------------------------------------------------------------------------------
# description  Adding gift options (wrap, greeting text) to the order container.
#----------------------------------------------------------------------------------------
# input        $AmazonOrder | Data from Amazon          | DE_EPAGES::Amazon::API::DTO::OrderReport
# input        $Container   | current LineItemContainer | DE_EPAGES::Order::API::Object::LineItemContainer
#========================================================================================
sub _addGiftOptions ($$) {
    my $self = shift;
    my ( $AmazonOrder, $Container ) = @_;

    my $giftWrapSum  = 0;
    my @GiftMessages = ();
    my $GiftWrap     = undef;
    foreach my $Item ( @{ $AmazonOrder->ItemArray } ) {
        push( @GiftMessages, $Item->GiftMessageText )
          if ( defined $Item->GiftMessageText
            and not scalar grep { $_ eq $Item->GiftMessageText } @GiftMessages );
        $GiftWrap = $Item->GiftWrapLevel if ( defined $Item->GiftWrapLevel );

        foreach my $itemPrice ( @{ $Item->ItemPriceArray } ) {
            $giftWrapSum += $itemPrice->Amount if ( $itemPrice->Type eq 'GiftWrap' );
        }
    }

    if ( scalar @GiftMessages ) {
        $Container->insertAmazonGiftMessage( { 'Comment' => join( "\n", @GiftMessages ) } );
    }
    if ( $GiftWrap or $giftWrapSum ) {
        $Container->insertAmazonGiftWrap( {
                'ChoiceName'    => $GiftWrap,
                'Comment'       => $GiftWrap,
                'BasePrice'     => $giftWrapSum,
                'LineItemPrice' => $giftWrapSum,
            }
        );
    }
    return;
}

#========================================================================================
# function     _getShippingName
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Name = $Object->_getShippingName( $ServiceLevel, $hOptions );
#----------------------------------------------------------------------------------------
# description  Returns the localized name for expedited or regular shipping.
#----------------------------------------------------------------------------------------
# input        $ServiceLevel | 'Standard' or 'Expedited' | string
# input        $hOptions     | Options | hash.ref
#               used keys are:
#               <ul>
#                 <li>LangCode | 2-digit language code | string
#               </ul>
# return       $Name | localized shipping level string | string
#========================================================================================
sub _getShippingName ($$) {
    my $self = shift;
    my ( $ServiceLevel, $hOptions ) = @_;

    TestString( 'LangCode', $hOptions->{'LangCode'}, 2, 2 );

    ## list only known service levels.
    $ServiceLevel = 'Standard' if ( $ServiceLevel ne 'Standard' and $ServiceLevel ne 'Expedited' );

    return ProcessFileTemplate(
        $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Amazon/Templates/MBO/Order/LineItemShipping-' . $ServiceLevel . '.txt',
        $hOptions->{'LangCode'}, {}
    );
}

#========================================================================================
# function     _getProductForSKU
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Product = $Object->_getProductForSKU( $Shop, $SKU );
#----------------------------------------------------------------------------------------
# description  Fetches Product (if existing) for a given SKU/Alias
#               Throws "NoProductForSKU" if product does not exist.
#----------------------------------------------------------------------------------------
# input        $Shop | Shop       | DE_EPAGES::Shop::API::Object::Shop
# input        $SKU  | SKU        | string
# return       $Product | Product | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub _getProductForSKU ($$) {
    my $self = shift;
    my ( $Shop, $SKU ) = @_;
    my $Product = GetProduct( $Shop, $SKU );

    Error( 'NoProductForSKU', { 'SKU' => $SKU, } ) unless ( defined $Product );

    return $Product;
}

#========================================================================================
# function     _createCustomerForTransaction
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Customer = $Object->_createCustomerForTransaction( $Shop, $Order, $AmazonLanguageID );
#----------------------------------------------------------------------------------------
# description  Creates a new customer
#----------------------------------------------------------------------------------------
# input        $Shop         | Shop              | DE_EPAGES::Shop::API::Object::Shop
# input        $AmazonOrder  | Amazon order data | DE_EPAGES::Amazon::API::DTO::OrderReport
# input        $EmailTLD     | Top level domain of customer's email address (de, uk, ...) | string
# return       $Customer     | customer object   | DE_EPAGES::Customer::API::Object::Customer
#========================================================================================
sub _createCustomerForTransaction {
    my $self = shift;
    my ( $Shop, $AmazonOrder, $EmailTLD ) = @_;

    # create customer
    my $folder = $Shop->folder('Customers');

    # get the customer class of the shop (with custom attributes)
    my $CustomerClass = $Shop->get('CustomerClass');

    my $Customer = $folder->insertChild(
        $CustomerClass,
        {
            'Alias'         => GetNextCustomerNo($Shop),
            'CustomerGroup' => $Shop->get('NonMemberCustomerGroup'),
            'Shop'          => $Shop,
        }
    );

    my ($Firstname, $Lastname) = SplitFullname($AmazonOrder->BuyerAddress->Name);

    # get invoice country.
    my $Country = $AmazonOrder->BuyerAddress->CountryCode;
    my $unknownCountryCodeNotice = '';
    my $CountryID = defined $Country ? eval{GetPKeyCountryByCode2($Country)} : undef;

    HandleError(
        OnError => sub {
            my ($Error) = @_;
            if ($Error->code eq 'NoResults_CountryByCode2') {
                $unknownCountryCodeNotice = " / COUNTRYCODE: $Country";
                Warning($Error);
                return 'FIXED';
            }
        }
    );

    my ($Company, $Street1, $Street2) = _GetStreetsAndCompany($AmazonOrder->BuyerAddress, $EmailTLD, $CountryID);

    my $CustomerAddress = $Customer->insertChild(
        'Address',
        {
            'Alias'       => $Customer->folder('Addresses')->nextAlias('BillingAddress %s'),
            'DisplayName' => $AmazonOrder->BuyerAddress->Name,
            'FirstName'   => $Firstname,
            'LastName'    => $Lastname,
            'City'        => $AmazonOrder->BuyerAddress->City.$unknownCountryCodeNotice,
            'CountryID'   => $CountryID,
            'Phone'       => $AmazonOrder->BuyerAddress->PhoneNumber,
            'Zipcode'     => $AmazonOrder->BuyerAddress->PostalCode,
            'State'       => $AmazonOrder->BuyerAddress->StateOrRegion,
            'Street'      => $Street1,
            'Street2'     => $Street2,
            'Company'     => $Company,
            'EMail'       => $AmazonOrder->BuyerEmailAddress,
        }
    );

    $Customer->set( { 'BillingAddress' => $CustomerAddress } );
    return $Customer;
}

sub _GetStreetsAndCompany {
    my ($Address, $EmailTLD, $CountryID) = @_;

    # Check if it is an order via amazon.de and also a german address (EPG-27659)
    if (defined $EmailTLD
        and lc($EmailTLD) eq 'de'
        and $CountryID == GetPKeyCountryByCode2('DE')
        and $Address->AddressFieldTwo
        and not $Address->AddressFieldTwo =~ m/Packstation/i
    ) {
        return ($Address->AddressFieldOne, $Address->AddressFieldTwo, '');
    }
    else {
        return ('', $Address->AddressFieldOne, $Address->AddressFieldTwo);
    }
}


#========================================================================================
# function     _setBillingAndShippingAddress
# state        private
#----------------------------------------------------------------------------------------
# syntax       _setBillingAndShippingAddress( $Customer ,$AmazonOrder, $Order, $AmazonLanguageID );
#----------------------------------------------------------------------------------------
# description  Creates the billing and shipping address of the order.
#----------------------------------------------------------------------------------------
# input        $Customer | customer | DE_EPAGES::Customer::API::Object::Customer
# input        $AmazonOrder | Amazon data | DE_EPAGES::Amazon::API::DTO::OrderReport
# input        $Order | local CustomerOrder | DE_EPAGES::Order::API::Object::CustomerOrder
# input        $AmazonLanguageID | languageid of the used amazon portal (amazon.de, amazon.co.uk, ...) | int
# input        $EmailTLD     | Top level domain of customer's email address (de, uk, ...) | string
#========================================================================================
sub _setBillingAndShippingAddress {
    my $self = shift;
    my ( $Customer, $AmazonOrder, $Order, $EmailTLD ) = @_;

    # Check if amazon has sent a billing address
    if( $Customer->get('BillingAddress')->get('DisplayName') &&
        $Customer->get('BillingAddress')->get('City') &&
        $Customer->get('BillingAddress')->get('Zipcode') ){
            # Copy the billing address from the customer
            my $CloneBillingAddress
              = $Customer->get('BillingAddress')->clone( { 'Parent' => $Order, 'Alias' => 'BillingAddress' } );
            $Order->set( { 'BillingAddress' => $CloneBillingAddress } );
    }

    if ( $AmazonOrder->FulfillmentAddress ) {

        my ($Firstname, $Lastname) = SplitFullname($AmazonOrder->FulfillmentAddress->Name);

        # get shipping country
        my $Country = $AmazonOrder->FulfillmentAddress->CountryCode;
        my $unknownCountryCodeNotice = '';
        my $CountryID = defined $Country ? eval{GetPKeyCountryByCode2($Country)} : undef;

        HandleError(
            OnError => sub {
                my ($Error) = @_;
                if ($Error->code eq 'NoResults_CountryByCode2') {
                    $unknownCountryCodeNotice = " / COUNTRYCODE: $Country";
                    $Order->set({'CustomerComment' => "UNKNOWN COUNTRYCODE $Country - Please check delivery address of this order in Amazon Seller Central! ".
                                 ($Order->get('CustomerComment') // '')});
                    Warning($Error);
                }
            }
        );

        my ($Company, $Street1, $Street2) = _GetStreetsAndCompany($AmazonOrder->FulfillmentAddress, $EmailTLD, $CountryID);

        my $ShippingAddress = $Order->insertChild(
            'Address',
            {
                'Alias'       => $Order->get('BillingAddress') ? 'ShippingAddress' : 'BillingAddress',
                'DisplayName' => $AmazonOrder->FulfillmentAddress->Name,
                'FirstName'   => $Firstname,
                'LastName'    => $Lastname,
                'City'        => $AmazonOrder->FulfillmentAddress->City.$unknownCountryCodeNotice,
                'CountryID'   => $CountryID,
                'Phone'       => $AmazonOrder->FulfillmentAddress->PhoneNumber,
                'Zipcode'     => $AmazonOrder->FulfillmentAddress->PostalCode,
                'State'       => $AmazonOrder->FulfillmentAddress->StateOrRegion,
                'Street'      => $Street1,
                'Street2'     => $Street2,
                'Company'     => $Company,
            }
        );

        if ($Order->get('BillingAddress')) {
           $Order->set( { 'ShippingAddress' => $ShippingAddress, } );
        } else {
           $Order->set( { 'BillingAddress' => $ShippingAddress, } );
        }

    }
    return;
}




#========================================================================================
# function     _addImportedLineItem
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object->_addImportedLineItem( $Input );
#----------------------------------------------------------------------------------------
# description  Saves imported LineItemAmazonTransaction for later use
#----------------------------------------------------------------------------------------
# input        $Item | LineItemTransaction | DE_EPAGES::Amazon::API::Object::LineItemAmazonTransaction
#========================================================================================
sub _addImportedLineItem ($) {
    my $self = shift;
    my ($Item) = @_;
    push @{ $self->{GeneratedLineItems} }, $Item;
}

#========================================================================================
# function     _getPaymentName
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemName = $Object->_getPaymentName( $hOptions );
#----------------------------------------------------------------------------------------
# description  Returns localized payment name for Amazon Payments.
#----------------------------------------------------------------------------------------
# input        $hOptions | options | hash.ref
#               used keys are:
#               <ul>
#                 <li>LangCode | 2-digit language code | string
#               </ul>
# return       $LineItemName | localized name | string
#========================================================================================
sub _getPaymentName ($) {
    my $self = shift;
    my ($hOptions) = @_;

    TestString( 'LangCode', $hOptions->{'LangCode'}, 2, 2 );

    my $LineItemName
      = ProcessFileTemplate(
        $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Amazon/Templates/MBO/Order/LineItemPayment-AmazonPayments.txt',
        $hOptions->{'LangCode'}, {} );

    return $LineItemName;
}

#========================================================================================
# function     _GetLocaleInfoFromEmailAddress
# state        private
#----------------------------------------------------------------------------------------
# syntax       ($LocaleID , $LanguageID) = _GetLocaleInfoFromEmailAddress( $Email, $Shop );
#----------------------------------------------------------------------------------------
# description  Return the LanguageID and LocaleID from the Email Attribute of Amazon order if that
#               language exists in MAP_LOCALE constant and is visible in the shop.
#----------------------------------------------------------------------------------------
# input        $Shop | Shop | DE_EPAGES::Shop::API::Object::Shop
# input        $AmazonOrder  | Amazon Order | object
# return       ($LocaleID , $LanguageID) | array with LocaleID and LanguageID | array
#========================================================================================
sub _GetLocaleInfoFromEmailAddress {
    my ( $Shop, $AmazonOrder ) = @_;

    if ( $AmazonOrder->BuyerEmailAddress =~ /\.([A-Z]{2,4})$/i )
    {    ## no critic qw(DE_EPAGES::Style::RequireExtendedFormatting)
        my $EmailTLD = $1;
        my $NewLocale = $MAP_LOCALE{$EmailTLD};
        if ($NewLocale) {
            foreach my $locale ( @{ $Shop->get('SiteLocales') } ) {
                if ( $locale->get('LocaleID') eq $NewLocale ) {
                    my $LanguageID = $locale->get('Language')->{'LanguageID'};
                    return ( $NewLocale, $LanguageID, $EmailTLD );
                }
            }
        }
    }
    return;
}


1;
