#========================================================================================
# package      DE_EPAGES::Coupon::API::Object::CouponCode
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for CouponCode.
#========================================================================================
package DE_EPAGES::Coupon::API::Object::CouponCode;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_NET TAX_MODEL_GROSS );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias  );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $CouponCode->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'CouponCode' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $CouponCode->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

sub ErrorDefinition {
    return {
        CouponCurrencyMismatch => 'The coupon #CouponAlias can\'t be used in the lineitem container, because the currency is different.',
    };
}

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $CouponCode = DE_EPAGES::Coupon::API::Object::CouponCode->insert($hInfo);
# example      my $Class = LoadClassByAlias('CouponCode');
#               my $CouponCode = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertCouponCode' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $CouponCode | new object | object
# hook         OBJ_InsertCouponCode | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>CouponCodeID  - CouponCode object id - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $Parent = $hInfo->{'Parent'};
    Error('OBJECT_EXISTS', { 'Alias' => $hInfo->{'Alias'} }) if $Parent->existsChild($hInfo->{'Alias'});

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertCouponCode', {'Object' => $self, 'CouponCodeID' => $self->id });
    delete $hInfo->{$_} foreach qw ( CouponCampaign RedeemCount Type IsInvalidated ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $CouponCode->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteCouponCode' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteCouponCode | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>CouponCodeID  - CouponCode object id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteCouponCode', {'Object' => $self, 'CouponCodeID' => $self->id });

    # update all line items that reference the deleted coupon code
    my $ClassLineItemCoupon = LoadClassByAlias('LineItemCoupon');
    $_->set({'CouponCode' => undef}) foreach @{ $ClassLineItemCoupon->getObjects('ByCouponCodeID', [$self->id]) };

    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $CouponCode->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateCouponCode' and 'OBJ_AfterUpdateCouponCode'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateCouponCode | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>CouponCodeID  - CouponCode object id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateCouponCode | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>CouponCodeID  - CouponCode object id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateCouponCode', {'Object'=>$self, 'Values'=> $hValues, 'CouponCodeID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateCouponCode',  {'Object'=>$self, 'Values'=> $hValues, 'CouponCodeID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     lineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hLineItem = $CouponCode->lineItem( $Container, $hInfo );
#----------------------------------------------------------------------------------------
# description  Builds a line item like hash for all line item information.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# input        $hInfo | keys <ul>
#                   <li>Quantity      - quantity            (optional default=MinOrder)- float
#                   <li>CurrencyID    - alpha currency code (optional default=Container.CurrencyID) - char(3)
#                   <li>TaxModel      - gros or net price   (optional default=Container.TaxModel)  - int
#                   <li>IsQuantityChanged - quantity was changed by product (optional default=0)  - int
#                   <li>.... (optional)- ...
#               </ul> | ref.hash
# return       $hLineItem |  hash for lineitem insert, keys <ul>
#                   <li>SKU
#                   <li>Name
#                   <li>TaxClass
#                   <li>Quantity
#                   <li>BasePrice
#                   <li>Discount
#                   <li>LineItemPrice
#                   <li>Product
#                   <li>TaxRate
#                   <li>IsQuantityChanged
#               </ul> | ref.hash
#========================================================================================
sub lineItem {
    my $self = shift;
    my ($Container, $hInfo) = @_;

    my $LineItemPrice = 0;
    my $BasePrice     = undef;

    $hInfo->{'CurrencyID'} = $Container->get('CurrencyID') unless defined $hInfo->{'CurrencyID'};
    my $CouponCampaign = $self->get('CouponCampaign');

    if( $CouponCampaign->get('CurrencyID') ne $hInfo->{'CurrencyID'} ) {
        # Can't use this coupon,because th coupaign has a different currency,even
        # for type PERCENT and SHIPPING which don't really depend on the currency.
        Error('CouponCurrencyMismatch', {
            LineItemContainerCurrencyID => $hInfo->{'CurrencyID'},
            CouponCampaignCurrencyID => $CouponCampaign->get('CurrencyID'),
            LineItemContainer => $Container->get('Path'),
            CouponAlias => $self->alias,
        });
    }

    my $CouponLineItem = $Container->get('Coupon');
    my $aValidationErrors = ( defined $CouponLineItem ) ? $CouponLineItem->get('ValidationErrors') : [];

    # coupon is for the basket incl. shipping
    my $InclShippingCosts = $CouponCampaign->get('InclShippingCosts');

    # get the price for shipping
    my $ShippingPrice = 0;
    my $ShippingLineItem = $Container->get('Shipping');
    if (defined $ShippingLineItem) {
        #recalculate the shipping lineitem because the shipping price can be changed by changing other prices (e.g. weight of basket)
        $ShippingLineItem->recalculate($Container);
        # get the correct price now for setting the coupon price
        $ShippingPrice = $ShippingLineItem->get('LineItemPrice');
    }

    my $Type =  $CouponCampaign->get('Type');
    my $TaxClass = undef;
    if ( $Type eq  'FIX') {

        # test if we have a gift  coupon (use grand total from container for calculation)
        # gift coupons are not included in tax calculation, they are only reduce the grand total price
        my $IsGiftCoupon = $CouponCampaign->get('WithoutTax');

        $BasePrice = $IsGiftCoupon ? $Container->get('GrandTotal') : $Container->get('LineItemsSubTotalDiscountable');

        # add shipping costs to the base price
        # normal coupon handling
        if ($InclShippingCosts and not $IsGiftCoupon) {
            $BasePrice += $ShippingPrice;
        }


        # we have a gift coupon and he is not valid for shipping costs
        if ($IsGiftCoupon and not $InclShippingCosts) {
            my $ShippingGrossPrice;

            # we have to substract the shipping gross price if tax model is net
            if ($Container->get('TaxModel') == TAX_MODEL_NET and defined $ShippingLineItem->get('TaxClass')) {
                $ShippingGrossPrice =  $Container->get('TaxMatrix')->net2gross($hInfo->{'CurrencyID'}, $ShippingPrice,
                                                          {'TaxModel' => TAX_MODEL_GROSS,
                                                           'TaxArea' =>  $Container->get('TaxArea'),
                                                           'TaxClass' => $ShippingLineItem->get('TaxClass'),},1);
                # substract shipping gross price from gift coupon if shipping costs are not included
                $BasePrice -= $ShippingGrossPrice;
            } else {
                # substract shipping from gift coupon if shipping costs are not included
                $BasePrice -= $ShippingPrice ;
            }
        }

        # set the coupon lineitem pice to the base price if coupon price is higher (you can not get cash back with a coupon)
        my $Value  = $CouponCampaign->get('Value');
        $LineItemPrice = $Value > $BasePrice ? $BasePrice : $Value;

    } elsif ( $Type eq  'PERCENT') {

        my $LocaleID = $Container->get('LocaleID');
        my $Locale = $Container->getSite()->siteLocale($LocaleID);

        #-- 'Use percentage discount of X% only for these products?' = NO
        #-- The percentage discount value is calculated and placed into the line item price of the coupon line item
        if (!$CouponCampaign->get('UseCouponOnlyForSelectedProducts')) {
            $BasePrice = $InclShippingCosts ? $Container->get('LineItemsSubTotalDiscountable') : $Container->get('LineItemsSubTotal');

            if ($InclShippingCosts ) {
                $BasePrice += $ShippingPrice;
            }
            my $Value  = $CouponCampaign->get('Value');
            $LineItemPrice = $BasePrice * $Value / 100; ## no critic qw(ProhibitMagicNumbers)
            if ( $LineItemPrice ) {
                $LineItemPrice = $Locale->roundMoney($LineItemPrice, $hInfo->{'CurrencyID'});
            }
        } elsif ( $CouponCampaign->get('UseCouponOnlyForSelectedProducts') and $InclShippingCosts ) {
          #-- 'Use percentage discount of X% only for these products?' is set to YES
          #-- Discounts are calculated and set per line item product that is valid for the coupon and shown in its 'Discount' column
          #-- If 'also valid for delivery costs' is NO, we don't need to worry
          #-- If YES, we need to do the discount calculation here for just the shipping line item
          #--   this means that the line item price for the coupon line is the calculated percentage discount value for just our shipping line
          #--
          #-- Note: shipping line templates do not consistently show a value in the 'discount' column across all sf templates (if at all)
          #--   probably because there are so many shipping classes/lineItem templates to account for
          #--
          #-- Comparative logic can also be placed in DE_EPAGES::Coupon::Hooks::LineItemContainer, where we can modify the line item price directly
          #-- However, we think it would be better here as it will be easier for the customer to see the actual discount applied
          #-- Although it would probably be ideal if it showed in its own discount column, but see note above
          my $CouponDiscount = ($CouponCampaign->get('Value') / 100) * ($ShippingPrice);

          if ( $CouponDiscount && $CouponDiscount > 0 ) {
            $LineItemPrice = $Locale->roundMoney($CouponDiscount, $hInfo->{'CurrencyID'});
          }
        }

    } elsif ( $Type eq  'SHIPPING') {
        # set the price to 0 because the coupon is for free shipping
        $BasePrice = 0 ;
        $InclShippingCosts = undef;
    } else {
        Error('UnknownCouponType', {'Type' => $Type });
    }


    $LineItemPrice = scalar(@$aValidationErrors)
        ? 0
        : $Container->formatter->round( $LineItemPrice * -1 ),

    my %Info = (
        'SKU'                   => $self->alias,
        'Name'                  => $CouponCampaign->get('NameOrAlias', $Container->get('LanguageID')),
        'Quantity'              => 1,
        'BasePrice'             => $BasePrice,
        'LineItemPrice'         => $LineItemPrice,
        'CouponCode'            => $self,
        'Position'              => 1000,
        'TaxClass'              => $TaxClass,
        'IsForShipping'         => $Type eq  'SHIPPING' ? 1 : 0,
        'InclShippingCosts'     => $InclShippingCosts,
    );

    return \%Info;
}


#========================================================================================
# function     addToContainer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $CouponCode->addToContainer( $Container, $hInfo );
#----------------------------------------------------------------------------------------
# description  Adds a line item to container after checking if the Coupon is invalidated,
#               if the valid coupon count is reached, if the coupon is effective/valid
#               and if there is a coupon in the basket already.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# input        $hInfo |  |hash
# return       $LineItem | coupon lineitem | object
#========================================================================================
sub addToContainer {
    my $self = shift;
    my ($Container, $hInfo) = @_;

    Error('CouponIsInvalidated') if $self->get('IsInvalidated');
    my $CouponCampaign = $self->get('CouponCampaign');
    Error('ValidCouponCountReached') if (defined $CouponCampaign->get('ValidCouponCount') and $CouponCampaign->get('RedeemCount') >= $CouponCampaign->get('ValidCouponCount'));
    Error('CouponNotEffective') unless $CouponCampaign->get('IsEffective');

    my $CouponLineItem = $Container->get('Coupon');
    Error('AlreadyCouponAtBasket') if defined $CouponLineItem;

    # test if we have a gift coupon or a normal coupon (depends on the attribute WithoutTax on the coupon campaign  )
    my $LineItemClass = LoadClassByAlias( $CouponCampaign->get('WithoutTax') ? 'LineItemGiftCoupon' : 'LineItemCoupon');

    return $Container->insertLineItem( $LineItemClass, $self->lineItem($Container, $hInfo) );
}

#========================================================================================
# function     recalculateLineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CouponCode->recalculateLineItem($LineItem);
#----------------------------------------------------------------------------------------
# description  Coupon line items call this function to recalculate the item.
#----------------------------------------------------------------------------------------
# input        $LineItem | line item which will be recalculated | ref.hash
#========================================================================================
sub recalculateLineItem {
    my $self = shift;
    my ($LineItem) = @_;
    $LineItem->set($self->lineItem($LineItem->container));
    return;
}

#========================================================================================
# function     redeem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CouponCode->redeem;
#----------------------------------------------------------------------------------------
# description  Redeems a coupon and increases the count of redeemed coupons. If a coupon
#               is redeemable only once then it'll be invalidated.
#========================================================================================
sub redeem {
    my $self = shift;
    my $CouponCampaign= $self->get('CouponCampaign');
    my %Vars;
    $Vars{'RedeemCount'} = $self->get('RedeemCount') + 1;
    if ( $CouponCampaign->get('IsRedeemableOnlyOnce') ) {
        $Vars{'IsInvalidated'} = 1;
    }
    $self->set(\%Vars);
    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Coupon::API::Object::CouponCode->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'RedeemCount' => 0,
        'IsInvalidated' => 0,
    };
}

1;
