use strict;
use Getopt::Long;
use Time::HiRes qw (time);

use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadObjectByPath );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Shop::API::System qw( GetUnityShopIDs );

#========================================================================================
# function         PatchUseCouponOnlyForSelectedProducts
# state            public
#----------------------------------------------------------------------------------------
# syntax           PatchUseCouponOnlyForSelectedProducts();
#----------------------------------------------------------------------------------------
# description      Adapts the UseCouponOnlyForSelectedProducts attribute
#                   in CouponCampaign so it's true for Unity Shops
#========================================================================================
sub PatchAttributeUseCouponOnlyForSelectedProducts {
    my ($ShopAlias) = @_;

    if (defined $ShopAlias) {
        # single shop run
        my $Shop = LoadObjectByPath('/Shops/'.$ShopAlias);
        if ($Shop->isUnity()) {
            # Unity shop
            _PatchAttributeUseCouponOnlyForSelectedProducts($Shop);
        } else {
            # non Unity shop
            print 'Shop ' . $ShopAlias . " is not a Unity shop.\n";
        }
    } else {
        # all Unity shops
        my $aShopIDs = GetUnityShopIDs();

        foreach my $ShopID (@$aShopIDs) {
            eval {
                my $Shop = LoadObject($ShopID);
                _PatchAttributeUseCouponOnlyForSelectedProducts($Shop);
            };
            warn('Patch of the UseCouponOnlyForSelectedProducts failed for shop ID ' . $ShopID . " with error \"" . GetError() . "\".")
                if ExistsError();
        }
    }

    return;
}

sub _PatchAttributeUseCouponOnlyForSelectedProducts {
    my ($Shop) = @_;

    my $aCouponCampaigns = $Shop->get('CouponCampaigns');
    foreach my $CouponCampaign (@$aCouponCampaigns) {
        if ($CouponCampaign->get('Type') eq 'PERCENT' and (
            scalar( grep{ $_->get('Type') eq 'PRODUCT' } @{$CouponCampaign->folder('CouponConditions')->children()})
        )
        ) {
            $CouponCampaign->set({ 'UseCouponOnlyForSelectedProducts' => 1 });
        } else {
            $CouponCampaign->set({ 'UseCouponOnlyForSelectedProducts' => 0 });
        }
    }
    return;
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shopalias=s'    => \$ShopAlias,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { PatchAttributeUseCouponOnlyForSelectedProducts($ShopAlias); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Adapts the UseCouponOnlyForSelectedProducts attribute in CouponCampaign attribute so it's true for Unity Shops
Usage:
    perl $0 [options] [flags] datafiles

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
    -shopalias  alias of a single shop                      (optional)
                (if not given, the script is preformed for all Unity shops)
flags:
    -help       show the command line options

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );
