#==============================================================================
# package      DE_EPAGES::Coupon::UI::CouponCampaign
# state        private
#==============================================================================
package DE_EPAGES::Coupon::UI::CouponCampaign;
use base qw( DE_EPAGES::Presentation::UI::Object );

use strict;
use DE_EPAGES::Coupon::API::Constants qw ( COUPON_CODE_LENGTH_UNITY );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Presentation::API::Template qw ( Translate );
use DE_EPAGES::Presentation::UI::Tools qw( CheckLengthLocalizedFieldsOnUpdate );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

#==============================================================================
# Event handlers
#==============================================================================
sub NewCouponCampaign {
    my $self    = shift;
    my ($Servlet) = @_;
    my $Shop    = $Servlet->object;

    my $Folder  = $Shop->folder('CouponCampaigns');
    my $Class   = LoadClassByAlias ('CouponCampaign');
    return $self->SUPER::NewObject($Servlet, $Folder, $Class);
}

sub Save {
    my $self = shift;
    my ($Servlet) = @_;

    my $CouponCampaign = $Servlet->object;

    my $Form = $Servlet->form;
    my $type = $Form->formValue('Type');
    my $formName = ($type eq 'FIX')?'CouponCampaign_MoneyValue':'CouponCampaign_FloatValue';

    # load form field definitions and set CurrencyID from the coupon campaign
    my $hFormData = $Form->formData( $CouponCampaign->class, $formName );
    foreach my $hFieldDef ( grep { $_->{'Type'} eq 'reg_money' } values %{$hFormData->{'Fields'}} ) {
        $hFieldDef->{'CurrencyID'} = $CouponCampaign->get('CurrencyID');
    }

    my $hFormValues = $Form->checkValues( $CouponCampaign, $hFormData );
    my $hFormValuesLanguage = $Form->form($CouponCampaign, 'CouponCampaignLanguages');

    # check if another object exists with the new alias
    my $NewAlias = $hFormValues->{Alias};
    if (defined $NewAlias &&
        $CouponCampaign->alias !~ /^\Q$NewAlias\E$/i &&
        $CouponCampaign->parent->existsChild($NewAlias)) {
        $Servlet->vars('OBJECT_EXISTS',1);
        $Servlet->form->addFormError({
           'Form'  => 'Save',
           'Reason'=> 'OBJECT_EXISTS',
           'Name'  => 'Alias',
           'Value' => $NewAlias,
        });
    }

    # set default value if form field was left empty
    if(not $hFormValues->{'MinLineItemsSubTotal'}){
        $hFormValues->{'MinLineItemsSubTotal'} = 0;
    }

    # some extra tests for time
    if ($hFormValues->{'TimeTo'} and not $hFormValues->{'TimeFrom'}) {
        $Servlet->form->addFormError({
            'Reason'=> 'MissingData',
            'Name'  => 'TimeFrom',
       }) ;
    }
    if ($hFormValues->{'TimeFrom'} and not $hFormValues->{'TimeTo'}) {
        $Servlet->form->addFormError({
            'Reason'=> 'MissingData',
            'Name'  => 'TimeTo',
        }) ;
    }
    if ($hFormValues->{'TimeFrom'} and $hFormValues->{'TimeTo'} and
        $hFormValues->{'TimeFrom'} > $hFormValues->{'TimeTo'})
    {
        $Servlet->form->addFormError({
            'Reason'=> 'FromLaterThanTo',
            'Name'  => 'TimeFrom',
        }) ;
        $Servlet->form->addFormError({
            'Reason'=> 'FromLaterThanTo',
            'Name'  => 'TimeTo',
        }) ;
    }
    # extra tests for mandatroy fields depending on Type
    if ($hFormValues->{'Type'} eq 'FIX' and not $hFormValues->{'Value'}) {
        $Servlet->form->addFormError({
            'Reason'=> 'MissingData',
            'Name'  => 'FixValue',
        }) ;
    }
    if ($hFormValues->{'Type'} eq 'PERCENT' and not $hFormValues->{'Value'}) {
        $Servlet->form->addFormError({
            'Reason'=> 'MissingData',
            'Name'  => 'PercentValue',
        }) ;
    }

    CheckLengthLocalizedFieldsOnUpdate($Servlet);

    $Servlet->form->executeFormError;

    if ($hFormValues->{'Type'} ne 'PERCENT') {
        $hFormValues->{'UseCouponOnlyForSelectedProducts'} = 0;
    }

    # coupon can not be ignored if type is not FIX
    $hFormValues->{'WithoutTax'} = 0 unless $hFormValues->{'Type'} eq 'FIX';

    $CouponCampaign->set($hFormValues);
    foreach my $hLanguage (@{$hFormValuesLanguage->{'Languages'}}) {
        for (qw(Name Description)) {
            $CouponCampaign->set({$_ => $hLanguage->{$_}}, $hLanguage->{'LanguageID'});
        }
    }
}

#========================================================================================
# function     ViewPagedChildren
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hVars = $Handler->ViewPagedChildren( $Servlet );
#----------------------------------------------------------------------------------------
# description  Creates a pager for CouponCampaigns.
#
#----------------------------------------------------------------------------------------
# input        $Servlet | current servlet (contains form data) | object
# return       $hVars | TLE variables for 'Pager' | ref.hash
#========================================================================================
sub ViewPagedChildren {
    my $self = shift;
    my ($Servlet) = @_;

    # trigger hook to allow the adding of context sensitive apps
    TriggerHook('UI_Shop_UnityViewCouponCampaigns', { 'Servlet' => $Servlet });

    my $Object = $Servlet->object->folder('CouponCampaigns');
    my $Pager = DE_EPAGES::Presentation::API::Object::PagerChildren->new({'Servlet' => $Servlet, 'Object' => $Object});
    return $Pager->page;
}

sub DeleteCouponConditions {
    my $self = shift;
    my ($Servlet) = @_;
    my $Campaign = $Servlet->object;
    my $CouponConditionClass = LoadClassByAlias('CouponCondition');

    $self->SUPER::DeleteListedObjects( $Servlet, $CouponConditionClass );

    # test if we have at least 1 condition of type PRODUCT
    # UseCouponOnlyForSelectedProducts to 0 if there are not product in the condition list,
    # otherwise the coupon can not be used for other products
    return unless $Campaign->get('UseCouponOnlyForSelectedProducts');

    foreach my $Condition (@{$Campaign->get('CouponConditions')}) {
        return if $Condition->get('Type') eq 'PRODUCT';
    }
    $Campaign->set({'UseCouponOnlyForSelectedProducts' => 0});

}

sub UnityViewNewCouponCampaign {
    my $self = shift;
    my ($Servlet) = @_;

    my $Folder = $Servlet->object;
    return unless $Folder->alias eq 'CouponCampaigns';

    $Servlet->params('IsNewCampaign', 1);
    $Servlet->params('Title', Translate('Coupons'));

    # load new product default attributes
    $Servlet->vars('Class', LoadClassByAlias('CouponCampaign'));
    return $self->ViewNewObject($Servlet);
}

sub SaveUnityNewCouponCampaign {
    my $self = shift;
    my ($Servlet) = @_;

    my $Parent  = $Servlet->object;
    my $Form    = $Servlet->form;

    my $Name = $Form->value('Name');
    $Form->executeFormError({
       'Form'  => 'Save',
       'Reason'=> 'FORM_NO_NAME',
       'Name'  => 'Name'
    }) unless $Name;

    my $Class = LoadClassByAlias('CouponCampaign');
    $Servlet->vars('Class', $Class);

    # insert new object
    my $TempObject = $self->SUPER::NewObject($Servlet, $Parent, $Class);

    # save form on new object
    $Servlet->params('ObjectID', $TempObject->id);
    $Servlet->params('ViewAction', 'UnityMBO-ViewGeneral');
    $Servlet->params('ViewObjectID', $TempObject->id);

    $self->SaveUnityCouponCampaign($Servlet);
}

sub SaveUnityCouponCampaign {
    my $self = shift;
    my ($Servlet) = @_;
    my $CouponCampaign = $Servlet->object;
    my $Form = $Servlet->form;
    my $CouponCodeExistsError = 0;

    $Form->executeFormError({
       'Form'  => 'Save',
       'Reason'=> 'FORM_NO_NAME',
       'Name'  => 'Name'
    }) unless $Form->value('Name');

    #set default values
    my $hValues = $Form->values();
    $hValues->{'Alias'} = $CouponCampaign->get('Alias');
    $hValues->{'WithoutTax'} = 0;
    $hValues->{'InclShippingCosts'} = 0;
    $hValues->{'CouponCodeLength'} = COUPON_CODE_LENGTH_UNITY;
    $Form->setValues($hValues);

    $self->Save($Servlet);

    # percentage discount for complete basket. The feature only for products ist not available yet in now
    $CouponCampaign->set({'UseCouponOnlyForSelectedProducts' => 1})
            if $Form->value('Type') eq 'PERCENT' and
            scalar( grep{ $_->get('Type') eq 'PRODUCT' } @{$CouponCampaign->folder('CouponConditions')->children()});


    #Save coupon Codes (autogenerated and manually)

    # generate coupon codes automatically
    if ($Form->value('GenerateCouponCodes')) {

        my $hGenerateForm = $Servlet->form->form($CouponCampaign, 'GenerateCouponCodes');
        eval {
            $CouponCampaign->generateNewCouponCodes( $hGenerateForm->{'NumberOfCodes'} );
        };
        if (ExistsError()) {
            $Servlet->form->mapFormError({'Form' => 'CouponForm', 'Name' => 'GenerateCoupons'}, GetError());
        }

    } else {
    # create code(s) manually
        my $hFormValues = $Servlet->form->form($CouponCampaign, 'ManuallyGenerateCouponCodes');
        my $ahCouponCodes = $hFormValues->{'CouponCodes'};
        my $Folder = $CouponCampaign->getSite->folder('CouponCodes', { Position=> 10 });


        my $Index = 0;
        foreach (@$ahCouponCodes) {
            # test if already exists and add form error
            next unless $_->{'CouponCode'} =~ /\S/;
            if($Folder->existsChild($_->{'CouponCode'})){
                $Form->addFormError({
                   'Form'  => 'Save',
                   'Reason'=> 'COUPON_CODE_EXISTS',
                   'Index' => $Index,
                   'LoopName' => 'CouponCodes',
                   'Name'  => 'CouponCode',
                   'Value' =>  $_->{'CouponCode'}
                });
                $CouponCodeExistsError = 1;
                next;
            }
            # create codes
            $Folder->insertChild('CouponCode', {
                'Alias' => $_->{'CouponCode'},
                'Type'  => 'MANUAL',
                'CouponCampaign' => $CouponCampaign,
            });
            $Index ++;
        }
    }

    Error('INVALID_FORM', {
        'ViewVars'=> {
            'CouponCodeExistsError' => $CouponCodeExistsError,
            'FormError' => 1,
            'ViewObjectID' => $CouponCampaign->id,
            'ViewAction' => 'UnityMBO-ViewGeneral',
            'ErrorObjectID' => $CouponCampaign->id,
            'ErrorAction' => 'UnityMBO-ViewGeneral',
        },
    }) if $Servlet->form->existsFormError();

    #$Servlet->form->executeFormError();

}

sub UnityView {
    my $self = shift;
    my ($Servlet) = @_;

    # enable language switch
    $Servlet->vars('ShowLanguageSwitch', 1);

    $self->View($Servlet);
}

1;

