#========================================================================================
# package      DE_EPAGES::Product::API::Resource::ProductCrossSelling
# state        public
# base         DE_EPAGES::WebService::API::Resource::BaseObjectResource
# description  implements the REST resource for crossselling.
#========================================================================================
package DE_EPAGES::CrossSelling::API::Resource::CrossSelling;
use base DE_EPAGES::WebService::API::Resource::BaseObjectResource;

use DE_EPAGES::Core::API::Url qw (ParseQueryString);
use DE_EPAGES::Object::API::Factory qw (LoadObjectByPath);
use DE_EPAGES::Object::API::Object::Locale;
use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Shop::API::Resource::Formatter;
use DE_EPAGES::Tax::API::Shop qw ( TaxInfoBySellingCountryID );
use DE_EPAGES::Product::API::Resource::Product;
use DE_EPAGES::Product::API::ProductService;
use DE_EPAGES::WebService::API::ResourcePager;

use feature 'switch';

use strict;



sub service {
    return LoadObjectByPath('/RestServices/CrossSelling');
}



sub class {
    return 'Product';
}



sub producedContentTypes {
    shift;
    my ($RequestMethod) = @_;

    return {
        GET => [
            ['get', 'application/vnd.epages.v1+json'],
            ['get', 'application/json'],
        ],
    }->{$RequestMethod};
}



sub match {
    my $self = shift;
    my ($Request) = @_;

    return unless $Request->uri->path =~ m|^/api/products/([^/]+)/crossselling/?$|;

    return $1;
}



sub params {
    my $self = shift;
    my ($ActionName, $Object, $Request) = @_;

    given( $ActionName ) {
        when( 'get' ) {
            my $hQuery = ParseQueryString( $Request->uri->query, 'utf-8' );
            my $LocaleID = $hQuery->{'locale'}->[0];
            my $CurrencyID = $hQuery->{'currency'}->[0];
            my $hPagerParams = {'PageSize' => $hQuery->{'resultsPerPage'}->[0],
                                'Page'     => $hQuery->{'page'}->[0]
                               };
            my $IncludeInvisible = $hQuery->{'includeInvisible'}->[0];
            my $SellingCountryID = $Request->header('X-EPAGES-SELLINGCOUNTRYID');
            my $SessionID = $Request->header('X-EPAGES-SESSIONID');

            return [ $Object, $LocaleID, $CurrencyID, $hPagerParams, $IncludeInvisible, $SellingCountryID, $SessionID ];
        }
        default {
            return [ $Object ];
        }
    }
}



sub href {
    my $self = shift;
    my ($APIBaseURL, $ProductGUID) = @_;
    return "$APIBaseURL/api/products/$ProductGUID/crossselling/";
}



sub hrefToProduct {
    my $self = shift;
    my ( $APIBaseURL, $GUID ) = @_;

    return DE_EPAGES::Product::API::Resource::Product->new()->href($APIBaseURL, $GUID);

}



sub get {
    my $self = shift;
    my ($Product, $LocaleID, $CurrencyID, $hPagerParams, $IncludeInvisible, $SellingCountryID, $SessionID) = @_;

    my $Shop = $Product->getSite;
    my $Context = DE_EPAGES::Shop::API::Resource::ShopRequestContext->new($Shop);

    $LocaleID = $Context->validLocale($LocaleID);
    $CurrencyID = $Context->validCurrency($CurrencyID);

    my $Locale= DE_EPAGES::Object::API::Object::Locale->load($LocaleID);
    my $LanguageID = $Locale->get('LanguageID');

    my $APIBaseURL = $Context->apiBaseUrl();

    # we fetch the crossselling products for the product here
    my $aCrossSellingProducts = [] ;
    my $AttributeName = defined $IncludeInvisible && lc($IncludeInvisible) eq 'true' ? 'CrossSelling' : 'VisibleCrossSelling';
    # use cross-selling products of the master product for variations (UNITY-4850)
    foreach my $CrossSellingObject (@{$Product->getWithParent($AttributeName)}){
        push @$aCrossSellingProducts, $CrossSellingObject->get('TargetProduct');
    }

    my $hPParams = {'PageSize' => $hPagerParams->{'PageSize'} // 100,
                    'Page'     => $hPagerParams->{'Page'} // 1
                   };
    my $Pager = DE_EPAGES::WebService::API::ResourcePager->new($hPParams);

    my $ProductService = $self->_createProductService();

    my $hShopData = $ProductService->getShopData($Shop,$LocaleID, $CurrencyID);
    my $hSellingCountryData = TaxInfoBySellingCountryID($Shop, $SellingCountryID );

    my $PagedItems =  $Pager->pagedItems($aCrossSellingProducts, sub {
        $ProductService->getProductData($_[0], $hShopData, $SessionID, $hSellingCountryData);
    });

    my $hResult = {};
    $hResult->{'manualCrossSellingTitle'} = $Shop->get('IsUnityRegistered') ?
                                            $Shop->get('UnityCrossSellingBoxHeadline',$LanguageID) : undef ;
    # add paged items to result
    $hResult->{'embedded'} = $PagedItems;

    # add a link to self.
    push @{$hResult->{'links'}}, {
        rel => 'self',
        title => undef,
        href => $self->href($APIBaseURL, $Product->guid),
    };

    # add a link to product.
    push @{$hResult->{'links'}}, {
        rel => 'product',
        title => undef,
        href => $self->hrefToProduct($APIBaseURL, $Product->guid),
    };

    return $hResult;
}



sub _createProductService {
    return DE_EPAGES::Product::API::ProductService->new();
}

1;