#========================================================================================
# package      DE_EPAGES::Database::API::DBISQLite
# base         DE_EPAGES::Database::API::DBI
# state        public
# description  central database interface for sqlite databases.
#========================================================================================
package DE_EPAGES::Database::API::DBISQLite;
use base DE_EPAGES::Database::API::DBI;

use strict;

use DateTime::Format::MySQL;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Constants qw(PERL516);
use DE_EPAGES::Core::API::Error qw( Error ExistsError GetError HandleError );
use DE_EPAGES::Core::API::Hook qw (TriggerCoreHook);
use DE_EPAGES::Core::API::Log qw( GetLog LogDebug );
use DE_EPAGES::Core::API::PerlTools qw( IsA );
use DE_EPAGES::Core::API::Statistics qw( AddStatistic );
use DE_EPAGES::Core::API::Warning qw( Warning );

sub ErrorDefinition {
    return {
        'DBIPrepare'    => 'prepare failed : #message at #statement',
        'DBIExecute'    => 'execute failed : #message at #statement',
    };
}

sub _envvar { ## no critic qw(ProhibitUnusedPrivateSubroutines)
    my ($var) = @_;
    return $ENV{$var} ? $ENV{$var} : "_${var}_";
}
sub new ($$$$$) {
    my $class = shift;
    my ($datasource, $user, $password, $cachekey) = @_;

    # replace each _VAR_ by $ENV{VAR} (like _EPAGES_CONFIG_
    # by $ENV{EPAGES_CONFIG}) if $ENV{VAR} is true
    $datasource =~ s/_(\w+)_/&_envvar($1)/eg;

    my $self = DE_EPAGES::Database::API::DBI::new($class, $datasource, $user, $password, $cachekey);
    $self->{'datetime_parser'} = DateTime::Format::MySQL->new;

    return $self;
}

#========================================================================================
# function     disconnect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $dbi->disconnect;
#----------------------------------------------------------------------------------------
# description  Disconnects the underlying database handle.
#========================================================================================
sub disconnect ($) {
    my $self = shift;

    # clean statement handle cache
    if( ref $self->{'STATEMENT_STH'} ) {
        $_->[0]->finish foreach values %{$self->{'STATEMENT_STH'}};
        %{$self->{'STATEMENT_STH'}} = ();
    }

    $self->SUPER::disconnect;
}

# retry up to 5 times on 'disk I/O error' (produced by the NFS at Amen)
sub _prepare {
    my $self = shift;
    my ($statement) = @_;

    my $dbh = $self->dbh;
    my $TryCounter = 0;
    while( 1 ) {
        my $sth = $dbh->prepare($statement);
        return $sth if $sth;

        $TryCounter++;
        my $ErrorMessage = $dbh->errstr || 'unknown error';

        if( $dbh->{'AutoCommit'} &&
            $TryCounter < 5 &&
            $ErrorMessage =~ /disk I\/O error/
        ) {
            GetLog->info( sprintf 'SQLite prepare error %d "%s" for "%s". Trying again.',
                $TryCounter, $ErrorMessage, $statement );
            sleep 1;
            next;
        }

        Error( 'DBIPrepare', {
            'message'   => $dbh->errstr,
            'statement' => $statement,
            'state'     => $dbh->state,
            'code'      => $dbh->err,
            'source'    => $self->source,
            'TryCounter' => $TryCounter,
        });
    }
}

# retry up to 5 times on 'disk I/O error' (produced by the NFS at Amen)
sub _execute {
    my $self = shift;
    my ($sth, $aParameter, $statement) = @_;


    my $dbh = $self->dbh;
    my $TryCounter = 0;
    while( !$sth->execute(@$aParameter) ) {
        $TryCounter++;
        my $ErrorMessage = $dbh->errstr || 'unknown error';

        if( ! $self->inTransaction &&
            $TryCounter < 5 &&
            $ErrorMessage =~ /disk I\/O error/
        ) {
            GetLog->info( sprintf 'SQLite execute error %d "%s" for "%s". Trying again.',
                $TryCounter, $ErrorMessage, $statement );
            sleep 1;
            next;
        }

        Error ('DBIExecute', {
            'message'   => $sth->errstr,
            'statement' => $statement,
            'parameter' => join (',', map { $_ // 'undef' } @$aParameter),
            'state'     => $sth->state,
            'code'      => $sth->err,
            'source'    => $self->source,
            'TryCounter' => $TryCounter,
        });
    }
}

sub ExecuteStatement {
    my $self = shift;
    my ($statement, $aParameter, $with_output) = @_;
    $aParameter = [] unless defined $aParameter;
    $_ = $self->quoteDateTime($_) foreach grep { IsA( $_, 'DateTime') } @$aParameter;
    $with_output = 1 unless defined $with_output;

    my $time1 = time;
    LogDebug('Statement', $statement, __PACKAGE__ . '::ExecuteStatement::Statement');
    LogDebug('Parameter', $aParameter, __PACKAGE__ . '::ExecuteStatement::Parameter');

    my $cachedstatement = 0;
    my $sth;

    if (scalar @$aParameter && $self->{'cache_sth'}) {
        my $cacheentry = $self->{'STATEMENT_STH'}{$statement};
        if (defined $cacheentry) {
            $sth = $cacheentry->[0];
            $cacheentry->[1]++;
        } else {
            $sth = $self->_prepare($statement);
            $cacheentry = [$sth, 1];
            $self->{'STATEMENT_STH'}{$statement} = $cacheentry;
        }
        $cachedstatement = 1;
    } else {
        $sth = $self->_prepare($statement);
    }
    eval { $self->_execute( $sth, $aParameter, $statement ); };
    HandleError( OnError => sub {
        delete $self->{'STATEMENT_STH'}{$statement};
        $sth->finish;
    });

    my @result;
    if ($with_output) {
        my $aValues;
        while( $aValues = $sth->fetchrow_arrayref ) {
            push @result, [@$aValues];
        }
    }
    $sth->finish;

    my $timediff = time - $time1;
    if ($timediff > 0.1) { ## no critic qw(ProhibitMagicNumbers)
        LogDebug('LongStatement', $statement, __PACKAGE__ . '::ExecuteStatement::LongStatement::'.$timediff);
    }
    LogDebug('Result', \@result, __PACKAGE__ . '::ExecuteStatement::Result');
    AddStatistic($self->driverName . 'Execute', $timediff);
    AddStatistic('DBISQLite:'.$statement, $timediff);

    return \@result;
}

sub execute {
    my $self = shift;
    my ($statement, $raParameter, $aDttps, $aColumnNames) = @_;
    return $self->unquoteResultSet($self->ExecuteStatement($statement, $raParameter), $aDttps, $aColumnNames );
}

#========================================================================================
# function     do
# state        public
#----------------------------------------------------------------------------------------
# syntax       $dbi->do( $SQL, $aParams, $aTypes );
#----------------------------------------------------------------------------------------
# description  Executes the given statement and logs it via
#               <function DE_EPAGES::Database::API::DBI::logTransaction>.
#----------------------------------------------------------------------------------------
# input        $SQL | SQL statement | string
# input        $aParams | list of parameters to replace '?' placeholders in the SQL
#                          string | ref.array.*
# input        $aTypes | list of parameters types | ref.array.string
#========================================================================================
sub do {
    my $self = shift;
    my ($SQL, $aParams, $aTypes) = @_;
    $self->transaction(sub{
        $self->logTransaction($SQL, $aParams, $aTypes);
        $self->ExecuteStatement($SQL, $aParams, 0 );
    });
    return;
}

sub existsTableColumn {
    my $self = shift;
    my ($table, $column) = @_;

    my $sql = "SELECT * FROM $table LIMIT 0,0";
    my $sth = $self->_prepare( $sql );
    $self->_execute( $sth, [], $sql );
    my $aColumns = $sth->{NAME};
    my $Exists = ( grep { $_ eq $column } @$aColumns ) ? 1 : 0;
    $sth->finish;
    return $Exists;
}

#========================================================================================
# function     dropIndex
#----------------------------------------------------------------------------------------
# syntax       $dbi->dropIndex($table, $indexname);
#----------------------------------------------------------------------------------------
# example      $dbi->dropIndex('product', 'i_product_super');
#----------------------------------------------------------------------------------------
# description  Drops index $indexname from table $tablename.
#----------------------------------------------------------------------------------------
# input        $table | table name | string
# input        $indexname | index name | string
#========================================================================================
sub dropIndex {
    my $self = shift;
    my ($table, $indexname) = @_;

    $self->do("DROP INDEX $indexname");
    return;
}

#========================================================================================
# function     existsIndex
#----------------------------------------------------------------------------------------
# syntax       $exists = $dbi->existsIndex($Table, $IndexName);
#----------------------------------------------------------------------------------------
# example      if( $dbi->existsIndex('product', 'i_product_super') ) { ... }
#----------------------------------------------------------------------------------------
# description  Returns true if the index exists.
#----------------------------------------------------------------------------------------
# input        $Table | table name | string
# input        $IndexName | index name | string
# return       $exists | true if index exists | boolean
#========================================================================================
sub existsIndex {
    my $self = shift;
    my ($Table, $IndexName) = @_;

    my $sql = 'SELECT sql FROM sqlite_master WHERE type = \'index\' AND name = ? AND tbl_name = ?';
    my $aaRows = $self->execute( $sql, [$IndexName, $Table] );
    return scalar( @$aaRows ) ? 1 : 0;
}

#========================================================================================
# function     transaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       $dbi->transaction( $cCode, $cRollback );
# example      $dbi->transaction( sub {
#                   $dbi->do( "DELETE FROM objects WHERE classid = 25" );
#                   $dbi->do( "DELETE FROM classes WHERE classid = 25" );
#               } );
#----------------------------------------------------------------------------------------
# description  Executes a code block in a transaction.
#----------------------------------------------------------------------------------------
# input        $cCode | code block | code reference
# input        $cRollback | roll back code block (optional) | code reference
#========================================================================================
sub transaction {
    my $self = shift;
    my ($cCode, $cRollback) = @_;

    my $dbh = $self->dbh;
    if( !$self->inTransaction ) {
        my $TryCounter = 0;
        my $IOError = 0;
        my $Error;
        do {
            $IOError = 0;
            $TryCounter++;
            eval {
                local $self->{'transactionLog'} = [];
                my $LogTransaction = 0;
                {
                    $dbh->do('BEGIN EXCLUSIVE TRANSACTION') or Error ('DBIExecute', {
                        'message'   => $dbh->errstr,
                        'statement' => 'BEGIN EXCLUSIVE TRANSACTION',
                        'state'     => $dbh->state,
                        'code'      => $dbh->err,
                        'source'    => $self->source,
                        'TryCounter' => $TryCounter,
                    });
                    # disable auto-commit for old perl (with old DBD::SQLite)
                    local $dbh->{'AutoCommit'} = 0 if $] < PERL516;
                    eval {
                        $cCode->();# execute the code block

                        if( $self->isTransactionLogEnabled && scalar @{$self->{'transactionLog'}} ) {

                            # record SQL statements for replication
                            foreach my $sql ( @{ $self->{'transactionLog'} } ) {
                                # Don't use $self->do(), because it would append another statement to the log
                                $self->ExecuteStatement( 'INSERT INTO transactionlog (sql) VALUES (?)', [ $sql ] );
                                $self->{'transactionIndex'} = $self->dbh->last_insert_id('', '', 'transactionlog', 'transactionid');
                            }

                            # send changes to MC
                            $LogTransaction = 1;
                        }
                        $dbh->do('COMMIT') or Error ('DBIExecute', {
                            'message'   => $dbh->errstr,
                            'statement' => 'COMMIT',
                            'state'     => $dbh->state,
                            'code'      => $dbh->err,
                            'source'    => $self->source,
                            'TryCounter' => $TryCounter,
                        });
                        $dbh->{'AutoCommit'} = 1;
                    };
                    if( ExistsError() ) {
                        my $InnerError = GetError();
                        eval { ## no critic qw(RequireExistsError)
                            $dbh->do('ROLLBACK') or warn $dbh->errstr;
                            $cRollback->() if defined $cRollback;
                        }; # ignore follow-up errors
                        Error( $InnerError );
                    }
                }

                if( $LogTransaction ) {
                    # send changes to MC
                    TriggerCoreHook('DBISQLiteTransaction', {
                        Object=>$self, OnDestroy=>0, Index=>$self->{'transactionIndex'}
                    });
                }
            };
            if (ExistsError()) {
                $Error = GetError();
                my $ErrorMessage = $Error->vars->{'message'};
                if (defined $ErrorMessage && $ErrorMessage =~ /disk I\/O error|database is locked/ ) { ##no critic (RequireExtendedFormatting)
                    Warning($Error);
                    sleep(1);
                    $IOError = 1;
                } else {
                    Error( $Error );
                }
            }
        } while ($IOError and $TryCounter < 5);
        Error($Error) if $IOError;
    }
    else {
        $cCode->();
    }

    return;
}

1;
