use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::File qw (DeleteFile);
use DE_EPAGES::Database::API::Config qw( DatabaseConnectionData );
use DE_EPAGES::Database::API::Connection qw( RunOnStore GetCurrentDBHandle );

sub Main {

    local $| = 1;

    my $Password;
    my $ConnectionName;
    my $StoreName;
    my $DBName;
    my $Help;
    my $Verbose = 1;

    GetOptions(
        'connection=s'  => \$ConnectionName,
        'passwd=s'      => \$Password,
        'storename=s'   => \$StoreName,
        'dbname=s'      => \$DBName,
        'help'          => \$Help,
        'verbose'       => \$Verbose,
    );

    usage() if $Help;
    unless ($StoreName || $DBName) {
        print STDERR "Missing parameter -storename or -dbname\n\n"; usage();
    }
    unless ($StoreName || $ConnectionName) {
        print STDERR "Missing parameter -storename or -connection\n\n"; usage();
    }

    my $time1 = time();

    my $hConnection = $StoreName ?
        DatabaseConnectionData($StoreName, { incomplete => 1 }) : {};
    $ConnectionName = $hConnection->{'Backup'} // 'Backup' unless $ConnectionName;
    if ($hConnection->{'datasource.driver'} eq 'SQLite') {
        my $DBFile = $hConnection->{'datasource.params'}{'dbfile'};
        print "removing SQLite DB: $DBFile\n" if $Verbose;
        DeleteFile($DBFile);
    } else {
        print "using connection $ConnectionName\n" if $Verbose;
        RunOnStore(
            Store => $ConnectionName,
            DBPassword => $Password,
            Sub => sub {

                my $dbi = GetCurrentDBHandle();

                # Read login data from conf file if not set via command line.
                $DBName //= $hConnection->{'Database'};

                # create the database name from the store name, for example:
                # StoreES2 -> storeesdb2
                $DBName //= ($dbi->createUserAndStoreDBName( $StoreName ))[0];

                if( $dbi->existsDatabase($DBName) ) {
                    print "drop database '$DBName'\n" if $Verbose;
                    $dbi->dropDatabase($DBName);
                } else {
                    print "Database '$DBName' does not exist.\n" if $Verbose;
                }
            }
        );
    }

    my $time2 = time;
    my $seconds = $time2-$time1;
    printf "need %.3f seconds\n", $seconds if $Verbose;
}

sub usage {
    print <<END_USAGE;
Usage:
    perl $0 [options] [flags]

options:
    -connection    store name for master database                   (Backup)
    -passwd        database user password                         (optional)
    -storename     store name for the database connection         (required)
    -dbname        database name                                  (optional)

flags:
    -help          show the command line options
    -nooutput      do not print any output

Example:
    perl $0 -storename Store2
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => sub {
        Main();
    },
    Message => 'long',
);

1;
