use Getopt::Long;
use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::File qw ( ExistsFile  );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath LoadObject JoinObjectPath );

use DateTime;
use feature 'switch';

use strict;

my $NoOutput;
my $IsDryRun;
my $StoreName;

sub Main {
    local $| = 1;

    my $Help;
    my $Password;
    my $ListFile;
    my $Payment;

    GetOptions(
        'help'          => \$Help,
        'nooutput'      => \$NoOutput,
        'dryrun'        => \$IsDryRun,
        'passwd=s'      => \$Password,
        'storename=s'   => \$StoreName,
        'list=s'        => \$ListFile,
        'payment=s'     => \$Payment,
    );
    usage() if $Help;
    usage() unless ( defined($ListFile) || defined($Payment) );
    usage() unless defined $StoreName;

    my @Lines = ();

    if (defined $ListFile) {
        fileNotFound($ListFile) unless (ExistsFile($ListFile));

        open my $handle, '<', $ListFile or die "Unable to open $ListFile\n";
        chomp(@Lines = <$handle>);
        close $handle or die "Unable to close $ListFile\n";
    }

    if (defined $Payment) {
        push(@Lines, $Payment);
    }

    my $time1 = time;

    RunOnStore( Store => $StoreName, DBPassword => $Password, Sub => sub {
        foreach my $PaymentInfoLine (@Lines) {
            my ($PaymentTypeName, $CartridgeName) = split(/:/, $PaymentInfoLine);
            unless (ExistsCartridgeByPackage('DE_EPAGES::' . $CartridgeName)) {
                print "Store \"$StoreName\" : Cartridge \"DE_EPAGES::$CartridgeName\" is not installed (SKIP).\n" unless $NoOutput;
                next;
            }

            my $PaymentType = LoadObjectByPath(JoinObjectPath(['PaymentTypes', $PaymentTypeName]));
            my $DeprecationDate = $PaymentType->get('DeprecationDate');
            unless (defined($DeprecationDate) && (DateTime->compare($DeprecationDate, DateTime->now()) < 0)) {
                print "Store \"$StoreName\" : Payment type $PaymentTypeName is not deprecated (SKIP).\n" unless $NoOutput;
                next;
            }
            my $rAreValidSettings;

            given ($PaymentTypeName) {
                when ('4B') {
                    $rAreValidSettings = \&AreValidSettingsFor4B;
                }
                when ('Acceptance') {
                    $rAreValidSettings = \&AreValidSettingsForAcceptance;
                }
                when ('DocdataPay') {
                    $rAreValidSettings = \&AreValidSettingsForDocdataPay;
                }
                when ('PayPal') {
                    $rAreValidSettings = \&AreValidSettingsForPayPalStandardAndIntegralEvolution;
                }
                when ('PayPalIntegralEvolution') {
                    $rAreValidSettings = \&AreValidSettingsForPayPalStandardAndIntegralEvolution;
                }
                when ('PayPalPro') {
                    $rAreValidSettings = \&AreValidSettingsForPayPalPro;
                }
                when ('Saferpay') {
                    $rAreValidSettings = \&AreValidSettingsForSaferpay;
                }
                when ('SaferpayCreditCard') {
                    $rAreValidSettings = \&AreValidSettingsForSaferpayElvAndCreditCard;
                }
                when ('SaferpayELV') {
                    $rAreValidSettings = \&AreValidSettingsForSaferpayElvAndCreditCard;
                }
                when ('AmazonPayments') {
                    $rAreValidSettings = \&AreValidSettingsForAmazonPayments;
                }

                default {
                    print "Store \"$StoreName\" : Payment type \"$PaymentTypeName\" is not covered (SKIP).\n" unless $NoOutput;
                    next;
                }
            }

            RemoveNotSetPaymentMethods($rAreValidSettings, $PaymentType);
        }
    });

    my $time2 = time;
    my $seconds = $time2-$time1;
    my $logmsg = sprintf ('need %.3f seconds', $seconds);
    print $logmsg."\n" unless $NoOutput;
}

sub RemoveNotSetPaymentMethods {
    my ($rAreValidSettings, $PaymentType) = @_;

    my $hPaymentType = $PaymentType->get(['MethodClass', 'ID', 'Alias']);
    my $PaymentMethodClass = $hPaymentType->{'MethodClass'};
    my $PaymentTypeID = $hPaymentType->{'ID'};
    my $PaymentTypeAlias = $hPaymentType->{'Alias'};

    print "Store \"$StoreName\" : Process payment methods of \"$PaymentTypeAlias\" payment type\n" unless $NoOutput;

    my $aPaymentMethods = $PaymentMethodClass->findObjects('ByPaymentTypeID', [$PaymentTypeID]);
    my $Counter = 0;
    foreach my $PaymentMethodID (@$aPaymentMethods) {
        my $PaymentMethod = LoadObject($PaymentMethodID);
        unless ($rAreValidSettings->($PaymentMethod)) {
            if ($IsDryRun) {
                print "\t\tFOUND '" . $PaymentMethod->get('Path') . "' payment method\n" unless $NoOutput;
            } else {
                $PaymentMethod->delete();
                print "\t\tDELETED '" . $PaymentMethod->get('Path') . "' payment method\n" unless $NoOutput;
            }
            $Counter++;
        }
    }

    print "\tTotal payment methods of \"$PaymentTypeAlias\" payment type -> $Counter\n" unless $NoOutput;
}

#=======================================================
# PaymentMethod checkers
#=======================================================
sub AreValidSettingsFor4B {
    my ($PaymentMethod) = @_;

    my $FourBMerchant = $PaymentMethod->get('4B_merchant');
    return isValid($FourBMerchant);
}
sub AreValidSettingsForAcceptance {
    my ($PaymentMethod) = @_;

    my $MerchantName = $PaymentMethod->get('MerchantName');
    my $MerchantPassword = $PaymentMethod->get('MerchantPassword');
    my $Passphrase = $PaymentMethod->get('Passphrase');
    return isValid($MerchantName) && isValid($MerchantPassword) && isValid($Passphrase);
}

sub AreValidSettingsForDocdataPay {
    my ($PaymentMethod) = @_;

    my $MerchantName = $PaymentMethod->get('MerchantName');
    my $MerchantPassword = $PaymentMethod->get('MerchantPassword');
    my $ProfileName = $PaymentMethod->get('ProfileName');
    return isValid($MerchantName) && isValid($MerchantPassword) && isValid($ProfileName);
}

sub AreValidSettingsForPayPalStandardAndIntegralEvolution {
    my ($PaymentMethod) = @_;

    my $Account = $PaymentMethod->get('Account');
    return isValid($Account);
}

sub AreValidSettingsForPayPalPro {
    my ($PaymentMethod) = @_;

    my $APIUsername = $PaymentMethod->get('APIUsername');
    my $APIPassword = $PaymentMethod->get('APIPassword');
    my $APISignatur = $PaymentMethod->get('APISignatur');
    return isValid($APIUsername) && isValid($APIPassword) && isValid($APISignatur);
}

sub AreValidSettingsForSaferpay {
    my ($PaymentMethod) = @_;

    my $AccountID = $PaymentMethod->get('AccountID');
    return isValid($AccountID);
}

sub AreValidSettingsForSaferpayElvAndCreditCard {
    my ($PaymentMethod) = @_;

    my $AccountID = $PaymentMethod->get('AccountID');
    my $AccountPassword = $PaymentMethod->get('AccountPassword');
    return isValid($AccountID) && isValid($AccountPassword);
}

sub AreValidSettingsForAmazonPayments {
    my ($PaymentMethod) = @_;

    my $MerchantID = $PaymentMethod->get('MerchantID');
    my $AccessKeyID = $PaymentMethod->get('AccessKeyId');

    return isValid($MerchantID) && isValid($AccessKeyID);
}

#========================================================
sub usage {
    print <<END_USAGE;
Usage:
    perl $0 [options] flags objects

options:
    -passwd     database user password                     ()
    -storename  name of store                              ()
    -list       file with deprecated payments in the form:
                <payment_type_name>:<cartridge_name>       ()
    -payment    single deprecated payment in the form:
                <payment_type_name>:<cartridge_name>       ()

flags:
    -help       show the command line options
    -nooutput   do not print any output
    -dryrun     do not remove any payment method

Examples:
    perl $0 -storename Store -list /var/tmp/payment_cartridges.list
    perl $0 -storename Store -payment Acceptance:Ogone -dryrun

END_USAGE
    exit 2;
}

sub fileNotFound {
    my ($path) = @_;

    print "file \"$path\" not found\n";
    exit 2;
}

RunScript(
    'Sub' => \&Main,
);