#========================================================================================
# package      DE_EPAGES::ExternalShipping::API::Object::ShippingMethodWebservice
# base         DE_EPAGES::Shipping::API::Object::ShippingMethod
# state        public
#----------------------------------------------------------------------------------------
# description  Base class for webservice-based shipping price calculation
#----------------------------------------------------------------------------------------
# example      #TODO
#========================================================================================
package DE_EPAGES::ExternalShipping::API::Object::ShippingMethodWebservice;
use base DE_EPAGES::Shipping::API::Object::ShippingMethod;

use strict;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::Log qw ( LogDebug );
use DE_EPAGES::Core::API::TestParameter qw ( TestString TestArray TestInteger );

sub ErrorDefinition {
    return {
            'NotImplemented' => 'Webservice call not implemented.',
    };
}

#========================================================================================
# function     price
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingPrice = $Object->price( $hVars );
#----------------------------------------------------------------------------------------
# description  Returns the shipping price by the given requirements.
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#                   <li>ShippingElements
#                       <ul>
#                          <li>Element   - e.g. Products       - object
#                          <li>Quantity  - quantity of element - float
#                      </ul> | ref.array.hash
#               </ul> | ref.hash
# return       $ShippingPrice | price of shipping | money
#========================================================================================
sub price {
    my $self = shift;
    my ($hValues) = @_;

    #--- get shipping expenses
    my $Amount = $self->callShippingWebservice($hValues);
    return $self->_price($hValues->{'CurrencyID'}, $hValues->{'TaxModel'}, $Amount);
}

#========================================================================================
# function     callShippingService
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingMethod->callShippingService( $LineItemContainer );
#----------------------------------------------------------------------------------------
# description  Base method for webservice based shippingcost calculations
#----------------------------------------------------------------------------------------
# input        $LineItemContainer | LineItemContainer | object
# return       $ShippingCost | shipping cost | float
#========================================================================================
sub callShippingWebservice {
    my $self = shift;
    Error('NotImplemented');
}

#========================================================================================
# function     calculateShippingPrice
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $ShippingPrice = $ShippingMethod->calculateShippingPrice($Container);
#----------------------------------------------------------------------------------------
# description  Returns the shipping price by the given requirements.
#----------------------------------------------------------------------------------------
# input        $Container | LineItemContainer | ref.hash
# return       $ShippingPrice | price of shipping | money
#========================================================================================
sub calculateShippingPrice {
    my $self = shift;
    my ($Container) = @_;

    my $Amount = $self->callShippingWebservice($Container);

    my $hInfo = $Container->get([qw (CurrencyID TaxModel TaxMatrix)]);

    return $self->_price($hInfo->{'CurrencyID'}, $hInfo->{'TaxModel'}, $Amount, $hInfo->{'TaxMatrix'});
}

#========================================================================================
# function     _price
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $ShippingPrice = $ShippingMethod->_price($CurrencyID, $TaxModel, $Weight, $TaxMatrix);
#----------------------------------------------------------------------------------------
# description  Selects the corresponding shipping level and calculates the shipping
#               price for the given total weight.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency id | string(3)
# input        $TaxModel | tax model (0=net, 1=gross) | int
# input        $PriceFromWebservice | price returned by webservice | float
# input        $TaxMatrix | tax matrix | object
# return       $ShippingPrice | price of shipping | money
#========================================================================================
sub _price {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $PriceFromWebservice, $TaxMatrix) = @_;

    my $ShippingPrice;

    if ($self->get('UseWebserviceResult')) {
        $ShippingPrice = $PriceFromWebservice;
    } else {
        my $ahLevels = $self->getLevels($CurrencyID, $TaxModel, {'TaxMatrix' => $TaxMatrix});
        foreach my $hLevel (@{$ahLevels}) {
            if ($hLevel->{'LowerBound'} <= $PriceFromWebservice && $PriceFromWebservice < $hLevel->{'UpperBound'}) {
                $ShippingPrice = $hLevel->{'Multiplier'} * $PriceFromWebservice + $hLevel->{'BaseValue'};
                last;
            }
        }
    }

    Error('ShippingMethodNoPrice', { 'ShippingMethodID'=>$self->id }) unless defined $ShippingPrice;

    # shipping price should never be negative
    $ShippingPrice = 0 if $ShippingPrice < 0;
    return $ShippingPrice;
}

#========================================================================================
# function     getLevels
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahLevels = $Object->getLevels($CurrencyID, $TaxModel, $hParams);
# syntax       $ahLevels = $Object->getLevels($CurrencyID, $TaxModel, {'ExtraPrecision' => 1});
#----------------------------------------------------------------------------------------
# description  Return the shipping levels for the given currency in the requested tax
#               model.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency  id | char(3)
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $hParams | (optional) <ul>
#                   <li>ExtraPrecision  | (optional, default=0) not used | boolean
#                   <li>TaxMatrix      - tax matrix - object
#               </ul> | ref.hash
# return       $ahLevels | list of hashes, keys:
#                           BaseValue, LowerBound, UpperBound, Multiplier, ShippingLevelID
#                           | int
#========================================================================================
sub getLevels { ## no critic (RequireArgUnpacking)
    my $self = shift;
    my ($CurrencyID, $TaxModel, $hParams) = @_;

    my $Levels = $self->SUPER::getLevels(@_);

    LogDebug('getLevels = ',$Levels);

    my $TaxMatrix = $hParams->{'TaxMatrix'} // $self->getSite()->get('TaxMatrix');

    my $TaxClass = $self->get('TaxClass') // $TaxMatrix->defaultTaxClass;
    $TaxClass =  $TaxMatrix->getTaxClassByAlias($TaxClass->alias);

    foreach my $hLevel (@{$Levels}) {
        $hLevel->{'Multiplier'} = $TaxMatrix->net2gross($CurrencyID, $hLevel->{'Multiplier'}, {
            'TaxModel' => $TaxModel,
            'TaxClass' => $TaxClass,
            'Precision' => 4, # round to 4 digits like in 5.04 to avoid extra digits (0.999984 instead of 1.0)
        });
    }
    return $Levels;
}

#========================================================================================
# function     setLevels
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object->setLevels($CurrencyID, $TaxModel, $ahLevels);
#----------------------------------------------------------------------------------------
# description  Sets all shipping levels for the given currency.
#               Deletes shipping levels that are not included in the list.
#               Calculates net prices from gross prices if $TaxModel is TAX_MODEL_GROSS.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency  id | char(3)
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $ahLevels | list of hashes, keys:
#                           BaseValue, LowerBound, UpperBound, Multiplier
#                           | int
#========================================================================================
sub setLevels { ## no critic (RequireArgUnpacking)
    my $self = shift;
    my ($CurrencyID, $TaxModel, $ahLevels) = @_;
    TestString('CurrencyID', $CurrencyID);
    TestInteger('TaxModel', $TaxModel);
    TestArray('ahLevels', $ahLevels);

    # convert net prices to gross
    my $TaxMatrix = $self->getSite()->get('TaxMatrix');
    my $TaxClass = $self->get('TaxClass');

    foreach my $hLevel (@{$ahLevels}) {
        $hLevel->{'Multiplier'} = $TaxMatrix->gross2net($CurrencyID, $hLevel->{'Multiplier'}, {
            'TaxModel' => $TaxModel,
            'TaxClass' => $TaxClass,
            'Precision' => 9,
        });
    }

    $self->SUPER::setLevels(@_);
}

1;