#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::API::DialogBuilder
# state        public
#----------------------------------------------------------------------------------------
# description  Model module to get requests from UI and build dialogs by requesting services.
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::API::DialogBuilder;

use DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService;
use DE_EPAGES::GoogleSmartShopping::API::Tools qw ( LogGoogleSmartShopping );

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Presentation::API::Object::TrayBuilder;

use POSIX qw ( ceil );

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::GoogleSmartShopping::API::DialogBuilder->new( $hSettings );
#----------------------------------------------------------------------------------------
# description  Instantiates a new DialogBuilder object.
#----------------------------------------------------------------------------------------
# return       DialogBuilder | object | DE_EPAGES::GoogleSmartShopping::API::DialogBuilder
#========================================================================================
sub new {
    my $class = shift;
    my ( $hSettings ) = @_;

    my $self = bless {
        'Shop'       => $hSettings->{'Shop'},
        'FormValues' => $hSettings->{'FormValues'},
        'LanguageID' => $hSettings->{'LanguageID'}
    }, $class;

    return $self;
}

#========================================================================================
# function     fetchCampaignProductsPage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hProductsCampaignPage = $self->fetchCampaignProductsPage( );
#----------------------------------------------------------------------------------------
# description  Gets and returns campaign products page.
#----------------------------------------------------------------------------------------
# return       $hProductsCampaignPage | hash with campaign products | hash.ref
#========================================================================================
sub fetchCampaignProductsPage {
    my $self = shift;

    # Calculate remote page number and results per page to requested
    my $CurrentPage = $self->{'FormValues'}->{'currentPage'};
    my $EntriesPerPage = $self->{'FormValues'}->{'pageEntries'};
    my $NumberOfPagesInBuffer = $self->{'FormValues'}->{'pages'};

    my $ResultsPerPageRequest = $EntriesPerPage * $NumberOfPagesInBuffer;
    my $PageRequest = ceil( $CurrentPage / $NumberOfPagesInBuffer );

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({ Shop => $self->{'Shop'} });
    my $hProductsCampaignPage;
    eval {
        $hProductsCampaignPage = $service->getCampaignShoppingPerformance(
            $self->{'FormValues'}->{'campaign'}, $PageRequest, $ResultsPerPageRequest
        );
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Error getting campaign products performance metrics: ', $Error->message());
        return;
    }

    return $hProductsCampaignPage;
}

#========================================================================================
# function     buildCampaignProductsDialog
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hDialogData = $self->buildCampaignProductsDialog( $hProductsCampaignPage );
#----------------------------------------------------------------------------------------
# description  Builds and returns dialog data.
#----------------------------------------------------------------------------------------
# input        $hProductsCampaignPage | hash with products campaign page | hash.ref
# return       $hDialogData | hash with data to build dialog | hash.ref
#========================================================================================
sub buildCampaignProductsDialog {
    my $self = shift;
    my ($hProductsCampaignPage) = @_;

    my $NumberOfPagesInBuffer = $self->{'FormValues'}->{'pages'};
    my $EntriesPerPage = $self->{'FormValues'}->{'pageEntries'};
    my $LastPage = ceil( $hProductsCampaignPage->{'results'} / $EntriesPerPage );

    my $hJSONResults = {};
    # if there is a fetched result.
    if ( $hProductsCampaignPage->{'results'} > 0 ) {

        my $CurrentPage = $self->{'FormValues'}->{'currentPage'} // 1;

        if ( $CurrentPage > $LastPage or $CurrentPage < 0 ) {
            $CurrentPage = 1;
        }
        my $PagerSearchPage = ceil( $CurrentPage / $NumberOfPagesInBuffer );
        my $aFoundObjects = $hProductsCampaignPage->{'items'};

        my $Counter = 1;
        my $BufferPageIndex   = ( $PagerSearchPage - 1 ) * $NumberOfPagesInBuffer + 1;

        # build up table content.
        foreach my $hFoundObject (@$aFoundObjects) {
            my $hRowData = {
                'Name' => $hFoundObject->{'productName'},
                'Clicks' => $hFoundObject->{'clicks'},
                'Cost' => $hFoundObject->{'cost'},
                'Conversions' => $hFoundObject->{'conversions'},
                'ConversionValue' => $hFoundObject->{'conversionValue'},
            };
            push @{ $hJSONResults->{$BufferPageIndex} }, $hRowData;

            if ( $Counter % $EntriesPerPage == 0 ) {
                $BufferPageIndex++;
            }
            $Counter++;
        }
    }

    my $Builder = DE_EPAGES::Presentation::API::Object::TrayBuilder->new({LanguageID => $self->{'LanguageID'}, TranslationPackage => 'DE_EPAGES::Product' });
    # $Id, [$IsSelectable, $IsSortable, $IsLinkable, $IsRightAligned], $TranslationKey, $TranslationPackage
    $Builder->addColumnHeader('Name', [0, 0, 0, 0], 'Product', 'DE_EPAGES::Dictionary')
        ->addColumnHeader('Clicks', [0, 0, 0, 1], 'PerformanceDialogColumnClicks', 'DE_EPAGES::Dictionary')
        ->addColumnHeader('Cost', [0, 0, 0, 1], 'PerformanceDialogColumnCost', 'DE_EPAGES::Dictionary')
        ->addColumnHeader('Conversions', [0, 0, 0, 1], 'PerformanceDialogColumnNumberOfConversions', 'DE_EPAGES::Dictionary')
        ->addColumnHeader('ConversionValue', [0, 0, 0, 1], 'PerformanceDialogColumnValueOfConversions','DE_EPAGES::Dictionary');
    return {
        'totalNumber'  => $hProductsCampaignPage->{'results'},
        'lastPage'     => $LastPage,
        'tableContent' => $hJSONResults,
        'tableHead'    => $Builder->getHeaderList(),
    };
}

1;

