#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder
# state        public
#----------------------------------------------------------------------------------------
# description  Build return URL for Backlink
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Url qw ( ParseQueryString );
use DE_EPAGES::Object::API::Factory qw(LoadRootObject);

use DE_EPAGES::GoogleSmartShopping::API::Constants qw (
    GOOGLE_SMART_SHOPPING_APP_SCOPES
);
use URI;

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       new( );
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::GoogleSmartShopping::API::GoogleSmartShoppingUrlBuilder
#========================================================================================
sub new {
    my $class = shift;
    my $self  = bless {}, $class;
    return $self;
}

#========================================================================================
# function     buildConnectOAuthUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       buildConnectOAuthUrl( $StateURL );
#----------------------------------------------------------------------------------------
# description  Build the connect URL used for on-boarding
#----------------------------------------------------------------------------------------
# input        $StateURL | Current URL in the MBO  | string
# return       $URL->as_string() | Mollie connect url | string
#========================================================================================
sub buildConnectOAuthUrl {
    my $self = shift;
    my ($StateURL) = @_;

    my $System = LoadRootObject();
    my $hQueryParameters = $self->_buildQueryParameters($System, $StateURL);

    my $Url = $self->_buildEncodedUrl($System, $hQueryParameters);

    return $Url;
}

#----------------------------------------------------------------------------------------
sub _buildQueryParameters {
    my $self = shift;
    my ($System, $StateURL) = @_;
    # AD-10371 introduced 'prompt' => 'consent' to make sure we receive always a refresh_token on the OAuth flow
    # We might reconsider this thinking from the Merchant's perspective, if it makes sense
    # to force the Merchant to always confirm the scopes
    return {
        'client_id'       => $System->get('GoogleSmartShopping_AppClientID'),
        'redirect_uri'    => $System->get('GoogleSmartShopping_AppRedirectURL'),
        'response_type'   => 'code',
        'scope'           => join(' ', GOOGLE_SMART_SHOPPING_APP_SCOPES),
        'access_type'     => 'offline',
        'prompt'          => 'consent',
        'state'           => $StateURL
    };
}

#----------------------------------------------------------------------------------------
sub _buildEncodedUrl {
    my $self = shift;
    my ($System, $hQueryParameters) = @_;

    my $AppURL =  $System->get('GoogleSmartShopping_GoogleAppURL');
    my $Url = URI::URL->new($AppURL);
    $Url->query_form($hQueryParameters);
    my $hBacklinkUrlParameters = ParseQueryString($Url->equery(), 'utf-8');
    $Url->query_form( $hBacklinkUrlParameters );

    return $Url;
}

#========================================================================================
# function     buildRedirectToMboUrl
# state        public
#----------------------------------------------------------------------------------------
# syntax       buildRedirectToMboUrl( $hQueryParameters );
#----------------------------------------------------------------------------------------
# description  Build redirect to MBO URL using returned data from GoogleSmartShopping
#----------------------------------------------------------------------------------------
# input        $hQueryParameters | Hashref which is returned from Google | hashref
# return       $URL->as_string() | BacklinkUrl with AuthorizationCode or Error | string
#========================================================================================
sub buildRedirectToMboUrl {
    my $self = shift;
    my ($hQueryParameters) = @_;
    my $aBacklinkUrl = $hQueryParameters->{'state'};

    if ( (!isValid($aBacklinkUrl)) or (!isValid($aBacklinkUrl->[0])) ) {
        Error('NoURLFound');
    }

    return $self->_urlBuilder($hQueryParameters);
}

#----------------------------------------------------------------------------------------
sub _urlBuilder {
    my $self = shift;
    my ($hQueryParameters) = @_;

    # Original Change action must be removed
    delete($hQueryParameters->{'ChangeAction'});

    # State contains the URL which is backlink to MBO. No need to send it with QueryString again
    my $BackLinkUrl = delete($hQueryParameters->{'state'});

    my $URL = URI->new($BackLinkUrl->[0]);
    my $hBacklinkUrlParameters = ParseQueryString($URL->query(), 'utf-8');
    my $hQueryParam = {%{$hBacklinkUrlParameters}, %{$hQueryParameters}};
    $URL->query_form($hQueryParam);

    return $URL->as_string();
}

1;