#==============================================================================
# package      DE_EPAGES::GoogleSmartShopping::UI::Campaigns
# state        private
#==============================================================================
package DE_EPAGES::GoogleSmartShopping::UI::Campaigns;
use base DE_EPAGES::Presentation::UI::Object;

use DE_EPAGES::Core::API::PerlTools qw ( isValid GetCurrentDateTime );
use DE_EPAGES::Core::API::Object::DateTimeFormatter;
use DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService;
use DE_EPAGES::GoogleSmartShopping::API::ScriptTagService;
use DE_EPAGES::GoogleSmartShopping::API::Tools qw ( LogGoogleSmartShopping GetErrorCode );
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError Error );
use DE_EPAGES::GoogleSmartShopping::API::Constants qw (
    ERROR_FORM_SUBMIT
    ERROR_ACCEPT_TERMS
    ERROR_VALIDATION
);
use HTTP::Status qw ( HTTP_NOT_FOUND HTTP_OK );

use DateTime;

use strict;

#==============================================================================
# Event handlers
#==============================================================================
sub UnityViewGoogleSmartShoppingCampaigns {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();

    $Servlet->vars('CampaignCreationAllowed', 0);

    if ($self->_isCampaignCreationAllowed($Shop)) {
        my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
        my $aGoogleAdsCampaigns = $self->GetGoogleAdsCampaigns($Shop, $GoogleSmartShoppingConfig);
        unless (defined($aGoogleAdsCampaigns)) {
            $Servlet->vars('GoogleAdsCampaignsListError', 1);
            return $self->SUPER::View($Servlet);
        }
        $Servlet->vars('GoogleAdsCampaigns', $aGoogleAdsCampaigns);
        if(scalar(@$aGoogleAdsCampaigns) == 0) {
            $Servlet->vars('CampaignCreationAllowed', 1);
        }
    }

    return $self->SUPER::View($Servlet);
}

#========================================================================================
# function     CreateGoogleAdsCampaign
# state        public
#----------------------------------------------------------------------------------------
# syntax       CreateGoogleAdsCampaign( $Servlet );
#----------------------------------------------------------------------------------------
# description  Create a new Google Ads Campaign with the provided budget and an
#               associated conversion action.
#----------------------------------------------------------------------------------------
# input        $Servlet | Presentation Object | object
#========================================================================================
sub CreateGoogleAdsCampaign {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    my $Form = $Servlet->form();
    my $hFormValues = $Form->formValues();

    unless ($self->_isCampaignCreationAllowed($Shop)) {
        $Form->executeFormError({
            'Name'   => 'CreationFailed',
            'Reason' => 'Campaign creation not possible for this Shop'
        });
    }

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    my $FormErrorReason = $self->_validateCreateGoogleAdsCampaignRequest($hFormValues, $GoogleSmartShoppingConfig);
    if (defined $FormErrorReason) {
        $Form->executeFormError({
            'Name'   => 'CreationFailed',
            'Reason' => $FormErrorReason
        });
    }

    my $GoogleAdsApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({Shop => $Shop});
    my $ScriptTagService = DE_EPAGES::GoogleSmartShopping::API::ScriptTagService->new({Shop => $Shop});

    my $hAdsCampaignResponse = $self->_createCampaign($GoogleAdsApiService, $Shop, $GoogleSmartShoppingConfig, $Form);

    if (defined($hAdsCampaignResponse) && defined($hAdsCampaignResponse->{'id'})) {
        my $AdsCampaignId = $hAdsCampaignResponse->{'id'};
        $self->_createConversionAction($GoogleAdsApiService, $Form, $AdsCampaignId);
        $self->_createRemarketingAction($GoogleAdsApiService, $Form, $AdsCampaignId);
        $ScriptTagService->addScriptTag();
    }

    return;
}

#========================================================================================
# function     GetGoogleAdsCampaigns
# state        public
#----------------------------------------------------------------------------------------
# syntax       GetGoogleAdsCampaigns( $Shop, $GoogleSmartShoppingConfig );
#----------------------------------------------------------------------------------------
# description  Get a list of Smart Shopping Ads campaigns
#----------------------------------------------------------------------------------------
# input        $Shop | Shop Object | object
# input        $GoogleSmartShoppingConfig | GoogleSmartShoppingConfig Object | object
#----------------------------------------------------------------------------------------
# input        $Shop | Shop Object | object
# return       $aAdsCampaigns | list of Smart Shopping Ads campaigns | ref.array
#========================================================================================
sub GetGoogleAdsCampaigns {
    my $self = shift;
    my ($Shop, $GoogleSmartShoppingConfig) = @_;

    my $service = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({Shop => $Shop});
    my $AdsSubAccountId = $GoogleSmartShoppingConfig->get('GoogleAdsAccountID');
    my $aAdsCampaigns;
    my $hParams = {
        'AdsSubAccountId' => $AdsSubAccountId,
    };
    eval {
        $aAdsCampaigns = $service->getSmartShoppingCampaigns($hParams);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Error getting AdsCampaigns: ', $Error->message());
    }

    my $Formatter = DE_EPAGES::Core::API::Object::DateTimeFormatter->new();
    my $CurrentDateTime = GetCurrentDateTime();
    foreach my $Campaign (@$aAdsCampaigns) {
        $Campaign->{'startDate'} = $Formatter->parse_datetime( $Campaign->{'startDate'}, 'date', '%Y-%m-%d' );
        if (DateTime->compare($Campaign->{'startDate'}, $CurrentDateTime) == 1) {
            $Campaign->{'isFuture'} = 1;
        }
    }

    return $aAdsCampaigns;
}

#========================================================================================
# function     UpdateGoogleAdsCampaign
# state        public
#----------------------------------------------------------------------------------------
# syntax       UpdateGoogleAdsCampaign( $Servlet );
#----------------------------------------------------------------------------------------
# description  Update a Google Ads Campaign with the provided budget.
#----------------------------------------------------------------------------------------
# input        $Servlet | Presentation Object | object
#========================================================================================
sub UpdateGoogleAdsCampaign {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    Error('Undefined configuration for GoogleSmartShopping') unless defined $GoogleSmartShoppingConfig;

    my $Form = $Servlet->form();
    my $hFormValues = $Form->formValues();
    LogGoogleSmartShopping('Google Ads Campaign update form values', { 'FormValues' => $hFormValues });

    my $FormErrorReason = $self->_validateUpdateGoogleAdsCampaignRequest($hFormValues, $GoogleSmartShoppingConfig);
    if (defined $FormErrorReason) {
        LogGoogleSmartShopping("Update Ads campaign validation error: ", $FormErrorReason);
        return;
    }

    my $DailyBudget = $self->_formatMoneyField($Shop, $Form, $hFormValues->{'DailyBudget'}->[0]);
    my $hCampaignBudgetUpdate = {
        'AdsCampaignId'       => $hFormValues->{'AdsCampaignId'}->[0],
        'DailyBudget'         => $DailyBudget,
        'DailyBudgetCurrency' => $hFormValues->{'DailyBudgetCurrency'}->[0]
    };

    my $GoogleAdsApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({Shop => $Shop});
    eval {
        $GoogleAdsApiService->updateGoogleAdsCampaignBudget($hCampaignBudgetUpdate);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Ads Campaign budget update failed', { 'Error' => $Error->message() });
        return;
    }

    return;
}

#========================================================================================
# function     SwitchCampaignStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       SwitchCampaignStatus( $Servlet );
#----------------------------------------------------------------------------------------
# description  Toggle campaign status between "ENABLED" and "PAUSED".
#----------------------------------------------------------------------------------------
# input        $Servlet | Presentation Object | object
#========================================================================================
sub SwitchCampaignStatus {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();
    return unless ($self->_existsGoogleSmartShoppingConfig($Shop));

    my $Form = $Servlet->form();
    my $hPostData = $Form->formValues();

    my $CampaignId = $hPostData->{'CampaignId'}->[0];
    my $CampaignStatusEnabledString = $hPostData->{'CampaignStatusEnabled'}->[0];
    unless (isValid($CampaignStatusEnabledString)
            && ($CampaignStatusEnabledString ~~ [qw(true false)])
            && isValid($CampaignId)) {
        $Servlet->response->code(HTTP_NOT_FOUND);
        return;
    }

    my $CampaignStatusEnabledValue;
    if ($CampaignStatusEnabledString eq 'true') {
        $CampaignStatusEnabledValue = 1;
    } else {
        $CampaignStatusEnabledValue = 0;
    }
    my $hCampaignStatusUpdate = {
        'CampaignId'            => $CampaignId,
        'CampaignStatusEnabled' => $CampaignStatusEnabledValue
    };
    my $GoogleAdsApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({ Shop => $Shop });
    eval {
        $GoogleAdsApiService->updateGoogleAdsCampaign($hCampaignStatusUpdate);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Ads Campaign update failed', {
            'Error' => $Error->message(),
        });
        $Servlet->writeContentOutput({
            'Error' => $Error->message(),
            'CampaignId' => $CampaignId
        });
        return;
    }

    LogGoogleSmartShopping("SwitchCampaignStatus response: ", $hCampaignStatusUpdate);

    $Servlet->params('ContentType', 'application/json');
    $Servlet->response->code(HTTP_OK);
    $Servlet->writeContentOutput($hCampaignStatusUpdate);

    return;
}

#========================================================================================
# function     DeleteAdsCampaign
# state        public
#----------------------------------------------------------------------------------------
# syntax       DeleteAdsCampaign( $Servlet );
#----------------------------------------------------------------------------------------
# description  Delete a Google Ads Campaign.
#----------------------------------------------------------------------------------------
# input        $Servlet | Presentation Object | object
#========================================================================================
sub DeleteAdsCampaign {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();

    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');
    Error('Undefined configuration for GoogleSmartShopping') unless defined $GoogleSmartShoppingConfig;

    my $Form = $Servlet->form();
    my $hFormValues = $Form->formValues();
    LogGoogleSmartShopping('Google Ads Campaign delete form values', { 'FormValues' => $hFormValues });

    my $FormErrorReason = $self->_validateDeleteAdsCampaignRequest($hFormValues, $GoogleSmartShoppingConfig);
    if (defined $FormErrorReason) {
        LogGoogleSmartShopping("Delete Ads campaign validation error: ", $FormErrorReason);
        $Form->executeFormError({
            'Name'   => 'DeletionFailed',
            'Reason' => 'Validation error'
        });
    }

    my $GoogleAdsApiService = DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService->new({Shop => $Shop});
    eval {
        $GoogleAdsApiService->deleteAdsCampaign($hFormValues->{'AdsCampaignId'}->[0]);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Delete Ads Campaign failed', { 'Error' => $Error->message() });
        $Form->executeFormError({
            'Name'   => 'DeletionFailed',
            'Reason' => GetErrorCode($Error)
        });
    }

    $GoogleSmartShoppingConfig->deleteAdsCampaign();
    LogGoogleSmartShopping("GoogleSmartShoppingConfig AdsCampaign deleted", {});

    return;
}

#----------------------------------------------------------------------------------------
sub _createCampaign {
    my $self = shift;
    my ($GoogleAdsApiService, $Shop, $GoogleSmartShoppingConfig, $Form) = @_;

    my $DefaultLanguageID = $Shop->get('LanguageID');
    my $AdsCampaignName = $Shop->get('NameOrAlias', $DefaultLanguageID);
    my $AdsSubAccountId = $GoogleSmartShoppingConfig->get('GoogleAdsAccountID');
    my $MerchantCenterAccountId = $GoogleSmartShoppingConfig->get('MerchantCenterAccountID');
    my $FormValues = $Form->values();

    my $DailyBudget = $self->_formatMoneyField($Shop, $Form, $FormValues->{'DailyBudget'}->[0]);

    my $hParams = {
        'AdsSubAccountId'         => $AdsSubAccountId,
        'MerchantCenterAccountId' => $MerchantCenterAccountId,
        'DailyBudget'             => $DailyBudget,
        'AdsCampaignName'         => $AdsCampaignName,
    };

    my $hAdsCampaignResponse;
    eval {
        $hAdsCampaignResponse = $GoogleAdsApiService->createGoogleAdsCampaign( $hParams );
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Ads Campaign creation failed', {
            'Error' => $Error->message(),
        });

        $Form->executeFormError({
            'Name'   => 'CreationFailed',
            'Reason' => GetErrorCode($Error)
        });
    }

    return $hAdsCampaignResponse;
}

#----------------------------------------------------------------------------------------
sub _createConversionAction {
    my $self = shift;
    my ($GoogleAdsApiService, $Form, $AdsCampaignId) = @_;

    eval {
        $GoogleAdsApiService->createGoogleConversionAction( $AdsCampaignId );
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Ads Campaign conversion action creation failed', {
            'Error' => $Error->message(),
        });
        $Form->executeFormError({
            'Name'   => 'ConversionActionCreationFailed',
            'Reason' => GetErrorCode($Error)
        });
    }

    return;
}
#----------------------------------------------------------------------------------------
sub _createRemarketingAction {
    my $self = shift;
    my ($GoogleAdsApiService, $Form, $AdsCampaignId) = @_;

    eval {
        $GoogleAdsApiService->createGoogleRemarketingAction($AdsCampaignId);
    };
    if (ExistsError()) {
        my $Error = GetError();
        LogGoogleSmartShopping('Ads Campaign remarketing action creation failed', {
            'Error' => $Error->message(),
        });
        $Form->executeFormError({
            'Name'   => 'RemarketingActionCreationFailed',
            'Reason' => GetErrorCode($Error)
        });
    }

    return;
}

#----------------------------------------------------------------------------------------
sub _isCampaignCreationAllowed {
    my $self = shift;
    my ($Shop) = @_;

    return unless ($self->_existsGoogleSmartShoppingConfig($Shop));
    my $GoogleSmartShoppingConfig = $Shop->get('GoogleSmartShoppingConfig');

    # most important is that we have both Merchant Center and Ads accounts connected
    return unless (isValid($GoogleSmartShoppingConfig->get('MerchantCenterAccountID')));
    return unless (isValid($GoogleSmartShoppingConfig->get('GoogleAdsAccountID')));

    return 1;
}

#----------------------------------------------------------------------------------------
sub _validateDeleteAdsCampaignRequest {
    my $self = shift;
    my ($hFormValues, $GoogleSmartShoppingConfig) = @_;

    my $ValidationError = $self->_validateCampaignRequest($hFormValues, $GoogleSmartShoppingConfig);
    return $ValidationError if (defined($ValidationError));

    return ERROR_FORM_SUBMIT unless (
        isValid($hFormValues->{'AdsCampaignId'}) && isValid($hFormValues->{'AdsCampaignId'}->[0])
        && ($hFormValues->{'AdsCampaignId'}->[0] eq $GoogleSmartShoppingConfig->get('AdsCampaignId'))
    );

    return;
}

#----------------------------------------------------------------------------------------
sub _validateCreateGoogleAdsCampaignRequest {
    my $self = shift;
    my ($hFormValues, $GoogleSmartShoppingConfig) = @_;

    my $ValidationError = $self->_validateCampaignRequest($hFormValues, $GoogleSmartShoppingConfig);
    return $ValidationError if (defined($ValidationError));

    return ERROR_FORM_SUBMIT unless (
        isValid($hFormValues->{'DailyBudget'}) && isValid($hFormValues->{'DailyBudget'}->[0])
    );
    return;
}

#----------------------------------------------------------------------------------------
sub _validateUpdateGoogleAdsCampaignRequest {
    my $self = shift;
    my ($hFormValues, $GoogleSmartShoppingConfig) = @_;

    my $ValidationError = $self->_validateCreateGoogleAdsCampaignRequest($hFormValues, $GoogleSmartShoppingConfig);
    return $ValidationError if (defined($ValidationError));

    return ERROR_FORM_SUBMIT unless (
        isValid($hFormValues->{'AdsCampaignId'}) && isValid($hFormValues->{'AdsCampaignId'}->[0])
    );
    return ERROR_FORM_SUBMIT unless (
        isValid($hFormValues->{'DailyBudgetCurrency'}) && isValid($hFormValues->{'DailyBudgetCurrency'}->[0])
    );

    return;
}

#----------------------------------------------------------------------------------------
sub _validateCampaignRequest {
    my $self = shift;
    my ($hFormValues, $GoogleSmartShoppingConfig) = @_;

    return ERROR_ACCEPT_TERMS unless ($GoogleSmartShoppingConfig->areGoogleMerchantCenterRequirementsMet());
    return ERROR_VALIDATION unless (isValid($GoogleSmartShoppingConfig->get('GoogleAdsAccountID')));
    return ERROR_VALIDATION unless (isValid($GoogleSmartShoppingConfig->get('MerchantCenterAccountID')));

    return;
}

#----------------------------------------------------------------------------------------
sub _existsGoogleSmartShoppingConfig {
    my $self = shift;
    my ($Shop) = @_;

    my $CallingFunction = ( caller(0) )[3];
    unless ($Shop->existsChild('GoogleSmartShoppingConfig')) {
        LogGoogleSmartShopping("GoogleSmartShoppingConfig shop object not found in ", $CallingFunction);
        return 0;
    }
    return 1;
}

sub _formatMoneyField {
    my $self = shift;
    my ($Shop, $Form, $FieldValue) = @_;

    my $hFieldData = {
        'Name'       => 'Amount',
        'Type'       => 'reg_money',
        'Mandatory'  => 1,
        'MinValue'   => 0,
        'MaxValue'   => undef,
        'CurrencyID' => $Shop->get('CurrencyID')
    };

    return $Form->field($Shop, $FieldValue, $hFieldData);
}

1;
