#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::t::unit::DialogBuilder
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for the GoogleSmartShopping Dialog Builder operations
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::t::unit::DialogBuilder;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;

use Test::Mockify;
use Test::Mockify::Matcher qw ( ArrayRef HashRef Number String );
use Test::Mockify::Sut;
use Test::Mockify::Verify qw ( WasCalled );

use Test::Exception;
use Test::More;

use strict;

use constant CAMPAIGN_ID => 'campaign-id-number';

use constant DIALOG_RESULTS_PER_PAGE => 10;
use constant DIALOG_NUMBER_OF_BUFFERED_PAGES => 5;
use constant FORM_VALUES => {
    'currentPage' => 11, # 11th dialog page (10 items per page) corresponds with 3rd EpagesJ page (50 items per page)
    'pageEntries' => DIALOG_RESULTS_PER_PAGE,
    'pages'       => DIALOG_NUMBER_OF_BUFFERED_PAGES,
    'campaign'    => CAMPAIGN_ID
};

use constant TOTAL_CAMPAIGN_PRODUCTS => 112;
use constant DIALOG_LAST_PAGE => 12; # 11 pages of 10 items and 12th page of 2 items.

use constant CAMPAIGN_PRODUCTS_PAGE => {
    'results' => TOTAL_CAMPAIGN_PRODUCTS, # total number of available items (not the amount returned)
    'page' => 3, # 11th dialog page (10 items per page) corresponds with 3rd EpagesJ page (50 items per page)
    'resultsPerPage' => 50, # DIALOG_RESULTS_PER_PAGE * DIALOG_NUMBER_OF_BUFFERED_PAGES
    'items' => [ # Real number of items should be 12; provided two for brevity.
        { 'productName' => '01-Lifestyle Headphones', 'clicks' => '54', 'cost' => '14.00', 'conversions' => '2', 'conversionValue' => '35.00' },
        { 'productName' => '02-Navy Boots', 'clicks' => '34', 'cost' => '8.00', 'conversions' => '6', 'conversionValue' => '48.00' },
    ]
};

use constant LANGUAGE_ID => 1;

sub test {
    my $self = shift;

    $self->test_fetchCampaignProductsPage();
    $self->test_buildCampaignProductsDialog();

    done_testing();
}

sub test_fetchCampaignProductsPage {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $hServiceResponses = {
        getCampaignShoppingPerformance => CAMPAIGN_PRODUCTS_PAGE
    };
    my $GoogleAdsApiService = $self->_createGoogleAdsApiService($hServiceResponses);

    my $Shop = $self->_createShop();

    my $SUT = $self->_createSUT(
        { 'Shop' => $Shop, 'FormValues' => FORM_VALUES, 'LanguageID' => LANGUAGE_ID },
        $GoogleAdsApiService
    );

    lives_ok( sub { $SUT->fetchCampaignProductsPage(); }, $SubTestName . ' - Campaign fetching call is carried on.' );

    ok(
        WasCalled($GoogleAdsApiService, 'getCampaignShoppingPerformance'),
        $SubTestName . ' - Google Ads API service to get campaign shopping performance metrics is called.'
    );

}

sub test_buildCampaignProductsDialog {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $DialogTrayBuilder = $self->_createDialogTrayBuilder();

    my $Shop = $self->_createShop();

    my $SUT = $self->_createSUT(
        { 'Shop' => $Shop, 'FormValues' => FORM_VALUES, 'LanguageID' => LANGUAGE_ID },
        undef,
        $DialogTrayBuilder
    );

    my $DialogData = $SUT->buildCampaignProductsDialog(CAMPAIGN_PRODUCTS_PAGE);

    ok(
        WasCalled($DialogTrayBuilder, 'addColumnHeader'),
        $SubTestName . ' - Dialog tray builder to generate column headers is called.'
    );
    is( $DialogData->{'lastPage'}, DIALOG_LAST_PAGE, $SubTestName . ' - Dialog last page has been properly recalculated.');
    is(
        $DialogData->{'totalNumber'},
        TOTAL_CAMPAIGN_PRODUCTS,
        $SubTestName . ' - Dialog total products number has been properly passed.'
    );

}

#----------------------------------------------------------------------------------------
# Mockery
#----------------------------------------------------------------------------------------
sub _createSUT {
    my $self = shift;
    my ($hSettings, $GoogleAdsApiService, $DialogTrayBuilder) = @_;

    my $MockifySut = Test::Mockify::Sut->new('DE_EPAGES::GoogleSmartShopping::API::DialogBuilder', [ $hSettings ]);

    if (defined($GoogleAdsApiService)) {
        $MockifySut->mockStatic('DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService::new')
            ->when(String(), HashRef())
            ->thenReturn($GoogleAdsApiService);
    }
    if (defined($DialogTrayBuilder)) {
        $MockifySut->mockStatic('DE_EPAGES::Presentation::API::Object::TrayBuilder::new')
            ->whenAny()
            ->thenReturn($DialogTrayBuilder);
    }

    $MockifySut->mockImported('DE_EPAGES::GoogleSmartShopping::API::Tools', 'LogGoogleSmartShopping')
        ->whenAny()
        ->thenReturnUndef();

    $MockifySut->mockImported('DE_EPAGES::Core::API::Error', 'ExistsError')
        ->whenAny()
        ->thenReturn(0);

    return $MockifySut->getMockObject();
}

sub _createShop {
    my $self = shift;

    my $hAttributes = {};
    my $MockifyShop = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', $hAttributes);

    return $MockifyShop->getMockObject();
}

sub _createGoogleAdsApiService {
    my $self = shift;
    my ($hServiceResponse) = @_;

    my $Mockify = Test::Mockify->new('DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService');

    $Mockify->mock('getCampaignShoppingPerformance')
        ->when(String(CAMPAIGN_ID), Number(1), Number(50))
        ->thenReturn($hServiceResponse);

    return $Mockify->getMockObject();
}

sub _createDialogTrayBuilder {
    my $self = shift;

    my $Mockify = Test::Mockify->new('DE_EPAGES::Presentation::API::Object::TrayBuilder');

    $Mockify->mock('addColumnHeader')
        ->when(String(), ArrayRef(),String(), String())
        ->thenReturn($Mockify->getMockObject());

    return $Mockify->getMockObject();
}

__PACKAGE__->runScript();


