#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::t::unit::GoogleApiClient
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for DE_EPAGES::GoogleSmartShopping::API::GoogleApiClient
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::t::unit::GoogleApiClient;
use base DE_EPAGES::Core::API::Test;

use DE_EPAGES::Test::Mock::Mockify qw ( WasCalled );
use Test::Mockify::Matcher qw ( Object HashRef String );
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::MockifyServlet;
use DE_EPAGES::Test::Mock::EpagesObject;

use JSON::MaybeXS qw ( encode_json );
use Test::Exception;
use Test::Mockify::Sut;
use Test::MockObject::Extends;
use Test::More;
use HTTP::Request;
use HTTP::Status qw ( HTTP_OK HTTP_NO_CONTENT HTTP_UNAUTHORIZED HTTP_INTERNAL_SERVER_ERROR );

use DE_EPAGES::GoogleSmartShopping::API::Constants qw ( EPAGESJ_BASE_URI );

use strict;

use constant ACCESS_TOKEN => 'FAKE_ACCESS_TOKEN';
use constant EXPIRED_ACCESS_TOKEN => 'FAKE_EXPIRED_ACCESS_TOKEN';
use constant BAD_ACCESS_TOKEN => 'BAD_ACCESS_TOKEN';
use constant NEW_ACCESS_TOKEN => 'FAKE_NEW_ACCESS_TOKEN';
use constant REFRESH_TOKEN => 'FAKE_REFRESH_TOKEN';
use constant HEADER_WITH_EXPIRED_TOKEN => { 'AccessToken' => EXPIRED_ACCESS_TOKEN };
use constant AUTH_TOKEN_URL => 'https://oauth2.googleapis.com/token';
use constant GOOGLE_OPENID_USERINFO_URL => 'https://openidconnect.googleapis.com/v1/userinfo';
use constant HTTP_HEADERS => {
    'Accept'  => 'application/json'
};
use constant POST_PAYLOAD_CLAIM => {
    'merchantAccountId' => 4253424325,
    'multiClientAccountId' => 1234567
};
use constant EXTERNAL_POST_PAYLOAD => {
    'merchantId' => 4253424325,
    'aggregatorId' => 1234567
};
use constant EXPECTED_ACCESS_TOKEN_RESPONSE => {
    'access_token' => 'FAKE_ACCESS_TOKEN'
};
use constant EXPECTED_ACCESS_TOKEN_ERROR_RESPONSE => {
    "errorId" => "google-api-error",
    "message" => "Invalid Credentials",
    "details" => {
        "code" => 401,
        "errors" => [
            {
                "domain" => "global",
                "location" => "Authorization",
                "locationType" => "header",
                "message" => "Invalid Credentials",
                "reason" => "authError"
            }
        ],
        "message" => "Request had invalid authentication credentials. Expected OAuth 2 access token, login cookie or other valid authentication credential. See https://developers.google.com/identity/sign-in/web/devconsole-project.",
        "status" => "UNAUTHENTICATED"
    }
};
use constant EXPECTED_EPAGESJ_ERROR_RESPONSE => {
  "errorId" => "unknown-error",
  "message" => "An unknown error occurred. The user should be requested to retry and contact the support if this doesn\'t help.",
  "details" => {
    "exceptionType" => "java.lang.NullPointerException",
    "exceptionMessage" => "Just a test",
    "className" => "com.epages.google.smart.shopping.accounts.AccountsResource",
    "methodName" => "createSubAccount",
    "lineNumber" => 60
  }
};
use constant EXPECTED_EXTERNAL_ACCESS_TOKEN_ERROR_RESPONSE => {
    'error' => 'invalid_request',
    'error_description' => 'Invalid Credentials'
};
use constant EXPECTED_EXTERNAL_OTHER_ERROR_RESPONSE => {
    'error' => 'unknown_error',
    'error_description' => 'Unknown Error'
};
use constant USER_ACCOUNT_RESPONSE => {
    'accountIdentifiers' => [
        {
            merchantId => 444950401 
        }
    ],
    'kind' => 'content#accountsAuthInfoResponse',
};
use constant EXPECTED_CLAIM_WEBSITE_SUCCESS_RESPONSE => {
    'kind' => 'content#accountsClaimWebsiteResponse'
};
use constant EXPECTED_USERINFO_RESPONSE => {
    'sub'            => '109729977304043127834',
    'picture'        => 'https://lh3.googleusercontent.com/a/default-user=s96-c',
    'email'          => 'jdoyle@epages.com',
    'email_verified' => 'true',
    'hd'             => 'epages.com'
};

use constant UPDATE_SHOPURL_REQUEST => {
    'websiteUrl' => "https://kenneth-dow.com"
};

use constant UPDATE_SHOPURL_RESPONSE => {


    "id" => "462008688",

    "name" => "MySubAccount1",

    "websiteUrl"  => "https://example4.com",

    "adultContent" => "false",

    "emails" => 'epagestesttwo@gmail.com'


};


sub test {
    my $self = shift;

    $self->test_doget_epagesj_request_success();
    $self->test_doget_epagesj_request_expired_token();
    $self->test_doget_epagesj_request_failure();
    $self->test_dopost_epagesj_request_success();
    $self->test_dopost_epagesj_request_expired_token();
    $self->test_dopost_epagesj_request_failure();
    $self->test_dodelete_epagesj_request_success();
    $self->test_dodelete_epagesj_request_expired_token();
    $self->test_dodelete_epagesj_request_failure();
    $self->test_doput_epagesj_request_success();
    $self->test_doput_epagesj_request_expired_token();
    $self->test_doput_epagesj_request_failure();
    $self->test_doget_external_request_success();
    $self->test_doget_external_request_expired_token();
    $self->test_doget_external_request_failure();
    $self->test_dopost_external_request_success();
    $self->test_dopost_external_request_failure();
    done_testing();
}

#----------------------------------------------------------------------------------------
# Tests
#----------------------------------------------------------------------------------------
sub test_doget_epagesj_request_success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doGet(epagesj) --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $SUT = $self->_createSUT({ 'Shop' => $self->_createShop($MockedGoogleConfig) });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doGet($Url, HTTP_HEADERS, undef); },
        "$SubTestName - doGet can be called without errors."
    );
    ok(!WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(!WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, USER_ACCOUNT_RESPONSE, "$SubTestName - ACCESS_TOKEN returns correct USER_ACCOUNT_RESPONSE.");
}

sub test_doget_epagesj_request_expired_token {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doGet(epagesj) ExpiredToken --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => EXPIRED_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_UNAUTHORIZED, EXPECTED_ACCESS_TOKEN_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doGet($Url, HTTP_HEADERS, undef); },
        "$SubTestName - doGet can be called without errors."
    );
    ok(WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was called.");
    ok(WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was called.");
    is_deeply($hResponse, USER_ACCOUNT_RESPONSE, "$SubTestName - EXPIRED_ACCESS_TOKEN is renovated and returns correct USER_ACCOUNT_RESPONSE.");
}

sub test_doget_epagesj_request_failure {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doGet(epagesj) Other Error --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => BAD_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_INTERNAL_SERVER_ERROR, EXPECTED_EPAGESJ_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    throws_ok(
        sub { $hResponse = $SUT->doGet($Url, HTTP_HEADERS, undef); },
        qr/Error/,
        "$SubTestName - doGet(epagesj) rethrows non-token-expired error"
    );
}

sub test_dopost_epagesj_request_success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPost(epagesj) --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $SUT = $self->_createSUT({ 'Shop' => $self->_createShop($MockedGoogleConfig) });
    my $Url = EPAGESJ_BASE_URI . 'accounts/claim-website';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doPost($Url, HTTP_HEADERS, POST_PAYLOAD_CLAIM); },
        "$SubTestName - doPost can be called without errors."
    );
    ok(!WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(!WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, EXPECTED_CLAIM_WEBSITE_SUCCESS_RESPONSE, "$SubTestName - ACCESS_TOKEN returns empty hash.");
}

sub test_dopost_epagesj_request_expired_token {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPost(epagesj) ExpiredToken --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => EXPIRED_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_UNAUTHORIZED, EXPECTED_ACCESS_TOKEN_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doPost($Url, HTTP_HEADERS, POST_PAYLOAD_CLAIM); },
        "$SubTestName - doPost can be called without errors."
    );
    ok(WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was called.");
    ok(WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was called.");
    is_deeply($hResponse, EXPECTED_CLAIM_WEBSITE_SUCCESS_RESPONSE, "$SubTestName - EXPIRED_ACCESS_TOKEN is renovated and returns correct success response.");
}

sub test_dopost_epagesj_request_failure {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPost(epagesj) Other Error --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => BAD_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_INTERNAL_SERVER_ERROR, EXPECTED_EPAGESJ_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    throws_ok(
        sub { $hResponse = $SUT->doPost($Url, HTTP_HEADERS, POST_PAYLOAD_CLAIM); },
        qr/Error/,
        "$SubTestName - doPost(epagesj) rethrows non-token-expired error"
    );
}

sub test_dodelete_epagesj_request_success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doDelete(epagesj) --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $SUT = $self->_createSUT({ 'Shop' => $self->_createShop($MockedGoogleConfig) });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doDelete($Url, HTTP_HEADERS, undef); },
        "$SubTestName - doDelete can be called without errors."
    );
    ok(!WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(!WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, {}, "$SubTestName - success return value is empty hash.");
}

sub test_dodelete_epagesj_request_expired_token {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doDelete(epagesj) ExpiredToken --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => EXPIRED_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_UNAUTHORIZED, EXPECTED_ACCESS_TOKEN_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doDelete($Url, HTTP_HEADERS, undef); },
        "$SubTestName - doDelete can be called without errors."
    );
    ok(WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was called.");
    ok(WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was called.");
    is_deeply($hResponse, {}, "$SubTestName - success returns an empty hash.");
}

sub test_dodelete_epagesj_request_failure {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doDelete(epagesj) Other Error --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => BAD_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_INTERNAL_SERVER_ERROR, EXPECTED_EPAGESJ_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/user-account-info';
    my $hResponse;
    throws_ok(
        sub { $hResponse = $SUT->doDelete($Url, HTTP_HEADERS, undef); },
        qr/Error/,
        "$SubTestName - doDelete(epagesj) rethrows non-token-expired error"
    );
}

sub test_doput_epagesj_request_success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPut --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop });
    my $Url = EPAGESJ_BASE_URI . 'accounts/multiClientAccountId/sub-accounts/merchantAccountId';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doPut($Url, HTTP_HEADERS, UPDATE_SHOPURL_REQUEST); },
        "$SubTestName - doPut can be called without errors."
    );
    ok(!WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(!WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, UPDATE_SHOPURL_RESPONSE, "$SubTestName - Call returns correct UPDATE_SHOPURL_RESPONSE.");
}

sub test_doput_epagesj_request_expired_token {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPut --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => EXPIRED_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $MockedError = $self->_createError(HTTP_UNAUTHORIZED, EXPECTED_ACCESS_TOKEN_ERROR_RESPONSE);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop, 'Error' => => $MockedError });
    my $Url = EPAGESJ_BASE_URI . 'accounts/multiClientAccountId/sub-accounts/merchantAccountId';
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doPut($Url, HTTP_HEADERS, UPDATE_SHOPURL_REQUEST); },
        "$SubTestName - doPut can be called without errors."
    );
    ok(WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, UPDATE_SHOPURL_RESPONSE, "$SubTestName - Call returns correct UPDATE_SHOPURL_RESPONSE.");
}

sub test_doput_epagesj_request_failure {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPut --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => BAD_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop });
    my $Url = EPAGESJ_BASE_URI . 'accounts/multiClientAccountId/sub-accounts/merchantAccountId';
    my $hResponse;
    throws_ok(
        sub { $hResponse = $SUT->doPut($Url, HTTP_HEADERS, UPDATE_SHOPURL_REQUEST); },
        qr/Error/,
        "$SubTestName - doPut(epagesj) rethrows non-token-expired error"
    );
}

our $aResponseMocks;
our $ResponseIndex;

sub test_doget_external_request_success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doGet(external) --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $SUT = $self->_createSUT({
        'Shop'    => $self->_createShop($MockedGoogleConfig),
        'Status'  => HTTP_OK,
        'Response' => EXPECTED_USERINFO_RESPONSE
    });
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doGet(GOOGLE_OPENID_USERINFO_URL, HTTP_HEADERS, undef); },
        "$SubTestName - doGet can be called without errors."
    );
    ok(!WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(!WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, EXPECTED_USERINFO_RESPONSE, "$SubTestName - ACCESS_TOKEN returns correct USER_ACCOUNT_RESPONSE.");
}

sub test_doget_external_request_expired_token {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doGet(external) ExpiredToken --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => EXPIRED_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $SUT = $self->_createSUT({ 'Shop' => $MockedShop });
    $aResponseMocks = [
        $self->_createHttpResponse(HTTP_UNAUTHORIZED, EXPECTED_EXTERNAL_ACCESS_TOKEN_ERROR_RESPONSE),
        $self->_createHttpResponse(HTTP_OK, EXPECTED_ACCESS_TOKEN_RESPONSE),
        $self->_createHttpResponse(HTTP_OK, EXPECTED_USERINFO_RESPONSE)
    ];
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doGet(GOOGLE_OPENID_USERINFO_URL, HTTP_HEADERS, undef); },
        "$SubTestName - doGet can be called without errors."
    );
    ok(WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was called.");
    ok(WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was called.");
    is_deeply($hResponse, EXPECTED_USERINFO_RESPONSE, "$SubTestName - EXPIRED_ACCESS_TOKEN is renovated and returns the success response second try.");
}

sub test_doget_external_request_failure {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doGet(external) Other Error --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => BAD_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $SUT = $self->_createSUT({
        'Shop' => $MockedShop,
        'Status' => HTTP_INTERNAL_SERVER_ERROR,
        'Response' => EXPECTED_EXTERNAL_OTHER_ERROR_RESPONSE
    });
    my $hResponse;
    throws_ok(
        sub { $hResponse = $SUT->doGet(GOOGLE_OPENID_USERINFO_URL, HTTP_HEADERS, undef); },
        qr/Error/,
        "$SubTestName - doGet(external) rethrows non-token-expired error"
    );
}

sub test_dopost_external_request_success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPost(external) --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $SUT = $self->_createSUT({ 'Shop' => $self->_createShop($MockedGoogleConfig) });
    my $hResponse;
    lives_ok(
        sub { $hResponse = $SUT->doPost(AUTH_TOKEN_URL, HTTP_HEADERS, EXTERNAL_POST_PAYLOAD); },
        "$SubTestName - doPost can be called without errors."
    );
    ok(!WasCalled($SUT, '_isTokenExpired'), "$SubTestName - _isTokenExpired was NOT called.");
    ok(!WasCalled($SUT, '_doRefreshAccessToken'), "$SubTestName - _doRefreshAccessToken was NOT called.");
    is_deeply($hResponse, EXPECTED_ACCESS_TOKEN_RESPONSE, "$SubTestName - call succeeds.");
}

sub test_dopost_external_request_failure {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    note("-------- doPost(external) Other Error --------");
    my $hTokens = { 'RefreshToken' => REFRESH_TOKEN, 'AccessToken' => BAD_ACCESS_TOKEN };
    my $MockedGoogleConfig = $self->_createGoogleConfig($hTokens);
    my $MockedShop = $self->_createShop($MockedGoogleConfig);
    my $SUT = $self->_createSUT({
        'Shop' => $MockedShop,
        'Status' => HTTP_INTERNAL_SERVER_ERROR,
        'Response' => EXPECTED_EXTERNAL_OTHER_ERROR_RESPONSE
    });
    my $hResponse;
    throws_ok(
        sub { $hResponse = $SUT->doPost(AUTH_TOKEN_URL, HTTP_HEADERS, EXTERNAL_POST_PAYLOAD); },
        qr/Error/,
        "$SubTestName - doPost(external) rethrows non-token-expired error"
    );
}

#----------------------------------------------------------------------------------------
# Helpers
#----------------------------------------------------------------------------------------

sub _createGoogleConfig {
    my $self = shift;
    my ($hParams) = @_;

    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::GoogleSmartShopping::API::Object::GoogleSmartShoppingConfig', $hParams
    );

    return $Mockify->getMockObject();
}

sub _createShop {
    my $self = shift;
    my ($MockedGoogleConfig) = @_;

    my $hAttributes = {
        'Alias' => 'MyShop',
        'ShopTypeRef' => $self->_createShopTypeRef(),
        'GoogleSmartShoppingConfig' => $MockedGoogleConfig
    };
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::Shop::API::Object::Shop', $hAttributes
    );

    $Mockify->addMockWithReturnValueAndParameterCheck('existsChild', 1, [{'string'=>'GoogleSmartShoppingConfig'}]);

    return $Mockify->getMockObject();
}

sub _createSUT {
    my $self = shift;
    my ($hSUTParams) = @_;

    my $hServiceParams = { 'Shop' => $hSUTParams->{'Shop'} };
    my $MockifySut = Test::Mockify::Sut->new('DE_EPAGES::GoogleSmartShopping::API::GoogleApiClient', [ $hServiceParams ]);

    $MockifySut->mockImported('DE_EPAGES::Object::API::Factory', 'LoadRootObject')
        ->whenAny()
        ->thenReturn($self->_createSystem());

    $MockifySut->mockImported('DE_EPAGES::GoogleSmartShopping::API::Tools', 'LogGoogleSmartShopping')
        ->whenAny()
        ->thenReturnUndef();

    $MockifySut->mockStatic('DE_EPAGES::EPagesJ::API::Object::RestClient::create')
        ->whenAny()
        ->thenReturn($self->_createRestClient());

    $MockifySut->spy('_doRefreshAccessToken')
        ->when();

    $MockifySut->spy('_isTokenExpired')
        ->when(Object());

    if (defined($hSUTParams->{'Error'})) {
        $MockifySut->mock('_GetError')->whenAny()->thenReturn($hSUTParams->{'Error'});
    }
    
    $ResponseIndex = 0;
    $aResponseMocks = [
        $self->_createHttpResponse(
            $hSUTParams->{'Status'} // HTTP_OK,
            $hSUTParams->{'Response'} // EXPECTED_ACCESS_TOKEN_RESPONSE
        )
    ];
    my $UserAgent = Test::Mockify->new('LWP::UserAgent');
    $UserAgent->mock('request')->whenAny()->thenCall(sub {
        my $Total = scalar @$aResponseMocks;
        my $Response = $aResponseMocks->[$ResponseIndex];
        if ($ResponseIndex < $Total) {
            $ResponseIndex += 1;
        }
        return $Response;
    });
    $MockifySut->mockStatic('LWP::UserAgent::new')->whenAny()->thenReturn($UserAgent->getMockObject());

    my $Request = HTTP::Request->new();
    my $RequestMock = Test::MockObject::Extends->new($Request);
    $MockifySut->mockStatic('HTTP::Request::new')->whenAny()->thenReturn($RequestMock);
    return $MockifySut->getMockObject();
}

sub _createError {
    my $self = shift;
    my ($Status, $hResponse) = @_;

    my $Error = DE_EPAGES::Core::API::Error->new(
        Code => 'Error',
        Vars => {
            'Status' => $Status,
            'Response' => encode_json($hResponse)
        }
    );
    return $Error;
}

sub _createHttpResponse {
    my $self = shift;
    my ($Status, $hResponse) = @_;

    my $MockifyResponse = Test::Mockify->new('HTTP::Response');
    $MockifyResponse->mock('code')->when()->thenReturn($Status);
    $MockifyResponse->mock('decoded_content')->when()->thenReturn(encode_json($hResponse));
    return $MockifyResponse->getMockObject();
}

sub _createSystem {
    my $self = shift;

    my $hAttributes = {
        'EPagesJBaseURL' => 'https://shop.example.com:8088/rs',
        'OAuthServiceBaseUrl' => undef,
        'ProviderID' => 'epages',
        'GoogleSmartShopping_AuthURL' => 'epages',
        'GoogleSmartShopping_AuthTokenURL' => AUTH_TOKEN_URL,

        'GoogleSmartShopping_AppClientID' => 'CLIENT_ID',
        'GoogleSmartShopping_AppClientSecret' => 'CLIENT_SECRET',
        'GoogleSmartShopping_AppRedirectURL' => 'CLIENT_REDIRECT_URI',
        'GoogleSmartShopping_GoogleAppURL' => 'GOOGLE_APP_URI',
        'GoogleSmartShopping_AdminRefreshToken' => REFRESH_TOKEN,
        'GoogleSmartShopping_AdsDeveloperToken' => undef,
        'GoogleSmartShopping_MerchantCenterMultiClientAccountID' => 12345,
        'GoogleSmartShopping_AdsManagerAccountId' => 333444,

    };
    my $MockifySystem = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Object::API::Object::Site', $hAttributes
    );

    return $MockifySystem->getMockObject();
}

sub _createRestClient {
    my $self = shift;

    my $aParameters = [];

    my $MockifyRestClient = DE_EPAGES::Test::Mock::Mockify->new(
            'DE_EPAGES::EPagesJ::API::Object::RestClient', $aParameters
    );
    $MockifyRestClient->addMockWithReturnValue('client', $self->_createHttpClient());
    # $MockifyRestClient->addMockWithReturnValue('encodeRequest', sub { encode_json($_); });
    return $MockifyRestClient->getMockObject();
}

sub _createHttpClient {
    my $self = shift;

    my $Mockify = Test::Mockify->new('DE_EPAGES::EPagesJ::API::Object::HTTPClient', []);

    $Mockify->mock('get')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => ACCESS_TOKEN}))
            ->thenReturn($self->_createHttpResponse(HTTP_OK, USER_ACCOUNT_RESPONSE));

    $Mockify->mock('get')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => EXPIRED_ACCESS_TOKEN}))
            ->thenThrowError('UNAUTHORIZED');

    $Mockify->mock('get')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => BAD_ACCESS_TOKEN}))
            ->thenThrowError('OTHER');

    $Mockify->mock('post')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => ACCESS_TOKEN}), String(encode_json(POST_PAYLOAD_CLAIM)))
            ->thenReturn($self->_createHttpResponse(HTTP_OK, EXPECTED_CLAIM_WEBSITE_SUCCESS_RESPONSE));

    $Mockify->mock('post')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => EXPIRED_ACCESS_TOKEN}), String(encode_json(POST_PAYLOAD_CLAIM)))
            ->thenThrowError('UNAUTHORIZED');

    $Mockify->mock('post')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => BAD_ACCESS_TOKEN}), String(encode_json(POST_PAYLOAD_CLAIM)))
            ->thenThrowError('OTHER');

    $Mockify->mock('delete')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => ACCESS_TOKEN}))
            ->thenReturn($self->_createHttpResponse(HTTP_NO_CONTENT, {}));

    $Mockify->mock('delete')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => EXPIRED_ACCESS_TOKEN}))
            ->thenThrowError('UNAUTHORIZED');

    $Mockify->mock('delete')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => BAD_ACCESS_TOKEN}))
            ->thenThrowError('OTHER');

    $Mockify->mock('put')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => ACCESS_TOKEN}), String(encode_json(UPDATE_SHOPURL_REQUEST)))
            ->thenReturn($self->_createHttpResponse(HTTP_OK, UPDATE_SHOPURL_RESPONSE));

    $Mockify->mock('put')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => EXPIRED_ACCESS_TOKEN}), String(encode_json(UPDATE_SHOPURL_REQUEST)))
            ->thenThrowError('UNAUTHORIZED');

    $Mockify->mock('put')
            ->when(Object(), HashRef({'Accept'  => 'application/json', 'X-AccessToken' => BAD_ACCESS_TOKEN}), String(encode_json(UPDATE_SHOPURL_REQUEST)))
            ->thenThrowError('OTHER');

    return $Mockify->getMockObject();
}

sub _createShopTypeRef {
    my $self = shift;

    my $hAttributes = {
        'Alias' => 'UnityDemo'
    };
    my $MockifySystem = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
            'DE_EPAGES::Shop::API::Object::ShopTypeRef', $hAttributes
    );
    return $MockifySystem->getMockObject();
}


#========================================================================================
# Main
#========================================================================================

__PACKAGE__->runScript();
