#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::t::unit::UI_Shop
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for the GoogleSmartShopping view setup
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::t::unit::UI_Shop;
use base DE_EPAGES::Core::API::Test;

use strict;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::MockifyServlet;

use Test::Mockify;
use Test::Mockify::Matcher qw ( HashRef Object String Number );
use Test::Mockify::Sut;
use Test::Mockify::Verify qw ( WasCalled GetParametersFromMockifyCall );
use Test::More;

use DE_EPAGES::GoogleSmartShopping::API::Constants qw ( ERROR_ACCEPT_TERMS ERROR_FORM_SUBMIT ERROR_TOO_MANY_SUBACCOUNTS );

use HTTP::Status qw ( HTTP_OK HTTP_BAD_REQUEST HTTP_FORBIDDEN );
use JSON::MaybeXS qw ( encode_json );

use constant MERCHANT_EMAIL => 'example@email.com';
use constant CONNECT_OAUTH_URL => 'https://www.example.com';
use constant GOOGLE_OAUTH_ERROR => 'access_denied';
use constant GOOGLE_SINGLE_QUOTE_ERROR => 'quo\'te';
use constant GOOGLE_OAUTH_TOKENS_EXCHANGE_ERROR => 'INTERNAL_SERVER_ERROR';
use constant SUCCESSFUL_GOOGLE_OAUTH_CALLBACK_RESPONSE => {
    code  => [ 'some-google-authorization-code' ],
    scope => [ 'space-separated-list-of-scopes-string']
};
use constant UNSUCCESSFUL_GOOGLE_OAUTH_CALLBACK_RESPONSE => {
    error  => [ GOOGLE_OAUTH_ERROR ]
};
use constant GOOGLE_OAUTH_CALLBACK_UNKNOWN_PARAMETER => {
    unknown => 'asdf'
};
use constant GOOGLE_OAUTH_CALLBACK_ERROR_SINGLE_QUOTE => {
    error  => [ GOOGLE_SINGLE_QUOTE_ERROR ]
};
use constant SUCCESSFUL_GOOGLE_OAUTH_TOKENS_EXCHANGE_RESPONSE => {
    access_token  => 'some-google-access-token',
    refresh_token => 'some-google-refresh-token',
    id_token      => 'some-long-encrypted-key',
    expires_in    => 3599,
    token_type    => 'Bearer',
    scope         => 'space-separated-list-of-scopes-string'
};
use constant UNSUCCESSFUL_GOOGLE_OAUTH_TOKENS_EXCHANGE_RESPONSE => {
    "error" => "invalid_grant",
    "error_description" => "Malformed auth code."
};
use constant TOKENS_EXCHANGE_SINGLE_QUOTE_ERROR_RESPONSE => {
    error => [ GOOGLE_SINGLE_QUOTE_ERROR ]
};
use constant EXAMPLE_EMAIL => 'google-account-email';
use constant SUCCESSFUL_USER_INFO_RESPONSE => {
    email => EXAMPLE_EMAIL
};
use constant SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE => {
    name => 'google-merchant-center-account-name',
    id   => 123456789
};
use constant SUCCESSFUL_GOOGLE_ADS_ACCOUNTS_RESPONSE => {
    name => 'google-ads-account-name',
    id   => 555778899
};
use constant FAILED_USER_INFO_RESPONSE => { };
use constant FAILED_USER_INFO_RESPONSE_WITH_EMAIL => {
    is_token_expired => 1,
    email => EXAMPLE_EMAIL,
};
use constant SUCCESSFUL_CALLBACK_JS_CONTENT_RESPONSE => "<script type=\"text/javascript\">\n    window.opener.ep.fn.callbackSuccess();\n    window.close();\n</script>\n";
use constant ACCESS_DENIED_CALLBACK_JS_CONTENT_RESPONSE => "<script type=\"text/javascript\">\n    window.opener.ep.fn.showError('access_denied');\n    window.close();\n</script>\n";
use constant BAD_REQUEST_JS_CONTENT_RESPONSE => "<script type=\"text/javascript\">\n    window.opener.ep.fn.showError('unknown_error');\n    window.close();\n</script>\n";


use constant REVOKE_TOKEN_RESPONSE_ERROR => {
    'error' => 'invalid_token',
    'error_description' => 'poggers'
};

use constant ACCOUNT_ID => 89879879;
use constant MULTI_CLIENT_ACCOUNT_ID => 398123456;
use constant MERCHANT_CENTER_ID => 458599058;
use constant GOOGLE_ADS_ID => 555778899;

use constant MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_HOME_PAGE => {
    "id" => "home_page_issue",
    "documentation"=> "https://support.google.com/merchants/answer/176793",
    "severity"=> "error",
    "title"=> "Website not claimed"
};

use constant MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_QUALITY_LOW => {
    "documentation"=> "https://support.google.com/merchants/answer/9849758",
    "id" => "merchant_quality_low",
    "severity"=> "error",
    "title"=> "Account isn't eligible for enhanced free listings"
};

use constant MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_MISSING_AD_WORDS_LINK => {
    "documentation"=> "https://support.google.com/merchants/answer/6159060",
    "id" => "missing_ad_words_link",
    "severity"=> "error",
    "title"=> "No Google Ads account linked"
};

use constant MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE => {
    'id' => 'pending_address_and_phone',
    'documentation' => 'https://support.google.com/merchants/answer/10139041',
    'severity' => 'critical',
    'title' => 'You need to enter your business address and phone number',
    'detail' => 'Go to the business information page to enter your business address and phone number',
};

use constant MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_UNKNOWN => {
    "documentation"=> "https://support.google.com/merchants/answer/123456789",
    "id" => "unknown_issue",
    "severity"=> "error",
    "title"=> "Something is wrong but we don't know what is it"
};

use constant MERCHANT_CENTER_ACCOUNT_STATUSES_UNKNOWN_ISSUE_RESPONSE => {
    "accountId"=> "414922848",
    "accountLevelIssues"=> [
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_UNKNOWN
    ],
    "kind"=> "content#accountStatus",
    "websiteClaimed"=> "true",
    "accountManagement"=> "manual"
};
use constant UNSUCCESSFUL_MERCHANT_CENTER_ACCOUNT_STATUSES_RESPONSE => {
    "accountId"=> "414922848",
    "accountLevelIssues"=> [
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_QUALITY_LOW,
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE,
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_MISSING_AD_WORDS_LINK,
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_HOME_PAGE
    ],
    "kind"=> "content#accountStatus",
    "websiteClaimed"=> "false",
    "accountManagement"=> "manual"
};

use constant EXPECTED_TOO_MANY_SUBACCOUNTS_RESPONSE => {
    "errorId" => "google-api-error",
    "message" => "Maximum number of subaccounts reached",
    "details" => {
        "code" => 403,
        "errors" => [
            {
                "domain" => "content.ContentErrorDomain",
                "message" => "Maximum number of subaccounts reached",
                "reason" => "quota/too_many_subaccounts"
            }
        ],
        "message" => "Maximum number of subaccounts reached"
    }
};

use constant EXPECTED_ACCESS_DENIED_RESPONSE => {
    "errorId" => "google-api-error",
    "message" => "access denied",
    "details" => {
        "code" => 401,
        "errors" => [
            {
                "domain" => "global",
                "message" => "access denied",
                "reason" => "auth/account_access_denied"
            }
        ],
        "message" => "access denied"
    }
};
use constant ADS_ACCOUNT_ID => 899423423432879;

sub test {
    my $self = shift;

    $self->test_ViewGoogleSmartShopping_On_OnboardingStart_NoGoogleSmartShoppingConfig();
    $self->test_ViewGoogleSmartShopping_On_OnboardingStart_NoAccessToken();
    $self->test_ViewGoogleSmartShopping_On_OnboardingStart_NoRefreshToken();
    $self->test_ViewGoogleSmartShopping_On_GettingConnectedAccountEmail();
    $self->test_ViewGoogleSmartShopping_On_ConnectedAccountEmailFailsFirstTime();
    $self->test_ViewGoogleSmartShopping_On_ConnectedAccountEmailFailsTokenExpiredOrRevoked();
    $self->test_ViewGoogleSmartShopping_On_ConnectedAdsAccount_BillingApproved();
    $self->test_ViewGoogleSmartShopping_On_ConnectedAdsAccount_BillingUnapproved();
    $self->test_ViewGoogleSmartShopping_On_ConnectedAdsAccount_BillingError();
    $self->test_DisconnectConnectedAccount_RevokeToken_NoConfig();
    $self->test_DisconnectConnectedAccount_RevokeToken_Successful();
    $self->test_DisconnectConnectedAccount_RevokeToken_Error();
    $self->test_DisconnectMerchantAccount_Successful_SubAccount();
    $self->test_DisconnectMerchantAccount_Successful_Not_SubAccount();
    $self->test_DisconnectMerchantAccount_Failed_SubAccount_delete();
    $self->test_DisconnectAdsAccount_Successful_No_Campaign();
    $self->test_DisconnectAdsAccount_Successful_With_Campaign();
    $self->test_DisconnectAdsAccount_Failed_deleteAdsAccount();
    $self->test_DisconnectAdsAccount_Failed_deleteAdsCampaign();
    $self->test_GoogleOAuthCallback_On_OnboardingCallbackSuccess_And_SuccessfulTokensExchange();
    $self->test_GoogleOAuthCallback_On_OnboardingCallbackError();
    $self->test_GoogleOAuthCallback_On_OnboardingCallbackSuccess_And_UnsuccessfulTokensExchange();
    $self->test_GoogleOAuthCallback_On_OnboardingCallbackUnknownParameter();
    $self->test_ConnectedAccount_MerchantCenterRequirements_Not_Met();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccounts();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccounts_NoAccounts();
    $self->test_CreateMerchantCenterAccount_Success();
    $self->test_CreateMerchantCenterAccount_Fail_Form();
    $self->test_CreateMerchantCenterAccount_Fail_Api();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Subaccount_Success();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Standalone_Merchant_Account_Success();
    $self->test_CreateGoogleAdsAccount_Success();
    $self->test_CreateGoogleAdsAccount_Fail_Form_Requirements_Not_Met();
    $self->test_CreateGoogleAdsAccount_Fail_Form_No_MerchantCenter();
    $self->test_CreateGoogleAdsAccount_Fail_Form_Already_Has_Ads_Account();
    $self->test_CreateGoogleAdsAccount_Fail_Api();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Unknown_Issue();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Success_When_Ads_Account_IS_Connected();
    $self->test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Success_When_No_Ads_Account();


    done_testing();
}

sub test_ViewGoogleSmartShopping_On_OnboardingStart_NoGoogleSmartShoppingConfig {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = undef;
    my $AlreadyExistingConfig = 0;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $SUT = $self->_createSUT();

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    is(
        $Servlet->vars('GoogleOAuthConnect'),
        CONNECT_OAUTH_URL,
        "$SubTestName - GoogleSmartShopping setup view has OAuth URL when there is no GoogleSmartShoppingConfig object"
    );
    ok( WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was built' );
}

sub test_ViewGoogleSmartShopping_On_OnboardingStart_NoAccessToken {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => undef,
        RefreshToken => undef,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $SUT = $self->_createSUT();

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    is(
        $Servlet->vars('GoogleOAuthConnect'),
        CONNECT_OAUTH_URL,
        "$SubTestName - GoogleSmartShopping setup view has OAuth URL when GoogleSmartShoppingConfig access token is empty"
    );
    ok( WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was built' );
}

sub test_ViewGoogleSmartShopping_On_OnboardingStart_NoRefreshToken {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => undef,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $SUT = $self->_createSUT();

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    is(
        $Servlet->vars('GoogleOAuthConnect'),
        CONNECT_OAUTH_URL,
        "$SubTestName - GoogleSmartShopping setup view has OAuth URL when GoogleSmartShoppingConfig refresh token is empty"
    );
    ok( WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was built' );
}

sub test_ViewGoogleSmartShopping_On_GettingConnectedAccountEmail {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );
    ok(
        WasCalled($ApiService, 'getMerchantCenterAccounts'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );
}

sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccounts {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 0,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );
    ok(
        WasCalled($ApiService, 'getMerchantCenterAccounts'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    is(
        $Servlet->vars('MerchantCenterAccounts'),
        SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE,
        "$SubTestName - Verify that MerchantCenterAccounts correctly returned."
    );
    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );
}


sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccounts_NoAccounts {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 0,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => []
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );
    ok(
        WasCalled($ApiService, 'getMerchantCenterAccounts'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    is_deeply(
        $Servlet->vars('MerchantCenterAccounts'),
        [],
        "$SubTestName - Verify that MerchantCenterAccounts returns empty response when no accounts found."
    );
    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );
}

sub test_ViewGoogleSmartShopping_On_ConnectedAccountEmailFailsFirstTime {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => FAILED_USER_INFO_RESPONSE
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that lookup of Google account email is attempted'
    );
}

sub test_ViewGoogleSmartShopping_On_ConnectedAccountEmailFailsTokenExpiredOrRevoked {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 0,
        AccountEmail => EXAMPLE_EMAIL,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => FAILED_USER_INFO_RESPONSE_WITH_EMAIL
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that lookup of Google account email is attempted'
    );
    ok(
        !WasCalled($ApiService, 'getMerchantCenterAccounts'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was not called when AccountInfo fails'
    );
    ok( WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was built' );
    is(
        $Servlet->vars('TokenExpiredOrRevoked'),
        1,
        $SubTestName . ' - Ensure servlet flag is set for expired or revoked RefreshToken.'
    );
    is(
        $Servlet->vars('ConnectedGoogleAccountEmail'),
        EXAMPLE_EMAIL,
        $SubTestName . ' - Check that saved email was retrieved.'
    );
}

sub test_ViewGoogleSmartShopping_On_ConnectedAdsAccount_BillingApproved {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        AccountEmail => EXAMPLE_EMAIL,
        GoogleAdsAccountID => ADS_ACCOUNT_ID,
    });

    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
    });
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            thenReturn => 1
        }
    });
    my $SUT = $self->_createSUT($ApiService, undef, $AdsApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($AdsApiService, 'isAdsAccountBillingApproved'),
        $SubTestName . ' - Verify that isAdsAccountBillingApproved was called'
    );

    is(
        $Servlet->vars('IsAdsAccountBillingApproved'),
        1,
        $SubTestName . ' - Check that correct IsAdsAccountBillingApproved is sent to the template.'
    );
}

sub test_ViewGoogleSmartShopping_On_ConnectedAdsAccount_BillingUnapproved {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        AccountEmail => EXAMPLE_EMAIL,
        GoogleAdsAccountID => ADS_ACCOUNT_ID,
    });

    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
    });
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            thenReturn => 0
        }
    });
    my $SUT = $self->_createSUT($ApiService, undef, $AdsApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($AdsApiService, 'isAdsAccountBillingApproved'),
        $SubTestName . ' - Verify that isAdsAccountBillingApproved was called'
    );

    is(
        $Servlet->vars('IsAdsAccountBillingApproved'),
        0,
        $SubTestName . ' - Check that correct IsAdsAccountBillingApproved is sent to the template.'
    );
}

sub test_ViewGoogleSmartShopping_On_ConnectedAdsAccount_BillingError {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        AccountEmail => EXAMPLE_EMAIL,
        GoogleAdsAccountID => ADS_ACCOUNT_ID,
    });

    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
    });
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            throw_error => 1
        }
    });
    my $SUT = $self->_createSUT($ApiService, undef, $AdsApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    ok(
        WasCalled($AdsApiService, 'isAdsAccountBillingApproved'),
        $SubTestName . ' - Verify that isAdsAccountBillingApproved was called'
    );

    is(
        $Servlet->vars('IsAdsAccountBillingApproved'),
        0,
        $SubTestName . ' - Check that correct IsAdsAccountBillingApproved is sent to the template.'
    );
}

sub test_GoogleOAuthCallback_On_OnboardingCallbackSuccess_And_SuccessfulTokensExchange {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
    });
    my $AlreadyExistingConfig = 0;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = SUCCESSFUL_GOOGLE_OAUTH_CALLBACK_RESPONSE;
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        OAuthTokens => SUCCESSFUL_GOOGLE_OAUTH_TOKENS_EXCHANGE_RESPONSE
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->GoogleOAuthCallback($Servlet);
    my $aWriteHTTPResponseParams = GetParametersFromMockifyCall($SUT, '_writeHTTPResponse');
    my $Code = $aWriteHTTPResponseParams->[1];
    my $hParams = $aWriteHTTPResponseParams->[2];
    my $Content = $aWriteHTTPResponseParams->[3];

    ok(
        WasCalled($Shop, 'existsChild'),
        $SubTestName . ' - Verify that access token is being saved into GoogleSmartShoppingConfig object '
    );
    ok(
        WasCalled($Shop, 'insertChild'),
        $SubTestName . ' - Verify that GoogleSmartShoppingConfig object gets created because did not exist before'
    );
    is(
        $Code,
        HTTP_OK,
        $SubTestName . ' - Ensure HTTP response type is OK'
    );
    is_deeply(
        $hParams,
        { 'ContentType' => 'text/html' },
        $SubTestName . ' - Ensure HTTP response with ContentType set to text/html'
    );
    is(
        $Content,
        SUCCESSFUL_CALLBACK_JS_CONTENT_RESPONSE,
        $SubTestName . ' - Ensure $Servlet->writeContentOutput() was called with the correct JS to refresh the parent window'
    );
}

sub test_GoogleOAuthCallback_On_OnboardingCallbackError {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = undef;
    my $AlreadyExistingConfig = 0;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = UNSUCCESSFUL_GOOGLE_OAUTH_CALLBACK_RESPONSE;
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $SUT = $self->_createSUT();

    $SUT->GoogleOAuthCallback($Servlet);
    my $aWriteHTTPResponseParams = GetParametersFromMockifyCall($SUT, '_writeHTTPResponse');
    my $Code = $aWriteHTTPResponseParams->[1];
    my $hParams = $aWriteHTTPResponseParams->[2];
    my $Content = $aWriteHTTPResponseParams->[3];

    is(
        $Code,
        HTTP_FORBIDDEN,
        $SubTestName . ' - Ensure HTTP response type is 403 Forbidden'
    );
    is_deeply(
        $hParams,
        { 'ContentType' => 'text/html' },
        $SubTestName . ' - Ensure HTTP response with ContentType set to text/html'
    );
    is(
        $Content,
        ACCESS_DENIED_CALLBACK_JS_CONTENT_RESPONSE,
        $SubTestName . ' - Ensure $Servlet->writeContentOutput() was called with the correct JS to send access_denied error to the parent window'
    );
}

sub test_GoogleOAuthCallback_On_OnboardingCallbackSuccess_And_UnsuccessfulTokensExchange {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
    });
    my $AlreadyExistingConfig = 0;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = SUCCESSFUL_GOOGLE_OAUTH_CALLBACK_RESPONSE;
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        OAuthTokens => {'throw_error' => 1}
    });
    my $MockedError = $self->_createError(HTTP_BAD_REQUEST, UNSUCCESSFUL_GOOGLE_OAUTH_TOKENS_EXCHANGE_RESPONSE);
    my $SUT = $self->_createSUT($ApiService, $MockedError);

    $SUT->GoogleOAuthCallback($Servlet);
    my $aWriteHTTPResponseParams = GetParametersFromMockifyCall($SUT, '_writeHTTPResponse');
    my $Code = $aWriteHTTPResponseParams->[1];
    my $hParams = $aWriteHTTPResponseParams->[2];
    my $Content = $aWriteHTTPResponseParams->[3];

    is(
        $Code,
        HTTP_BAD_REQUEST,
        $SubTestName . ' - Ensure HTTP response type is 400 Internal server error'
    );
    is_deeply(
        $hParams,
        { 'ContentType' => 'text/html' },
        $SubTestName . ' - Ensure HTTP response with ContentType set to text/html'
    );
    is(
        $Content,
        BAD_REQUEST_JS_CONTENT_RESPONSE,
        $SubTestName . ' - Ensure $Servlet->writeContentOutput() was called with the correct JS to send unknown_error to the parent window'
    );
}

sub test_GoogleOAuthCallback_On_OnboardingCallbackUnknownParameter {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
    });
    my $AlreadyExistingConfig = 0;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = GOOGLE_OAUTH_CALLBACK_UNKNOWN_PARAMETER;
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $SUT = $self->_createSUT();

    $SUT->GoogleOAuthCallback($Servlet);
    my $aWriteHTTPResponseParams = GetParametersFromMockifyCall($SUT, '_writeHTTPResponse');
    my $Code = $aWriteHTTPResponseParams->[1];
    my $hParams = $aWriteHTTPResponseParams->[2];
    my $Content = $aWriteHTTPResponseParams->[3];

    is(
        $Code,
        HTTP_BAD_REQUEST,
        $SubTestName . ' - Ensure HTTP response type is 400 Bad Request for unknown parameter to callback ChangeAction'
    );
    is_deeply(
        $hParams,
        { 'ContentType' => 'text/html' },
        $SubTestName . ' - Ensure HTTP response with ContentType set to text/html'
    );
    is(
        $Content,
        BAD_REQUEST_JS_CONTENT_RESPONSE,
        $SubTestName . ' - Ensure content contains JS for window close - consistency with other cases where popup is closed'
    );
}


sub test_DisconnectConnectedAccount_RevokeToken_NoConfig {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = undef;
    my $AlreadyExistingConfig = 0;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService(undef);

    # given: 'On-boarding NOT done, GoogleSmartShoppingConfig not defined '
    my $SUT = $self->_createSUT($ApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectPersonalAccount($Servlet);

    # then: 'Make request to google API to revoke the token, delete gss config from shop'
    ok(!WasCalled($ApiService, 'revokeAccessToken'),
        $SubTestName . ' - Verify service revokeAccessToken was NOT called');
    ok(!WasCalled($ApiService, 'deleteMerchantCenterSubAccount'),
        $SubTestName . ' - Verify service deleteMerchantCenterSubAccount was NOT called');

}

sub test_DisconnectConnectedAccount_RevokeToken_Successful {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        MerchantCenterAccountID => ACCOUNT_ID,
        IsMerchantCenterSubAccount => 1,
        IsAdsAccountSubAccount => 1,
        GoogleAdsAccountID => GOOGLE_ADS_ID,
        AdsCampaignId => 743242423
    });
    my $AlreadyExistingConfig = 1;

    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({ deleteMerchantCenterSubAccount => {} });
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            thenReturn => 0
        },
    });

    # setup
    my $SUT = $self->_createSUT($ApiService, undef, $AdsApiService);

    # given: 'On-boarding done, connected account email successful retrieved '
    $SUT->UnityViewGoogleSmartShopping($Servlet);
    ok(WasCalled($ApiService, 'getAndSaveUserInfo'), $SubTestName . ' - Verify getAndSaveUserInfo was called');
    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectPersonalAccount($Servlet);

    # then: 'Make request to google API to revoke the token, delete gss config from shop'
    ok(WasCalled($ApiService, 'revokeAccessToken'),
        $SubTestName . ' - Verify service revokeAccessToken was called');
    ok(WasCalled($ApiService, 'deleteMerchantCenterSubAccount'),
        $SubTestName . ' - Verify service deleteMerchantCenterSubAccount was called');
    ok(WasCalled($AdsApiService, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify Ads service deleteAdsCampaign was called');
    ok(WasCalled($AdsApiService, 'unlinkGoogleAdsAccount'),
        $SubTestName . ' - Verify Ads service unlinkGoogleAdsAccount was called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'delete'),
        $SubTestName . ' - Verify $GoogleSmartShoppingConfig->delete was called');
}

sub test_DisconnectConnectedAccount_RevokeToken_Error {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        MerchantCenterAccountID => ACCOUNT_ID,
        IsMerchantCenterSubAccount => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService( { 'revokeAccessToken' => { 'throw_error' => 1 } });

    # given: 'On-boarding done, connected account email successful retrieved '
    my $SUT = $self->_createSUT($ApiService, $self->_createError(HTTP_BAD_REQUEST, REVOKE_TOKEN_RESPONSE_ERROR));
    $SUT->UnityViewGoogleSmartShopping($Servlet);
    ok(WasCalled($ApiService, 'getAndSaveUserInfo'), $SubTestName . ' - Verify getAndSaveUserInfo was called');
    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );

    # when: 'Merchant clicks Disconnect button', service call returns an error
    $SUT->DisconnectPersonalAccount($Servlet);

    # then: Delete config, call revokeAccessToken, executeFormError should be called
    ok(WasCalled($GoogleSmartShoppingConfig, 'delete'),
        $SubTestName . ' - Verify GoogleSmartShoppingConfig->delete was called');
    ok(WasCalled($ApiService, 'revokeAccessToken'),
        $SubTestName . ' - Verify service revokeAccessToken was called');
    ok(WasCalled($ApiService, 'deleteMerchantCenterSubAccount'),
        $SubTestName . ' - Verify service deleteMerchantCenterSubAccount was called');
    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");
}

sub test_DisconnectMerchantAccount_Successful_SubAccount {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        MerchantCenterAccountID => ACCOUNT_ID,
        IsMerchantCenterSubAccount => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({  });

    # setup
    my $SUT = $self->_createSUT($ApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectMerchantCenterAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(WasCalled($ApiService, 'deleteMerchantCenterSubAccount'),
        $SubTestName . ' - Verify service deleteMerchantCenterSubAccount was called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteMerchantCenterAccount'),
        $SubTestName . ' - Verify deleteMerchantCenterAccount was called');
}

sub test_DisconnectMerchantAccount_Successful_Not_SubAccount {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        MerchantCenterAccountID => ACCOUNT_ID,
        IsMerchantCenterSubAccount => 0,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({  });

    # setup
    my $SUT = $self->_createSUT($ApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectMerchantCenterAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(!WasCalled($ApiService, 'deleteMerchantCenterSubAccount'),
        $SubTestName . ' - Verify service deleteMerchantCenterSubAccount was NOT called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteMerchantCenterAccount'),
        $SubTestName . ' - Verify deleteMerchantCenterAccount was called');
}

sub test_DisconnectMerchantAccount_Failed_SubAccount_delete {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        MerchantCenterAccountID => ACCOUNT_ID,
        IsMerchantCenterSubAccount => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);

    my $ApiService = $self->_createGoogleApiService({
        'deleteMerchantCenterSubAccount' => {
            'throw_error' => 1
        }
    });

    # setup
    my $SUT = $self->_createSUT($ApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectMerchantCenterAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(WasCalled($ApiService, 'deleteMerchantCenterSubAccount'),
        $SubTestName . ' - Verify service deleteMerchantCenterSubAccount was NOT called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteMerchantCenterAccount'),
        $SubTestName . ' - Verify deleteMerchantCenterAccount was called');
    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");
}

sub test_DisconnectAdsAccount_Successful_No_Campaign {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        GoogleAdsAccountID => ACCOUNT_ID,
        IsAdsAccountSubAccount => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            thenReturn => 1
        }
    });

    # setup
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectGoogleAdsAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(!WasCalled($AdsApiService, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify service deleteAdsCampaign was NOT called');
    ok(WasCalled($AdsApiService, 'unlinkGoogleAdsAccount'),
        $SubTestName . ' - Verify service unlinkGoogleAdsAccount was called');

    ok(!WasCalled($GoogleSmartShoppingConfig, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify deleteAdsCampaign was NOT called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteAdsAccount'),
        $SubTestName . ' - Verify deleteAdsAccount was called');
}

sub test_DisconnectAdsAccount_Successful_With_Campaign {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        GoogleAdsAccountID => ACCOUNT_ID,
        IsAdsAccountSubAccount => 1,
        AdsCampaignId => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            thenReturn => 1
        }
    });

    # setup
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectGoogleAdsAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(WasCalled($AdsApiService, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify service deleteAdsCampaign was called');
    ok(WasCalled($AdsApiService, 'unlinkGoogleAdsAccount'),
        $SubTestName . ' - Verify service unlinkGoogleAdsAccount was called');

    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify deleteAdsCampaign was called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteAdsAccount'),
        $SubTestName . ' - Verify deleteAdsAccount was called');
}

sub test_DisconnectAdsAccount_Successful {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        GoogleAdsAccountID => ACCOUNT_ID,
        IsAdsAccountSubAccount => 0,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService({
        isAdsAccountBillingApproved => {
            thenReturn => 1
        }
    });

    # setup
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectGoogleAdsAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(!WasCalled($AdsApiService, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify service deleteAdsCampaign was NOT called');
    ok(WasCalled($AdsApiService, 'unlinkGoogleAdsAccount'),
        $SubTestName . ' - Verify service unlinkGoogleAdsAccount was called');

    ok(!WasCalled($GoogleSmartShoppingConfig, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify deleteAdsCampaign was NOT called');
    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteAdsAccount'),
        $SubTestName . ' - Verify deleteAdsAccount was called');
}

sub test_DisconnectAdsAccount_Failed_deleteAdsAccount {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        GoogleAdsAccountID => ACCOUNT_ID,
        IsAdsAccountSubAccount => 0,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService({
        deleteAdsAccount => {
            throw_error => 1
        }
    });

    # setup
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectGoogleAdsAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(!WasCalled($AdsApiService, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify service deleteAdsCampaign was NOT called');
    ok(WasCalled($AdsApiService, 'unlinkGoogleAdsAccount'),
        $SubTestName . ' - Verify service unlinkGoogleAdsAccount was called');

    ok(!WasCalled($GoogleSmartShoppingConfig, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify deleteAdsCampaign was NOT called');
    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");
}

sub test_DisconnectAdsAccount_Failed_deleteAdsCampaign {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        GoogleAdsAccountID => ACCOUNT_ID,
        IsAdsAccountSubAccount => 0,
        AdsCampaignId => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService({
        deleteAdsCampaign => {
            throw_error => 1
        }
    });

    # setup
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    # when: 'The Merchant clicks disconnect button'
    $SUT->DisconnectGoogleAdsAccount($Servlet);

    # then: 'Make request to google API to delete the subAccount, delete gss config from shop'
    ok(WasCalled($AdsApiService, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify service deleteAdsCampaign was called');
    ok(WasCalled($AdsApiService, 'unlinkGoogleAdsAccount'),
        $SubTestName . ' - Verify service unlinkGoogleAdsAccount was called');

    ok(WasCalled($GoogleSmartShoppingConfig, 'deleteAdsCampaign'),
        $SubTestName . ' - Verify deleteAdsCampaign was called');
    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");
}

sub test_ConnectedAccount_MerchantCenterRequirements_Not_Met {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        MerchantCenterAccountName => 'MerchantCenterAccountName',
        MerchantCenterAccountID => ACCOUNT_ID,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);

    is(
        $Servlet->vars('GoogleSmartShoppingConfig')->{'MerchantCenterAccountName'},
        "MerchantCenterAccountName",
        "$SubTestName - MerchantCenterAccountName is assigned to the Servlet"
    );

    is(
        $Servlet->vars('GoogleSmartShoppingConfig')->{'MerchantCenterAccountID'},
        ACCOUNT_ID,
        "$SubTestName - MerchantCenterAccountID is assigned to the Servlet"
    );

    is(
        $Servlet->vars('GoogleSmartShoppingConfig')->{'ChecklistBillingTOC'},
        0,
        "$SubTestName - Google Merchant Center billing terms and conditions requirement is not met"
    );
    is(
        $Servlet->vars('GoogleMerchantCenterRequirementsMet'),
        0,
        "$SubTestName - Google Merchant Center requirements not met warning is shown"
    );
}

sub test_CreateMerchantCenterAccount_Success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {
        HasAdultContent => [0],
        newMerchantAccountAcceptTerms => [1],
    };
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        createMerchantCenterSubAccount => {
            id => ACCOUNT_ID
        }
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->CreateMerchantCenterAccount($Servlet);

    is_deeply(
        $Servlet->vars('ResponseContent'),
        {id => ACCOUNT_ID},
        "$SubTestName - MerchantCenter SubAccount is successfully created"
    );
}

sub test_CreateMerchantCenterAccount_Fail_Form {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {
        HasAdultContent => [""],
        newMerchantAccountAcceptTerms => [1],
    };
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        createMerchantCenterSubAccount => {
            id => ACCOUNT_ID
        }
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->CreateMerchantCenterAccount($Servlet);

    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");

    my $aExecuteFormErrorParams = GetParametersFromMockifyCall($Servlet->form(), 'executeFormError');

    is_deeply(
        $aExecuteFormErrorParams->[0],
        {
            'Name'   => 'CreationFailed',
            'Reason' => ERROR_FORM_SUBMIT
        },
        $SubTestName . ' - Ensure Form error gets thrown'
    );

}

sub test_CreateMerchantCenterAccount_Fail_Api {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {
        HasAdultContent => [1],
        newMerchantAccountAcceptTerms => [1],
    };
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        createMerchantCenterSubAccount => {
            id => ACCOUNT_ID,
            throw_error => 1
        }
    });
    my $SUT = $self->_createSUT($ApiService, $self->_createError(HTTP_BAD_REQUEST, EXPECTED_TOO_MANY_SUBACCOUNTS_RESPONSE));

    $SUT->CreateMerchantCenterAccount($Servlet);

    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");

    my $aExecuteFormErrorParams = GetParametersFromMockifyCall($Servlet->form(), 'executeFormError');

    is_deeply(
        $aExecuteFormErrorParams->[0],
        {
            'Name'   => 'CreationFailed',
            'Reason' => ERROR_TOO_MANY_SUBACCOUNTS
        },
        $SubTestName . ' - Ensure Api error too_many_accounts gets thrown'
    );
}

sub test_CreateGoogleAdsAccount_Success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        AccountEmail                           => MERCHANT_EMAIL,
    });
    my $AlreadyExistingConfig = 1;
    my $hQueryParameters = {};
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $ServletForm = $self->_createServletForm($hQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService();
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    $SUT->CreateGoogleAdsAccount($Servlet);

    ok(!WasCalled($Servlet->form(), 'executeFormError'), "$SubTestName - ensure executeFormError was NOT called");
    ok(WasCalled($AdsApiService, 'createGoogleAdsSubAccount'), "$SubTestName - ensure createGoogleAdsLink was called");
    ok(WasCalled($AdsApiService, 'createAndEnableGoogleAdsLink'), "$SubTestName - ensure createAndEnableGoogleAdsLink was called");
}

sub test_CreateGoogleAdsAccount_Fail_Form_Requirements_Not_Met {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 0,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        AccountEmail                           => MERCHANT_EMAIL,
    });
    my $AlreadyExistingConfig = 1;
    my $hQueryParameters = {};
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $ServletForm = $self->_createServletForm($hQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService();
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    $SUT->CreateGoogleAdsAccount($Servlet);

    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");

    if (WasCalled($Servlet->form(), 'executeFormError')) {
        my $aExecuteFormErrorParams = GetParametersFromMockifyCall($Servlet->form(), 'executeFormError');

        is_deeply(
            $aExecuteFormErrorParams->[0],
            {
                'Name'   => 'CreationFailed',
                'Reason' => ERROR_ACCEPT_TERMS
            },
            $SubTestName . ' - Ensure form error not_accepted_terms gets thrown if google requirements are not met'
        );
    }
}

sub test_CreateGoogleAdsAccount_Fail_Form_No_MerchantCenter {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken => 'some-access-token',
        RefreshToken => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
    });
    my $AlreadyExistingConfig = 1;
    my $hQueryParameters = {};
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $ServletForm = $self->_createServletForm($hQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService();
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    $SUT->CreateGoogleAdsAccount($Servlet);

    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");

    if (WasCalled($Servlet->form(), 'executeFormError')) {
        my $aExecuteFormErrorParams = GetParametersFromMockifyCall($Servlet->form(), 'executeFormError');

        is_deeply(
            $aExecuteFormErrorParams->[0],
            {
                'Name'   => 'CreationFailed',
                'Reason' => ERROR_FORM_SUBMIT
            },
            $SubTestName . ' - Ensure Api error form_submit_failed gets thrown if no Merchant Center account'
        );
    }
}

sub test_CreateGoogleAdsAccount_Fail_Form_Already_Has_Ads_Account {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken        => 'some-access-token',
        RefreshToken       => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        GoogleAdsAccountID => 12342314,
    });
    my $AlreadyExistingConfig = 1;
    my $hQueryParameters = {};
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $ServletForm = $self->_createServletForm($hQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleAdsApiService();
    my $SUT = $self->_createSUT(undef, undef, $AdsApiService);

    $SUT->CreateGoogleAdsAccount($Servlet);

    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");

    if (WasCalled($Servlet->form(), 'executeFormError')) {
        my $aExecuteFormErrorParams = GetParametersFromMockifyCall($Servlet->form(), 'executeFormError');

        is_deeply(
            $aExecuteFormErrorParams->[0],
            {
                'Name'   => 'CreationFailed',
                'Reason' => ERROR_FORM_SUBMIT
            },
            $SubTestName . ' - Ensure Api error form_submit_failed gets thrown if user already has Ads account'
        );
    }
}

sub test_CreateGoogleAdsAccount_Fail_Api {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        AccountEmail                           => MERCHANT_EMAIL,
    });
    my $AlreadyExistingConfig = 1;
    my $hQueryParameters = {};
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $ServletForm = $self->_createServletForm($hQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $AdsApiService = $self->_createGoogleApiService({
        createGoogleAdsSubAccount => {
            throw_error => 1
        }
    });
    my $SUT = $self->_createSUT(undef, $self->_createError(HTTP_BAD_REQUEST, EXPECTED_TOO_MANY_SUBACCOUNTS_RESPONSE), $AdsApiService);

    $SUT->CreateGoogleAdsAccount($Servlet);

    ok(WasCalled($Servlet->form(), 'executeFormError'),
        "$SubTestName - Verify form error was executed.");

    if (WasCalled($Servlet->form(), 'executeFormError')) {
        my $aExecuteFormErrorParams = GetParametersFromMockifyCall($Servlet->form(), 'executeFormError');

        is_deeply(
            $aExecuteFormErrorParams->[0],
            {
                'Name'   => 'CreationFailed',
                'Reason' => ERROR_TOO_MANY_SUBACCOUNTS
            },
            $SubTestName . ' - Ensure Api error too_many_accounts gets thrown'
        );
    }
}

sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Subaccount_Success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        MerchantCenterMultiClientAccountID  => MULTI_CLIENT_ACCOUNT_ID,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE,
        MerchantCenterAccountStatuses => $self->_createSuccessfulMerchantCenterAccountStatusesResponse()
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);


    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );

    ok(
        WasCalled($SUT, '_getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that _getMerchantCenterAccountStatuses was called'
    );

    ok(
        WasCalled($ApiService, 'getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    my $ExpectedAccountLevelIssues = [ MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE ];

    my $aGetMerchantCenterAccountStatusesParams = GetParametersFromMockifyCall($ApiService, 'getMerchantCenterAccountStatuses');
    is_deeply(
        $aGetMerchantCenterAccountStatusesParams,
        [ MULTI_CLIENT_ACCOUNT_ID, MERCHANT_CENTER_ID ],
        "$SubTestName - Verify that getMerchantCenterAccountStatuses is called with params: MCA, merchantId if subaccount."
    );

    is_deeply($Servlet->vars('MerchantCenterAccountStatuses')->{'AccountLevelIssues'}, $ExpectedAccountLevelIssues,
        "$SubTestName - Verify that MerchantCenterAccountLevelIssues correctly set.");

    is($Servlet->vars('MerchantCenterAccountStatuses')->{'IsWebsiteClaimed'}, 1,
        "$SubTestName - Verify that IsWebsiteClaimed set as true in servlet.");

    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );
    ok( WasCalled($SUT, '_removeIgnoredIssues'), $SubTestName . ' - Verify that _removeIgnoredIssues was called' );
}

sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Standalone_Merchant_Account_Success {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        MerchantCenterMultiClientAccountID  => undef,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE,
        MerchantCenterAccountStatuses => $self->_createSuccessfulMerchantCenterAccountStatusesResponse()
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);


    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );

    ok(
        WasCalled($SUT, '_getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that _getMerchantCenterAccountStatuses was called'
    );

    ok(
        WasCalled($ApiService, 'getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    my $ExpectedAccountLevelIssues = [ MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE ];

    my $aGetMerchantCenterAccountStatusesParams = GetParametersFromMockifyCall($ApiService, 'getMerchantCenterAccountStatuses');
    is_deeply(
        $aGetMerchantCenterAccountStatusesParams,
        [ MERCHANT_CENTER_ID, MERCHANT_CENTER_ID ],
        "$SubTestName - Verify that getMerchantCenterAccountStatuses is called with params: merchantId, merchantId if NOT subaccount."
    );

    is_deeply($Servlet->vars('MerchantCenterAccountStatuses')->{'AccountLevelIssues'}, $ExpectedAccountLevelIssues,
        "$SubTestName - Verify that MerchantCenterAccountLevelIssues correctly set.");

    is($Servlet->vars('MerchantCenterAccountStatuses')->{'IsWebsiteClaimed'}, 1,
        "$SubTestName - Verify that IsWebsiteClaimed set as true in servlet.");

    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );
}

sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Unknown_Issue {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        MerchantCenterMultiClientAccountID  => MULTI_CLIENT_ACCOUNT_ID,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE,
        MerchantCenterAccountStatuses => MERCHANT_CENTER_ACCOUNT_STATUSES_UNKNOWN_ISSUE_RESPONSE
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);


    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );

    ok(
        WasCalled($SUT, '_getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that _getMerchantCenterAccountStatuses was called'
    );

    ok(
        WasCalled($ApiService, 'getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    my $ExpectedAccountLevelIssues = [ MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_UNKNOWN ];

    is_deeply($Servlet->vars('MerchantCenterAccountStatuses')->{'AccountLevelIssues'}, $ExpectedAccountLevelIssues,
        "$SubTestName - Verify that MerchantCenterAccountLevelIssues correctly set (no translationKey should be available).");

    is($Servlet->vars('MerchantCenterAccountStatuses')->{'IsWebsiteClaimed'}, 1,
        "$SubTestName - Verify that IsWebsiteClaimed set as true in servlet.");

    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );
}



sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Success_When_Ads_Account_IS_Connected {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        MerchantCenterMultiClientAccountID     => MULTI_CLIENT_ACCOUNT_ID,
        GoogleAdsAccountID                     => ADS_ACCOUNT_ID,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE,
        MerchantCenterAccountStatuses => $self->_createSuccessfulMerchantCenterAccountStatusesResponse()
    });

    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);


    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );

    ok(
        WasCalled($SUT, '_getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that _getMerchantCenterAccountStatuses was called'
    );

    ok(
        WasCalled($ApiService, 'getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    my $ExpectedAccountLevelIssues = [
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_MISSING_AD_WORDS_LINK,
        MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE,

    ];

    ok( WasCalled($SUT, '_removeIgnoredIssues'), $SubTestName . ' - Verify that _removeIgnoredIssues was called' );
    ok( WasCalled($SUT, '_isGoogleAdsAccountConnected'), $SubTestName . ' - Verify that _isGoogleAdsAccountConnected was called' );

    is_deeply($Servlet->vars('MerchantCenterAccountStatuses')->{'AccountLevelIssues'}, $ExpectedAccountLevelIssues,
        "$SubTestName - Verify missing_ad_words_link IS NOT removed when Ads account IS connected ");

    is($Servlet->vars('MerchantCenterAccountStatuses')->{'IsWebsiteClaimed'}, 1,
        "$SubTestName - Verify that IsWebsiteClaimed set as true in servlet.");

    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );

}

sub test_ViewGoogleSmartShopping_On_Getting_MerchantCenterAccount_Statuses_Success_When_No_Ads_Account {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $GoogleSmartShoppingConfig = $self->_createGoogleSmartShoppingConfig({
        AccessToken                            => 'some-access-token',
        RefreshToken                           => 'some-refresh-token',
        AreGoogleMerchantCenterRequirementsMet => 1,
        MerchantCenterAccountID                => MERCHANT_CENTER_ID,
        MerchantCenterMultiClientAccountID     => MULTI_CLIENT_ACCOUNT_ID,
    });
    my $AlreadyExistingConfig = 1;
    my $Shop = $self->_createShop($GoogleSmartShoppingConfig, $AlreadyExistingConfig);
    my $hFormQueryParameters = {};
    my $ServletForm = $self->_createServletForm($hFormQueryParameters);
    my $Servlet = $self->_createServlet($Shop, $ServletForm);
    my $ApiService = $self->_createGoogleApiService({
        UserInfo => SUCCESSFUL_USER_INFO_RESPONSE,
        MerchantCenterAccounts => SUCCESSFUL_MERCHANT_CENTER_ACCOUNTS_RESPONSE,
        MerchantCenterAccountStatuses => $self->_createSuccessfulMerchantCenterAccountStatusesResponse()
    });
    my $SUT = $self->_createSUT($ApiService);

    $SUT->UnityViewGoogleSmartShopping($Servlet);


    ok(
        WasCalled($ApiService, 'getAndSaveUserInfo'),
        $SubTestName . ' - Verify that Google account email is being retrieved'
    );

    ok(
        WasCalled($SUT, '_getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that _getMerchantCenterAccountStatuses was called'
    );

    ok(
        WasCalled($ApiService, 'getMerchantCenterAccountStatuses'),
        $SubTestName . ' - Verify that getMerchantCenterAccounts was called when AccountInfo call is successful'
    );
    my $ExpectedAccountLevelIssues = [ MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE ];

    ok( WasCalled($SUT, '_removeIgnoredIssues'), $SubTestName . ' - Verify that _removeIgnoredIssues was called' );
    ok( WasCalled($SUT, '_isGoogleAdsAccountConnected'), $SubTestName . ' - Verify that _isGoogleAdsAccountConnected was called' );
    is_deeply($Servlet->vars('MerchantCenterAccountStatuses')->{'AccountLevelIssues'}, $ExpectedAccountLevelIssues,
        "$SubTestName - Verify missing_ad_words_link is removed when Ads account is not connected ");

    is($Servlet->vars('MerchantCenterAccountStatuses')->{'IsWebsiteClaimed'}, 1,
        "$SubTestName - Verify that IsWebsiteClaimed set as true in servlet.");

    ok( !WasCalled($SUT, '_buildConnectOAuthURL'), $SubTestName . ' - Verify that OAuth connect URL was NOT built' );

}

#----------------------------------------------------------------------------------------
# Mockery
#----------------------------------------------------------------------------------------
sub _createSUT {
    my $self = shift;
    my ($ApiService, $MockedError, $AdsApiService) = @_;

    my $MockifySut = Test::Mockify::Sut->new('DE_EPAGES::GoogleSmartShopping::UI::Shop');
    $MockifySut->mock('_buildConnectOAuthURL')
        ->when(Object('DE_EPAGES::Presentation::Servlet::Permission'), Object('DE_EPAGES::Shop::API::Object::Shop'))
        ->thenReturn(CONNECT_OAUTH_URL);
    $MockifySut->mock('_writeHTTPResponse')
        ->whenAny()
        ->thenReturnUndef();

    $MockifySut->mock('_getGoogleAdsAccounts')
        ->when(Object())
        ->thenReturn(SUCCESSFUL_GOOGLE_ADS_ACCOUNTS_RESPONSE);
    $MockifySut->mockImported('DE_EPAGES::GoogleSmartShopping::API::Tools', 'LogGoogleSmartShopping')
        ->whenAny()
        ->thenReturnUndef();
    if (defined($MockedError)) {
        $MockifySut->mockImported('DE_EPAGES::Core::API::Error', 'GetError')
            ->whenAny()
            ->thenReturn($MockedError);
    }
    $MockifySut->spy('_getMerchantCenterAccountStatuses')->whenAny();
    $MockifySut->spy('_removeIgnoredIssues')->whenAny();
    $MockifySut->spy('_isGoogleAdsAccountConnected')->whenAny();

    if (defined($ApiService)) {
        $MockifySut->mockStatic('DE_EPAGES::GoogleSmartShopping::API::GoogleApiService::new')
            ->when(HashRef())
            ->thenReturn($ApiService);
    }

    if (defined($AdsApiService)) {
        $MockifySut->mockStatic('DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService::new')
            ->when(HashRef())
            ->thenReturn($AdsApiService);
    }

    return $MockifySut->getMockObject();
}

sub _createGoogleSmartShoppingConfig {
    my $self = shift;
    my ($hConfig) = @_;

    $hConfig->{'AreGoogleMerchantCenterRequirementsMet'} //= 0;

    my $hAttributes = {
        AccountEmail               => $hConfig->{'AccountEmail'} // undef,
        AccessToken                => $hConfig->{'AccessToken'},
        RefreshToken               => $hConfig->{'RefreshToken'},
        ChecklistBillingTOC        => $hConfig->{'AreGoogleMerchantCenterRequirementsMet'},
        ChecklistContactInfo       => $hConfig->{'AreGoogleMerchantCenterRequirementsMet'},
        ChecklistGoogleAdsPolicies => $hConfig->{'AreGoogleMerchantCenterRequirementsMet'},
        ChecklistCookieNotice      => $hConfig->{'AreGoogleMerchantCenterRequirementsMet'},
        ChecklistReturnPolicy      => $hConfig->{'AreGoogleMerchantCenterRequirementsMet'},
        MerchantCenterAccountName => $hConfig->{'MerchantCenterAccountName'},
        MerchantCenterAccountID => $hConfig->{'MerchantCenterAccountID'},
        IsMerchantCenterSubAccount => $hConfig->{'IsMerchantCenterSubAccount'},
        MerchantCenterMultiClientAccountID => $hConfig->{'MerchantCenterMultiClientAccountID'},
        GoogleAdsAccountID => $hConfig->{'GoogleAdsAccountID'} // undef ,
        GoogleAdsAccountName => $hConfig->{'GoogleAdsAccountName'} // undef ,
        IsAdsAccountSubAccount => $hConfig->{'IsAdsAccountSubAccount'} // undef ,
        AdsCampaignId => $hConfig->{'AdsCampaignId'} // undef ,
    };
    my $MockifyConfig = DE_EPAGES::Test::Mock::MockifyEpagesObject->new(
        'DE_EPAGES::GoogleSmartShopping::API::Object::GoogleSmartShoppingConfig', $hAttributes
    );
    $MockifyConfig->addMockWithReturnValueAndParameterCheck('set', undef, [ 'hashref' ]);

    $MockifyConfig->addMockWithReturnValue('delete', undef);
    $MockifyConfig->addMockWithReturnValue('deleteMerchantCenterAccount', undef);
    $MockifyConfig->addMockWithReturnValue('deleteAdsAccount', undef);
    $MockifyConfig->addMockWithReturnValue('deleteAdsCampaign', undef);

    return $MockifyConfig->getMockObject();
}

sub _createGoogleApiService {
    my $self = shift;
    my ($hServiceResponse) = @_;

    my $Mockify = Test::Mockify->new('DE_EPAGES::GoogleSmartShopping::API::GoogleApiService');
    if ($hServiceResponse->{'OAuthTokens'}->{'throw_error'}) {
        $Mockify->mock('getOAuthTokens')
            ->when(String())
            ->thenThrowError('oauth_failed');
    } else {
        $Mockify->mock('getOAuthTokens')
            ->when(String())
            ->thenReturn($hServiceResponse->{'OAuthTokens'} // {});
    }
    $Mockify->mock('getAndSaveUserInfo')
        ->when()
        ->thenReturn($hServiceResponse->{'UserInfo'} // {});
    if ($hServiceResponse->{'revokeAccessToken'}->{'throw_error'}) {
        $Mockify->mock('revokeAccessToken')
            ->when(String())
            ->thenThrowError('lol');
    } else {
        $Mockify->mock('revokeAccessToken')
            ->when(String())
            ->thenReturn($hServiceResponse // {});
    }
    $Mockify->mock('getMerchantCenterAccounts')
        ->when()
        ->thenReturn($hServiceResponse->{'MerchantCenterAccounts'} // {});
    $Mockify->mock('getMerchantCenterAccountStatuses')
        ->when(Number(), Number())
        ->thenReturn($hServiceResponse->{'MerchantCenterAccountStatuses'} // {});

    if ($hServiceResponse->{'deleteMerchantCenterSubAccount'}->{'throw_error'}) {
        $Mockify->mock('deleteMerchantCenterSubAccount')
            ->when(Number(ACCOUNT_ID))
            ->thenThrowError('access_denied');
    } else {
        $Mockify->mock('deleteMerchantCenterSubAccount')
            ->when(Number(ACCOUNT_ID))
            ->thenReturnUndef();
    }

    if ($hServiceResponse->{'createMerchantCenterSubAccount'}->{'throw_error'}) {
        $Mockify->mock('createMerchantCenterSubAccount')
            ->when(Number())
            ->thenThrowError('throw!');
    } else {
        $Mockify->mock('createMerchantCenterSubAccount')
            ->when(Number())
            ->thenReturn($hServiceResponse->{'createMerchantCenterSubAccount'} // {});
    }

    return $Mockify->getMockObject();
}

sub _createGoogleAdsApiService {
    my $self = shift;
    my ($hServiceResponse) = @_;

    my $Mockify = Test::Mockify->new('DE_EPAGES::GoogleSmartShopping::API::GoogleAdsApiService');

    if ($hServiceResponse->{'isAdsAccountBillingApproved'}->{'throw_error'}) {
        $Mockify->mock('isAdsAccountBillingApproved')
            ->when()
            ->thenThrowError('access_denied');
    } elsif (defined $hServiceResponse->{'isAdsAccountBillingApproved'}->{'thenReturn'}) {
        $Mockify->mock('isAdsAccountBillingApproved')
            ->when()
            ->thenReturn($hServiceResponse->{'isAdsAccountBillingApproved'}->{'thenReturn'});
    }

    if ($hServiceResponse->{'createGoogleAdsSubAccount'}->{'throw_error'}) {
        $Mockify->mock('createGoogleAdsSubAccount')
            ->whenAny()
            ->thenThrowError('access_denied');
    } else {
        $Mockify->mock('createGoogleAdsSubAccount')
            ->whenAny()
            ->thenReturnUndef();
    }

    if ($hServiceResponse->{'createAndEnableGoogleAdsLink'}->{'throw_error'}) {
        $Mockify->mock('createAndEnableGoogleAdsLink')
            ->whenAny()
            ->thenThrowError('access_denied');
    } else {
        $Mockify->mock('createAndEnableGoogleAdsLink')
            ->whenAny()
            ->thenReturnUndef();
    }

    if ($hServiceResponse->{'unlinkGoogleAdsAccount'}->{'throw_error'}) {
        $Mockify->mock('unlinkGoogleAdsAccount')
            ->whenAny()
            ->thenThrowError('access_denied');
    } else {
        $Mockify->mock('unlinkGoogleAdsAccount')
            ->whenAny()
            ->thenReturnUndef();
    }

    if ($hServiceResponse->{'deleteAdsCampaign'}->{'throw_error'}) {
        $Mockify->mock('deleteAdsCampaign')
            ->whenAny()
            ->thenThrowError('access_denied');
    } else {
        $Mockify->mock('deleteAdsCampaign')
            ->whenAny()
            ->thenReturnUndef();
    }

    return $Mockify->getMockObject();
}

sub _createError {
    my $self = shift;
    my ($Status, $hResponse) = @_;

    my $Error = DE_EPAGES::Core::API::Error->new(
        'Code' => 'Error',
        'Vars' => {
            'Status' => $Status,
            'Response' => encode_json($hResponse)
        }
    );
    return $Error;
}

sub _createServlet {
    my $self = shift;
    my ($Shop, $ServletForm) = @_;

    my $MockifyServlet = DE_EPAGES::Test::Mock::MockifyServlet->new($Shop);
    $MockifyServlet->addMockWithReturnValue('form', $ServletForm);

    return $MockifyServlet->getMockObject();
}

sub _createShop {
    my $self = shift;
    my ($GoogleSmartShoppingConfig, $AlreadyExistingConfig) = @_;

    my $hAttributes = {
        'ID' => undef,
        'GoogleSiteVerificationToken' => 'VerificationToken',
        'GoogleSmartShoppingConfig' => $GoogleSmartShoppingConfig
    };
    my $MockifyShop = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', $hAttributes);
    $MockifyShop->addMockWithReturnValueAndParameterCheck(
        'existsChild', $AlreadyExistingConfig, [ {'string' => 'GoogleSmartShoppingConfig'} ]
    );
    $MockifyShop->addMockWithReturnValueAndParameterCheck(
        'insertChild', $GoogleSmartShoppingConfig, [ {'string' => 'GoogleSmartShoppingConfig'}, 'hashref' ]
    );

    return $MockifyShop->getMockObject();
}

sub _createServletForm {
    my $self = shift;
    my ($hQueryParameters) = @_;

    my $Mockify = Test::Mockify->new( 'DE_EPAGES::Presentation::API::Form');
    $Mockify->mock('queryValues')
        ->when()
        ->thenReturn($hQueryParameters);

    $Mockify->mock('formValues')
        ->when()
        ->thenReturn($hQueryParameters);


    $Mockify->mock('executeFormError')
        ->when()
        ->thenReturnUndef();

    $Mockify->mock('executeFormError')
        ->when(HashRef())
        ->thenReturnUndef();

    $Mockify->mock('addFormError')
        ->when(HashRef())
        ->thenReturnUndef();

    $Mockify->mock('existsFormError')
        ->when()
        ->thenReturnUndef();

    return $Mockify->getMockObject();
}

sub _createSuccessfulMerchantCenterAccountStatusesResponse {
    my $self = shift;

    return {
        "accountId"=> "414922848",
        "accountLevelIssues"=> [
            MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_QUALITY_LOW,
            MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_MISSING_AD_WORDS_LINK,
            MERCHANT_CENTER_ACCOUNT_LEVEL_ISSUE_PENDING_ADDRESS_AND_PHONE
        ],
        "kind"=> "content#accountStatus",
        "websiteClaimed"=> "true",
        "accountManagement"=> "manual"
    };
}

__PACKAGE__->runScript();

