#========================================================================================
# package      DE_EPAGES::Installer::API::PatchBuilder
# state        public
# description  This modul generates and change all files necessary for patch.
#               These files can be find under ../Patches of each cartridge.
#               The patch change the file ../API/Cartridge.pm in each cartridge.
#               Each patch increase the version number of each cartridge, even if there are no changes.
#               So the ePages version number and the cartridge version numbers are linked.
#
#========================================================================================
package DE_EPAGES::Installer::API::PatchBuilder;

use strict;
use feature 'switch';

use Config;
use Getopt::Long;
use Data::Dumper;
use DE_EPAGES::Core::API::CVS;
use DE_EPAGES::Core::API::File qw (
    CopyFile
    CreateDirectory
    DeleteDirectory
    DeleteFile
    DetectFileEncoding
    ExistsDirectory
    ExistsFile
    GetFileContent
    GetFileMD5Digest
    GetFileNames
    GetSubDirectories
    RunOnDirectory
    WriteFile);
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Installer::API::CVSTagHandler;
use DE_EPAGES::Installer::API::Object::PatchCodeGenerator;
use DE_EPAGES::Installer::API::XMLFilePatcher;
use DE_EPAGES::Installer::API::PatchXMLExportDriver;
use DE_EPAGES::Object::API::Factory qw(SplitObjectPath);
use DE_EPAGES::Object::API::XMLCompare qw (CompareXML);
use DE_EPAGES::Object::API::XMLStructureExportDriver;

## no critic qw(RequireExtendedFormatting ProhibitLongLines)

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder = DE_EPAGES::Installer::API::PatchBuilder->new(
#                   'OldReleaseTag' => $OldReleaseTag,
#                   'NewReleaseTag' => $NewReleaseTag,
#                   'Verbose' => 1,
#                   'ForceReview' => 1,
#                   'DontExportDeletedDirectories' => 0,
#               );
#----------------------------------------------------------------------------------------
# description  Creates a new PatchBuilder object
#----------------------------------------------------------------------------------------
# input        $hOptions | keys
#                   <ul>
#                       <li>OldReleaseTag - old release tag, e.g. 'release_6_9_1'
#                       <li>NewReleaseTag - new release tag, e.g. 'release_6_9_2'
#                       <li>Verbose       - print debug information (0 or 1)
#                       <li>ForceReview   - add dies to Cartridge.pm to force manual check
#                       <li>DontExportDeletedDirectories - set to 1 if deleted cartridge
#                           directories should not exported, e.g. for language pack patch
#                       <li>reconstruct   - 1 to reconstruct from CVS
#                                           ignore already existing patch files
#                       <li>CommitBranch - name of branch to commit
#                       <li>AutoMode - PatchBuilder operates in HEAD Branch
#                       <li>UseGeneratedBranch - ony allowed in combination with AutoMode, operates in branches HEAD and B_generated_patch_code
#                   </ul>
#               | hash
# return       $PatchBuilder | PatchBuilder object | object
#========================================================================================
sub new {
    my $class = shift;
    my ($hOptions) = @_;
    TestString( 'OldReleaseTag', $hOptions->{'OldReleaseTag'} );
    TestString( 'NewReleaseTag', $hOptions->{'NewReleaseTag'} );
    die "\nmissing option 'AutoMode'" unless (defined $hOptions->{'AutoMode'});
    die "\n'UseGeneratedBranch' only allowed with option 'AutoMode'" if (
        (defined $hOptions->{'UseGeneratedBranch'}) and
        ($hOptions->{'UseGeneratedBranch'}) and
        not ($hOptions->{'AutoMode'})
    );
    my $self = bless $hOptions, $class;

    $self->{'CompanyName'} = 'DE_EPAGES' if (not defined($self->{'CompanyName'}));
    $self->{'PatchStructure'} = {};

    #set old tag
    $self->{'OldReleaseOrHotFixTag'} = $self->{'OldReleaseTag'};

    return $self;
}

#========================================================================================
# function     exportReleaseFolders
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportReleaseFolders($CVSPath, $OldDir, $NewDir);
#----------------------------------------------------------------------------------------
# description  step 1
#               export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $CVSPath | cvs path from cartridge that should be patcht | string
# input        $OldDir | directory to export cartridge version tagged with old release tag | string
# input        $NewDir | directory to export cartridge version tagged with new release tag | string
#========================================================================================
sub exportReleaseFolders {
    my $self = shift;
    my ($CVSPath, $OldDir, $NewDir) = @_;

    my $OriginalOldReleaseOrHotFixTag = $self->{'OldReleaseOrHotFixTag'};

    #print message
    $self->printMessage( "\t>step-1: export old and new version of cartridge to\n\told: $OldDir\n\tnew: $NewDir" );

    #check if to use hot fix tag
    if ($self->{'UseOldHotFixTag'}){

        #check for hotfix tag
        my $CVSTagHandler = $self->DE_EPAGES::Installer::API::CVSTagHandler->new();

        #print message
        my $OldTag = $self->{'OldReleaseOrHotFixTag'};
        $self->printMessage( "\t>WARNING: changed old target tag from '$OriginalOldReleaseOrHotFixTag' to '$OldTag'" );

        #set tag to build from
        $self->{'OldReleaseOrHotFixTag'} = $CVSTagHandler->getLatestCorrespondingHotFixTag(
            $CVSPath . '/API/Cartridge.pm',
            $self->{'OldReleaseTag'}
        );
    }

    #export cartridge versions tagged with old and new release
    $self->exportReleaseFolder($CVSPath, $OldDir, $self->{'OldReleaseOrHotFixTag'});
    $self->exportReleaseFolder($CVSPath, $NewDir, $self->{'NewReleaseTag'});

    return;
}

#========================================================================================
# function     generatePatchFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->generatePatchFiles($OldFolder, $NewFolder, $CartridgeFolder);
#----------------------------------------------------------------------------------------
# description  step 2
#               - compare old and new version of cartridge
#               - generates all patch files under ../Patches/Patch../Database/XML
#               - (if patch files already builded, only additional tags where added
#                  manually changes with stay)
#----------------------------------------------------------------------------------------
# input        $OldFolder | folder of cartridgeversion tagged with old release tag | string
# input        $NewFolder | folder of cartridgeversion tagged with new release tag | string
# input        $CartridgeFolder | folder of cartridge, where should commit | string
# input        $useGit | 1, if you use the Git version, 0 otherwise | bool
#========================================================================================
sub generatePatchFiles {
    my $self = shift;
    my ($OldFolder, $NewFolder, $CartridgeFolder, $useGit) = @_;

    $self->printMessage("\t##### Generate patch files #####");

    #get name of folder to save patch files
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);
    my $CommitBranch = $self->{'CommitBranch'};

    #Git replacement does not seem necessary at this point
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    #create directory unless exists
    CreateDirectory($TargetFolder) unless (ExistsDirectory($TargetFolder));

    if(!$useGit) {
        #add target folder to cvs
        RunOnDirectory(
            $CartridgeFolder,
            sub {
                my $RelativeTargetName = $TargetFolder;
                $RelativeTargetName =~ s,\Q$CartridgeFolder\E/,,;

                #add folder Patches
                $CVS->command('add', undef, ['Patches'], {}, {'skiperrors' => 1});

                #add target folder
                $CVS->command('add', undef, [$RelativeTargetName], {}, {'skiperrors' => 1});
            }
        );

        #update target folder if commit branch is set
        RunOnDirectory(
            $TargetFolder,
            sub {
                if (
                    defined $CommitBranch and
                    $CommitBranch ne ''
                ){
                    #message
                    $self->printMessage("\tCommit Branch selected, updating file to branch '$CommitBranch':\n\t$TargetFolder");

                    #update Cartridge.pm to special commit branch
                    $CVS->command( 'update', { '-d', '-P', '-r', $CommitBranch }, undef ) ;
                }
            }
        );
    }

    #remove all previous patch files if reconstruct is enabled
    $self->_cleanPreviousPatchFiles($TargetFolder) if ($self->{'reconstruct'});

    #print Message
    $self->printMessage("\t>step-2: collect information from old and new version of cartridge, save patch files in:\n\t$TargetFolder");

    #handle files that changed from old to new release tag
    $self->_handleChangedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle files that only can find in new cartridge version
    $self->_handleAddedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle files that only can find in old cartridge version
    $self->_handleDeletedFiles($CartridgeFolder, $OldFolder, $NewFolder);

    #handle directories that only can find in old cartridge version
    $self->_handleDeletedDirectories($CartridgeFolder, $OldFolder, $NewFolder)
    unless (exists $self->{'DontExportDeletedDirectories'} and $self->{'DontExportDeletedDirectories'} == 1);
}

sub _cleanPreviousPatchFiles {
    my $self = shift;
    my ($Folder) = @_;

    if(ExistsDirectory($Folder)){
        foreach my $File (@{GetFileNames($Folder, {'recursive' => 1})}){
            #ignore files that ignored by cvs
            next if ($File =~ /\/CVS\//);
            print "\n  delete file '$File'";
            DeleteFile($File) if ExistsFile($File);
            print "\n";
        }
    }
}

#========================================================================================
# function     createPatchXMLFromPatchFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->createPatchXMLFromPatchFiles($CartridgePath, $CartridgePackage, $BaseName );
#----------------------------------------------------------------------------------------
# description  step 3
#               creates the patch.xml file from the files in the patch directory.
#----------------------------------------------------------------------------------------
# input        $CartridgePath | target directory for the cartridge | string
# input        $CartridgePackage | cartridge package, e.g. 'DE_EPAGES::Object' | string
# input        $BaseName | (optional; default: 'patch.xml') output file name | string
# return       $Created | true if an XML file was created | boolean
#========================================================================================
sub createPatchXMLFromPatchFiles{
    my $self = shift;
    my ($CartridgePath, $CartridgePackage, $BaseName) = @_;

    #generate names
    $BaseName =  'patch.xml' unless defined $BaseName;
    my $CartridgePatchPath = _GetPatchTargetName($CartridgePath, $self->oldVersion);
    my $FileName = $CartridgePatchPath .'/'. $BaseName;
    my $CartridgePatchDBXMLPath = $CartridgePatchPath . '/Database/XML';

    #print message
    $self->printMessage( "\t>step-3: create or update:\n\t$CartridgePatchPath/patch.xml" );

    #save information in hash structure, necessary for export patch.xml
    $self->{'PatchStructure'}->{'PatchCartridge'} = {
            'OldRelease' => $self->{'OldReleaseOrHotFixTag'},
            'NewRelease' => $self->{'NewReleaseTag'},
            'NewVersion' => $self->newVersion,
            'Cartridge'  => $CartridgePackage,
            'OldVersion' => $self->oldVersion,
    };

    #check if patch folder exists
    if (ExistsDirectory($CartridgePatchDBXMLPath)){

        #for each file in patch folder
        foreach my $DBXMLFile (@{GetFileNames($CartridgePatchDBXMLPath)}){
            #ignore files that ignored by cvs
            next if ($DBXMLFile =~ /\/\.\#[\w.]*/);

            my $DBXMLFilePatchPath = $DBXMLFile;
            $DBXMLFile =~ s,^\Q$CartridgePatchDBXMLPath\E/,,;
            $DBXMLFilePatchPath =~ s,^\Q$CartridgePath\E/,,;

            #save information in hash structure, necessary for export patch.xml
            $self->_collectPatchXMLInfoFromFiles($DBXMLFile, $DBXMLFilePatchPath);
        }
    }
    else {
        #no patch folder found, no db content changes have to add to patch.xml
        $self->printMessage("\t(no patch folder found for $CartridgePath)");
    }

    #stop if nothing is to do
    return 0 if (
        (#$self->{'PatchStructure'}->{'PatchDBStructure'} doesnt exists
            (not exists $self->{'PatchStructure'}->{'PatchDBStructure'}) or
            (#only warnings exists in $self->{'PatchStructure'}->{'PatchDBStructure'}
                exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'} and
                scalar keys %{$self->{'PatchStructure'}->{'PatchDBStructure'}} == 1
            )
        ) and
        (not exists $self->{'PatchStructure'}->{'Patch'}) and
        (not exists $self->{'PatchStructure'}->{'PatchCartridgeDir'})
    );

    #export patch.xml
    my $PatchExportDriver = DE_EPAGES::Installer::API::PatchXMLExportDriver->new($self->{'PatchStructure'});
    if (not ExistsFile($FileName)){#patch.xml doesnt exists, patcht first time
        $PatchExportDriver->exportXML($FileName);
    }
    else{#an older version of patch.xml file already exists
        #generate temporary new name
        my $NewFileName = $FileName;
        $NewFileName =~ s/(\Q$BaseName\E)$/\.\#\(new\)$1/;

        #export to temporary file
        $PatchExportDriver->exportXML($NewFileName);

        if (CompareXML($NewFileName, $FileName) and CompareXML($FileName,$NewFileName)){
            #no changes detected - delete the new file
            DeleteFile($NewFileName);
        }
        elsif (CompareXML($NewFileName, $FileName)){#new patch.xml has no additional tags or changed values, nothing to do
            print "    [Warning:] difference in patch.xml file detected, please check manually\n";
            print "       -original $FileName\n";
            print"        -new      $NewFileName\n";
        }
        else{#new patch.xml has additional tags or attributes
            my $OldFileName = $FileName;
            $OldFileName =~ s/(\Q$BaseName\E)$/\.\#\(old\)$1/;

            #get xml structure of old and new patch.xml files
            my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
            my $OrigXMLStruct = $ImportHandler->getXMLStructure($FileName);
            my $AddXMLStruct = $ImportHandler->getXMLStructure($NewFileName);

            #merge new and old patch.xml, positions could be destroyed, it is necessary to correct the positions
            $OrigXMLStruct->assimilateXMLStructure($AddXMLStruct,{'overwrite' => '1'});
            _CorrectPositionsOfPatchXMLXMLStruct($OrigXMLStruct);

            #save old file
            CopyFile($FileName,$OldFileName);
            DeleteFile($NewFileName);


            #export combined patch.xml
            my $ExportDriver = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
            $ExportDriver->exportXML($FileName, $OrigXMLStruct);
            print "    [Warning:] difference in patch.xml file detected, please check manually\n";
            print "       -original $OldFileName\n";
            print "       -new      $FileName\n";
        }
    }

    return 1;
}

#========================================================================================
# function     patchCartridgePm
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->patchCartridgePm($CartridgePath)
#----------------------------------------------------------------------------------------
# description  step 4
#               patch the ../API/Cartridge.pm file
#----------------------------------------------------------------------------------------
# input        $CartridgePath | target directory for the cartridge | string
# input        $GenerateWarnings | if 0 no warnings will generate in Cartridge.pm | bool
# input        $useGit | 1, if Git version is used, 0 otherwise | bool
#========================================================================================
sub patchCartridgePm{
    my $self = shift;
    my ($CartridgePath,$GenerateWarnings,$useGit) = @_;

    #get Info
    my $CommitBranch = $self->{'CommitBranch'};
    my $CartridgePmPath = $CartridgePath . '/API/Cartridge.pm';
    my $sCartridgePm = GetFileContent($CartridgePmPath, {'usebom' => 1 }); #use byte order mark
    my $Encoding = DetectFileEncoding($CartridgePmPath);
    my $OriginalContent = $$sCartridgePm;
    my $OldVersion = $self->oldVersion;
    my $NewVersion = $self->newVersion;
    my $OldVersionUnderscored = $OldVersion;
    my $NewVersionUnderscored = $NewVersion;
    my $CodeGenerator = DE_EPAGES::Installer::API::Object::PatchCodeGenerator->new();
    my $ControlCode = $CodeGenerator->getControlCode();
    $OldVersionUnderscored =~ s/\./_/g;
    $NewVersionUnderscored =~ s/\./_/g;


    if(!$useGit) {
        if (
            defined $CommitBranch and
            $CommitBranch ne ''
        ){
            #message
            $self->printMessage("\tCommit branch selected, updating file to branch '$CommitBranch':\n\t$CartridgePath/API/Cartridge.pm");

            #update Cartridge.pm to special commit branch
            $self->cleanUpdateToBranch($CartridgePath . '/API', $CommitBranch, 'Cartridge.pm');
        }
    }

    $GenerateWarnings = 1 unless defined $GenerateWarnings;

    #print message
    $self->printMessage("\t>step-4: patch $CartridgePmPath");

    #paste control code
    $$sCartridgePm =~ s{(
        \#[\=]+\n                 #comment line
        \#\s\function\s+new   #'# function     new'
    )}
    {$ControlCode$1}x if ($self->{'ForceReview'} eq '1');

    #register versions
    if (not $$sCartridgePm =~ /\s+\'Version\'\s+\=\>\s\'\Q$NewVersion\E/){
        #update Version
        $$sCartridgePm =~ s/(\s+\'Version\'\s+\=\>\s\')[0-9.]+(\'\,)/$1$NewVersion$2/;
    }
    if (not $$sCartridgePm =~ /\s+\'Patches\'\s+\=\>\s\[[0-9.\,\']*\Q$OldVersion\E/){
        #register old version
        $$sCartridgePm =~ s/(\s+\'Patches\'\s+\=\>\s\[[0-9.\,\']+)(\]\,)/$1\,\'$OldVersion\'$2/;
        $$sCartridgePm =~ s/(\s+\'Patches\'\s+\=\>\s\[)(\])/$1\'$OldVersion\'$2/;
    }

    #check for changes
    if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}){
        my $ExecutePatchXML = '';

        #create warning
        my $PatchBuilderWarning = '';
        if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}){

            if ($GenerateWarnings){
                foreach my $DontKnowWhatToDoFile (keys %{$self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}}){

                    if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} and
                        $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} eq 'Changed')
                    {
                        $PatchBuilderWarning = $PatchBuilderWarning . "    die \"changes in $DontKnowWhatToDoFile cant automate at these time.\"\;\n";
                    }
                    elsif (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} and
                        $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}->{'CVS'} eq 'Deleted')
                    {
                        $PatchBuilderWarning = $PatchBuilderWarning . "    die \"delete $DontKnowWhatToDoFile cant automate at these time.\"\;\n";
                    }
                    else
                    {
                        die 'cant generate warning for constalation warning ' .
                            $DontKnowWhatToDoFile . ' -> ' .
                            Dumper($self->{'PatchStructure'}->{'PatchDBStructure'}->
                                {'PatchBuilder-Warning'}->{$DontKnowWhatToDoFile}) . "\n";
                    }
                }
            }
        }

        #create code generator object and placeholder
        my $DBStructureChangeCode = '';

        #check if patch.xml should be executed
        if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}){
            #define function call to execute patch.xml
            $ExecutePatchXML = "\n    \$self->importPatchDir('Patch_$OldVersionUnderscored', 'PatchDBStructure');\n";
        }

        #generate code to handle changes in sql files
        if (scalar keys %{$self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'mysql'}} >= 1){
            $DBStructureChangeCode .= $CodeGenerator->getCodeForDiffInSQLFiles($self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'mysql'}, 'mysql');
        }
        $DBStructureChangeCode =~ s/\s+my\s+\$dbi\s+\=\s+GetCurrentDBHandle\(\)\;/FIRST_DBI_POSITION/;
        $DBStructureChangeCode =~ s/\s+my\s+\$dbi\s+\=\s+GetCurrentDBHandle\(\)\;//g;
        $DBStructureChangeCode =~ s/FIRST_DBI_POSITION/    my \$dbi \= GetCurrentDBHandle\(\)\;\n/;

        my $FunctionHeader = "\n#========================================================================================\n" .
                "# function     patchDBStructure_$OldVersionUnderscored\n" .
                "# state        private\n" .
                "#----------------------------------------------------------------------------------------\n" .
                "# syntax       \$Installer->patchDBStructure_$OldVersionUnderscored;\n" .
                "#----------------------------------------------------------------------------------------\n" .
                "# description  Patches cartridge db structure in active store (database connection)\n" .
                "#               from version '$OldVersion' to version '$NewVersion'.\n" .
                "#----------------------------------------------------------------------------------------\n" .
                "# return       \$Version | version of cartridge after patch function, '$NewVersion' | string\n" .
                "#========================================================================================\n";

        #check for rebuild from CVS
        if ($self->{'reconstruct'}){
            #remove old patchDBStructure function
            $$sCartridgePm =~ s/\Q$FunctionHeader\E//g;
            $$sCartridgePm =~ s/sub\spatchDBStructure_\Q$OldVersionUnderscored\E\s\{.*?return\s\'\Q$NewVersion\E\'\;.*?\}\n//s;
        }

        if (not $$sCartridgePm =~ /sub\spatchDBStructure_\Q$OldVersionUnderscored\E\s\{/){
            #define function
            my $FunctionCode =
                $FunctionHeader .
                "sub patchDBStructure_$OldVersionUnderscored {\n" .
                "    my \$self = shift;\n" .
                "    #dont use API function, to avoid inconistencies between perl modules and database content\n" .
                "$PatchBuilderWarning\n" .
                "$DBStructureChangeCode\n" .
                "$ExecutePatchXML\n" .
                "    return '$NewVersion';\n" .
                "}\n";

            #write function to Cartridge.pm file content
            $$sCartridgePm =~ s/(\n1\;)/$FunctionCode$1/ if (exists $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'});
        }
    }

    #generate options for encoding
    my $hFlags = undef;
    if(defined $Encoding){
        $hFlags = { 'usebom' => 1, 'encoding' => $Encoding };
    }

    #write Cartridge.pm file
    WriteFile($CartridgePmPath,$sCartridgePm, $hFlags) if (
        ($OriginalContent ne $$sCartridgePm)

        and not(
            #exists in original Cartridge.pm
            ($OriginalContent =~ /sub\spatchDBStructure_\Q$OldVersionUnderscored\E\s\{/)
            and
            #not exists in generated new Cartridge.pm
            (not($$sCartridgePm =~ /sub\spatchDBStructure_\Q$OldVersionUnderscored\E\s\{/))
        )
    );
}

#========================================================================================
# function     cleanUpdateToBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->cleanUpdateToBranch($Path,$Branch)
#----------------------------------------------------------------------------------------
# description  clean update to branch or tag
#----------------------------------------------------------------------------------------
# input        $Path | path to update | string
# input        $Branch | branch to update from | string
# input        $Part | files or folders | string
#========================================================================================
sub cleanUpdateToBranch{
    my $self = shift;
    my ($Path,$Branch,$Part) = @_;

    #get cvs handler
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    RunOnDirectory(
        $Path,
        sub {

            #message
            $self->printMessage("\tupdating folder to branch '$Branch':\n\t$Path");

            #update Cartridge.pm to special commit branch
            $CVS->command( 'update', { '-d', '-C', '-P', '-r', $Branch }, [$Part] ) ;
        }
    );
}

#========================================================================================
# function     updateToBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->updateToBranch($Path,$Branch)
#----------------------------------------------------------------------------------------
# description  update to branch or tag
#----------------------------------------------------------------------------------------
# input        $Path | path to update | string
# input        $Branch | branch to update from | string
# input        $Part | files or folders | string
#========================================================================================
sub updateToBranch{
    my $self = shift;
    my ($Path,$Branch,$Part) = @_;

    #get cvs handler
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    RunOnDirectory(
        $Path,
        sub {

            #message
            $self->printMessage("\nupdating folder to branch '$Branch'\n:$Path");

            #update Cartridge.pm to special commit branch
            $CVS->command( 'update', { '-d', '-P', '-r', $Branch }, [$Part] ) ;
        }
    );
}

#========================================================================================
# function     setCommitBranch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->setCommitBranch($Branch)
#----------------------------------------------------------------------------------------
# description  actualize the commit branch
#----------------------------------------------------------------------------------------
# input        $Branch | branch to commit | string
#========================================================================================
sub setCommitBranch {
    my $self = shift;
    my ($CommitBranch) = @_;

    $self->{'CommitBranch'} = $CommitBranch;

    return;
}

#========================================================================================
# function     addPatchFolderToCVS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->addPatchFolderToCVS($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 5
#               add files under ../Patches to cvs
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub addPatchFolderToCVS {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    #get info
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    if (ExistsDirectory($TargetFolder)){
        RunOnDirectory(
            $CartridgeFolder,
            sub {


                my $RelativeTargetName = $TargetFolder;
                $RelativeTargetName =~ s,\Q$CartridgeFolder\E/,,;


                $CVS->command('add', undef, [$RelativeTargetName], {}, {'skiperrors' => 1});

                foreach my $Folder (@{GetSubDirectories($TargetFolder, {'recursive' => 1})}){
                    next if ($Folder =~ /\/CVS/);
                    my $RelativeName = $Folder;
                    $RelativeName =~ s,\Q$CartridgeFolder\E/,,;

                    $CVS->command('add', undef, [$RelativeName], {}, {'skiperrors' => 1});
                }

                foreach my $File (@{GetFileNames($TargetFolder, {'recursive' => 1})}){
                    next if ($File =~ /\/CVS\//);
                    next if ($File =~ /\.\#/);
                    my $RelativeFileName = $File;
                    $RelativeFileName =~ s,\Q$CartridgeFolder\E/,,;

                    $CVS->command('add', undef, [$RelativeFileName], {}, {'skiperrors' => 1});
                    $CVS->command('update', undef, [$RelativeFileName], {}, {'skiperrors' => 1});
                }
                return;
            }
        );
    }
    return;
}

sub _getPatchTag{
    my $self = shift;
    my ($ReleaseTag) = @_;

    $ReleaseTag =~ s/release/patch/;

    return $ReleaseTag;
}

#========================================================================================
# function     commitPatchFolderToCVS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->commitPatchFolderToCVS($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 5
#               add files under ../Patches to cvs
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub commitPatchFolderToCVS {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot add patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    my $CommitString = "automated patch building from '$self->{OldReleaseTag}' to '$self->{NewReleaseTag}'";

    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {

                #commit
                $CVS->command( 'commit', { '-m', $CommitString } );

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     updateCVSPatchReleaseTagsOnPatchFolder
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->updateCVSPatchReleaseTagsOnPatchFolder($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 9
#               update tags
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub updateCVSPatchReleaseTagsOnPatchFolder {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot add patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);
    my $CVSPath = _GetCVSPath($TargetFolder);

    #get tag handler
    my $CVSTagHandler = DE_EPAGES::Installer::API::CVSTagHandler->new();

    #get control tag
    my $NewReleaseTag = $self->{NewReleaseTag};
    my $PatchTag = $self->_getPatchTag($NewReleaseTag);

    die "\n patch and release tag are same '$PatchTag'" if ($PatchTag eq $NewReleaseTag);

    my $ReleaseControlTag = $CVSTagHandler->getControlTag($NewReleaseTag);
    my $PatchControlTag = $CVSTagHandler->getControlTag($PatchTag);
    my $CommitBranch = $self->{'CommitBranch'};
    $CommitBranch = 'HEAD' if ($CommitBranch eq 'B_generated_patch_code');

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {
                #set necessary tags on generated files
                $CVSTagHandler->setTagByServer($NewReleaseTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($ReleaseControlTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchControlTag, $CommitBranch, $CVSPath);

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     updateCVSPatchReleaseTagsOnCartridgePM
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->updateCVSPatchReleaseTagsOnCartridgePM($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 9
#               tag Cartridge.pm file
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub updateCVSPatchReleaseTagsOnCartridgePM {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot commit patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = $CartridgeFolder . '/API';

    #get tag handler
    my $CVSTagHandler = DE_EPAGES::Installer::API::CVSTagHandler->new();

    #get control tag
    my $NewReleaseTag = $self->{NewReleaseTag};
    my $PatchTag = $self->_getPatchTag($NewReleaseTag);

    die "\n patch and release tag are same '$PatchTag'" if ($PatchTag eq $NewReleaseTag);

    my $ReleaseControlTag = $CVSTagHandler->getControlTag($NewReleaseTag);
    my $PatchControlTag = $CVSTagHandler->getControlTag($PatchTag);
    my $CommitBranch = $self->{'CommitBranch'};
    $CommitBranch = 'HEAD' if ($CommitBranch eq 'B_generated_patch_code');

    my $CVSPath = _GetCVSPath($TargetFolder . '/Cartridge.pm');

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {

                #set necessary tags on generated files
                $CVSTagHandler->setTagByServer($NewReleaseTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($ReleaseControlTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchTag, $CommitBranch, $CVSPath);
                $CVSTagHandler->setTagByServer($PatchControlTag, $CommitBranch, $CVSPath);

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     commitCartridgePMToCVS
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PatchBuilder->commitCartridgePMToCVS($CartridgeFolder)
#----------------------------------------------------------------------------------------
# description  step 5
#               add files under ../Patches to cvs
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | cartridge directory | string
#========================================================================================
sub commitCartridgePMToCVS {
    my $self = shift;
    my ($CartridgeFolder) = @_;

    die "\nno commit branch known, cannot commit patch folder to CVS\n" unless ($self->{'CommitBranch'});

    #get info
    my $TargetFolder = $CartridgeFolder . '/API';
    my $CommitString = "automated patch building from '$self->{OldReleaseTag}' to '$self->{NewReleaseTag}'";

    #get cvs handler
    my $CVS = DE_EPAGES::Core::API::CVS->new({}, {'print' => 1});

    if (ExistsDirectory($TargetFolder)){

        RunOnDirectory(
            $TargetFolder,
            sub {
                #cvs add unless is added
                $CVS->command('update', undef, [] );

                #commit
                $CVS->command( 'commit', { '-m', $CommitString }, ['Cartridge.pm'] );

                return;
            }
        );
    }

    return;
}

#========================================================================================
# function     exportReleaseFolder
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportReleaseFolder($CVSPath, $Dir, $Tag);
#----------------------------------------------------------------------------------------
# description  part of step 1
#               export two versions of one cartridge, one tagged with old release and
#               one tagged with the new release tag to different folders
#----------------------------------------------------------------------------------------
# input        $CVSPath | cvs path from cartridge that should be patcht | string
# input        $OldDir | directory to export cartridge version tagged with $Tag | string
# input        $Tag | tag of cartridge, most any release tag | string
#========================================================================================
sub exportReleaseFolder {
    my $self = shift;
    my ($CVSPath, $Dir, $Tag) = @_;

    DeleteDirectory($Dir, { 'recursive'=>1});
    CreateDirectory($Dir);

    my $OS = $Config{'osname'};
    my $CVS = DE_EPAGES::Core::API::CVS->new({ '-q' => '' }, {'print' => 1});

    #check for operating system
    if ($OS eq 'MSWin32'){
        #use absolute pathes
        $CVS->command('export', {'-d' => $Dir, '-r'=>$Tag}, [ $CVSPath ]);
    }
    else{
        #use relative pathes

        #get pathes
        my $PathEnd = SplitObjectPath($Dir)->[-1];
        my $PathMainPart = $Dir;
        $PathMainPart =~ s/[\\|\/]\Q$PathEnd\E$//;

        #run on directory
        RunOnDirectory(
            $PathMainPart,
            sub {
                $CVS->command('export', {'-d' => $PathEnd, '-r'=>$Tag}, [ $CVSPath ]);
            }
        );
    }

    return;
}

#========================================================================================
# function     verbose
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->verbose()
#----------------------------------------------------------------------------------------
# description  return object attribute verbose
#----------------------------------------------------------------------------------------
# return       $Verbose | object attribute | boolean
#========================================================================================
sub verbose {
    return shift->{'Verbose'};
}

#========================================================================================
# function     printMessage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->printMessage('message to test');
#----------------------------------------------------------------------------------------
# description  print message if verbose is true
#----------------------------------------------------------------------------------------
# input        $Message | message to print | string
#========================================================================================
sub printMessage {
    my $self = shift;
    my ($Message) = @_;

    print $Message, "\n" if $self->verbose;
}

#========================================================================================
# function     newVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $NewVersion = $self->newVersion();
#----------------------------------------------------------------------------------------
# description  return object attribute, new version
#----------------------------------------------------------------------------------------
# return       $NewVersion | object attribute | string
#========================================================================================
sub newVersion {
    my $self = shift;
    return GetVersionByReleaseTag( $self->{'NewReleaseTag'} );
}

#========================================================================================
# function     oldVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $OldVersion = $self->oldVersion();
#----------------------------------------------------------------------------------------
# description  return object attribute, old version
#----------------------------------------------------------------------------------------
# return       $OldVersion | object attribute | string
#========================================================================================
sub oldVersion {
    my $self = shift;
    return GetVersionByReleaseTag( $self->{'OldReleaseOrHotFixTag'} );
}

#========================================================================================
# function     _handleChangedFiles
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_handleChangedFiles($CartridgeFolder, $OldFolder, $NewFolder);
#----------------------------------------------------------------------------------------
# description  part of step 2
#               handle changed files, generates patch xml files and
#               collect information under $self->{'PatchStructure'}, necessary to export patch.xml
#----------------------------------------------------------------------------------------
# input        $CartridgeFolder | folder of cartridge, where should commit | string
# input        $OldFolder | folder of cartridgeversion tagged with old release tag | string
# input        $NewFolder | folder of cartridgeversion tagged with new release tag | string
#========================================================================================
sub _handleChangedFiles {
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    #get old files, by run on old directory
    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    #get new files, by run on new directory
    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    #get all filenames that can find in the new and in the old directory
    my $aIntersectionItems = _GetIntersectionItems($aOldFiles,$aNewFiles);

    #remove files that are equal in old and new folder
    my $aChangedFiles = _RemoveFilesNotChanged($aIntersectionItems, $OldFolder, $NewFolder);

    #get patch target folder
    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    #get cartridge name
    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    foreach my $FileName (@$aChangedFiles) {#for each changed file
        $FileName=~ s,^/,,;

        given ($FileName) {
            when (/^Database\/XML\/.*\.xml/) {
                my $OldFile = $OldFolder . '/' . $FileName;
                my $NewFile = $NewFolder . '/' . $FileName;
                $self->_handleChangedXMLFile( $CartridgeName, $OldFile, $NewFile, $FileName, $TargetFolder );
            }
            when (/^Database\/Sybase\/Procedures\//) {
                #generate relative name to patch file
                my $RelTargetName = $TargetFolder . '/' . $FileName;
                $RelTargetName =~ s/.*\/(Patches\/Patch_.*)$/$1/;

                #register files for handle in patch.xml file
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}->{'02-ProceduresAndOther'}->{$RelTargetName}->{'CVS'} = 'Changed';

                #copy changed procedure file in patch folder
                CopyFile(
                    $NewFolder . '/' . $FileName,
                    $TargetFolder . '/' . $FileName,
                    {'create_directory' => 1}
                );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\// ) {
                next if($FileName =~ /delete_.*?\.sql$/);

                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$FileName}->{'CVS'} = 'Changed';

                #register files, that have to be code generated for
                if ($FileName =~ /^Database\/Sybase\//){
                    $self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'Sybase'}->{$OldFolder . '/' . $FileName} = $NewFolder . '/' . $FileName;
                }
                elsif ($FileName =~ /^Database\/mysql\//){
                    $self->{'PatchStructure'}->{'PatchCartridgePM'}->{'OldNewSQLFiles'}->{'mysql'}->{$OldFolder . '/' . $FileName} = $NewFolder . '/' . $FileName;
                }
            }
            when ( /^Data\/WebRoot\// ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyWebRootFile'}->{$FileName}->{'CVS'} = 'Changed';
            }
            when ( /^Data\/javascript\// ) {
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyJavascriptFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Public\// ) {
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyPublicFile'}->{$FileName}->{'CVS'} = 'Changed';

                unless( $FileName =~ /\bjavascript\b/ ||
                    $FileName =~ /\.js$/ ||
                    $FileName =~ /\.css$/
                ) {
                    my $OldFile = $OldFolder . '/' . $FileName;
                    my $MD5 = GetFileMD5Digest($OldFile);
                    $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyPublicFile'}->{$FileName}->{'OldMD5Digest'} = $MD5;
                }
            }
            when ( /^Data\/Private\// ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyPrivateFiles'}->{$FileName}->{'CVS'} = 'Changed';
            }
            when ( /^Data\/Scheduler\// ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySchedulerFiles'}->{$FileName}->{'CVS'} = 'Changed';
            }
            when ( /^Data\/SharedConfig\// ) {
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySharedConfigFiles'}->{$FileName}->{'CVS'} = 'Changed';
            }
        }
    }
}

#funktion handhabt alle hinzugefgten Files
sub _handleAddedFiles{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    my $aImportFiles = _GetMissingItems($aOldFiles,$aNewFiles);

    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    foreach my $FileName (@$aImportFiles) {
        $FileName=~ s,^/,,;
        given ($FileName) {
            when ( /^Database\/XML\/.*\.xml/ ) {
                my $NewFile = $NewFolder . '/' . $FileName;
                $self->_handleNewXMLFile( $CartridgeName, $NewFile, $FileName, $TargetFolder );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\/Tables/ ) { #tables should come before procedures in patch.xml file
                next if($FileName =~ /delete_.*?\.sql$/);

                #generate relative name to patch file
                my $RelTargetName = $TargetFolder . '/' . $FileName;
                $RelTargetName =~ s/.*\/(Patches\/Patch_.*)$/$1/;

                #register files for handle in patch.xml file
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}->{'01-Tables'}->{$RelTargetName}->{'CVS'} = 'New';

                #copy added file in patch folder
                CopyFile(
                    $NewFolder . '/' . $FileName,
                    $TargetFolder . '/' . $FileName,
                    {'create_directory' => 1}
                );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\// ) {
                #generate relative path to patch file
                my $RelTargetName = $TargetFolder . '/' . $FileName;
                $RelTargetName =~ s/.*\/(Patches\/Patch_.*)$/$1/;

                #register files for handle in patch.xml file
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'onPatchProcessSingleFile'}->{'02-ProceduresAndOther'}{$RelTargetName}->{'CVS'} = 'New';

                #copy added file in patch folder
                CopyFile(
                    $NewFolder . '/' . $FileName,
                    $TargetFolder . '/' . $FileName,
                    {'create_directory' => 1}
                );
            }
            when ( /^Data\/WebRoot\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyWebRootFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/javascript\//){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyJavascriptFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Public\//){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchCopyPublicFile'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Private\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopyPrivateFiles'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/Scheduler\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySchedulerFiles'}->{$FileName}->{'CVS'} = 'New';
            }
            when ( /^Data\/SharedConfig\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchCopySharedConfigFiles'}->{$FileName}->{'CVS'} = 'New';
            }
        }
    }
}

sub _displacedFile {
    my ($FileName, $CartridgeName) = @_;

    foreach my $Displaced (qw( DHLProductList DefaultShop DefaultShopUnity DemoContent Distributor Distributors
                               DistributorStore EbayRegions Providers ShopTypes StandardsLanguagePacks )) {
        if ( $FileName =~ /\/$Displaced\.xml$/ or $FileName =~ /\/ShopTypes.*\.xml$/ ) {
            print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
            return 1;
        }
    }

    #special case for FlexProvider
    if ($FileName =~ /\/DistributorShopTypes.*\.xml$/){
        print "\n    Warning: better move file $FileName to -> ../Data/XML/\n";
        return 1 unless $CartridgeName eq 'FlexProvider';
    }
    return 0;
}

sub _handleNewXMLFile {
    my $self = shift;
    my ($CartridgeName, $NewFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /\/\.\#/);

    #check for displaced files
    return if _displacedFile($FileName, $CartridgeName);

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $OldFile = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Installer/Scripts/empty.xml';

    my $PatchFile = $FileName;
    $PatchFile =~ s,Database/XML/,,;

    #make import file
    my $OrigPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    my $TmpPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    my $OldPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $TmpPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 0);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 0);

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -previous $OrigPatchFile\n      -new $TmpPatchFile\n";
            }
        }
    }
}

sub _handleChangedXMLFile {
    my $self = shift;
    my ($CartridgeName, $OldFile, $NewFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /\/\.\#/);

    #check for displaced files
    return if _displacedFile($FileName, $CartridgeName);

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $PatchFile = $FileName;
    $PatchFile =~ s,Database/XML/,,;

    #make delete file
    my $OrigPatchFile = $TargetFolder . '/Database/XML/del_' . $PatchFile;
    my $TmpPatchFile = $TargetFolder . '/Database/XML/del_' . $PatchFile;
    my $OldPatchFile = $TargetFolder . '/Database/XML/del_' . $PatchFile;
    $TmpPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 1);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 1);

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                #no changes have to add
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -previous $OldPatchFile\n      -new $OrigPatchFile\n";
            }
        }
    }

    #make import file
    $OrigPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $TmpPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $OldPatchFile = $TargetFolder . '/Database/XML/imp_' . $PatchFile;
    $TmpPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 0);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 0);

        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    [Warning:] new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one
            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Warning:] patch already build, please check changes\n      -new $OrigPatchFile\n      -previous $OldPatchFile\n";
            }
        }
    }
}

#function handles all deleted files
sub _handleDeletedFiles{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFiles = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetFileNames( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFiles = RunOnDirectory( $NewFolder, sub { GetFileNames( undef, {'recursive' => 1} ); } );

    my $aDeleteFiles = _GetMissingItems($aNewFiles,$aOldFiles);

    my $TargetFolder = _GetPatchTargetName($CartridgeFolder, $self->oldVersion);

    my $CartridgeName = $CartridgeFolder;
    my $CompanyName = $self->{'CompanyName'}; ## no critic qw( ProhibitUnusedVariablesEnhanced )
    $CartridgeName =~ s/^.*[\/\\]Cartridges[\/\\]$CompanyName[\/\\]//;

    foreach my $FileName (@$aDeleteFiles) {
        $FileName=~ s,^/,,;
        given( $FileName ) {
            when ( /^Database\/XML\/.*\.xml/ ) { #../Database/XML/*.xml files
                my $OldFile = $OldFolder . '/' . $FileName;
                #register file name for delete in patch.xml
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';

                #build patch.xml files
                $self->_handleDeletedXMLFile( $CartridgeName, $OldFile, $FileName, $TargetFolder );
            }
            when ( /^Database\/(?:SQL|Sybase|mysql)\// ) {
                $self->{'PatchStructure'}->{'PatchDBStructure'}->{'PatchBuilder-Warning'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/WebRoot\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteWebRootFile'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/javascript\//){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchDeleteJavascriptFile'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/Public\//){
                $self->{'PatchStructure'}->{'PatchStoreTypeDir'}->{'onPatchDeletePublicFile'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/Private\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeletePrivateFiles'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\/Scheduler\//){
                $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteSchedulerFiles'}->{$FileName}->{'CVS'} = 'Changed';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            when ( /^Data\//){
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDirPriorPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
            default {
                $self->{'PatchStructure'}->{'PatchCartridgeDirFinalPart'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
                $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$FileName}->{'CVS'} = 'Deleted';
            }
        }
    }
}

sub _handleDeletedXMLFile {
    my $self = shift;
    my ($CartridgeName, $OldFile, $FileName, $TargetFolder) = @_;

    return if ($FileName =~ /\/SQL\/.*\.xml/);
    return if ($FileName =~ /\/mysql\/.*\.xml/);
    return if ($FileName =~ /\/\.\#/);

    return if _displacedFile($FileName, $CartridgeName);

    my $NewFile = $ENV{'EPAGES_CARTRIDGES'} . '/DE_EPAGES/Installer/Scripts/empty.xml';

    my $FilePatcher = DE_EPAGES::Installer::API::XMLFilePatcher->new();

    my $PatchFile = $FileName;
    $PatchFile =~ s,Database/XML/,,;

    #make delete file
    my $OrigPatchFile = "$TargetFolder/Database/XML/del_$PatchFile";
    my $TmpPatchFile = "$TargetFolder/Database/XML/del_$PatchFile";
    my $OldPatchFile = "$TargetFolder/Database/XML/del_$PatchFile";
    $TmpPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(tmp\)$1/;
    $OldPatchFile =~ s/\/([\w.]*\.xml)$/\/\.\#\(old\)$1/;

    #check if patch is already build
    if (not ExistsFile($OrigPatchFile)){
        #patch is build first time, build patch file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $OrigPatchFile, 1);
    }
    else{
        #patch is already build, save new patch file as temporary file
        $FilePatcher->patchXMLFiles($OldFile, $NewFile, $TmpPatchFile, 1);
        if (not ExistsFile($TmpPatchFile)){
            #new patch have no difference in file, print message
            print "\n    Warning: new version of patch dont have $OrigPatchFile included\n" ;
        }
        else{#a new and a old patch file exists look for witch is the right one

            if(CompareXML($TmpPatchFile, $OrigPatchFile, {'checkreference' => '1', 'checkdelete' => '1'})){
                DeleteFile($TmpPatchFile);
            }
            else{
                my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
                my $OrigPatchFileXMLStruct = $ImportHandler->getXMLStructure($OrigPatchFile);
                my $TmpPatchFileXMLStruct = $ImportHandler->getXMLStructure($TmpPatchFile);

                $OrigPatchFileXMLStruct->assimilateXMLStructure($TmpPatchFileXMLStruct,{'overwrite' => '1'});
                my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
                CopyFile($OrigPatchFile,$OldPatchFile);
                $XMLStructExporter->exportXML($OrigPatchFile,$OrigPatchFileXMLStruct);
                DeleteFile($TmpPatchFile);

                print "\n    [Conflict:] patch already build\n      -previous $OrigPatchFile\n      -new $TmpPatchFile\n";
            }
        }
    }
}

sub _handleDeletedDirectories{
    my $self = shift;
    my ($CartridgeFolder, $OldFolder, $NewFolder) = @_;

    my $aOldFolders = ExistsDirectory( $OldFolder )
        ? RunOnDirectory( $OldFolder, sub { GetSubDirectories( undef , {'recursive' => 1} ) } )
        : [];

    my $aNewFolders = RunOnDirectory( $NewFolder, sub { GetSubDirectories( undef, {'recursive' => 1} ) } );

    my $aDeleteFolders = _GetMissingItems($aNewFolders,$aOldFolders);
    foreach my $FolderName (@$aDeleteFolders){
        $FolderName =~ s,^/,,;
        $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}->{$FolderName}->{'CVS'} = 'Deleted';
    }
    #ignore subfolders
    if (exists $self->{'PatchStructure'}->{'PatchCartridgeDir'} and $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}){
        foreach my $FolderName (@$aDeleteFolders){
            foreach my $SubFolder (keys %{$self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}}){
                delete $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}->{$SubFolder}
                if $SubFolder =~ /^\Q$FolderName\E\/.+/;
            }
        }
    }
    #ignore files from deleted folders
    if (exists $self->{'PatchStructure'}->{'PatchCartridgeDir'} and $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}){
        foreach my $FolderName (keys %{$self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeDirectory'}}){
            $FolderName =~ s,^/,,;
            foreach my $File (keys %{$self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}}){
                delete $self->{'PatchStructure'}->{'PatchCartridgeDir'}->{'onPatchDeleteCartridgeFile'}->{$File}
                if $File =~ /^\Q$FolderName\E\//;
            }
        }
    }

}
#HilfsFunktion ermittelt alle Namen der Ordner die fehlen
sub _GetMissingItems {
    my ($aReferenceNames,$aControlNames) = @_;

    my $hRef;
    my @Result = ();
    my $CleanName;

    foreach my $RefName (@$aReferenceNames){
        $CleanName = $RefName;
        $CleanName =~ s/^\.//;
        $hRef->{$CleanName} = '1';
    }

    foreach my $ControlName (@$aControlNames){
        $CleanName = $ControlName;
        $CleanName =~ s/^\.//;
        push(@Result, $CleanName) if ($hRef->{$CleanName} ne '1');
    }

    return \@Result;
}

#return names of elements exists in both arrays
sub _GetIntersectionItems {
    my ($aReferenceNames,$aControlNames) = @_;

    my $hRef;
    my @Result = ();
    my $CleanName;

    foreach my $RefName (@$aReferenceNames){
        $CleanName = $RefName;
        $CleanName =~ s/^\.//;
        $hRef->{$CleanName} = '1';
    }

    foreach my $ControlName (@$aControlNames){
        $CleanName = $ControlName;
        $CleanName =~ s/^\.//;
        push(@Result, $CleanName) if ($hRef->{$CleanName} eq '1');
    }

    return \@Result;
}

sub _RemoveFilesNotChanged{
    my ($aIntersectionItems, $OldDir, $NewDir) = @_;

    my @Result;

    foreach my $FileName (@$aIntersectionItems){
        my $CheckSumOld = GetFileMD5Digest($OldDir . $FileName);
        my $CheckSumNew = GetFileMD5Digest($NewDir . $FileName);
        push(@Result, $FileName) if ($CheckSumOld ne $CheckSumNew);
    }
    return \@Result;
}

sub _collectPatchXMLInfoFromFiles{
    my $self = shift;
    my ($DBXMLFile, $DBXMLFilePatchPath) = @_;

    #onPatchImportObjects
    given ($DBXMLFile) {
        when ( /^imp_Attributes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'01-Attributes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_LastUpdateProfiles.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'02-LastUpdateProfiles'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Actions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }

        #exceptions please remove if files renamed
        when ( /^imp_GoogleAnalyticsActions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_EtrackerActions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Permissions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'04-Permissions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_PageTypes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'05-PageTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_TemplateTypes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'06-TemplateTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Forms.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'07-Forms'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Features.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'08-Features'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_ClientApplication.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'09-ClientApplication'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_WidgetGroups.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'10-WidgetGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_WidgetTypes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'11-WidgetTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_NavBars.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'13-NavBars'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_NavElement.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'14-NavElements'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Search.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'15-Searchs'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_MailType.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'16-MailTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Shipping.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'17-Shipping'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_PaymentLogo.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'19-PaymentLogo'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_PaymentType.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'20-PaymentTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_PortalSites.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'21-PortalSites'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_System.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'22-Systems'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Units.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'23-Units'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Taxes.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'24-Taxes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_StyleGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'26-StyleGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_CountrySubdivisions_.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'28-CountrySubdivisions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_EcoParticipation.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_ImageDBGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_GadgetGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_EbayCallTypes.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_PaymentGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_FlexShopType.*\.xml$/){
            #ignore
        }

        when ('imp_StyleTemplates.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'25-StyleTemplates'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Countries_3166.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'27-Countries'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_TaxMatrixDigital.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'29-TaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_DefaultTaxMatrixDigital.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'30-DefaultTaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_TaxMatrixUS.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'31-TaxMatrixUS'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Helptopics.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_EtrackerActions.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_EbayEnvironment.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Locales.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_TaxTypes.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_CreditCheckProviders.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Currencies_4217.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_OrderProcessSteps.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Languages.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Locales.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_CustomizableProducts.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_DistributorShopTypes.xml'){
            #$self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_AvataxTaxExemption.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_ERPSystemSoap.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_ERPSystemLexware.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_ERPSystem.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_UnityMBOGuidedTourGeneralConfig.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_UnityMBOGuidedTourLegalConfig.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_UnityMBOGuidedTourProductsConfig.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_DefaultSystemTemplate.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_ShopNotifications.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_DHLIntrashipService.xml') {
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Fonts.xml') {
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        # ADD NEW onPatchImportObjects 32-Other *BEFORE* THIS LINE

        when (/^imp_ShopNotification.*\.xml$/){
            # ignore
        }

        #EPG-15720 ignore Translation.*.xml files in patch.xml, automatic delete and import now in Installer.pm
        when ( /^imp_Translation\..*\.(\w\w|\w\w-\w\w)\.xml$/){
            #$self->{'PatchStructure'}->{'Patch'}->{'onPatchImportObjects'}->{'30-Translations'}->{$DBXMLFilePatchPath} = {};
        }

        #onPatchDeleteObjects
        when ( /^del_Attributes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'01-Attributes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_LastUpdateProfiles.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'02-LastUpdateProfiles'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Actions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }

        #exceptions please remove if files are renamed
        when ( /^del_GoogleAnalyticsActions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_EtrackerActions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'03-Actions'}->{$DBXMLFilePatchPath} = {};
        }

        when ( /^del_Permissions.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'04-Permissions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_PageTypes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'05-PageTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_TemplateTypes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'06-TemplateTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Forms.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'07-Forms'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Features.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'08-Features'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_ClientApplication.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'09-ClientApplication'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_WidgetGroups.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'10-WidgetGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_WidgetTypes.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'11-WidgetTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_NavBars.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'13-NavBars'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_NavElement.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'14-NavElements'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Search.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'15-Searchs'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_MailType.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'16-MailTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Shipping.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'17-Shipping'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_DefaultShipping.*\.xml/){
           $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'18-Shipping'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_PaymentLogo.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'19-PaymentLogo'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_PaymentType.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'20-PaymentTypes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_PortalSites.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'21-PortalSites'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_System.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'22-Systems'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Units.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'23-Units'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Taxes.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'24-Taxes'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_StyleTemplates\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'25-StyleTemplates'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_StyleGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'26-StyleGroups'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_CountrySubdivisions_.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'28-CountrySubdivisions'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_Countries_3166\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'27-Countries'}->{$DBXMLFilePatchPath} = {};
        }

        when ('del_TaxMatrixDigital.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'29-TaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_DefaultTaxMatrixDigital.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'30-DefaultTaxMatrixDigital'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_TaxMatrixUS.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'31-TaxMatrixUS'}->{$DBXMLFilePatchPath} = {};
        }

        # ADD NEW onPatchDeleteObjects 32-Other *AFTER* THIS LINE
        when ( /^del_FlexShopType.*\.xml$/){
            #ignore
        }
        when ( /^del_PaymentGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_EcoParticipation.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_ImageDBGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_GadgetGroups.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^del_EbayCallTypes.*\.xml$/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }

        when ('del_DHLIntrashipService.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_ShopNotifications.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_UnityMBOGuidedTourLegalConfig.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_UnityMBOGuidedTourProductsConfig.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_UnityMBOGuidedTourGeneralConfig.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_ERPSystem.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_ERPSystemLexware.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_ERPSystemSoap.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_AvataxTaxExemption.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_Helptopics.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_EtrackerActions.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_EtrackerActions.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_TaxTypes.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_CreditCheckProviders.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_Currencies_4217.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_OrderProcessSteps.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_Languages.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_Locales.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_CustomizableProducts.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_DistributorShopTypes.xml'){
            #$self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_EbayEnvironment.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_DefaultSystemTemplate.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'32-Other'}->{$DBXMLFilePatchPath} = {};
        }
        #EPG-15720 ignore Translation.*.xml files in patch.xml, automatic delete and import now in Installer.pm
        when ( /^del_Translation\..*\.(\w\w|\w\w-\w\w)\.xml$/){
        #    $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteObjects'}->{'30-Translations'}->{$DBXMLFilePatchPath} = {};
        }

        when ( /^del_Hooks.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteHooks'}->{$DBXMLFilePatchPath} = {};
        }
        when ( /^imp_Hooks.*\.xml/){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportHooks'}->{$DBXMLFilePatchPath} = {};
        }

        when ('del_ShopChildDependencies.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteShopChildDependencies'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_ShopChildDependencies.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportShopChildDependencies'}->{$DBXMLFilePatchPath} = {};
        }
        when ('del_Dependencies.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchDeleteDependencies'}->{$DBXMLFilePatchPath} = {};
        }
        when ('imp_Dependencies.xml'){
            $self->{'PatchStructure'}->{'Patch'}->{'onPatchImportDependencies'}->{$DBXMLFilePatchPath} = {};
        }

        default {
            die "can't interpret $DBXMLFile";
        }
    }
}

sub _CorrectPositionsOfPatchXMLXMLStruct{
    my ($XMLStruct) = @_;

    foreach my $Element (@{$XMLStruct->getElements()}){
        given ($Element) {
            ## no critic qw( ProhibitComplexRegexes )
            when (/^\/PatchCartridge\{OldRelease\-\>(?:lp_|cs_|test_)?release_[0-9_]*\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '000001')
            }
            when (/\/PatchDBStructure\{PatchDBStructure\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010000')
            }
            when (/\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010100')
            }
            when (/\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}.*\/\"{0,1}File\{.*\/Tables\/.*\.sql/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010101')
            }
            when (/\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}.*\/\"{0,1}File\{.*\/Procedures\/.*\.sql/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010102')
            }
            when (/\/PatchFunction\{Method\-\>onPatchProcessSingleFile\}.*\/\"{0,1}File\{.*\/Views\/.*\.sql/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '010102')
            }
            when (/\/Patch\{Patch\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020000')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020100')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteDependencies\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteShopChildDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020200')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteShopChildDependencies\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteHooks\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020300')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020400')
            }

            # ADD NEW onPatchDeleteObjects *AFTER* THIS LINE
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*DHLIntrashipService\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020401')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PaymentGroups\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020402')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ShopNotifications\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020403')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourProductsConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020404')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourLegalConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020405')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourGeneralConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020406')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ERPSystem\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020407')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ERPSystemLexware\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020408')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*ERPSystemSoap\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020409')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*AvataxTaxExemption\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020410')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*EbayCallTypes\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020411')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Countries_3166\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020412')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*CustomizableProducts\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020413')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Locales\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020414')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Languages\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020415')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*OrderProcessSteps\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020416')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*CreditCheckProviders\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020417')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Currencies_4217\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020418')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*EtrackerActions\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020419')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Helptopics\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020420')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*EcoParticipation\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020421')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Taxes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020422')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Units\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020423')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*System\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020424')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PortalSites\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020425')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PaymentType\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020426')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PaymentLogo\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020427')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Shipping\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020428')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*MailType\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020429')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Search\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020430')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*NavElement\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020431')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*NavBars\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020432')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Features\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020433')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*WidgetTypes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020434')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*WidgetGroups\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020435')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Forms\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020436')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*StyleGroups\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020437')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*StyleTemplates\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020438')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*TemplateTypes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020439')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*PageTypes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020440')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Permissions\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020441')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Actions\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020442')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*LastUpdateProfiles\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020443')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteObjects\}.*\/\"{0,1}File\{.*Attributes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020444')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteHooks\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020500')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportDependencies\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportHooks\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020600')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportHooks\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020700')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Attributes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020701')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*LastUpdateProfiles\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020702')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Actions\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020703')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Permissions\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020704')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PageTypes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020705')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*TemplateTypes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020706')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*StyleTemplates\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020707')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*StyleGroups\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020708')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Forms\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020709')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Features\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020710')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*WidgetGroups\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020711')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*WidgetTypes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020712')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*NavBars\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020713')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*NavElement\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020714')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Search\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020715')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*MailType\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020716')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Shipping\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020717')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PaymentLogo\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020718')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PaymentType\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020719')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PortalSites\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020720')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*System\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020721')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Units\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020722')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Taxes\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020723')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EcoParticipation\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020724')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Helptopics\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020725')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EtrackerActions\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020726')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EbayEnvironment\.xml/){ #"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020727')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Locales\.xml/){ #"
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020728')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*CreditCheckProviders\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020729')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Currencies_4217\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020730')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*OrderProcessSteps\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020731')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Languages\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020732')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*Countries_3166\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020733')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*CustomizableProducts\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020734')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*EbayCallTypes\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020735')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*AvataxTaxExemption\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020736')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ERPSystemSoap\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020737')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ERPSystemLexware\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020738')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ERPSystem\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020739')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourGeneralConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020740')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourLegalConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020741')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*UnityMBOGuidedTourProductsConfig\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020742')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotifications\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020743')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*PaymentGroups\w*\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '020744')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*DHLIntrashipService\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207445')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationPendingOrder\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207446')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationEbayNoOAuthToken\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207447')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationCookieNotice\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207448')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationLegalDocuments\.xml/) {
                $XMLStruct->setXMLStructAttribute($Element, 'Position', '0207449')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationProductSlider\.xml/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '0207450')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationFacetedSearch\.xml/) {
                $XMLStruct->setXMLStructAttribute($Element, 'Position', '0207451')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationSpreedlyStripePaymentIntents\.xml/) {
                $XMLStruct->setXMLStructAttribute($Element, 'Position', '0207452')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationPayPalPlus\.xml/) {
                $XMLStruct->setXMLStructAttribute($Element, 'Position', '0207453')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportObjects\}.*\/\"{0,1}File\{.*ShopNotificationDeliveryThreshold\.xml/) {
                $XMLStruct->setXMLStructAttribute($Element, 'Position', '0207454')
            }
            # ADD NEW onPatchImportObjects *BEFORE* THIS LINE

            when (/\/PatchFunction\{Method\-\>onPatchImportShopChildDependencies\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021000')
            }
            when (/\/PatchFunction\{Method\-\>onPatchImportShopChildDependencies\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteWebRootFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021100')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteWebRootFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeletePrivateFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021300')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeletePrivateFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteSchedulerFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021400')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteSchedulerFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteSharedConfigFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021450')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteSharedConfigFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyWebRootFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021500')
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyWebRootFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyPrivateFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021600')
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyPrivateFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopySchedulerFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021700')
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopySchedulerFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopySharedConfigFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '021750')
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopySharedConfigFile\}/){
                #ignore
            }
            when (/\/PatchCartridgeDir\{PatchCartridgeDir\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '030000')
            }
            when (/\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '030100')
            }
            when (/\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}/){
                #ignore
            }
            when (/\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '030200')
            }
            when (/\/PatchCartridgeDir\{PatchCartridgeDir\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}/){
                #ignore
            }

            when (/\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '040000')
            }
            when (/\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '040100')
            }
            when (/\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}/){
                #ignore
            }
            when (/\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '040200')
            }
            when (/\/PatchCartridgeDirPriorPart\{PatchCartridgeDirPriorPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}/){
                #ignore
            }

            when (/\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '050000')
            }
            when (/\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '050100')
            }
            when (/\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeDirectory\}/){
                #ignore
            }
            when (/\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '050200')
            }
            when (/\/PatchCartridgeDirFinalPart\{PatchCartridgeDirFinalPart\}\/PatchFunction\{Method\-\>onPatchDeleteCartridgeFile\}/){
                #ignore
            }

            when (/\/PatchStoreTypeDir\{PatchStoreTypeDir\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060000')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeletePublicFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060100')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeletePublicFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteJavascriptFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060200')
            }
            when (/\/PatchFunction\{Method\-\>onPatchDeleteJavascriptFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyPublicFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060300')
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyPublicFile\}/){
                #ignore
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyJavascriptFile\}$/){
                $XMLStruct->setXMLStructAttribute($Element,'Position', '060400')
            }
            when (/\/PatchFunction\{Method\-\>onPatchCopyJavascriptFile\}/){
                #ignore
            }
            default{
                die "\nunknown element $Element in xml structure of patch.xml"
            }
        }
    }
    return;
}

sub GetVersionByReleaseTag{
    my ($ReleaseTag) = @_;
    my $CartridgeVersion = $ReleaseTag;
    $CartridgeVersion =~ s/^(?:lp_|cs_|test_)?release_//;
    $CartridgeVersion =~ s/_/\./g;

    return $CartridgeVersion;
}

sub _GetPatchTargetName {
    my ($CartridgeFolder, $OldCartridgeVersion) = @_;

    $OldCartridgeVersion =~ s/\./_/g;

    return $CartridgeFolder . '/Patches/Patch_' . $OldCartridgeVersion;
}

sub _GetCVSPath {
    my ($Folder) = @_;

    my $CVSPath =$Folder;

    $CVSPath =~ s/.*(\\|\/)DE_EPAGES//g;

    die "\nunable to replace chunk from path, cannot create CVS path" if ($CVSPath eq $Folder);

    return 'RND/epages/epages5/AllOS/Cartridges/DE_EPAGES' . $CVSPath;
}

sub addCartridgeTag{
    my $self = shift;
    my ($PatchFile, $CartridgeName) = @_;

    my $ImportHandler = DE_EPAGES::Object::API::XMLStructureImportHandler->new();
    my $XMLStruct = $ImportHandler->getXMLStructure($PatchFile);

    foreach my $Element (@{$XMLStruct->getElements()}){
        next if($XMLStruct->getXMLStructAttribute($Element,'Position') eq '1');
        $XMLStruct->setXMLStructAttribute($Element,'Position',$XMLStruct->getXMLStructAttribute($Element,'Position') + 1 );
    }

    my $CompanyName = $self->{'CompanyName'};

    $XMLStruct->addElement(
        '/Cartridge',
        {},
        {
            'Position' => '2',
            'TagName' => 'Cartridge',
            'Class' => 'Cartridge',
            'NonEffectiveForSubPathes' => '1'
        },
        {'Package' => "$CompanyName::" . $CartridgeName}
    );

    my $XMLStructExporter = DE_EPAGES::Object::API::XMLStructureExportDriver->new();
    $XMLStructExporter->exportXML($PatchFile,$XMLStruct);
}

1;
