#========================================================================================
# package      DE_EPAGES::Order::API::SplitableTaxes
# state        public
#----------------------------------------------------------------------------------------
# description  Class to calculate the Splitable tax rates for payment and shipping.
#               This class is use on the REST API and takes LineItemContainer as main param
#========================================================================================
package DE_EPAGES::Order::API::SplitableTaxes;

use DE_EPAGES::Shop::API::Resource::ShopRequestContext;
use DE_EPAGES::Shop::API::Resource::Formatter;
use DE_EPAGES::Object::API::Factory qw ( LoadObject );
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_GROSS );
use DE_EPAGES::Object::API::Table::Currency qw (GetInfoCurrency);
use DE_EPAGES::Core::API::Object::NumberFormat;
use List::Util qw( reduce );

use strict;

sub new {
    my $class = shift;
    my $self  = bless { }, $class;
    return $self;
}

sub calculatePaymentTax {
    my $self = shift;
    my ($LineItemContainer, $LanguageID, $Formatter) = @_;

    my $LineItemPaymentDiscount = $LineItemContainer->get('PaymentDiscount');
    return unless defined $LineItemPaymentDiscount;

    my $LineItemPaymentPrice = $LineItemPaymentDiscount->get('LineItemPrice');
    if(defined $LineItemContainer->get('Payment')->get('LineItemPrice')) {
        $LineItemPaymentPrice += $LineItemContainer->get('Payment')->get('LineItemPrice');
    }
    my $hNetPriceForTax = $LineItemContainer->netPricesForTax();
    my $hLineItem = $hNetPriceForTax->{'LineItem'};

    return $self->_calculateTax($LineItemContainer, $LanguageID, $Formatter, $LineItemPaymentPrice, $hLineItem);
}


sub calculateShippingTax {
    my $self = shift;
    my ($LineItemContainer, $LanguageID, $Formatter) = @_;

    my $LineItemShipping = $LineItemContainer->get('Shipping');
    return unless defined $LineItemShipping;

    my $LineItemPrice = $LineItemShipping->get('LineItemPrice');
    my $hNetPriceForTax = $LineItemContainer->netPricesForTax();
    my $hLineItemShipping = $hNetPriceForTax->{'LineItemShipping'};

    return $self->_calculateTax($LineItemContainer, $LanguageID, $Formatter, $LineItemPrice, $hLineItemShipping);
}

sub _calculateTax {
    my $self = shift;
    my ($LineItemContainer, $LanguageID, $Formatter, $LineItemPrice, $hLineItemPriceByTaxClassId) = @_;

    my $TotalPrice = 0;
    $TotalPrice = reduce { $a + $b } values %{$hLineItemPriceByTaxClassId};

    if( $TotalPrice == 0) {
        $TotalPrice = 1; #AD-6720: fix division by zero
    }

    my @TaxPrice;
    my $PartialGrossPrice;
    my $PartialNetPrice;
    foreach my $TaxClassId (keys %$hLineItemPriceByTaxClassId){
       my $TaxClass = $self->_LoadObject($TaxClassId);
       my $hTaxData = $self->_createTaxData($TaxClass, $LineItemContainer);

       my $PercentagePrice = $hLineItemPriceByTaxClassId->{$TaxClassId} / $TotalPrice;
       if($hTaxData->{'TaxModel'} == TAX_MODEL_GROSS) {
           $PartialGrossPrice = $LineItemPrice * $PercentagePrice;
           $PartialNetPrice = $hTaxData->{'TaxMatrix'}->gross2net($LineItemContainer->get('CurrencyID'), $PartialGrossPrice, $hTaxData);
       } else {
           $hTaxData->{'TaxModel'} = TAX_MODEL_GROSS;
           $PartialNetPrice = $LineItemPrice * $PercentagePrice;
           $PartialGrossPrice = $hTaxData->{'TaxMatrix'}->net2gross($LineItemContainer->get('CurrencyID'), $PartialNetPrice, $hTaxData);
       }

       my $CurrencyID = $LineItemContainer->get('CurrencyID');
       my $Precision = $self->_GetInfoCurrency($CurrencyID)->{'MinorUnit'};
       my $NumberFormatter = $self->_createNumberFormat();
       my $TaxAmount = $hTaxData->{'TaxRate'} != 0 ? $NumberFormatter->round(($PartialGrossPrice - $PartialNetPrice), $Precision) : 0;
       my $hPrice = {
          'price' => $Formatter->price($TaxAmount, $CurrencyID, undef),
          'taxClass' => $self->_createTaxClassData($hTaxData, $LanguageID),
       };
       push (@TaxPrice, $hPrice);
    }
    return \@TaxPrice;
}

sub _createTaxData {
    my $self = shift;
    my ($TaxClass, $LineItemContainer) = @_;
    return unless (defined $TaxClass);

    my $TaxArea;
    my $TaxMatrix;
    my $TaxRate;
    if ( $TaxClass->instanceOf('TaxClassDigital') ) {
        $TaxMatrix = $LineItemContainer->get('TaxMatrixDigital');
        $TaxArea = $LineItemContainer->get('TaxAreaDigital');
        $TaxRate = $TaxMatrix->taxRateDigital($TaxArea, $TaxClass);
    } elsif( $TaxClass->instanceOf('TaxClass') ) {
        $TaxMatrix = $LineItemContainer->get('TaxMatrix');
        $TaxArea = $LineItemContainer->get('TaxArea');
        #correct TaxClass to the child of Taxmatrix
        my $CurrentTaxClass = $TaxClass;
        $TaxClass = $TaxMatrix->child($CurrentTaxClass->alias);

        $TaxRate = $TaxMatrix->taxRate($TaxArea, $TaxClass);
    }
    return {
       'TaxMatrix' => $TaxMatrix,
       'TaxClass' => $TaxClass,
       'TaxArea' => $TaxArea,
       'TaxRate' => $TaxRate,
       'TaxModel' => $LineItemContainer->get('TaxModel'),
    }
}

sub _createTaxClassData{
    my $self = shift;
    my ($hTaxData, $LanguageID) = @_;

    my $TaxClass = $hTaxData->{'TaxClass'};
    my $TaxRate = $hTaxData->{'TaxRate'};
    return unless defined $TaxClass;
    return {
        'taxClassId' => $TaxClass->guid(),
        'name' => $TaxClass->get('Name', $LanguageID),
        'percentage' => $TaxRate * 100,
    };
}

sub _GetInfoCurrency{
    my $self = shift;
    my ($CurrencyID) = @_;
    return GetInfoCurrency($CurrencyID)
}

sub _createNumberFormat{
    my $self = shift;
    return DE_EPAGES::Core::API::Object::NumberFormat->new();
}

sub _LoadObject {
    my $self = shift;
    my ($ObjectId) = @_;
    return LoadObject($ObjectId);
}


1;