#========================================================================================
# package      DE_EPAGES::Order::Hooks::Address
# state        public
# description  removes billing, shipping and shop address in baskets
#========================================================================================
package DE_EPAGES::Order::Hooks::Address;

use strict;
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Order::API::Shop qw (ProcessOrderDocumentTextAttributes);
use DE_EPAGES::Shop::API::Constants qw ( TAX_MODEL_NET );
use DE_EPAGES::WebInterface::API::ApplicationServer qw( GetAppServer );
use Scalar::Util qw( blessed );

#========================================================================================
# function     OnDeleteAddress
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnDeleteAddress( $hParams );
#----------------------------------------------------------------------------------------
# description  OnDeleteAddress is triggered, to remove references to address.
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used key is
#               <ul>
#                   <li>AddressID
#               </ul> | ref.hash.*
#========================================================================================
sub OnDeleteAddress {
    my ($hParams) = @_;

    my $AddressID = $hParams->{'AddressID'};
    my $BasketClass = LoadClassByAlias('Basket');
    $_->set({BillingAddress  => undef, 'IsAddressOK'=>0 })
            foreach @{$BasketClass->getObjects('ByBillingAddressID', [$AddressID])};
    $_->set({ShippingAddress => undef, 'IsAddressOK'=>0 })
            foreach @{$BasketClass->getObjects('ByShippingAddressID',[$AddressID])};
    $_->set({ShopAddress     => undef })
            foreach @{$BasketClass->getObjects('ByShopAddressID',[$AddressID])};
}

#========================================================================================
# function     OnAfterUpdateAddress
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnAfterUpdateAddress( $hParams );
#----------------------------------------------------------------------------------------
# description  OnAfterUpdateAddress (OBJ_AfterUpdateAddress) is triggered, to change 
#               the order document attributes for Unity and set IsAddressOK on baskets 
#               and change shipping and payment methods if not longer are available 
#               for the address e.g. a regions can prevent the usage af shipping 
#               and payment methods
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments, used key is
#               <ul>
#                   <li>Object
#                   <li>AddressID
#               </ul> | ref.hash.*
#========================================================================================
sub OnAfterUpdateAddress {
    my ($hParams) = @_;

    my $Address = $hParams->{'Object'};

    my $Shop = $Address->getSite();
    return unless $Shop->instanceOf('Shop');

    if($Shop->isUnity()){
        # set cookie for selling country by using the user/customer data
        my $UserOrBasket = $Address->parent->parent;
        if ($UserOrBasket->instanceOf('AnonymousUser') ||
            $UserOrBasket->instanceOf('User') ||
            $UserOrBasket->instanceOf('Basket')) {
            my $ShippingAddress = $UserOrBasket->get('ShippingAddress') // $UserOrBasket->get('BillingAddress');
            my $AppServer = GetAppServer();
            if(defined $AppServer){
                my $Servlet = $AppServer->currentServlet;
                if(blessed($Servlet) eq 'DE_EPAGES::Presentation::Servlet::PageType'){
                    $Servlet->setSellingCountryCookie($ShippingAddress->get('CountryID')) if defined $ShippingAddress;
                }
            }
        }

        # we set order document attributes only for unity and if the shop address was changed
        if($Address->instanceOf('ShopAddress')){
            ProcessOrderDocumentTextAttributes($Shop, 'UnityMBO-OrderDocumentText2', 1);
        }
    }

    my $AddressID = $hParams->{'AddressID'};
    my $BasketClass = LoadClassByAlias('Basket');
    $_->set({'IsAddressOK'=>0 }) foreach @{$BasketClass->getObjects('ByBillingAddressID', [$AddressID])};
    $_->set({'IsAddressOK'=>0 }) foreach @{$BasketClass->getObjects('ByShippingAddressID',[$AddressID])};


    foreach my $Basket ( @{$BasketClass->getObjects('ByBillingAddressID', [$AddressID])},
                         @{$BasketClass->getObjects('ByShippingAddressID',[$AddressID])}
    ){

        my $LineItemContainer = $Basket->container;

        # ignore ordered baskets because no LineItemContainer exists here
        next unless defined $LineItemContainer;

        $Basket->set({'IsAddressOK'=>0 });

        # correct shipping and payment if address was changed
        if ($LineItemContainer->needPayment) {
            $Basket->updatePaymentMethod();
        }
        if ($LineItemContainer->needShipping) {
            $Basket->updateShippingMethod();
        }

        $Basket->recalculate;

    }
}


#========================================================================================
# function     OnUpdateAddressSaveVATID
# state        public
#----------------------------------------------------------------------------------------
# syntax       OnUpdateAddressSaveVATID ($hParams);
#----------------------------------------------------------------------------------------
# description  Sets the VATID of the customers address if users VATID was changed
#               Switch the customer tax model to net if VATID was added
#               (Hook OBJ_AfterUpdateAddress)
#----------------------------------------------------------------------------------------
# input        $hParams | parameter hash with:
#               <ul>
#                   <li>Object - address - object
#               </ul> | ref.hash.*
#========================================================================================
sub OnUpdateAddressSaveVATID {
    my ($hParams) = @_;

    my $UserBillingAddress = $hParams->{'Object'};

    # test  if we have a billing address
    return unless $UserBillingAddress->alias =~ /^Billing/i;

    # test if we have updated a user address
    if ($UserBillingAddress->parent->parent->instanceOf('AnonymousUser') ||
        $UserBillingAddress->parent->parent->instanceOf('User')) {
        my $User = $UserBillingAddress->parent->parent;

        # test if a VATID is available on the user billing address
        my $UserVATID = $UserBillingAddress->get('VATID');

        return unless defined $UserVATID;
        return if $UserVATID eq '';

        #get the cutomer for the user
        my $Customer;
        if ($User and not $User->get('IsAnonymous')) {
            $Customer = $User->get('Customers')->[0];
        }

        return unless defined $Customer;

        my $CustomerBillingAddress = $Customer->get('BillingAddress');
        return unless $CustomerBillingAddress;

        my $CustomerVATID = $CustomerBillingAddress->get('VATID');
        return if defined $CustomerVATID && $CustomerVATID ne '';

        # update customer VATID and set customer to net
        $CustomerBillingAddress->set({'VATID' => $UserVATID});
        $Customer->set({'TaxModel' => TAX_MODEL_NET});
    }

}

1;





