#========================================================================================
# package      DE_EPAGES::Presentation::API::SiteTLEProcessor
# state        public
#----------------------------------------------------------------------------------------
# description  This module restricts the TLE processor to objects of a single site.
#               If used inside a shop, prevents access to the System object and to other
#               shops.
#========================================================================================
package DE_EPAGES::Presentation::API::SiteTLEProcessor;
use base qw( Exporter DE_EPAGES::TLE::API::BaseProcessor );

use strict;
our @EXPORT_OK = qw(
    ReplaceTLEs
);

use DE_EPAGES::Core::API::PerlTools qw ( ElementOf );
use DE_EPAGES::Presentation::API::Template qw (GetTLEProcessor);
use DE_EPAGES::TLE::API::TLE qw( GetByteCodeByText );

my @FORBIDDEN_TLES = qw (ParentObjects Parent System Shops Site AttributeValues SelectedBranch ParentStyleTemplate);
my @FORBIDDEN_TLEFUNCTIONS = qw (XPATH PARENTPATH SITELOCALE KEY ATTRIBUTEVALUE CHILDREN DUMP STYLEGROUPBYID STYLETEMPLATEBYID);

#========================================================================================
# function     ReplaceTLEs
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Content = ReplaceTLEs($TemplateString, $hVars)
#----------------------------------------------------------------------------------------
# description  Replaces TLE variables in the mail template. Adds formatters and TLE
#               extensions using the TLEProcessorRegistration hook.
#----------------------------------------------------------------------------------------
# input        $TemplateString | template | string
# input        $hVars | TLE variables | ref.hash
# return       $Content | content with replaced variables | string
#========================================================================================
sub ReplaceTLEs {
    my ($TemplateString, $hVars) = @_;

    my $cByteCode = GetByteCodeByText(\$TemplateString);
    my $TLEProcessor = GetTLEProcessor('SitePageProcessor',__PACKAGE__);

    return $TLEProcessor->replaceTLEs($cByteCode, $hVars);
}

#========================================================================================
# function     existsTLE
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Exists = $TLEProcessor->existsTLE( $Name );
#----------------------------------------------------------------------------------------
# description  Returns true if the TLE variable exists.
#----------------------------------------------------------------------------------------
# input        $Name   | the name of the TLE variable | string
# return       $Exists | 1/0, flag to indicate the existence | boolean
#========================================================================================
sub existsTLE {
    my $self = shift;
    my ($Name) = @_;

    return 1 if $self->existsTLEDefault($Name);
    return 1 if $self->existsTLEFormError($Name);

    my @Parts = split /\./, $Name;
    # find first entry
    my $FirstTLE = shift @Parts;
    my $aStack = $self->{'Context'}{'AccessParentVars'} ? [@{$self->{'Stack'}}] : [ $self->{'Stack'}[0] ];
    my $Context = undef;
    my $Found = 0;
    while (scalar @$aStack and not $Found) {
        my $StackEntry = shift @$aStack;
        next unless defined $StackEntry;
        if (ElementOf($FirstTLE, @FORBIDDEN_TLES)) {
            $Found =0;
            last;
        }
        if (exists $StackEntry->{$FirstTLE}) {
            $Found = 1;
            $Context = $StackEntry->{$FirstTLE};
        } elsif ($self->existsTleAtVariableHandler($FirstTLE)) {
            $Found = 1;
            $StackEntry->{$FirstTLE} = $self->tleFromVariableHandler($FirstTLE);
            $Context = $StackEntry->{$FirstTLE};
        }
    }
    return 0 unless $Found;
    # all other parts only in this context
    while (scalar @Parts and ref $Context) {
        my $Part = shift @Parts;
        return 0 unless exists $Context->{$Part};
        return 0 if ElementOf($Part, @FORBIDDEN_TLES);
        $Context = $Context->{$Part};
    }
    $self->{'LocalValue'} = $Context;
    return 1;
}

#========================================================================================
# function     callBlock
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = callBlock($FunctionName, $cParameter, $cTemplate);
#----------------------------------------------------------------------------------------
# description  Forwards template block requests to the function handler, see
#               <function DE_EPAGES::TLE::API::LoopHandler::JOIN> for an implementation
#               example. Forbidden functions are also tested.
#----------------------------------------------------------------------------------------
# input        $FunctionName | name of function (eg LOOP,JOIN) | string
# input        $cParameter | code that returns the parameter list for the block function
#                               as array ref | ref.code
# input        $cTemplate | template bytecode | ref.code
# return       $Value | content of tle $Name | string
#========================================================================================
sub callBlock {
    my $self = shift;
    my ($FunctionName, $cParameter, $cTemplate) = @_;

    # handle forbidden TLE functions
    if (ElementOf($FunctionName, @FORBIDDEN_TLEFUNCTIONS)){
        my $TLEProcessor = GetTLEProcessor('SitePageProcessor',__PACKAGE__);
        my $Object = $TLEProcessor->tle('Object');
        return $Object->tleHash() if defined $Object;
        return undef;
    }

    return $self->SUPER::callBlock($FunctionName, $cParameter, $cTemplate);
}


1;
