#========================================================================================
# package      DE_EPAGES::PriceList::API::Object::PriceList
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for PriceList.
#========================================================================================
package DE_EPAGES::PriceList::API::Object::PriceList;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Core::API::Error qw ( Error);
use DE_EPAGES::Database::API::Connection qw (GetCurrentDBHandle);
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias LoadObject Transaction );
use DE_EPAGES::PriceList::API::Table::PriceListCustomer qw (
    DeletePriceListCustomer
    InsertPriceListCustomer
    ExistsPriceListCustomer
    GetPKeyPriceListCustomersByPriceListID );
use DE_EPAGES::PriceList::API::Table::PriceListCustomerGroup qw (
    DeletePriceListCustomerGroup
    InsertPriceListCustomerGroup
    ExistsPriceListCustomerGroup
    GetPKeyPriceListCustomerGroupsByPriceListID );
use DE_EPAGES::PriceList::API::PriceList qw (IsTimeFrameDifferent);
## no critic qw( RequireCorrectImports )
use DE_EPAGES::PriceList::API::XML::PriceListExportDriver;
## use critic
use DE_EPAGES::Shop::API::Constants qw( TaxModel2Text Text2TaxModel );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::Core::API::TestParameter qw ( TestArray TestInteger TestObject TestString );
use DE_EPAGES::PriceList::API::Table::ScalePrice qw ( DeleteScalePrices GetScalePricesByPriceListID );
use DE_EPAGES::PriceList::API::Table::QuantityDiscount qw (GetQuantityDiscountsByPriceListID);
use DE_EPAGES::PriceList::API::Table::GlobalValueDiscount qw (GetGlobalValueDiscountsByPriceListID);

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $PriceList->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'PriceList' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $PriceList->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $PriceList = DE_EPAGES::PriceList::API::Object::PriceList->insert($hInfo);
# example      my $Class = LoadClassByAlias('PriceList');
#               my $PriceList = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertPriceList' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $PriceList | new object | object
# hook         OBJ_InsertPriceList | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>PriceListID   - price list identifier - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    # check feature
    my $Shop = $hInfo->{'Shop'};
    unless ($Shop->featureMaxValue('PriceLists')) {
        Error('FeatureLimitExceeded', {
                'Feature'  => 'PriceLists',
                'Store'    => $Shop->parent->parent->get('StoreName'),
                'Shop'     => $Shop->alias,
                'MaxValue' => 0,
                'NewValue' => 1,
            });
    }

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertPriceList', {'Object' => $self, 'PriceListID' => $self->id });
    delete $hInfo->{$_} foreach qw ( Shop CurrencyID ValidFrom ValidTo ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $PriceList->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeletePriceList' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeletePriceList | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>PriceListID   - price list identifier - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeletePriceList', {'Object' => $self, 'PriceListID' => $self->id });
    LoadClassByAlias('GlobalValueDiscount')->deleteObjects('ByPriceListID', [$self->id]);
    LoadClassByAlias('ValueDiscount')->deleteObjects('ByPriceListID', [$self->id]);
    LoadClassByAlias('QuantityDiscount')->deleteObjects('ByPriceListID', [$self->id]);

    DeletePriceListCustomer($_) foreach @{GetPKeyPriceListCustomersByPriceListID($self->id)};
    DeletePriceListCustomerGroup($_) foreach @{GetPKeyPriceListCustomerGroupsByPriceListID($self->id)};
    DeleteScalePrices($self->id);

    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $PriceList->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdatePriceList' and 'OBJ_AfterUpdatePriceList'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdatePriceList | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>PriceListID   - price list identifier - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdatePriceList | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>PriceListID   - price list identifier - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdatePriceList', {'Object'=>$self, 'Values'=> $hValues, 'PriceListID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdatePriceList',  {'Object'=>$self, 'Values'=> $hValues, 'PriceListID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::PriceList::API::Object::PriceList->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'Shop'       => $Parent->getSite(),
        'ValidFrom'  => GetCurrentDBHandle()->currentDateTime(),
        'ValidTo'    => GetCurrentDBHandle()->currentDateTime(),
        'CurrencyID' => 0,
    };
}

#========================================================================================
# function     addCustomerGroup
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PriceList->addCustomerGroup( $CustomerGroup );
#----------------------------------------------------------------------------------------
# description  Assigns a customer group to the price list.
#----------------------------------------------------------------------------------------
# input        $CustomerGroup | customer group | object
#========================================================================================
sub addCustomerGroup {
    my $self = shift;
    my ($CustomerGroup ) = @_;

    # Security check: pricelist and customer should belong to the same shop.
    Error('InvalidShop',{ a=>$CustomerGroup->getSite()->id, b=>$self->getSite()->id}) unless $CustomerGroup->getSite()->id == $self->getSite()->id;

    # Check if the customer is already assigned to the price list.
    return if $self->existCustomerGroup( $CustomerGroup );

    # Check if the customer is assigned to other price lists at the same time frame.
    my $From = $self->get('ValidFrom');
    my $To   = $self->get('ValidTo');
    my $CurrencyID = $self->get('CurrencyID');

    my @AssignedPriceLists = grep {$_->get('CurrencyID') eq $CurrencyID } @{$CustomerGroup->get('PriceLists')};
    foreach my $AssignedPriceList (@AssignedPriceLists ) {
        my $RecentFrom = $AssignedPriceList->get('ValidFrom');
        my $RecentTo   = $AssignedPriceList->get('ValidTo');

        unless (IsTimeFrameDifferent( {'ValidFrom' => $From , 'ValidTo' => $To },
                                      {'ValidFrom' => $RecentFrom , 'ValidTo' => $RecentTo } ) ) {
            Error('CustomerGroupAssignedToAnotherPriceListAtThisTimeFrame', {
                'PriceList'=> $self,
                'CustomerGroup' => $CustomerGroup,
                'CustomerGroupAlias' => $CustomerGroup->alias,
                'AssignedPriceList' => $AssignedPriceList,
                'AssignedPriceListAlias' => $AssignedPriceList->alias,
            });
        }
    }
    # Add the customer to the pricelist.
    InsertPriceListCustomerGroup( {
        PriceListID => $self->id,
        CustomerGroupID => $CustomerGroup->id,
    } );

    return;
}

#========================================================================================
# function     addCustomer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PriceList->addCustomer( $Customer );
#----------------------------------------------------------------------------------------
# description  Assigns a customer to the price list.
#----------------------------------------------------------------------------------------
# input        $Customer | customer | object
#========================================================================================
sub addCustomer {
    my $self = shift;
    my ($Customer) = @_;

    # Security check: pricelist and customer should belong to the same shop.
    Error('InvalidShop',{ a=>$Customer->getSite()->id, b=>$self->getSite()->id}) unless $Customer->getSite()->id == $self->getSite()->id;

    # Check if the customer is already assigned to the price list.
    return if $self->existCustomer( $Customer );

    # Check if the customer is assigned to other price lists at the same time frame.
    my $From = $self->get('ValidFrom');
    my $To   = $self->get('ValidTo');
    my $CurrencyID = $self->get('CurrencyID');

    my @AssignedPriceLists = grep {$_->get('CurrencyID') eq $CurrencyID } @{$Customer->get('PriceLists')};
    foreach my $AssignedPriceList (@AssignedPriceLists ) {
        my $RecentFrom = $AssignedPriceList->get('ValidFrom');
        my $RecentTo   = $AssignedPriceList->get('ValidTo');

        unless (IsTimeFrameDifferent( {'ValidFrom' => $From , 'ValidTo' => $To },
                                      {'ValidFrom' => $RecentFrom , 'ValidTo' => $RecentTo } ) ) {
            Error('CustomerAssignedToAnotherPriceListAtThisTimeFrame', {
                'PriceList' => $self,
                'Customer' => $Customer,
                'CustomerAlias' => $Customer->alias,
                'AssignedPriceList' => $AssignedPriceList,
                'AssignedPriceListAlias' => $AssignedPriceList->alias,
            });
        }
    }
    # Add the customer to the pricelist.
    InsertPriceListCustomer( {
        PriceListID => $self->id,
        CustomerID => $Customer->id,
    } );

    return;
}

#========================================================================================
# function     existCustomer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PriceList->existCustomer( $Customer );
#----------------------------------------------------------------------------------------
# description  check if a customer still assigned to the price list.
#----------------------------------------------------------------------------------------
# input        $Customer | customer | object
# return       $exists | 1/0  the Customer exists (yes/no) | boolean
#========================================================================================
sub existCustomer {
    my $self = shift;
    my ($Customer) = @_;

    return ExistsPriceListCustomer( $self->id, $Customer->id );
}

#========================================================================================
# function     existCustomerGroup
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PriceList->existCustomerGroup( $CustomerGroup );
#----------------------------------------------------------------------------------------
# description  check if a group still assigned to the price list.
#----------------------------------------------------------------------------------------
# input        $CustomerGroup | customer group| object
# return       $exists | 1/0  the CustomerGroup exists (yes/no) | boolean
#========================================================================================
sub existCustomerGroup {
    my $self = shift;
    my ($CustomerGroup) = @_;

    return ExistsPriceListCustomerGroup( $self->id, $CustomerGroup->id );
}

#========================================================================================
# function     removeCustomer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PriceList->removeCustomer( $Customer );
#----------------------------------------------------------------------------------------
# description  Removes the assignment of a customer to the price list.
#----------------------------------------------------------------------------------------
# input        $Customer | customer | object
#========================================================================================
sub removeCustomer {
    my $self = shift;
    my ($Customer) = @_;
    TestObject('Customer', $Customer, 'DE_EPAGES::Customer::API::Object::Customer');

    # security check: pricelist and customer should belong to the same shop
    Error('InvalidShop',{ a=>$Customer->getSite()->id, b=>$self->getSite()->id}) unless $Customer->getSite()->id == $self->getSite()->id;

    DeletePriceListCustomer( $self->id, $Customer->id ) if ( $self->existCustomer($Customer) );

    return;
}

#========================================================================================
# function     removeCustomerGroup
# state        public
#----------------------------------------------------------------------------------------
# syntax       $PriceList->removeCustomerGroup( $CustomerGroup );
#----------------------------------------------------------------------------------------
# description  Removes the assignment of a customer group to the price list.
#----------------------------------------------------------------------------------------
# input        $CustomerGroup | customer group | object
#========================================================================================
sub removeCustomerGroup {
    my $self = shift;
    my ($CustomerGroup) = @_;
    TestObject('CustomerGroup', $CustomerGroup, 'DE_EPAGES::Customer::API::Object::CustomerGroup');

    # security check: pricelist and customer group should belong to the same shop
    Error('InvalidShop',{ a=>$CustomerGroup->getSite()->id, b=>$self->getSite()->id}) unless $CustomerGroup->getSite()->id == $self->getSite()->id;

    DeletePriceListCustomerGroup( $self->id, $CustomerGroup->id ) if ( $self->existCustomerGroup($CustomerGroup) );

    return;
}

sub checkCanSetValidDates {
    my $self = shift;
    my ($ValidDates) = @_;
    my $CurrencyID = $self->get('CurrencyID');

    #test for all assigned customergroups
    my $AssignedCustomerGroups = $self->get('CustomerGroups');
    foreach my $CustomerGroup ( @$AssignedCustomerGroups) {
        #test all price lists with the same currency
        my @CustomerGroupPriceLists = grep {$_->get('CurrencyID') eq $CurrencyID } @{$CustomerGroup->get('PriceLists')};
        foreach my $CustomerGroupPriceList ( @CustomerGroupPriceLists) {
            next if $CustomerGroupPriceList->id == $self->id;
            my $CustomerGroupPriceListFrom = $CustomerGroupPriceList->get('ValidFrom');
            my $CustomerGroupPriceListTo   = $CustomerGroupPriceList->get('ValidTo');
            unless ( IsTimeFrameDifferent ( $ValidDates, {'ValidFrom' => $CustomerGroupPriceListFrom, 'ValidTo' => $CustomerGroupPriceListTo} ) ) {
                Error('CustomerGroupAssignedToAnotherPriceListAtThisTimeFrame', {
                    'PriceList' => $self,
                    'CustomerGroup' => $CustomerGroup,
                    'CustomerGroupAlias' => $CustomerGroup->alias,
                    'AssignedPriceList' => $CustomerGroupPriceList,
                    'AssignedPriceListAlias' => $CustomerGroupPriceList->alias,
                });
            }
        }
    }
    #test for all assigned customers
    my $AssignedCustomers = $self->get('Customers');
    foreach my $Customer ( @$AssignedCustomers) {
        #test all price lists with the same currency
        my @CustomerPriceLists = grep {$_->get('CurrencyID') eq $CurrencyID } @{$Customer->get('PriceLists')};
        foreach my $CustomerPriceList ( @CustomerPriceLists) {
            next if $CustomerPriceList->id == $self->id;
            my $CustomerPriceListFrom = $CustomerPriceList->get('ValidFrom');
            my $CustomerPriceListTo   = $CustomerPriceList->get('ValidTo');
            unless ( IsTimeFrameDifferent ( $ValidDates, {'ValidFrom' => $CustomerPriceListFrom, 'ValidTo' => $CustomerPriceListTo} ) ) {
                Error('CustomerAssignedToAnotherPriceListAtThisTimeFrame', {
                    'PriceList' => $self,
                    'Customer' => $Customer,
                    'CustomerAlias' => $Customer->alias,
                    'AssignedPriceList' => $CustomerPriceList,
                    'AssignedPriceListAlias' => $CustomerPriceList->alias,
                });
            }
        }
    }

    return 1;
}

#========================================================================================
# function     exportXMLAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hAttributes = $Object->exportXMLAttributes( $Driver, $hAllAttributes );
#----------------------------------------------------------------------------------------
# description  Returns plain attributes of object to XML export driver. This function
#               removes bit attributes with value 0.
#----------------------------------------------------------------------------------------
# input        $Driver | xml export driver with support function addObject | object
# input        $hAllAttributes | all attributes of object,
#                                 see $Object-><function exportableAttributes> | ref.hash
# return       $hAttributes | plain attributes | ref.hash.string
#========================================================================================
sub exportXMLAttributes {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;
    $hAttributes = $self->SUPER::exportXMLAttributes($Driver, $hAttributes);

    # map TaxModel name to Modell id
    if (exists $hAttributes->{'TaxModel'}) {
        $hAttributes->{'TaxModel'} = TaxModel2Text($hAttributes->{'TaxModel'});
    }
    return $hAttributes;
}

#========================================================================================
# function     importXML
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object->importXML($ObjectPackage, $Handler, $hElement, $Class, $hAttributes);
#----------------------------------------------------------------------------------------
# description  adds special handling for tax class
#----------------------------------------------------------------------------------------
# input        $ObjectPackage | package of object | string
# input        $Handler | xml import handler with support function addObject
#                       | DE_EPAGES::Object::API::XML::ObjectImportHandler
# input        $hElement | XML element structure, see XML::SAX | ref.hash
# input        $Class | class object; the new object will be an instance of this class
#                       | DE_EPAGES::Object::API::Object::Class
# input        $hAttributes | attributes hash | ref.hash
# return       $Object | the imported object | DE_EPAGES::Object::API::Object::Object or subclass
#========================================================================================
sub importXML {
    my ($ObjectPackage, $Handler, $hElement, $Class, $hAttributes) = @_;

    # map TaxModel name to Modell id
    if (exists $hAttributes->{'TaxModel'}) {
        $hAttributes->{'TaxModel'} = Text2TaxModel($hAttributes->{'TaxModel'});
    }

    return $ObjectPackage->SUPER::importXML($Handler, $hElement, $Class, $hAttributes);
}

#========================================================================================
# function     exportXMLElements
# state        private
#----------------------------------------------------------------------------------------
# syntax       $aElements = $Object->exportXMLElements($Driver, $hAllAttributes);
#----------------------------------------------------------------------------------------
# description  export xml data to export multilinguagal properties of this object.
#----------------------------------------------------------------------------------------
# input        $Driver | export driver | object
# input        $hAllAttributes | all attributes of object,
#                                 see $Object-><function allAttributes> | ref.hash
#========================================================================================
sub exportXMLElements {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;
    $self->SUPER::exportXMLElements($Driver, $hAttributes);
    if ($self->class->get('ExportLevel') == $Driver->exportLevel) {
        $Driver->PriceListGlobalValueDiscountsByPriceList();
    }
    return;
}

#========================================================================================
# function     scalePrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aScalePrices = $PriceList->scalePrices( $Product );
#----------------------------------------------------------------------------------------
# description  Returns all scale prices for the given product in this price list.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $aScalePrices | list of ScalePrice objects | ref.array.object
#========================================================================================
sub scalePrices {
    my $self = shift;
    my ($Product) = @_;

    return LoadClassByAlias('ScalePrice')->getObjects('ByProductIDPriceListID', {
        'ProductID' => $Product->id,
        'PriceListID' => $self->id,
    });
}

#========================================================================================
# function     getScalePrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahScalePrices = $Object->getScalePrices($Product, $TaxModel, );
# syntax       $ahScalePrices = $Object->getScalePrices($Product, $TaxModel, $ExtraPrecision);
#----------------------------------------------------------------------------------------
# description  Returns the scale prices for the given product and tax model in
#               the requested precision.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $ExtraPrecision | (optional, default=0) extra precision for net prices
#                           in back-office | boolean
# return       $ahScalePrices | list of hashes, keys:
#                           Price, Quantity
#                           | ref.array.hash
#========================================================================================
sub getScalePrices {
    my $self = shift;
    my ($Product, $TaxModel, $ExtraPrecision) = @_;

    TestObject('Product', $Product);
    TestInteger('TaxModel', $TaxModel);
    $ExtraPrecision = 0 unless defined $ExtraPrecision;

    my $TaxClass = $Product->getWithParent('TaxClass');

    return $self->_getScalePrices($Product->id, $TaxModel, $TaxClass, $ExtraPrecision);
}

#========================================================================================
# function     getScalePricesByProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahScalePrices = $Object->getScalePricesByProductID($ProductID, $TaxModel, $TaxCLass);
# syntax       $ahScalePrices = $Object->getScalePricesByProductID($ProductID, $TaxModel, $TaxCLass, $ExtraPrecision);
#----------------------------------------------------------------------------------------
# description  Returns the scale prices for the given product id, tax model and
#               tax class in the requested precision.
#----------------------------------------------------------------------------------------
# input        $ProductID | product id | String
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $TaxClass | tax class | TaxClass
# input        $ExtraPrecision | (optional, default=0) extra precision for net prices
#                           in back-office | boolean
# return       $ahScalePrices | list of hashes, keys:
#                           Price, Quantity
#                           | ref.array.hash
#========================================================================================
sub getScalePricesByProductID {
    my $self = shift;
    my ($ProductID, $TaxModel, $TaxClass, $ExtraPrecision) = @_;

    TestString('ProductID', $ProductID);
    TestInteger('TaxModel', $TaxModel);
    TestObject('TaxClass', $TaxClass);
    $ExtraPrecision = 0 unless defined $ExtraPrecision;

    return $self->_getScalePrices($ProductID, $TaxModel, $TaxClass, $ExtraPrecision);
}

#========================================================================================
# function     _getScalePrices
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ahScalePrices = $Object->_getScalePrices($ProductID, $TaxModel, $TaxCLass);
# syntax       $ahScalePrices = $Object->_getScalePrices($ProductID, $TaxModel, $TaxCLass, $ExtraPrecision);
#----------------------------------------------------------------------------------------
# description  Calculates the scale prices for the given product, tax model and
#               tax class in the requested precision.
#----------------------------------------------------------------------------------------
# input        $ProductID | product | object
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $TaxClass | tax class | taxClass
# input        $ExtraPrecision | (optional, default=0) extra precision for net prices
#                           in back-office | boolean
# return       $ahScalePrices | list of hashes, keys:
#                           Price, Quantity
#                           | ref.array.hash
#========================================================================================
sub _getScalePrices {
    my $self = shift;
    my ($ProductID, $TaxModel, $TaxClass, $ExtraPrecision) = @_;

    my $aScalePrices = LoadClassByAlias('ScalePrice')->getObjects('ByProductIDPriceListID', {
        'ProductID' => $ProductID,
        'PriceListID' => $self->id,
    });

    # convert objects to hashes
    my @ScalePrices = sort { $a->{'Price'} <=> $b->{'Price'} }
        map { $_->get([qw(Price Quantity)]) }
        @$aScalePrices;

    # get the TaxMatrix from class and not from the shop directly (important for delivery threshold!)
    my $TaxMatrix = $TaxClass->get('TaxMatrix');
    my $CurrencyID = $self->get('CurrencyID');
    # convert net prices to gross
    foreach my $hScalePrice (@ScalePrices) {
        $hScalePrice->{'Price'} = $TaxMatrix->net2gross($CurrencyID, $hScalePrice->{'Price'},
                                                       {'TaxModel' => $TaxModel,
                                                        'TaxClass' => $TaxClass
                                                       }, $ExtraPrecision);
    }

    return \@ScalePrices;
}

#========================================================================================
# function     setScalePrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object->setScalePrices($Product, $TaxModel, $ahScalePrices);
#----------------------------------------------------------------------------------------
# description  Sets all scale prices for the given product and tax model.
#               Deletes scale prices that are not included in the list.
#               Calculates net prices from gross prices if $TaxModel is TAX_MODEL_GROSS.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $ahScalePrices | list of hashes, keys:
#                           Price, Quantity
#                           | ref.array.hash
#========================================================================================
sub setScalePrices {
    my $self = shift;
    my ($Product, $TaxModel, $ahScalePrices) = @_;

    TestObject('Product', $Product);
    TestInteger('TaxModel', $TaxModel);
    TestArray('ahScalePrices', $ahScalePrices);


    # prepare values, make a local copy
    my %ScalePricesByQuantity;
    foreach my $hScalePrice (@$ahScalePrices) {
        my $Quantity = $hScalePrice->{'Quantity'};
        $Quantity = 0 unless defined $Quantity;
        my @Keys = qw(Price Quantity);
        $ScalePricesByQuantity{$Quantity}{$_} = $hScalePrice->{$_} foreach @Keys;
    }

    # convert gross prices to net
    my $TaxMatrix = $self->getSite()->get('TaxMatrix');
    my $TaxClass = $Product->getWithParent('TaxClass');
    my $CurrencyID = $self->get('CurrencyID');
    foreach my $hScalePrice (values %ScalePricesByQuantity) {
        $hScalePrice->{'Price'} = $TaxMatrix->gross2net($CurrencyID, $hScalePrice->{'Price'},
                                                       {'TaxModel' => $TaxModel,
                                                        'TaxClass' => $TaxClass,
                                                       });

    }

    # update existing scale prices, delete obsolete ones
    my $aScalePrices = $self->scalePrices($Product);
    Transaction( sub {
        foreach my $ScalePrice (@$aScalePrices) {
            my $Quantity = $ScalePrice->get('Quantity');
            $Quantity = 0 unless defined $Quantity;
            if( exists $ScalePricesByQuantity{$Quantity} ) {
                $ScalePrice->set( $ScalePricesByQuantity{$Quantity} );
                delete $ScalePricesByQuantity{$Quantity};
            }
            else {
                $ScalePrice->delete;
            }
        }

        # add new levels
        foreach my $hScalePrice (values %ScalePricesByQuantity) {
            LoadClassByAlias('ScalePrice')->insertObject({
                'PriceList' => $self,
                'Product'   => $Product,
                'Price'     => $hScalePrice->{'Price'},
                'Quantity'  => $hScalePrice->{'Quantity'},
            });
        }
    });

    return;
}

#========================================================================================
# function     hasScalePrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasScalePrices = $PriceList->hasScalePrices( $Product );
#----------------------------------------------------------------------------------------
# description  Returns true if the pricelist includes scale prices for the given
#               product.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $HasScalePrices | true if a scale price exists | boolean
#========================================================================================
sub hasScalePrices {
    my $self = shift;
    my ($Product) = @_;

    return scalar @{LoadClassByAlias('ScalePrice')->findObjects('ByProductIDPriceListID', [$Product->id, $self->id])} > 0 ? 1 : 0;
}

#========================================================================================
# function     hasScalePricesByProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasScalePrices = $PriceList->hasScalePricesByProductID( $ProductID );
#----------------------------------------------------------------------------------------
# description  Returns true if the pricelist includes scale prices for the given
#               product id.
#----------------------------------------------------------------------------------------
# input        $ProductID | id of the product | string
# return       $HasScalePrices | true if a scale price exists | boolean
#========================================================================================
sub hasScalePricesByProductID {
    my $self = shift;
    my ($ProductID) = @_;

    return scalar @{LoadClassByAlias('ScalePrice')->findObjects('ByProductIDPriceListID', [$ProductID, $self->id])} > 0 ? 1 : 0;
}

#========================================================================================
# function     quantityDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aQuantityDiscounts = $PriceList->quantityDiscounts( $Product );
#----------------------------------------------------------------------------------------
# description  Returns all quantity discounts for the given product in this price list.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $aQuantityDiscounts | list of QuantityDiscount objects | ref.array.object
#========================================================================================
sub quantityDiscounts {
    my $self = shift;
    my ($Product) = @_;

    return LoadClassByAlias('QuantityDiscount')->getObjects('ByProductIDPriceListID', {
        'ProductID' => $Product->id,
        'PriceListID' => $self->id,
    });
}

#========================================================================================
# function     getQuantityDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahQuantityDiscounts = $Object->getQuantityDiscounts($Product);
#----------------------------------------------------------------------------------------
# description  Returns the quantity discounts for the given product.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $ahQuantityDiscounts | list of hashes, keys:
#                           Discount - discount in percent,
#                           Quantity - product quantity
#                           | int
#========================================================================================
sub getQuantityDiscounts {
    my $self = shift;
    my ($Product) = @_;

    TestObject('Product', $Product);

    return $self->_getQuantityDiscounts($Product->id);
}

#========================================================================================
# function     getQuantityDiscountsByProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahQuantityDiscounts = $Object->getQuantityDiscountsByProductID($ProductID);
#----------------------------------------------------------------------------------------
# description  Returns the quantity discounts for the given product.
#----------------------------------------------------------------------------------------
# input        $ProductID | product id | String
# return       $ahQuantityDiscounts | list of hashes, keys:
#                           Discount - discount in percent,
#                           Quantity - product quantity
#                           | int
#========================================================================================
sub getQuantityDiscountsByProductID {
    my $self = shift;
    my ($ProductID) = @_;

    TestString('ProductID', $ProductID);

    return $self->_getQuantityDiscounts($ProductID);
}

#========================================================================================
# function     _getQuantityDiscounts
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ahQuantityDiscounts = $Object->_getQuantityDiscounts($ProductID);
#----------------------------------------------------------------------------------------
# description  Calculates the quantity discounts for the given product.
#----------------------------------------------------------------------------------------
# input        $ProductID | product | String
# return       $ahQuantityDiscounts | list of hashes, keys:
#                           Discount - discount in percent,
#                           Quantity - product quantity
#                           | int
#========================================================================================
sub _getQuantityDiscounts {
    my $self = shift;
    my ($ProductID) = @_;

    my $aQuantityDiscounts = LoadClassByAlias('QuantityDiscount')->getObjects('ByProductIDPriceListID', {
        'ProductID' => $ProductID,
        'PriceListID' => $self->id,
    });

    # convert objects to hashes
    my @QuantityDiscounts = sort { $a->{'Quantity'} <=> $b->{'Quantity'} }
        map { $_->get([qw(Discount Quantity)]) }
        @$aQuantityDiscounts;

    return \@QuantityDiscounts;
}

#========================================================================================
# function     setQuantityDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object->setQuantityDiscounts($Product, $TaxModel, $ahQuantityDiscounts);
#----------------------------------------------------------------------------------------
# description  Sets all quantity discounts for the given product and tax model.
#               Deletes quantity discounts that are not included in the list.
#               Calculates net prices from gross prices if $TaxModel is TAX_MODEL_GROSS.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $ahQuantityDiscounts | list of hashes, keys:
#                           Quantity, Discount
#                           | ref.array.hash
#========================================================================================
sub setQuantityDiscounts {
    my $self = shift;
    my ($Product, $TaxModel, $ahQuantityDiscounts) = @_;

    TestObject('Product', $Product);
    TestInteger('TaxModel', $TaxModel);
    TestArray('ahQuantityDiscounts', $ahQuantityDiscounts);

    # prepare values, make a local copy
    my %QuantityDiscountsByQuantity;
    foreach my $hQuantityDiscount (@$ahQuantityDiscounts) {
        my $Quantity = $hQuantityDiscount->{'Quantity'};
        $Quantity = 0 unless defined $Quantity;
        my @Keys = qw(Quantity Discount);
        $QuantityDiscountsByQuantity{$Quantity}{$_} = $hQuantityDiscount->{$_} foreach @Keys;
    }

    # update existing value discounts, delete obsolete ones
    my $aQuantityDiscounts = $self->quantityDiscounts($Product);
    Transaction( sub {
        foreach my $QuantityDiscount (@$aQuantityDiscounts) {
            my $Quantity = $QuantityDiscount->get('Quantity');
            $Quantity = 0 unless defined $Quantity;
            if( exists $QuantityDiscountsByQuantity{$Quantity} ) {
                $QuantityDiscount->set( $QuantityDiscountsByQuantity{$Quantity} );
                delete $QuantityDiscountsByQuantity{$Quantity};
            }
            else {
                $QuantityDiscount->delete;
            }
        }

        # add new value discounts
        foreach my $hQuantityDiscount (values %QuantityDiscountsByQuantity) {
            LoadClassByAlias('QuantityDiscount')->insertObject({
                'PriceList' => $self,
                'Product'   => $Product,
                'Quantity'  => $hQuantityDiscount->{'Quantity'},
                'Discount'  => $hQuantityDiscount->{'Discount'},
            });
        }
    });

    return;
}

#========================================================================================
# function     hasQuantityDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasDiscounts = $PriceList->hasQuantityDiscounts( $Product );
#----------------------------------------------------------------------------------------
# description  Returns true if the pricelist includes a quantity discount for the given
#               product.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $HasDiscounts | true if a discount exists | boolean
#========================================================================================
sub hasQuantityDiscounts {
    my $self = shift;
    my ($Product) = @_;

    return scalar @{LoadClassByAlias('QuantityDiscount')->findObjects('ByProductIDPriceListID', [$Product->id, $self->id])} > 0 ? 1 : 0;
}

#========================================================================================
# function     hasQuantityDiscountsByProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasDiscounts = $PriceList->hasQuantityDiscountsByProductID( $ProductID );
#----------------------------------------------------------------------------------------
# description  Returns true if the pricelist includes a quantity discount for the given
#               product id.
#----------------------------------------------------------------------------------------
# input        $ProductID | id of the product | string
# return       $HasDiscounts | true if a discount exists | boolean
#========================================================================================
sub hasQuantityDiscountsByProductID {
    my $self = shift;
    my ($ProductID) = @_;

    return scalar @{LoadClassByAlias('QuantityDiscount')->findObjects('ByProductIDPriceListID', [$ProductID, $self->id])} > 0 ? 1 : 0;
}

#========================================================================================
# function     valueDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aValueDiscounts = $PriceList->valueDiscounts( $Product );
#----------------------------------------------------------------------------------------
# description  Returns all value discounts for the given product in this price list.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $aValueDiscounts | list of ValueDiscount objects | ref.array.object
#========================================================================================
sub valueDiscounts {
    my $self = shift;
    my ($Product) = @_;

    return LoadClassByAlias('ValueDiscount')->getObjects('ByProductIDPriceListID', {
        'ProductID' => $Product->id,
        'PriceListID' => $self->id,
    });
}

#========================================================================================
# function     getValueDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahValueDiscounts = $Object->getValueDiscounts($Product, $TaxModel, );
# syntax       $ahValueDiscounts = $Object->getValueDiscounts($Product, $TaxModel, $ExtraPrecision);
#----------------------------------------------------------------------------------------
# description  Returns the value discounts for the given product and tax model in
#               the requested precision.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $ExtraPrecision | (optional, default=0) extra precision for net prices
#                           in back-office | boolean
# return       $ahValueDiscounts | list of hashes, keys:
#                           Discount, Value
#                           | ref.array.hash
#========================================================================================
sub getValueDiscounts {
    my $self = shift;
    my ($Product, $TaxModel, $ExtraPrecision) = @_;

    TestObject('Product', $Product);
    TestInteger('TaxModel', $TaxModel);
    $ExtraPrecision = 0 unless defined $ExtraPrecision;

    my $TaxClass = $Product->getWithParent('TaxClass');

    return $self->_getValueDiscounts($Product->id, $TaxModel, $TaxClass, $ExtraPrecision);
}

#========================================================================================
# function     getValueDiscountsByProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahValueDiscounts = $Object->getValueDiscountsByProductID($ProductID, $TaxModel, $TaxClass);
# syntax       $ahValueDiscounts = $Object->getValueDiscountsByProductID($ProductID, $TaxModel, $TaxClass, $ExtraPrecision);
#----------------------------------------------------------------------------------------
# description  Returns the value discounts for the given product, tax model, tax class in
#               the requested precision.
#----------------------------------------------------------------------------------------
# input        $ProductID | product id | String
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $TaxClass | tax class | taxClass
# input        $ExtraPrecision | (optional, default=0) extra precision for net prices
#                           in back-office | boolean
# return       $ahValueDiscounts | list of hashes, keys:
#                           Discount, Value
#                           | ref.array.hash
#========================================================================================
sub getValueDiscountsByProductID {
    my $self = shift;
    my ($ProductID, $TaxModel, $TaxClass, $ExtraPrecision) = @_;

    TestString('ProductID', $ProductID);
    TestInteger('TaxModel', $TaxModel);
    TestObject('TaxClass', $TaxClass);
    $ExtraPrecision = 0 unless defined $ExtraPrecision;

    return $self->_getValueDiscounts($ProductID, $TaxModel, $TaxClass, $ExtraPrecision);
}

#========================================================================================
# function     _getValueDiscounts
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ahValueDiscounts = $Object->_getValueDiscounts($ProductID, $TaxModel, $TaxClass);
# syntax       $ahValueDiscounts = $Object->_getValueDiscounts($ProductID, $TaxModel, $TaxClass, $ExtraPrecision);
#----------------------------------------------------------------------------------------
# description  Returns the value discounts for the given product, tax model and
#               tax class in the requested precision.
#----------------------------------------------------------------------------------------
# input        $ProductID | product id | String
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $TaxClass | tax class | taxClass
# input        $ExtraPrecision | (optional, default=0) extra precision for net prices
#                           in back-office | boolean
# return       $ahValueDiscounts | list of hashes, keys:
#                           Discount, Value
#                           | ref.array.hash
#========================================================================================
sub _getValueDiscounts {
    my $self = shift;
    my ($ProductID, $TaxModel, $TaxClass, $ExtraPrecision) = @_;

    my $aValueDiscounts =  LoadClassByAlias('ValueDiscount')->getObjects('ByProductIDPriceListID', {
        'ProductID' => $ProductID,
        'PriceListID' => $self->id,
    });

    # convert objects to hashes
    my @ValueDiscounts = sort { $a->{'Value'} <=> $b->{'Value'} }
        map { $_->get([qw(Discount Value)]) }
        @$aValueDiscounts;

    # get the TaxMatrix from class and not from the shop directly (important for delivery threshold!)
    my $TaxMatrix = $TaxClass->get('TaxMatrix');
    my $CurrencyID = $self->get('CurrencyID');
    # convert net prices to gross
    foreach my $hValueDiscounts (@ValueDiscounts) {
        $hValueDiscounts->{'Value'} = $TaxMatrix->net2gross($CurrencyID, $hValueDiscounts->{'Value'},
                                                           {'TaxModel' => $TaxModel,
                                                            'TaxClass' => $TaxClass
                                                           }, $ExtraPrecision);
    }

    return \@ValueDiscounts;
}

#========================================================================================
# function     setValueDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object->setValueDiscounts($Product, $TaxModel, $ahValueDiscounts);
#----------------------------------------------------------------------------------------
# description  Sets all value discounts for the given product and tax model.
#               Deletes value discounts that are not included in the list.
#               Calculates net prices from gross prices if $TaxModel is TAX_MODEL_GROSS.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# input        $TaxModel | tax model (1=gross/including tax, 0=net/without tax) | int
# input        $ahScalePrices | list of hashes, keys:
#                           Value, Discount
#                           | ref.array.hash
#========================================================================================
sub setValueDiscounts {
    my $self = shift;
    my ($Product, $TaxModel, $ahValueDiscounts) = @_;

    TestObject('Product', $Product);
    TestInteger('TaxModel', $TaxModel);
    TestArray('ahValueDiscounts', $ahValueDiscounts);

    # prepare values, make a local copy
    my %ValueDiscountsByValue;
    foreach my $hValueDiscount (@$ahValueDiscounts) {
        my $Value = $hValueDiscount->{'Value'};
        $Value = 0 unless defined $Value;
        my @Keys = qw(Value Discount);
        $ValueDiscountsByValue{$Value}{$_} = $hValueDiscount->{$_} foreach @Keys;
    }

    # convert gross prices to net
    my $TaxMatrix = $self->getSite()->get('TaxMatrix');
    my $TaxClass = $Product->getWithParent('TaxClass');
    my $CurrencyID = $self->get('CurrencyID');
    foreach my $hValueDiscount (values %ValueDiscountsByValue) {
        $hValueDiscount->{'Value'} = $TaxMatrix->gross2net($CurrencyID, $hValueDiscount->{'Value'},
                                                          {'TaxModel' => $TaxModel,
                                                           'TaxClass' => $TaxClass,
                                                          });

    }

    # update existing value discounts, delete obsolete ones
    my $aValueDiscounts = $self->valueDiscounts($Product);
    Transaction( sub {
        foreach my $ValueDiscount (@$aValueDiscounts) {
            my $Value = $ValueDiscount->get('Value');
            $Value = 0 unless defined $Value;
            if( exists $ValueDiscountsByValue{$Value} ) {
                $ValueDiscount->set( $ValueDiscountsByValue{$Value} );
                delete $ValueDiscountsByValue{$Value};
            }
            else {
                $ValueDiscount->delete;
            }
        }

        # add new value discounts
        my $Class = LoadClassByAlias('ValueDiscount');
        foreach my $hValueDiscount (values %ValueDiscountsByValue) {
            $Class->insertObject({
                'PriceList' => $self,
                'Product'   => $Product,
                'Value'     => $hValueDiscount->{'Value'},
                'Discount'  => $hValueDiscount->{'Discount'},
            });
        }
    });

    return;
}

#========================================================================================
# function     hasValueDiscounts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasDiscounts = $PriceList->hasValueDiscounts( $Product );
#----------------------------------------------------------------------------------------
# description  Returns true if the pricelist includes a value discount for the given
#               product.
#----------------------------------------------------------------------------------------
# input        $Product | product | object
# return       $HasDiscounts | true if a discount exists | boolean
#========================================================================================
sub hasValueDiscounts {
    my $self = shift;
    my ($Product) = @_;

    return scalar @{LoadClassByAlias('ValueDiscount')->findObjects('ByProductIDPriceListID', [$Product->id, $self->id])} > 0 ? 1 : 0;
}

#========================================================================================
# function     hasValueDiscountsByProductID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasDiscounts = $PriceList->hasValueDiscountsByProductID( $ProductID );
#----------------------------------------------------------------------------------------
# description  Returns true if the pricelist includes a value discount for the given
#               product id.
#----------------------------------------------------------------------------------------
# input        $ProductID | id of the product | string
# return       $HasDiscounts | true if a discount exists | boolean
#========================================================================================
sub hasValueDiscountsByProductID {
    my $self = shift;
    my ($ProductID) = @_;

    return scalar @{LoadClassByAlias('ValueDiscount')->findObjects('ByProductIDPriceListID', [$ProductID, $self->id])} > 0 ? 1 : 0;
}

#========================================================================================
# function     clone
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClonedPriceList = $PriceList->clone( $hParams, $Recursive );
#----------------------------------------------------------------------------------------
# description  Clone the given PriceList and insert ScalePrice, QuantityDiscounts
#               and GlobalValueDiscounts for cloned PriceList.
#----------------------------------------------------------------------------------------
# # input      $hParams | hash with mandatory attribute values.
#                       <ul>
#                         <li>Alias - alias - varchar(255)</li>
#                         <li>CustomerGroups - customergroups object - object</li>
#                       </ul> | hash ref
# input        $Recursive | clone child object too (optional, default is false) | boolean
# return       $NewPriceList | object of cloned pricelist | object
#========================================================================================
sub clone {
    my $self = shift;

    my ($hParams, $Recursive) = @_;
    my $Shop = $self->getSite();
    my $PriceList = $self;

    # Clone $PriceList as $NewPriceList
    my $NewPriceList = $self->SUPER::clone($hParams, $Recursive);

    my $ahPriceListScalePrices = GetScalePricesByPriceListID($Shop->id, $PriceList->id);

    # Import ScalePrice for cloned PriceList
    if (defined $ahPriceListScalePrices && scalar($ahPriceListScalePrices)) {
        foreach my $hPriceListScalePrice (@$ahPriceListScalePrices) {
            my $Product = LoadObject($hPriceListScalePrice->{'ProductID'});
            LoadClassByAlias('ScalePrice')->insertObject({
                'PriceList' => $NewPriceList,
                'Product'   => $Product,
                'Quantity'  => $hPriceListScalePrice->{'Quantity'},
                'Price'     => $hPriceListScalePrice->{'Price'},
            });
        }
    }

    # get QuantityDiscounts for current PriceList
    my $ahPriceListQuantityDiscounts = GetQuantityDiscountsByPriceListID($Shop->id, $PriceList->id);

    # Import GlobalValueDiscount for cloned PriceList
    if (defined $ahPriceListQuantityDiscounts && scalar($ahPriceListQuantityDiscounts)) {
        foreach my $hPriceListQuantityDiscount (@$ahPriceListQuantityDiscounts) {
            my $Product = LoadObject($hPriceListQuantityDiscount->{'ProductID'});
            LoadClassByAlias('QuantityDiscount')->insertObject({
                'PriceList'         => $NewPriceList,
                'Product'           => $Product,
                'Quantity'          => $hPriceListQuantityDiscount->{'Quantity'},
                'Discount'          => $hPriceListQuantityDiscount->{'Discount'},
            });
        }
    }

    # get GlobalValueDiscounts for current PriceList
    my $ahPriceListGlobalValueDiscounts = GetGlobalValueDiscountsByPriceListID($Shop->id, $PriceList->id);

    # Import GlobalValueDiscount for cloned PriceList
    if (defined $ahPriceListGlobalValueDiscounts && scalar($ahPriceListGlobalValueDiscounts)) {
        foreach my $hPriceListGlobalValueDiscount (@$ahPriceListGlobalValueDiscounts) {
            LoadClassByAlias('GlobalValueDiscount')->insertObject({
                'PriceList'         => $NewPriceList,
                'Value'             => $hPriceListGlobalValueDiscount->{'Value'},
                'Discount'          => $hPriceListGlobalValueDiscount->{'Discount'},
                'DiscountAbsolute'  => $hPriceListGlobalValueDiscount->{'DiscountAbsolute'},
            });
        }
    }
    return $NewPriceList;
}

1;
