#========================================================================================
# package      DE_EPAGES::PriceList::Hooks::Product
# state        public
# description  set prices and discounts
#========================================================================================
package DE_EPAGES::PriceList::Hooks::Product;

use strict;
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::PriceList::API::PriceList qw ( GetProductPriceList );

#========================================================================================
# function     ProductPrice
# state        public
#----------------------------------------------------------------------------------------
# syntax       ProductPrice( $hParams );
#----------------------------------------------------------------------------------------
# description  Hook function for the ProductPrice hook.
#               Sets the Price parameter if a scale price is applicable for the
#               given quantity of the given product.
#----------------------------------------------------------------------------------------
# input        $hParams |<ul>
#                   <li>Object - product - object</li>
#                   <li>Quantity - (optional, default=1) quantity - float</li>
#                   <li>CurrencyID - currency id - string</li>
#                   <li>Customer - (optional) customer - object</li>
#                   <li>Container - (optional) lineitem container - object</li>
#                   <li>LineItem - (optional) lineitem - object</li>
#                   </ul> | ref.hash
# output       $hParams | <ul><li>Price - new price - float</li></ul> | ref.hash
#========================================================================================
sub ProductPrice {
    my ($hParams) = @_;

    my $Product    = $hParams->{'Object'};
    my $CurrencyID = $hParams->{'CurrencyID'};
    my $Container  = $hParams->{'Container'};
    my $Customer   = $hParams->{'Customer'};
    my $User       = undef;

    #get customer by container if not defined yet
    if (defined $Container) {
        $Customer = $Container->customer unless defined $Customer;
        $User = $Container->user;
    }

    #is sub product but hasn't own prices
    my $SuperProduct = $Product->get('SuperProduct');
    $Product= $SuperProduct if defined $SuperProduct and not $Product->get('HasSubOwnPrices');

    my $Shop = $Product->getSite();
    my $PriceList = GetProductPriceList($Product, $Customer, $CurrencyID, $Shop, $User);
    return unless defined $PriceList;

    # get the last price with quantity <= wanted quantity
    my $SumQuantity = $Product->correctQuantity( $hParams->{'Quantity'} || 1 );
    $SumQuantity += _OtherLineItemQuantity( $Container, $hParams->{'LineItem'}, $Product );

    my $ProductTaxClass = $Product->getWithParent('TaxClass');
    $ProductTaxClass = $Product->getMappedTaxClassByTaxMatrix($Container->get('TaxMatrix')) if defined $Container;

    my $TaxModel = defined $Container ? $Container->get('TaxModel')
        : defined $Customer ? $Customer->get('TaxModel')
        : $Shop->get('TaxModel');
    ($hParams->{'Price'}) = map  {$_->{'Price'} }
                            sort {$b->{'Quantity'} <=> $a->{'Quantity'}}
                            grep {$_->{'Quantity'} <=  $SumQuantity }
                            @{$PriceList->getScalePricesByProductID($Product->id, $TaxModel, $ProductTaxClass)};

    return;
}

#========================================================================================
# function     ProductDiscount
# state        public
#----------------------------------------------------------------------------------------
# syntax       ProductDiscount( $hParams );
#----------------------------------------------------------------------------------------
# description  Hook function for the ProductDiscount hook.
#               Sets the Discount parameter if a quantity discount and/or value discount
#               is applicable for the given quantity of the given product.
#               If no line item container is specified, the absolute discount value
#               is set according to the price quantity of the product.
#----------------------------------------------------------------------------------------
# input        $hParams |<ul>
#                   <li>Object - product - object</li>
#                   <li>Quantity - (optional, default=1) quantity - float</li>
#                   <li>CurrencyID - currency id - string</li>
#                   <li>Customer - (optional) customer - object</li>
#                   <li>Container - (optional) lineitem container - object</li>
#                   <li>LineItem - (optional) lineitem - object</li>
#                   </ul> | ref.hash
# output       $hParams | <ul><li>Discount - discount value (money) - float</li></ul> | ref.hash
#========================================================================================
sub ProductDiscount {
    my ($hParams) = @_;

    my $Product      = $hParams->{'Object'};
    my $CurrencyID   = $hParams->{'CurrencyID'};
    my $Container    = $hParams->{'Container'};
    my $Customer     = $hParams->{'Customer'};
    my $User         = undef;

    #get customer by container if not defined yet
    if (defined $Container) {
        $Customer = $Container->customer unless defined $Customer;
        $User = $Container->user;
    }

    #is sub product but hasn't own prices
    my $SuperProduct = $Product->get('SuperProduct');
    $Product= $SuperProduct if defined $SuperProduct and not $Product->get('HasSubOwnPrices');

    my $Shop = $Product->getSite();
    my $PriceList = GetProductPriceList($Product, $Customer, $CurrencyID, $Shop, $User);
    return unless defined $PriceList;

    my $TaxModel = defined $Container ? $Container->get('TaxModel')
        : defined $Customer ? $Customer->get('TaxModel')
        : $Shop->get('TaxModel');
    my $ahQuantityDiscounts = $PriceList->getQuantityDiscounts( $Product, $TaxModel );
    my $ahValueDiscounts    = $PriceList->getValueDiscounts( $Product, $TaxModel );
    return if ( 0 == scalar(@$ahQuantityDiscounts) + scalar(@$ahValueDiscounts));

    # quantity of the current lineitem
    my $LineItemQuantity = $Product->correctQuantity( $hParams->{'Quantity'} || 1 );

    # quantity used for lookup in quantity discount table
    # and to calculate the total lineitem price for lookup in value disount table
    # = sum of quatities of all line items of the same product
    my $SumQuantity = $LineItemQuantity + _OtherLineItemQuantity( $Container, $hParams->{'LineItem'}, $Product );

    # get current product price
    my $ProductPrice = $Product->price($hParams);
    my $PriceQuantity = $Product->getWithParent('PriceQuantity');

    # get quantity discount for the sum of all lineitems of the product
    my $SumNonDiscountLineItemPrice = $ProductPrice  * ($SumQuantity / $PriceQuantity);
    my ($QuantityDiscount) = map  {$_->{'Discount'} }
        sort {$b->{'Quantity'} <=> $a->{'Quantity'}}
        grep {$_->{'Quantity'} <=  $SumQuantity }
        @$ahQuantityDiscounts;
    $QuantityDiscount //= 0;

    # get value discount for the sum of all lineitems of the product
    my $LineItemPrice = $SumNonDiscountLineItemPrice * (1 - $QuantityDiscount / 100); ## no critic qw(ProhibitMagicNumbers)
    my ($ValueDiscount) = map  {$_->{'Discount'} }
        sort {$b->{'Value'} <=> $a->{'Value'}}
        grep {$_->{'Value'} <=  $LineItemPrice}
        @$ahValueDiscounts;
    $ValueDiscount //= 0;

    # calc discount percentage
    my $DiscountPercent = $QuantityDiscount + $ValueDiscount - ($QuantityDiscount * $ValueDiscount) / 100; ## no critic qw(ProhibitMagicNumbers)

    # save discount amount at lineitem
    if ( $DiscountPercent ) {

        # price used to calculate the absolute discount
        my $Price = defined $Container
            ? $ProductPrice  * ($LineItemQuantity / $PriceQuantity)
            : $ProductPrice;

        # no problem to use the shop locale because roundMoney does not depend on the locale but on the currencyid only
        my $Locale = $Shop->siteLocale( $Shop->get('LocaleID') );
        $hParams->{'Discount'} = $Locale->roundMoney($Price * $DiscountPercent / 100, $CurrencyID); ## no critic qw(ProhibitMagicNumbers)
    }
    return;
}

sub OnDeleteProduct {
    my ($hParams) = @_;
    my $Product = $hParams->{'Object'};

    my $ScalePriceClass = LoadClassByAlias('ScalePrice');
    $_->delete foreach @{$ScalePriceClass->getObjects('ByProductID', { 'ProductID' => $Product->id })};

    my $QuantityDiscountClass = LoadClassByAlias('QuantityDiscount');
    $_->delete foreach @{$QuantityDiscountClass->getObjects('ByProductID', { 'ProductID' => $Product->id })};

    my $ValueDiscountClass = LoadClassByAlias('ValueDiscount');
    $_->delete foreach @{$ValueDiscountClass->getObjects('ByProductID', { 'ProductID' => $Product->id })};
}

#========================================================================================
# function     OnAfterProductClone
#----------------------------------------------------------------------------------------
# syntax       OnAfterProductClone( $hParams );
#----------------------------------------------------------------------------------------
# description  Duplicate pricelist after duplication of product. Triggered Hook AfterProductClone.
#----------------------------------------------------------------------------------------
# input        $hParams |
#               <ul>
#                   <li>Object - product object - object</li>
#                   <li>Clone - clone product object - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnAfterProductClone {
    my ($hParams) = @_;

    my $Product = $hParams->{'Object'};
    my $Clone = $hParams->{'Clone'};

    my $ScalePriceClass = LoadClassByAlias('ScalePrice');
    $_->clone({'Product' => $Clone}) foreach @{$ScalePriceClass->getObjects('ByProductID', { 'ProductID' => $Product->id })};

    my $QuantityDiscountClass = LoadClassByAlias('QuantityDiscount');
    $_->clone({'Product' => $Clone}) foreach @{$QuantityDiscountClass->getObjects('ByProductID', { 'ProductID' => $Product->id })};

    my $ValueDiscountClass = LoadClassByAlias('ValueDiscount');
    $_->clone({'Product' => $Clone}) foreach @{$ValueDiscountClass->getObjects('ByProductID', { 'ProductID' => $Product->id })};

    # we want to duplicate the pricelist entries for this product's subproducts, but
    # when OnInsertProduct is called the master has not yet cloned its pricelists, so we call it again
    for (@{$Clone->get('SubProducts')}) {
        my $hSubParams = {
            Object => $_,
        };
        OnInsertProduct($hSubParams);
    }
}


#========================================================================================
# function     OnInsertProduct
#----------------------------------------------------------------------------------------
# syntax       OnInsertProduct( $hParams );
#----------------------------------------------------------------------------------------
# description  Duplicate pricelist from super product. Triggered Hook AfterProductClone.
#----------------------------------------------------------------------------------------
# input        $hParams |
#               <ul>
#                   <li>Object - product object - object</li>
#               </ul> | ref.hash.*
#========================================================================================
sub OnInsertProduct {
    my ($hParams) = @_;

    my $Product = $hParams->{'Object'};
    my $Shop = $Product->getSite();
    if ($Shop->get('IsUnityRegistered')) {
        my $SuperProduct = $Product->get('SuperProduct');
        if ($SuperProduct) {
            my $ScalePriceClass = LoadClassByAlias('ScalePrice');
            $_->clone({ 'Product' => $Product }) foreach @{$ScalePriceClass->getObjects('ByProductID', { 'ProductID' => $SuperProduct->id })};

            my $QuantityDiscountClass = LoadClassByAlias('QuantityDiscount');
            $_->clone({ 'Product' => $Product }) foreach @{$QuantityDiscountClass->getObjects('ByProductID', { 'ProductID' => $SuperProduct->id })};

            my $ValueDiscountClass = LoadClassByAlias('ValueDiscount');
            $_->clone({ 'Product' => $Product }) foreach @{$ValueDiscountClass->getObjects('ByProductID', { 'ProductID' => $SuperProduct->id })};
        }
    }
}

# returns the quantity of other lineitems of the same product (or other variations without own prices)
sub _OtherLineItemQuantity {
    my ($Container, $LineItem, $Product) = @_;

    return 0 unless defined $Container;

    my $SumQuantity = 0;
    # find other lineitems of the same product (or other variations without own prices)
    # and add the line item quantity to $SumQuantity
    my $aProductLineItems = $Container->get('LineItems');
    foreach my $LineItemProduct ( @$aProductLineItems ) {
        next unless $LineItemProduct->instanceOf('LineItemProduct');
        next if defined $LineItem && $LineItemProduct->id == $LineItem->id;

        # other product lineitem found
        my $ProductOfLineItem = $LineItemProduct->get('Product');
        next unless defined $ProductOfLineItem;
        my $SuperProductOfLineItem = $ProductOfLineItem->get('SuperProduct');
        if( defined $SuperProductOfLineItem and not $ProductOfLineItem->get('HasSubOwnPrices') ) {
            $ProductOfLineItem = $SuperProductOfLineItem;
        }
        if( $ProductOfLineItem->id == $Product->id ) {
            $SumQuantity += $ProductOfLineItem->correctQuantity( $LineItemProduct->get('Quantity'));
        }
    }

    return $SumQuantity;
}

1;
