#========================================================================================
# package      DE_EPAGES::SEO::API::XMLSitemap
# state        public
# base         Exporter
# description  This package provides shop functions
#========================================================================================
package DE_EPAGES::SEO::API::XMLSitemap;
use base Exporter;
use strict;

our @EXPORT_OK = qw (
    ExportXMLSitemap
);

use DE_EPAGES::Content::API::Table::ShortURL qw ( ExistsShortURLByContentObjectID GetInfoShortURLByContentObjectID );
use DE_EPAGES::Core::API::Error              qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::File               qw ( DeleteFile ExistsFile );
use DE_EPAGES::Object::API::Factory          qw ( LoadObjectByPath LoadClassByAlias );
use DE_EPAGES::SEO::API::Constants           qw ( @BLACKLIST_MOBILECONTENT_CLASSES );
use DE_EPAGES::SEO::API::XML::SitemapExportDriver;
use DE_EPAGES::Trigger::API::Trigger         qw (TriggerHook);

use constant MAX_ENTRIES => 50000;

#========================================================================================
# function     ExportXMLSitemap
# state        public
#----------------------------------------------------------------------------------------
# syntax       ExportXMLSitemap($Shop);
#----------------------------------------------------------------------------------------
# description  calls export of sitemap xml files
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
#========================================================================================
sub ExportXMLSitemap {
    my ($Shop) = @_;
    #export as xml-file
    my $SitemapFileName = $Shop->get('SitemapPathFileName');
    my $TempSitemapFileName = "$SitemapFileName.tmp";
    my $MobileSitemapFileName = $Shop->get('MobileSitemapPathFileName');
    my $UseMobileSitemap = $Shop->featureMaxValue('MobileView') && $Shop->get('UseMobileSF');
    my $TempMobileSitemapFileName = "$MobileSitemapFileName.tmp";
    eval {
        DeleteFile($TempMobileSitemapFileName) if (!$UseMobileSitemap && ExistsFile($TempMobileSitemapFileName));
        _exportXMLSitemap( $Shop, $TempSitemapFileName, $UseMobileSitemap ? $TempMobileSitemapFileName : undef );
        rename $TempSitemapFileName, $SitemapFileName or Error('FileRenameFails', {
            'Source' => $TempSitemapFileName,
            'Target' => $SitemapFileName,
            'Code' => $!
        });
        if($UseMobileSitemap) {
            rename $TempMobileSitemapFileName, $MobileSitemapFileName or Error('FileRenameFails', {
                'Source' => $TempMobileSitemapFileName,
                'Target' => $MobileSitemapFileName,
                'Code' => $!
            });
        }
    };
    if (ExistsError()) {
        my $Error = GetError;
        DeleteFile($TempSitemapFileName) if ExistsFile($TempSitemapFileName);
        DeleteFile($TempMobileSitemapFileName) if ExistsFile($TempMobileSitemapFileName);
        Error($Error);
    }
}

#========================================================================================
# function     _exportXMLSitemap
# state        private
#----------------------------------------------------------------------------------------
# syntax       _exportXMLSitemap($Shop);
#----------------------------------------------------------------------------------------
# description  Exports sitemap xml
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $SitemapFileName | private path to sitemap.xml | string
# input        $MobileSitemapFileName | private path to mobilesitemap.xml | string
#========================================================================================
sub _exportXMLSitemap {
    my ($Shop, $SitemapFileName, $MobileSitemapFileName) = @_;

    my $aContentObjectsData = GetObjects($Shop);

    my $hClassesMap;
    foreach my $ClassAlias ( qw (StartPage Imprint CustomerInformation PrivacyPolicy TermsAndConditions)) {
        $hClassesMap->{$ClassAlias} = LoadClassByAlias($ClassAlias)->id;
    }

    my @MobileBlackListedClassIDs = map { LoadClassByAlias($_)->id } @BLACKLIST_MOBILECONTENT_CLASSES;
    my $Driver = DE_EPAGES::SEO::API::XML::SitemapExportDriver->new($hClassesMap);
    $Driver->exportXML($SitemapFileName, $MobileSitemapFileName, $Shop, $aContentObjectsData, \@MobileBlackListedClassIDs);

    return;
}

#========================================================================================
# function     GetObjects
# state        private
#----------------------------------------------------------------------------------------
# syntax       $aContentObjectData = GetObjects($Shop);
#----------------------------------------------------------------------------------------
# description  Returns all visible, enabled content objects
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# return       $ContentObjectsData | content object data | ref.array.hash
#========================================================================================
sub GetObjects {
    my ($Shop) = @_;

    my @ContentObjectHashs;
    my $maxEntries = MAX_ENTRIES;
    my $CountLocales = 0;

    my $hShopData = $Shop->get([ qw ( SiteLocales ProtocolAndServer WebServerScriptUrl DomainName LanguageID ) ]);

    # add the locale used for canonical url of these languages (first one by sort order) to a locale array
    my @LanguageData;
    my %ActiveLanguages;

    foreach my $Locale (@{$hShopData->{'SiteLocales'}}) {
        my $Language = $Locale->get('Language');
        my $LanguageID = $Language->{'LanguageID'};

        next if defined $ActiveLanguages{$LanguageID}; # e.g. iteration already did de_de and now does de_ch => only add once!
        next unless defined $Shop->activeLocaleByLanguage($LanguageID); #needed??

        $ActiveLanguages{$LanguageID} = 1;

        push(@LanguageData, {
            'LanguageID' => $LanguageID,
            'LocaleID' => $Locale->get('LocaleID'),
        });
        $CountLocales++
    }

    # find visible categories
    foreach my $Category ( @{ Categories($Shop) } ) {
        last if $maxEntries <= 0;
        push(@ContentObjectHashs, {
            'ObjectID' => $Category->id,    # todo: speed up finding the category ids
            'ClassID' => $Category->class->id,
        });
        $maxEntries -= $CountLocales; # each locale gets one entry in xml file
    }

    TriggerHook('API_GetObjectsForXMLSitemap', {
        'Object'             => $Shop,
        'ContentObjectHashs' => \@ContentObjectHashs,
        'MaxEntries'         => \$maxEntries,
        'CountLocales'       => $CountLocales,
        'ShopData'           => $hShopData,
    });

    # get only needed data (URLs) of the given objectids
    my @ContentObjectsData;
    my $StartPageClassID = LoadClassByAlias('StartPage')->id;

    foreach my $hContentObjectInfo (@ContentObjectHashs) {
        foreach my $hLang (@LanguageData) {
            my $Object = {
                %$hContentObjectInfo,
                'URL' => _getURLFromObjectData(
                    $hContentObjectInfo,
                    $hLang->{'LanguageID'},
                    $hShopData->{'DomainName'},
                    $hShopData->{'LanguageID'},
                    $hLang->{'LocaleID'},
                    $hShopData->{'ProtocolAndServer'},
                    $hShopData->{'WebServerScriptUrl'},
                    $StartPageClassID
                 )
            };
            push(@ContentObjectsData, $Object);
        }
    }
    my @URLs = ();
    TriggerHook('API_GetURLsForXMLSitemap', {
        'Object'             => $Shop,
        'URLs'               => \@URLs,
        'LanguageIDs'        => $Shop->languages
    });

    push(@ContentObjectsData, $_) foreach (@URLs);

    return \@ContentObjectsData;
}

sub _getURLFromObjectData {
    my ($hContentObjectInfo, $LanguageID, $ShopDomainName, $ShopLanguageID, $LocaleString,
        $ProtocolAndServer, $WebServerScriptUrl, $StartPageClassID) = @_;

    my $ContentObjectID = $hContentObjectInfo->{'ObjectID'};
    my $Uri;
    if (defined $ShopDomainName) {
        if ($hContentObjectInfo->{'ClassID'} == $StartPageClassID) {
            if ($ShopLanguageID != $LanguageID) {
                $Uri = $LocaleString;
            } else {
                $Uri = '';
            }
        } elsif(defined $LanguageID and ExistsShortURLByContentObjectID($ContentObjectID, $LanguageID)) {
            $Uri = GetInfoShortURLByContentObjectID($ContentObjectID, $LanguageID)->{'Uri'};
        }
    }
    if (defined $Uri) {
        my $CanonicalURL = $ProtocolAndServer . '/';   # EPG-18932 : ending slash for shop domain
        $CanonicalURL = $CanonicalURL . $Uri if $Uri ne '';
        return $CanonicalURL;
    } else {
        return $WebServerScriptUrl . '/' . $LocaleString . '/?ObjectID=' . $ContentObjectID;
    }
}

#========================================================================================
# function     Categories
# state        private
#----------------------------------------------------------------------------------------
# syntax       $aCategories = Categories($Shop);
#----------------------------------------------------------------------------------------
# description  Returns all visible, enabled categories
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# return       $aCategoryIDs | category identifier | ref.array.int
#========================================================================================
sub Categories {
    my ($Shop) = @_;

    my $aCategories;

    my $GroupEveryone = LoadObjectByPath( '/Groups/Everyone' );
    my $GroupEveryoneID = defined $GroupEveryone ? $GroupEveryone->id : undef;

    my $Startpage = $Shop->get('Categories');
    push @$aCategories, $Startpage;

    subcategories($aCategories, $Startpage);

    return [ grep { IsCategoryValid($_, $GroupEveryoneID) } @{ $aCategories } ];
}

sub subcategories {
    my ($aCategories, $Category) = @_;

    my @VisibleContentObjects = @{$Category->get('VisibleContentObjects')};
    push @$aCategories, @VisibleContentObjects;
    foreach my $ChildCategory (@VisibleContentObjects) {
        subcategories($aCategories, $ChildCategory);
    }
}

sub IsCategoryValid {
    my ($Category, $GroupEveryoneID) = @_;

    return 0 if $Category->instanceOf('Link');
    return 0 if $GroupEveryoneID and $Category->get('ReadPermissionFor') ne $GroupEveryoneID;

    return 1;
}


1;
