#==============================================================================
# package      DE_EPAGES::Saferpay::UI::Shop
# state        private
#==============================================================================
package DE_EPAGES::Saferpay::UI::Shop;
use base DE_EPAGES::Order::UI::Shop;

use strict;
use DateTime;
use DE_EPAGES::Core::API::Error qw ( Error GetError ExistsError );
use DE_EPAGES::Core::API::PerlTools qw (fcmp GetLocalTimeZone isValid);
use DE_EPAGES::Core::API::Warning qw (BlockWarnings Warning);
use DE_EPAGES::Database::API::Connection qw (GetCurrentDBHandle);
use DE_EPAGES::Saferpay::API::Log qw (SaferPayLog);
use DE_EPAGES::Object::API::Factory qw (ExistsObject ExistsObjectByGUID LoadObjectByGUID);
use DE_EPAGES::Object::API::Table::Currency qw (GetInfoCurrency);
use DE_EPAGES::Saferpay::API::SaferpayConnector qw (GetXMLAttributes VerifyPayConfirm PayComplete);
use DE_EPAGES::Saferpay::API::Constants qw (STATUS_AUTHORISED
                                            STATUS_CANCELLED
                                            STATUS_COMPLETE);
use DE_EPAGES::Saferpay::UI::BasketOrder;

#==============================================================================
# Event handlers
#==============================================================================
sub PaymentCallbackSaferpay {
    my $self = shift;
    my ($Servlet) = @_;

    my $FormValues = $Servlet->params;
    SaferPayLog('Saferpay', 'PaymentCallbackSaferpay_FormValues', $FormValues);

    # test the parameters of saferpay, show errors if something is wrong
    my ($hValues, undef) = TestSaferpayParameters($Servlet, $FormValues);

    # update the paymentlineitem
    my $PaymentLineItem = delete($hValues->{'PaymentLineItem'});
    $PaymentLineItem->set($hValues);

    my $PaymentObject;
    my $TransStatus = $PaymentLineItem->get('TransStatus');
    my $WaitForNotification = 0;
    if ($TransStatus eq STATUS_AUTHORISED  ||  $TransStatus eq STATUS_COMPLETE) {
        my $Status = $PaymentLineItem->get('TransStatus') // '';
        $PaymentObject = $PaymentLineItem->container->parent;
        return $PaymentObject if ($PaymentObject->instanceOf('CustomerOrder') && ($Status eq STATUS_AUTHORISED || $Status eq STATUS_COMPLETE)); # view order confirmation page
        $WaitForNotification = 1; # wait for notification in template
    }

    $PaymentObject = $PaymentLineItem->container->parent;

    # cancelled or declined
    if ($PaymentObject->instanceOf('CustomerOrder')) {
        $Servlet->vars('ViewObjectID', $Servlet->object->id);
        $Servlet->vars('PaymentObject', $PaymentObject->tleHash);
        $Servlet->form->executeFormError( {
            'Name'       => 'PaymentFailed',
            'Reason'     => 'PaymentDeclined',
            'ViewAction' => 'ViewPaymentSaferpayResult'
        });
        return;
    }
    else {
        if ($WaitForNotification) {
            $Servlet->vars('ViewAction', 'WaitForSaferpayNotify');
        }
        return $PaymentObject; # back to basket page or wait for notification
    }
}

#==============================================================================
sub PaymentNotifySaferpay {
    my $self = shift;
    my ($Servlet) = @_;

    my $FormValues = $Servlet->params;
    SaferPayLog('Saferpay', 'PaymentNotifySaferpay_FormValues', $FormValues);

    my $Object = eval { _ExecuteSaferpayCallback($Servlet, $FormValues, 0); };
    if (ExistsError()) {
        my $Error = GetError();
        Error($Error) if $Error->code ne 'INVALID_FORM';
        Error('INVALID_FORM', {
            'ViewVars'=> {
                'FormError' => 0,
                'IsBlocked' => 1,
                'ViewObjectID' => $Servlet->object->id,
            },
            'ViewAction' => 'View',
        });
    }
    $Servlet->vars('ExitAfterEvent',1);

    return $Object;
}

#==============================================================================
sub _ExecuteSaferpayCallback {
    my ($Servlet, $FormValues) = @_;

    # test the parameters of saferpay, show errors if something is wrong
    my ($hValues, $hCard) = TestSaferpayParameters($Servlet, $FormValues);

    # update the paymentlineitem
    my $PaymentLineItem = delete($hValues->{'PaymentLineItem'});
    $PaymentLineItem->set($hValues);

    # update card if returned
    if ($hCard->{'CardRefID'}) {
        my $CardRefID = delete($hCard->{'CardRefID'});
        my $CreditCard = LoadObjectByGUID($CardRefID);
        if ($CreditCard->get('LineItemPaymentSaferpay')->id == $PaymentLineItem->id) {
            $CreditCard->set($hCard);
        }
    }

    # complete Payment if transaction is a capture transaction and authorise was successful
    if ($PaymentLineItem->get('TransStatus') eq STATUS_AUTHORISED  &&  !$PaymentLineItem->get('TransactionType')) {
        if (!PayComplete($PaymentLineItem)) {
            SaferPayLog('Saferpay', 'PaymentCallbackSaferpay_PayCompleteError', $hValues);
        }
    }

    #--- create order etc.
    my $Order = DE_EPAGES::Saferpay::UI::BasketOrder->HandleAuthoriseResult($Servlet, $PaymentLineItem);
    return $Order if (defined $Order); # view order confirmation page


    # cancelled by user or payment declined
    my $Object = $PaymentLineItem->container->parent;
    if ($Object->instanceOf('CustomerOrder')) {
        $Servlet->vars('ViewAction', 'ViewPaymentSaferpayCancel');
        $Object->set({'CancelledOn' => GetCurrentDBHandle()->currentDateTime});
    }

    return $Object;
}


#==============================================================================
sub TestSaferpayParameters {
    my ($Servlet, $FormValues) = @_;
    my $Shop = $Servlet->object;
    my $Form = $Servlet->form;

    my %Values = ();

    my $PaymentLineItemGUID = $FormValues->{'MGUID'};
    $Form->executeFormError( {
            'Name'       => 'MGUID',
            'Reason'     => 'BadCallback_NoPaymentGUID',
            'ViewAction' => 'ViewPaymentSaferpayResult'
        } ) if (!isValid($PaymentLineItemGUID) || !(ExistsObjectByGUID($PaymentLineItemGUID)));


    my $PaymentLineItem = LoadObjectByGUID($FormValues->{'MGUID'});

    #Log Shopdata
    my %ShopData;
    $ShopData{'GrandTotal'}      = $PaymentLineItem->get('LineItemContainer')->get('GrandTotal');
    $ShopData{'CurrencyID'}      = $PaymentLineItem->get('CurrencyID');
    $ShopData{'TransID'}         = $PaymentLineItem->get('TransID');
    $ShopData{'ShopIDByPayment'} = $PaymentLineItem->getSite()->id;
    $ShopData{'ShopID'}          = $Shop->id;

    SaferPayLog('Saferpay', 'Saferpay Shopdata', \%ShopData);

    $Form->executeFormError( {
        'Name'       => 'MGUID',
        'Reason'     => 'BadCallback_ShopIDMisMatch',
        'ViewAction' => 'ViewPaymentSaferpayResult'
    } ) if ($ShopData{'ShopIDByPayment'} != $ShopData{'ShopID'});


    # cancelled Transaction
    if (!exists($FormValues->{'DATA'})) {
        $Values{'PaymentLineItem'} = $PaymentLineItem;
        $Values{'TransStatus'}     = STATUS_CANCELLED;
        return \%Values;
    }

    # extract Data from XML
    my $Result = GetXMLAttributes($FormValues->{'DATA'});
    SaferPayLog('Saferpay', 'Saferpay Result', $Result);

    # check currency
    $Form->executeFormError({
         'Name'       => 'currency',
         'Reason'     => 'BadCallback_AmountMisMatch',
         'ViewAction' => 'ViewPaymentSaferpayResult'
    }) if ($ShopData{'CurrencyID'} ne $Result->{'CURRENCY'});

    # check amount
    my $hCurrency = GetInfoCurrency($ShopData{'CurrencyID'});
    my $ResultAmount = $Result->{'AMOUNT'} / 10 ** $hCurrency->{'MinorUnit'}; ## no critic qw(ProhibitMagicNumbers)

    $Form->executeFormError({
         'Name'       => 'currency',
         'Reason'     => 'BadCallback_AmountMisMatch',
         'ViewAction' => 'ViewPaymentSaferpayResult'
    }) if (fcmp($ShopData{'GrandTotal'}, $ResultAmount) != 0);


    # check trans id
    $Form->executeFormError({
            'Name'            => 'transId',
            'Reason'          => 'BadCallback_DuplicateTransaction',
            'ViewAction'      => 'ViewPaymentSaferpayResult',
            'SavedTransID'    => $ShopData{'TransID'},
            'CallbackTransID' => $Result->{'ID'}
        }) if (defined($ShopData{'TransID'}) && $Result->{'ID'} ne $ShopData{'TransID'});


    # Verify Pay Confirm
    $Form->executeFormError({
            'Name'            => 'VerifyPayConfirm',
            'Reason'          => 'BadCallback_VerifyPayConfirm',
            'ViewAction'      => 'ViewPaymentSaferpayResult'
        }) if (!VerifyPayConfirm($PaymentLineItem, $FormValues));


    $Values{'PaymentLineItem'} = $PaymentLineItem;
    $Values{'TransID'}         = $Result->{'ID'};
    $Values{'ProviderID'}      = $Result->{'PROVIDERID'};
    $Values{'ProviderName'}    = $Result->{'PROVIDERNAME'};

    #--- 3D secure
    foreach my $key (qw(ECI XID)) {
        $Values{$key} = $Result->{$key} if (defined $Result->{$key});
    }
    $Values{'AvsCode'} = $Result->{'CAVV'} if (defined $Result->{'CAVV'});

    #--- secure card data
    my %CardValues = ();
    if ($Result->{'CARDREFID'}) {
        $CardValues{'CardRefID'}   = $Result->{'CARDREFID'};
        $CardValues{'CardMask'}    = $Result->{'CARDMASK'};
        $CardValues{'CardBrand'}   = $Result->{'CARDBRAND'};
        $CardValues{'CardType'}    = $Result->{'CARDTYPE'};
        $CardValues{'LifeTime'}    = $Result->{'LIFETIME'};
        $CardValues{'ExpiryMonth'} = $Result->{'EXPIRYMONTH'};
        $CardValues{'ExpiryYear'}  = $Result->{'EXPIRYYEAR'};
        $CardValues{'CCCountry'}   = $Result->{'CCOUNTRY'};
    }

    if (defined $FormValues->{'transTime'}) {
        $Values{'TransTime'} = DateTime->from_epoch(
                                          'epoch'     => int($FormValues->{'transTime'} / 1000), ## no critic qw(ProhibitMagicNumbers)
                                          'time_zone' => GetLocalTimeZone());
    }

    else {
        $Values{'TransTime'} = GetCurrentDBHandle()->currentDateTime;
    }

    SaferPayLog('Saferpay', 'TestSaferpayParameters ok for payment line item ' . $PaymentLineItem->get('GUID'), {});

    return (\%Values, \%CardValues);
}

#========================================================================================
# function     TransactionPaymentName
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Name = $Handler->TransactionPaymentName();
#----------------------------------------------------------------------------------------
# description  return Saferpay as string. Name of PaymentType for search/view
#                               transactions.
#----------------------------------------------------------------------------------------
# return       $Name | name of payment type | string
#========================================================================================
sub TransactionPaymentName { 'Saferpay' }

#==============================================================================
sub CaptureSaferpayTransactions {
    my $self = shift;
    my ($Servlet) = @_;
    my $Shop = $Servlet->object;

    my @ToCapture = $Servlet->form->value('ListObjectID');
    return if 0 == scalar @ToCapture;

    my $aWarnings = BlockWarnings( sub {
        foreach my $ObjectID(@ToCapture) {
            next unless(ExistsObject($ObjectID));
            my $PaymentLineItem = $Shop->loadObject($ObjectID, 'LineItemPaymentSaferpay');
            next if ($PaymentLineItem->get('TransStatus') ne STATUS_AUTHORISED);
            my $ret = PayComplete($PaymentLineItem);
            if ($ret != 1) {
                Warning('SaferpayCaptureFailed', {'TransID' => $PaymentLineItem->get('TransID'), 'ReturnCode' => $PaymentLineItem->get('ReturnCode')});
            }
            else {
                $PaymentLineItem->container->parent->set(({'PaidOn' => GetCurrentDBHandle()->currentDateTime}));
            }
        }
    }, ['SaferpayCaptureFailed']);

    $Servlet->form->mapFormError({
        'Form'  => 'CaptureSaferpayTransactions',
        'Name'  => 'CaptureSaferpayTransactions'
    }, $_) foreach @$aWarnings;

    $Servlet->form->executeFormError;
}

1;

