#========================================================================================
# package      DE_EPAGES::ShippingOption::API::Object::ShippingOption
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for ShippingOption.
#========================================================================================
package DE_EPAGES::ShippingOption::API::Object::ShippingOption;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::PerlTools qw (fcmp);
use DE_EPAGES::Core::API::TestParameter qw( TestInteger TestString TestFloat );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $ShippingOption->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'ShippingOption' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $ShippingOption->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingOption = DE_EPAGES::ShippingOption::API::Object::ShippingOption->insert($hInfo);
# example      my $Class = LoadClassByAlias('ShippingOption');
#               my $ShippingOption = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertShippingOption' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $ShippingOption | new object | object
# hook         OBJ_InsertShippingOption | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingOptionID - shipping option id   - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $Shop = $hInfo->{'Shop'} || $hInfo->{'Parent'}->getSite();
    my $newValue = scalar(@{$Shop->get('ShippingOptions')}) + 1;
    my $maxValue = $Shop->featureMaxValue('ShippingOption');
    if ($newValue > $maxValue) {
        Error('ShippingOptionFeatureLimitExceeded', {
              'Feature'  => 'ShippingOption',
              'MaxValue' => $maxValue,
              'NewValue' => $newValue,
              'Shop'     => $Shop
        });
    }
    my $self = $class->SUPER::insert($hInfo);


    $Shop->featureCurrentValue('ShippingOption', $newValue);

    TriggerHook('OBJ_InsertShippingOption', {'Object' => $self, 'ShippingOptionID' => $self->id });
    delete $hInfo->{$_} foreach qw ( Shop TaxClass IsVisible HasTextField ); ## no critic qw(ProhibitLongLines)
    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingOption->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteShippingOption' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteShippingOption | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingOptionID - shipping option id   - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteShippingOption', {'Object' => $self, 'ShippingOptionID' => $self->id });

    # get current feature value
    my $Shop = $self->getSite();
    my $newValue = scalar(@{$Shop->get('ShippingOptions')}) - 1;
    $newValue = 0 if ($newValue < 0);

    # delete all prices
    LoadClassByAlias('ShippingOptionPrice')->deleteObjects('ByShippingOptionID', [$self->id]);
    $_->set({'ShippingOption'=>undef}) foreach @{LoadClassByAlias('LineItemShippingOption')->getObjects('ByShippingOptionID', [$self->id])};
    LoadClassByAlias('ShippingOptionChoice')->deleteObjects('ByShippingOptionID', [$self->id]);
    $self->SUPER::delete;

    # update feature value
    $Shop->featureCurrentValue('ShippingOption', $newValue, 1);

    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingOption->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateShippingOption' and 'OBJ_AfterUpdateShippingOption'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateShippingOption | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingOptionID - shipping option id   - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateShippingOption | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingOptionID - shipping option id   - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateShippingOption', {'Object'=>$self, 'Values'=> $hValues, 'ShippingOptionID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateShippingOption',  {'Object'=>$self, 'Values'=> $hValues, 'ShippingOptionID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::ShippingOption::API::Object::ShippingOption->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'IsVisible' => 0,
        'HasTextField' => 0,
    };
}

#========================================================================================
# function     exportXMLAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hAttributes = $Object->exportXMLAttributes( $Driver, $hAllAttributes );
#----------------------------------------------------------------------------------------
# description  Returns plain attributes of object to XML export driver. This function
#               removes bit attributes with value 0.
#----------------------------------------------------------------------------------------
# input        $Driver | xml export driver with support function addObject | object
# input        $hAllAttributes | all attributes of object,
#                                 see $Object-><function exportableAttributes> | ref.hash
# return       $hAttributes | plain attributes | ref.hash.string
#========================================================================================
sub exportXMLAttributes {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;
    $hAttributes = $self->SUPER::exportXMLAttributes($Driver, $hAttributes);

    if (exists $hAttributes->{'TaxClass'}) {
        # map TaxClass object to TaxClass name
        my $TaxClass = $self->get('TaxClass');
        $hAttributes->{'TaxClass'} = defined $TaxClass ? $TaxClass->alias : undef;
    }
    return $hAttributes;
}

#========================================================================================
# function     exportXMLElements
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Object->exportXMLElements($Driver, $hAllAttributes);
#----------------------------------------------------------------------------------------
# description  Exports multilinguagal attribute value of this object to the xml SAX driver.
#----------------------------------------------------------------------------------------
# input        $Driver | export driver | object
# input        $hAllAttributes | all attributes of object,
#                                 see $Object-><function allAttributes> | ref.hash
#========================================================================================
sub exportXMLElements {
    my $self = shift;
    my ($Driver, $hAttributes) = @_;

    $self->SUPER::exportXMLElements($Driver, $hAttributes);

    # add prices
    my $Class = LoadClassByAlias('ShippingOptionPrice');
    my $aPrices = $Class->getObjects('ByShippingOptionID', {'ShippingOptionID'=>$self->id});
    $Driver->addObject( 'ShippingOptionPrice', $_ ) foreach @$aPrices;
}

sub importXML {
    my ($ObjectPackage, $Handler, $hElement, $Class, $hAttributes) = @_;

    if (not $Handler->topID('IsReference')) {

        my $Parent = $Handler->topID('Object');
        my $Shop = $Parent->getSite();

        # map TaxClass name to TaxClass object
        my $TaxMatrix = $Shop->get('TaxMatrix');
        $hAttributes->{'TaxClass'} = $TaxMatrix->child($hAttributes->{'TaxClass'});

        # set shop
        $hAttributes->{'Shop'} = $Shop;
    }

    return $ObjectPackage->SUPER::importXML($Handler, $hElement, $Class, $hAttributes);
}

#========================================================================================
# function     price
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Price = $ShippingOption->price( $hVars );
#----------------------------------------------------------------------------------------
# example      $Price = $ShippingOption->price( {
#                   CurrencyID => 'USD',
#                   TaxModel   => $ShippingOption->getSite()->get('TaxModel'),
#                   ExtraPrecision => 1,
#               } );
#----------------------------------------------------------------------------------------
# description  Returns the price by the given requirements. If no price exists, 0 will be returned.
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID      - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel        - gross or net price    - int
#                   <li>ExtraPrecision  - extra precision       - int
#                   <li>TaxMatrix (optional) - tax matrix    - object
#                   <li>TaxClass (optional) - tax class    - object
#               </ul> | ref.hash
# return       $Price | price of shipping option | money
#========================================================================================
sub price {
    my $self = shift;
    my ($hVars) = @_;

    my $TaxClass = $hVars->{'TaxClass'};
    my $TaxMatrix = $hVars->{'TaxMatrix'};
    my $CurrencyID = $hVars->{'CurrencyID'};
    my $TaxModel = $hVars->{'TaxModel'};
    my $ExtraPrecision = $hVars->{'ExtraPrecision'};
    $ExtraPrecision = 0 unless defined $ExtraPrecision;
    TestString('CurrencyID', $CurrencyID);
    TestInteger('TaxModel', $TaxModel);
    TestInteger('ExtraPrecision', $ExtraPrecision);

    my $Class = LoadClassByAlias('ShippingOptionPrice');
    my $aPrices = $Class->getObjects('ByShippingOptionIDCurrencyID', {
        'ShippingOptionID' => $self->id,
        'CurrencyID' => $hVars->{'CurrencyID'},
    });

    return 0 if 0 == scalar(@$aPrices);

    my $Price =  $aPrices->[0]->get('Price');

    # use the correct Taxmatrix
    $TaxMatrix = $self->getSite()->get('TaxMatrix') unless defined $TaxMatrix;

    # convert net price to gross if necessary
    $TaxClass = $self->get('TaxClass') unless defined $TaxClass;
    $Price = $TaxMatrix->net2gross($CurrencyID, $Price,{'TaxModel' => $TaxModel,
                                                        'TaxClass' => $TaxClass,
                                                        }, $ExtraPrecision);

    return $Price;
}

#========================================================================================
# function     setPrice
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Product->setPrice($CurrencyID, $TaxModel, $Value);
# example      use DE_EPAGES::Shop::API::Constants qw( TAX_MODEL_GROSS TAX_MODEL_NET );
#               $Product->setPrice('EUR', TAX_MODEL_GROSS, 19.99);
#----------------------------------------------------------------------------------------
# description  Sets the list price of this shipping option.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | currency code | string
# input        $TaxModel | tax model:
#                       TAX_MODEL_GROSS=including tax,
#                       TAX_MODEL_NET=without tax | int
#========================================================================================
sub setPrice {
    my $self = shift;
    my ($CurrencyID, $TaxModel, $Value) = @_;
    TestString('CurrencyID', $CurrencyID);
    TestInteger('TaxModel', $TaxModel);

    my $Class = LoadClassByAlias('ShippingOptionPrice');
    my $aPrices = $Class->getObjects('ByShippingOptionIDCurrencyID', {
        'ShippingOptionID' => $self->id,
        'CurrencyID' => $CurrencyID,
    });

    if( defined $Value ) {
        TestFloat('Value', $Value);
        # convert gross price to net if necessary
        my $TaxClass = $self->get('TaxClass');
        $Value = $self->getSite()->get('TaxMatrix')->gross2net($CurrencyID, $Value,
                                                                 {'TaxModel' => $TaxModel,
                                                                  'TaxClass' => $TaxClass,
                                                                 });
        if (scalar @$aPrices) {
            $aPrices->[0]->set({'Price' => $Value});
        }
        else {
            $Class->insertObject({
                'ShippingOption' => $self,
                'CurrencyID'     => $CurrencyID,
                'Price'          => $Value
            });
        }
    }
    else {
        if (scalar @$aPrices) {
            $aPrices->[0]->delete;
        }
    }
}

#========================================================================================
# function     canAddToBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShippingOption->canAddToBasket($hVars);
#----------------------------------------------------------------------------------------
# description  Tests if a shipping option can be added to the basket with the given currency.
#               Uses function <function DE_EPAGES::ShippingOption::API::Object::ShippingOption::price>
#               to evaluate if shipping option has a price (price can be 0).
#----------------------------------------------------------------------------------------
# input        $hVars | vars of container and basket | ref.hash
# return       $Can | can be added to the basket | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($hVars) = @_;

    return $self->get('IsVisible');
}

#========================================================================================
# function     addToContainer
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $Object->addToContainer( $Container, $hInfo );
#----------------------------------------------------------------------------------------
# description  Adds lineItem of shipping option to line item container.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# input        $hInfo    | hash with keys ShippingOptionImage, Comment
# return       $LineItem | lineitem | object
#========================================================================================
sub addToContainer {
    my $self = shift;
    my ($Container, $hInfo) = @_;

    my $TaxClass;
    my $ExistingTaxClass = $self->get('TaxClass');
    if($ExistingTaxClass) {
        $TaxClass = $Container->get('TaxMatrix')->getTaxClassByAlias($ExistingTaxClass->alias);
    }

    my $hInfoContainer = $Container->get(['CurrencyID', 'TaxModel', 'TaxMatrix']);
    $hInfoContainer->{'TaxClass'} = $TaxClass;

    $hInfo->{'SKU'}            = $self->alias;
    $hInfo->{'Name'}           = $self->get('NameOrAlias');
    $hInfo->{'Quantity'}       = 1;
    $hInfo->{'BasePrice'}      = $self->price($hInfoContainer);
    $hInfo->{'LineItemPrice'}  = $hInfo->{'BasePrice'};
    $hInfo->{'Position'}       = 900;
    $hInfo->{'ShippingOption'} = $self;
    $hInfo->{'ChoiceName'}     = $hInfo->{'Choice'}->get('Name') if (defined $hInfo->{'Choice'});
    $hInfo->{'TaxClass'}       = $TaxClass;
    $hInfo->{'TaxRate'}        = $Container->taxRate($TaxClass);

    return $Container->insertLineItem(LoadClassByAlias('LineItemShippingOption'), $hInfo);
}

#========================================================================================
# function     recalculateLineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object->recalculateLineItem( $LineItem, $hInfo );
#----------------------------------------------------------------------------------------
# description  Recalculate a shipping option line item when something has changed on basket,
#               for example if a different currency was chosen.
#----------------------------------------------------------------------------------------
# input        $LineItem | line item | object
# input        $hInfo    | hash with keys CurrencyID, TaxModel | ref.hash
#========================================================================================
sub recalculateLineItem {
    my $self = shift;
    my ($LineItem, $hValues) = @_;
    my $Container = $LineItem->container;
    my $CurrencyID = defined $hValues->{'CurrencyID'} ? $hValues->{'CurrencyID'} : $Container->get('CurrencyID');
    my $TaxModel = defined $hValues->{'TaxModel'} ? $hValues->{'TaxModel'} : $Container->get('TaxModel');
    my $TaxClass = $Container->get('TaxMatrix')->getTaxClassByAlias($self->get('TaxClass')->alias);
    my $NameOrAlias = $self->get('NameOrAlias');

    my $Price = $self->price({
        'CurrencyID' => $CurrencyID,
        'TaxModel'   => $TaxModel,
        'TaxMatrix'   => $Container->get('TaxMatrix'),
        'TaxClass'   => $TaxClass,
    });

    $LineItem->set({
        'Name'          => $NameOrAlias,
        'BasePrice'     => $Price,
        'LineItemPrice' => $Price,
        'TaxRate'       => $Container->taxRate($TaxClass),
    }) if fcmp($Price , $LineItem->get('BasePrice')) or $LineItem->set('Name') ne $NameOrAlias;
    return;
}

1;
