#========================================================================================
# package      DE_EPAGES::Shop::API::WebService::ShopService
# state        public
#----------------------------------------------------------------------------------------
# description  Provides web service access to shops.
#----------------------------------------------------------------------------------------
# example      use DE_EPAGES::WebService::API::Client;
#               my $FactoryService = DE_EPAGES::WebService::API::Client
#                   ->uri('urn://epages.de/WebService/FactoryService/2004/08')
#                   ->proxy("http://%2fUsers%2f$Login:$Password@localhost:8080/epages/Store.soap");
#               my $ShopID = $FactoryService->objectByPath( '/Shops/DemoShop' )->result;
#               my $ShopService = DE_EPAGES::WebService::API::Client
#                   ->uri('urn://epages.de/WebService/ShopService/2004/08')
#                   ->proxy("http://%2fUsers%2f$Login:$Password@localhost:8080/epages/Store.soap");
#               $ShopService->setFeatureValues( $ShopID, [
#                   { Feature => 'Products', Value => 1000 },
#                   { Feature => 'SlideShow', Value => 1 },
#               ]);
#========================================================================================
package DE_EPAGES::Shop::API::WebService::ShopService;
use base DE_EPAGES::WebService::API::WebService::ObjectService;

use strict;
use DE_EPAGES::Cartridge::API::Table::Cartridge qw(GetAllInfoCartridges);
use DE_EPAGES::Core::API::Error qw( Error HandleError );
use DE_EPAGES::Core::API::File qw( CopyDirectory IsInPath );
use DE_EPAGES::Core::API::Log qw (LogDebug);
use DE_EPAGES::Core::API::TestParameter qw( TestInteger TestArray TestHash );
use DE_EPAGES::MailType::API::NewPassword qw( SendNewPasswordToken );
use DE_EPAGES::Object::API::Factory qw( ExistsObject LoadClassByAlias LoadObject LoadRootObject LoadObjectByPath Transaction  );
use DE_EPAGES::Object::API::Language qw( GetCodeByLanguageID );
use DE_EPAGES::Presentation::API::User qw( GetUserFolder );
use DE_EPAGES::Presentation::API::Session qw( InsertSession );
use DE_EPAGES::XML::API::DateTimeFormatXML;

sub ErrorDefinition {
    return {
        'FeatureNotAvialable' => 'Feature #Feature not available at this database.',
        'NotSameProvider' => 'Provider with path #ProviderPath is not the same as #OldProviderPath at store #Store.',
    };
}

#========================================================================================
# function     create
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopID = $ShopService->create( $hOptions )->result;
# example      $ShopID = $ShopService({
#                   'Alias' => 'DemoShop'
#                   'ShopTypeRefID' => $ShopTypeRef->id,
#                   'ClosedByProvider' => 0,
#               })->result;
#----------------------------------------------------------------------------------------
# description  Inserts a new shop object.
#----------------------------------------------------------------------------------------
# input        $hOptions | Options;
#               used keys are:
#               <ul>
#                 <li>Alias | shop alias | string</li>
#                 <li>ShopTypeRefID | shoptyperef id | int</li>
#                 <li>ClosedByProvider | true if closed by provider | boolean</li>
#                 <li>(optional) ProviderPath | path to provider | string</li>
#                 <li>(optional) WebServerScriptNamePart | part of webserver script name | string</li>
#               </ul> | hash.ref
# return       $ShopID | shop id | integer
#========================================================================================
sub create {
    shift;
    my ($hOptions) = @_;
    LogDebug('create', $hOptions);
    TestHash('hOptions', $hOptions, {
        'Alias' => 'String;NotNull',
        'ShopTypeRefID' => 'Integer;NotNull',
        'ClosedByProvider' => 'Integer;NotNull',
        'ProviderPath' => 'String;Null',
        'WebServerScriptNamePart' => 'String;Null',
    }, 1);

    # create shop object
    my $Shop;
    eval {

        # set system attribute
        my $System = LoadRootObject();
        my $ProviderPath = $hOptions->{'ProviderPath'};
        if (defined $ProviderPath && $ProviderPath ne '') {
            my $OldProviderPath = $System->get('ProviderPath');
            Error('NotSameProvider', {
                'ProviderPath' => $ProviderPath,
                'OldProviderPath' => $OldProviderPath,
                'Store' => $System->get('StoreName'),
            }) if defined $OldProviderPath && $OldProviderPath ne $ProviderPath;
            $System->set({ 'ProviderPath' => $ProviderPath }) if $OldProviderPath ne $ProviderPath;
        }

        # create the Shop object
        my %ShopValues = (
            'Alias' => $hOptions->{'Alias'},
            'ClosedByProvider' => $hOptions->{'ClosedByProvider'},
            'ShopTypeRef' => LoadObject( $hOptions->{'ShopTypeRefID'} ),
        );
        
        my $WebServerScriptNamePart = $hOptions->{'WebServerScriptNamePart'};
        if (defined $WebServerScriptNamePart && $WebServerScriptNamePart ne '') {
            $ShopValues{'WebServerScriptNamePart'} = $WebServerScriptNamePart;
        }
            
        $Shop = $System->folder('Shops')->insertChild( 'Shop', \%ShopValues );
    };
    HandleError(
        OnError => sub {
            $Shop->delete if defined $Shop;
        }
    );

    return $Shop->id;
}

#========================================================================================
# function     getInfoSuperMerchant
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hInfo = $ShopService->getInfoSuperMerchant( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns login, email and other information of the primary administrator user.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# return       $hInfo | merchant info hash; keys:
#                       <ul>
#                       <li>Alias | login | string</li>
#                       <li>EMail | e-mail address | string</li>
#                       <li>LocaleID | locale id | string</li>
#                       <li>DeleteConfirmation | true if deletion should be confirmed | boolean</li>
#                       <li>IsLoginCookieAllowed | true if user automated login over cookies is enabled | boolean</li>
#                       <li>ShowFunctionHelp | true if function help is allowed | boolean</li>
#                       </ul> | ref.hash
#========================================================================================
sub getInfoSuperMerchant {
    my $self = shift;
    my ($ShopID) = @_;
    LogDebug('getInfoSuperMerchant', $ShopID);
    TestInteger('ShopID', $ShopID);

    my $Shop = LoadObject( $ShopID );
    my $SuperMerchant = $Shop->get('SuperMerchant');
    my $hValues =
      $SuperMerchant->get( [qw( Alias EMail Name LocaleID DeleteConfirmation IsLoginCookieAllowed ShowFunctionHelp )] );
    $hValues->{'Alias'}                = $self->soapString( $hValues->{'Alias'} );
    $hValues->{'EMail'}                = $self->soapString( $hValues->{'EMail'} );
    $hValues->{'Name'}                 = $self->soapString( $hValues->{'Name'} );
    $hValues->{'LocaleID'}             = $self->soapString( $hValues->{'LocaleID'} );
    $hValues->{'DeleteConfirmation'}   = $self->soapBool( $hValues->{'DeleteConfirmation'} );
    $hValues->{'IsLoginCookieAllowed'} = $self->soapBool( $hValues->{'IsLoginCookieAllowed'} );
    $hValues->{'ShowFunctionHelp'}     = $self->soapBool( $hValues->{'ShowFunctionHelp'} );

    return $hValues;
}

#========================================================================================
# function     updateSuperMerchant
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->updateSuperMerchant( $ShopID, {
#                   'Alias' => $Alias,
#                   'Password' => $Password,
#                   'EMail' => $EMail,
#                   'Name' => $Name,
#                   'LocaleID' => $LocaleID,
#                   'DeleteConfirmation' => $DeleteConfirmation,
#                   'IsLoginCookieAllowed' => $IsLoginCookieAllowed,
#                   'ShowFunctionHelp' => $ShowFunctionHelp,
#               });
#----------------------------------------------------------------------------------------
# description  Sets login, password, email and other settings of the primary administrator user.
#               All keys are optional.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $hOptions     | Options;
#               used keys are:
#               <ul>
#                 <li>Alias | login | string</li>
#                 <li>Password | password (encrypted) | string</li>
#                 <li>EMail | email address | string</li>
#                 <li>Name | name | string</li>
#                 <li>LocaleID | locale id | string</li>
#                 <li>DeleteConfirmation | true if deletion should be confirmed | boolean</li>
#                 <li>IsLoginCookieAllowed | true if user automated login over cookies is enabled | boolean</li>
#                 <li>ShowFunctionHelp | true if function help is allowed | boolean</li>
#               </ul> | hash.ref
#========================================================================================
sub updateSuperMerchant {
    shift;
    my ($ShopID, $hOptions) = @_;
    LogDebug('updateSuperMerchant', [$ShopID, $hOptions]);
    TestInteger('ShopID', $ShopID);
    TestHash('hOptions', $hOptions, {
        'Alias' => 'String',
        'Password' => 'String',
        'EMail' => 'String',
        'Name' => 'String',
        'LocaleID' => 'String',
        'DeleteConfirmation' => 'Integer',
        'IsLoginCookieAllowed' => 'Integer',
        'ShowFunctionHelp' => 'Integer',
    }, 1);

    my $Shop = LoadObject( $ShopID );
    my $SuperMerchant = $Shop->get('SuperMerchant');
    $SuperMerchant->set($hOptions) if 0 < scalar keys %$hOptions;
}

#========================================================================================
# function     sendSuperMerchantNewPassword
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->sendSuperMerchantNewPassword( $ShopID, $Login);
#----------------------------------------------------------------------------------------
# description  Send a new password to super merchant.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $Login | user login | string
#========================================================================================
sub sendSuperMerchantNewPassword {
    shift;
    my ($ShopID, $Login) = @_;

    my $Site = LoadObject($ShopID);
    my $UserFolder = GetUserFolder($Site);
    my $User = $UserFolder->child($Login);
    my $UserEMail = $User->get('EMail');

    my %MailParams;
    my $System   = LoadRootObject();
    my $MailType = $System->folder('MailTypes')->child('NewPassword');
    $MailParams{'FROM'} = $UserEMail;

    SendNewPasswordToken( $Site, $User, $MailType, \%MailParams );
}

sub _url {
    my ($URL, $ParameterString) = @_;
    if ($URL =~ /\/\?/) {
        $URL .= '&'.$ParameterString;
    } else {
        $URL .= '/?'.$ParameterString;
    }
}

#========================================================================================
# function     createSuperMerchantSession
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hValues = $ShopService->createSuperMerchantSession( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns hash with URL and valid session of the SuperMerchant for the MBO
#               of the given Shop ID. If the shop attribute "GrantServiceAccessUntil" is
#               a date in the past SessionId will be undef and Link to URL contains no session
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $AlreadyGranted | true if access is allowed without merchant grant| boolean
# return       $hValues | hash with valid session and URL to MBO; keys:
#                       <ul>
#                       <li>SessionId | a valid SuperMerchant session | string</li>
#                       <li>WebUrlAdminSSL | Link to MBO without SessionId | string</li>
#                       <li>WebUrlAdminSSLWithSessionId | Link to MBO with session | string</li>
#                       </ul> | ref.hash
#========================================================================================
sub createSuperMerchantSession {
    shift;
    my ($ShopID, $AlreadyGranted) = @_;
    LogDebug('createSuperMerchantSession', $ShopID);
    TestInteger('ShopID', $ShopID);

    my $Shop = LoadObject( $ShopID );
    my $GrantServiceAccess = $Shop->get('GrantServiceAccess');

    my $hValues->{'WebUrlAdminSSL'} = $Shop->get('WebUrlAdminSSL');

    if ($GrantServiceAccess || $AlreadyGranted) {
        my $SuperMerchant = $Shop->get('SuperMerchant');
        $hValues->{'SessionId'} = InsertSession( $SuperMerchant, 0 );
        $hValues->{'WebUrlAdminSSLWithSessionId'} =
          _url( $hValues->{'WebUrlAdminSSL'}, 'SessionID=' . $hValues->{'SessionId'} );
    }
    else {
        $hValues->{'SessionId'}                   = undef;
        $hValues->{'WebUrlAdminSSLWithSessionId'} = $hValues->{'WebUrlAdminSSL'};
    }
    return $hValues;
}

#========================================================================================
# function     setFeatureValues
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->setFeatureValues( $ShopID, $ahFeatureValues );
# example      $ShopService->setFeatureValues( 32082, [
#                   { Feature => 'Products', MaxValue => 1000 },
#                   { Feature => 'SlideShow', MaxValue => 1},
#                   { Feature => 'Forum', MaxValue => 0, IsVisible=>0 },
#               ]);
#----------------------------------------------------------------------------------------
# description  Sets the maximum feature values for a shop.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $ahFeatureValues | list of features and their values. Each list element
#                   is a hash with the keys <ul>
#                   <li>Feature | feature alias | string</li>
#                   <li>MaxValue | max. feature value for the shop | integer</li>
#                   <li>IsVisible | is visible for the shop | boolean</li>
#                   </ul> | ref.array.hash
#========================================================================================
sub setFeatureValues {
    shift;
    my ($ShopID, $ahFeatureValues) = @_;
    TestInteger( 'ShopID', $ShopID );
    TestArray( 'ahFeatureValues', $ahFeatureValues );

    my $Shop = LoadObject( $ShopID );
    my $Features = LoadObjectByPath( '/Features' );
    Transaction( sub {
        foreach my $hFeatureValue (@$ahFeatureValues) {
            my $MaxValue = $hFeatureValue->{MaxValue};
            my $IsVisible = exists $hFeatureValue->{IsVisible} ? $hFeatureValue->{IsVisible} : 1;

            # attension feature doesnt exists
            if (not $Features->existsChild( $hFeatureValue->{Feature} )) {
                if ($MaxValue) {
                    Error('FeatureNotAvialable', {
                        'Feature'=>$hFeatureValue->{Feature},
                     });
                } else { # ignore 0 value for non existing features
                    next;
                }
            }
            my $Feature = $Features->child( $hFeatureValue->{Feature} );
            $Shop->featureMaxValue( $Feature, $MaxValue, $IsVisible );
        }
    });
    return;
}

sub setShopsFeatureValues {
    my $self = shift;
    my ($ShopData) = @_;

    foreach my $ShopID (keys (%$ShopData)) {
        next unless $ShopID;
        next unless ExistsObject($ShopID);
        next unless LoadObject($ShopID)->instanceOf('Shop');
        LogDebug('update Features on Shop: ',$ShopID);
        setFeatureValues($self, $ShopID, $ShopData->{$ShopID});
    }
    return;
}

#========================================================================================
# function     getFeatureValues
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahFeatureValues = $ShopService->shopService->getFeatureValues( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns the allowed maximum feature values of the features of a shop.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# return       $ahFeatureValues | list of features and their values. Each list element
#                   is a hash with the keys <ul>
#                   <li>Feature | feature alias | string</li>
#                   <li>MaxValue | max. feature value for the shop | integer</li>
#                   </ul> | ref.array.hash
#========================================================================================
sub getFeatureValues {
    shift;
    my ($ShopID) = @_;
    TestInteger( 'ShopID', $ShopID );

    my $Shop = LoadObject( $ShopID );
    my $ahFeatureValues = $Shop->get('FeatureValues');
    $_->{'Feature'} = $_->{'Feature'}->alias foreach @{$ahFeatureValues};

    return $ahFeatureValues;
}

#========================================================================================
# function     getAllFeatures
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahFeatures = $ShopService->shopService->getAllFeatures()->result;
#----------------------------------------------------------------------------------------
# description  Returns informations of feature installed on the store.
#----------------------------------------------------------------------------------------
# return       $ahFeature | list of features. Each list element
#                   is a hash with the keys <ul>
#                     <li>Alias | feature alias | string</li>
#                   </ul> | ref.array.hash
#========================================================================================
sub getAllFeatures {
    shift;

    # get all Catridges.....
    my $ahCartriges = GetAllInfoCartridges();

    my $Class = LoadClassByAlias('Feature');
    my $ahFeatures;
    foreach my $hCartridge (@$ahCartriges) {
        next unless $hCartridge->{'InstallState'} == 2;
        my $aFeatures = $Class->getObjects('ByCartridgeID', [$hCartridge->{'CartridgeID'}]);
        push @$ahFeatures , map { {'Alias' => $_->alias} } @$aFeatures;
    }

    return $ahFeatures;
}

#========================================================================================
# function     getLastMerchantLogin
# state        public
#----------------------------------------------------------------------------------------
# syntax       $String = $ShopService->getLastMerchantLogin( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns the last login time and date of a shop merchant.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# return       $String | string containing date and time | string
#========================================================================================
sub getLastMerchantLogin {
    shift;
    my ($ShopID) = @_;

    my $Shop = LoadObject( $ShopID );
    my $LastMerchantLogin = $Shop->get('LastMerchantLogin');

    return unless $LastMerchantLogin;

    my $parser = DE_EPAGES::XML::API::DateTimeFormatXML->new();

    return $parser->format_datetime($LastMerchantLogin);
}

#========================================================================================
# function     getShopAddress
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hShopAddress = $ShopService->shopService->getShopAddress( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns the shop address as a tle hash.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# return       $hShopAddress | hash containing the address values | ref.hash
#========================================================================================
sub getShopAddress {
    my $self = shift;
    my ($ShopID) = @_;

    my $Shop = LoadObject( $ShopID );
    my $Address = $Shop->get('Address');
    return {} unless defined $Address;

    my @Attributes = keys %{$Address->exportableAttributes};
    push @Attributes, 'FullName';
    my $LanguageCode = GetCodeByLanguageID( $Shop->get('LanguageID') );
    my $hValues = $self->get($Address->id, \@Attributes, $LanguageCode);
    return $hValues;
}

#========================================================================================
# function     getShopAttributeAliases
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hShopAttributeAliases = $ShopService->shopService->getShopAttributeAliases( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns the shop attribute aliases as a tle hash.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# return       $hShopAttributeAliases | hash containing the attribute aliases | ref.hash
#========================================================================================
sub getShopAttributeAliases {
    my $self = shift;
    my ($ShopID) = @_;

    my $Shop = LoadObject( $ShopID );
    my $Class = $Shop->class;

    my $hAttributes = {};
    foreach my $Attribute (@{$Class->attributes}) {
           $hAttributes->{$Attribute->get('Alias')} = 1;
    }
    return $hAttributes;
}

#========================================================================================
# function     getMerchantEMails
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aShopMembers = $ShopService->shopService->getMerchantEMails( $ShopID )->result;
#----------------------------------------------------------------------------------------
# description  Returns logins and emails of merchant member (admins)
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# return       $aShopMembers | hash containing the Member 'Alias' and 'EMail' | ref.array.hash
#========================================================================================
sub getMerchantEMails {
    my $self = shift;
    my ($ShopID) = @_;

    my $Shop = LoadObject( $ShopID );
    my $aMerchantMembers = $Shop->child('Groups')->child('Merchants')->get('Members');

    my @Members;

    foreach my $Member (@$aMerchantMembers){
        if($Member->get('IsActivated') eq '1') {
            my $hTemp;
            $hTemp->{'Alias'} = $Member->alias;
            $hTemp->{'EMail'} = $Member->get('EMail');
            push(@Members, $hTemp);
        }
    }
    return \@Members;
}

#========================================================================================
# function     setMerchantEMails
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->shopService->setMerchantEMails( $ShopID, \@Admins );
#----------------------------------------------------------------------------------------
# description  Set EMails of merchant members (admins)
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $aMerchantEMails | list of merchant emails
#               i.e. {'MerchantAlias'=>'admin', 'MerchantEMail'=>'email@do.main'} | ref.array.hash
#========================================================================================
sub setMerchantEMails {
    my $self = shift;
    my ($ShopID, $aMerchantEMails) = @_;

    my $ShopObject = LoadObject( $ShopID );
    my $UsersFolder = $ShopObject->folder('Users');

    foreach my $hMerchantLoginEMail (@$aMerchantEMails){
        my $MerchantObject = $UsersFolder->child($hMerchantLoginEMail->{'MerchantAlias'});
        $MerchantObject->set({'EMail' => $hMerchantLoginEMail->{'MerchantEMail'}});
    }
    return;
}

#========================================================================================
# function     copyPublicFiles
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->copyPublicFiles( $ShopID, $SourceDir);
#----------------------------------------------------------------------------------------
# description  Copies all files from the $SourceDir directory to the public shop
#               directory. Source must be a folder from the shop's public path.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $SourceDir | source directory name | string
#========================================================================================
sub copyPublicFiles {
    my $self = shift;
    my ($ShopID,$SourceDir) = @_;
    TestInteger('ShopID', $ShopID);

    my $Shop = LoadObject( $ShopID );
    Error( 'INVALID_SHOP_ID', { 'ShopID' => $ShopID } ) unless $Shop->instanceOf('Shop');

    my $TargetDir = $Shop->get('PublicPath');

    my $User = $self->user();
    #-- Hidden feature: if calling using a 'System' user, allow ANY source directory
    #-- otherwise, check that the source directory is within the public path of the shop
    if ( $User->get('Site')->class->alias eq 'Shop' ) {
      if ( ! IsInPath($SourceDir, $TargetDir) ) {
        Error( 'INVALID_SOURCE', { 'Directory' => $SourceDir } );
      }
    }

    CopyDirectory( $SourceDir, $TargetDir, undef, { recursive => 1 } );
    return;
}

#========================================================================================
# function     getSecondaryDomains
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->getSecondaryDomains( $ShopID );
#----------------------------------------------------------------------------------------
# description  Return all secondary domains assigned to this shop.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
#----------------------------------------------------------------------------------------
# return       $aSecondaryDomains | secondary domain names | ref.array.string
#========================================================================================
sub getSecondaryDomains {
    my $self = shift;
    my ($ShopID) = @_;
    TestInteger('ShopID', $ShopID);

    return LoadObject($ShopID)->get('SecondaryDomains');
}

#========================================================================================
# function     setSecondaryDomains
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ShopService->setSecondaryDomains( $ShopID, $aDomainNames );
#----------------------------------------------------------------------------------------
# description  Set the additional domain names of this shop.
#----------------------------------------------------------------------------------------
# input        $ShopID | shop id | integer
# input        $aDomainNames | list of domain names to add | ref.array.string
#========================================================================================
sub setSecondaryDomains {
    my $self = shift;
    my ($ShopID, $aDomainNames) = @_;
    TestInteger('ShopID', $ShopID);
    TestArray('aDomainNames', $aDomainNames);

    my $Shop = LoadObject($ShopID);
    $Shop->set({'SecondaryDomains' => $aDomainNames});

    return;
}

1;
