#========================================================================================
# package      DE_EPAGES::ShopMigration::API::PriceListsCommon
# state        public
#----------------------------------------------------------------------------------------
# description common functions for exporting/importing PriceLists
#========================================================================================
package DE_EPAGES::ShopMigration::API::PriceListsCommon;
use base qw(Exporter);

use strict;
use feature 'switch';
use DE_EPAGES::ShopMigration::API::ShopCommonData qw(GetActiveShopLanguages);
use DE_EPAGES::Object::API::Language qw(GetPKeyLanguageByCode);
use DE_EPAGES::Object::API::Object::Object qw (clone);

our @EXPORT_OK = qw(
    SetPriceLists
    SetPriceListData
    GetPriceListData
    DeletePriceListsPrepShop
    CreateAdditionalPriceListsPrepShop

);
use DE_EPAGES::Core::API::PerlTools qw(ElementOf);
use DE_EPAGES::ShopMigration::API::Constants qw(
    PRICELIST_ATTRIBUTES
    PRICELIST_LOCALIZEDATTRIBUTES
);
use DE_EPAGES::Core::API::Error qw (Error);
use DE_EPAGES::Core::API::Log qw(LogDebug);

#========================================================================================
# function     GetPriceListData
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hPriceList = GetPriceListData($Shop, $PriceList);
#----------------------------------------------------------------------------------------
# description  Constructs a hash with the exportable fields for a single PriceList
#----------------------------------------------------------------------------------------
# input        $Shop | BaseShop (epages6) | object
# input        $PriceList | PriceList | object
#----------------------------------------------------------------------------------------
# return       $hPriceList | hash with following fields
#               <ul>
#                   <li>Attributes
#                       <ul>
#                           <li>Alias                            - string</li>
#                           <li>CurrencyID                       - string</li>
#                           <li>CustomerGroups          - array of aliases</li>
#                       </ul>
#                   </li>
#                    <li>LocalizedAttributes
#                       <ul>
#                           <li>LanguageCode2            - en|de|..</li>
#                               <ul>
#                                   <li>Name             - string</li>
#                               </ul>
#                           </li>
#                       </ul>
#                   </li>
#               </ul>| reference to hash
#========================================================================================
sub GetPriceListData {
    my ($Shop, $PriceList) = @_;

    my %Info = ();

    my $ActiveLanguages = GetActiveShopLanguages($Shop);
    # get baseattributes, reset some attributes to default in Prep2Base step (not used in now)
    my $Attributes = [ @{PRICELIST_ATTRIBUTES()} ];
    # Get customer groups assigned to price lists
    $Info{'Attributes'} = $PriceList->get($Attributes);
    $Info{'CustomerGroups'} =  $PriceList->{'CustomerGroups'} = _getPriceListCustomerGroupData($PriceList);
    # get localized attributes
    my $LocalizedAttributes = [ @{PRICELIST_LOCALIZEDATTRIBUTES()} ];
    foreach my $hLanguage (@$ActiveLanguages) {
        my $hLocalInfo = $PriceList->get($LocalizedAttributes, $hLanguage->{'LanguageID'});
        $Info{'LocalizedAttributes'}{$hLanguage->{'Code2'}} = $hLocalInfo;
    }
    return \%Info;
}

#========================================================================================
# function     SetPriceLists
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetPriceLists($Shop, $ahPriceLists);
#----------------------------------------------------------------------------------------
# description  sets attributes on PriceList as specified in the $ahPriceLists
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $ahPriceLists  | see SetPriceListData() | reference to array of hashes
#========================================================================================
sub SetPriceLists {
    my ($Shop, $ahPriceLists) = @_;

    foreach my $hPriceList (@$ahPriceLists) {
        SetPriceListData($Shop, $hPriceList);
    }
}

#========================================================================================
# function     SetPriceListData
# state        public
#----------------------------------------------------------------------------------------
# syntax       SetPriceListData($Shop, $hPriceList);
#----------------------------------------------------------------------------------------
# description  sets attributes on a PriceList as specified in the hash $hPriceList
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $hPriceList    | hash with following fields
#               <ul>
#                   <li>Attributes
#                       <ul>
#                           <li>Alias           - string</li>
#                           <li>CurrencyID      - string</li>
#                           <li>CustomerGroups          - array of aliases</li>
#                       </ul>
#                   </li>
#                    <li>LocalizedAttributes
#                       <ul>
#                           <li>LanguageCode2            - en|de|..</li>
#                               <ul>
#                                   <li>Name             - string</li>
#                               </ul>
#                           </li>
#                       </ul>
#                   </li>
#               </ul>| reference to hash
#========================================================================================
sub SetPriceListData {
    my ($Shop, $hPriceList) = @_;

    my @Attributes = @{PRICELIST_ATTRIBUTES()};
    my $LocalizedAttributes = [ @{PRICELIST_LOCALIZEDATTRIBUTES()}];

    foreach my $Key (keys %{$hPriceList->{Attributes}}) {
        die "Invalid Element for Attributes: $Key" unless ElementOf($Key, @Attributes);
    }

    #Set the date to the future
    $hPriceList->{"Attributes"}{'ValidTo'} = DateTime->new('year'=>2999, 'month'=>1,'day'=>1);

    if ($Shop->get('IsMigrationPrepShop')) {
        $hPriceList->{'Attributes'}{'Alias'} = "PL_epages6_".$hPriceList->{'Attributes'}{'Alias'};
    }

    my $NewPriceListAlias = $hPriceList->{'Attributes'}{Alias};
    #Set attributes
    my $NewPriceList = $Shop->folder('PriceLists')->existsChild($NewPriceListAlias) ?
        $Shop->folder('PriceLists')->child($NewPriceListAlias) :
        $Shop->folder('PriceLists')->insertChild('PriceList', {
            %{$hPriceList->{'Attributes'}},
        });

    # Set localized attributes
    foreach my $LanguageCode2 (keys %{$hPriceList->{LocalizedAttributes}}) {
        foreach my $Key (keys %{$hPriceList->{LocalizedAttributes}{$LanguageCode2}}) {
            die "Invalid Element for Attributes: $Key" unless ElementOf($Key, @$LocalizedAttributes);
            next unless $Key eq 'Name';
            my $Value = $hPriceList->{LocalizedAttributes}{$LanguageCode2}{$Key};
            if (defined $Value && $Value !~ /\w/) {
                Error('NAME_NOT_VALID', {
                    'Alias'      => $hPriceList->{'Attributes'}{'Alias'},
                    'LanguageCode2' => $LanguageCode2,
                    'Attribute'     => 'Name',
                    'Value'         => $Value,
                });
            }
        }

        my $LanguageID = GetPKeyLanguageByCode($LanguageCode2);
        $NewPriceList ->set($hPriceList->{LocalizedAttributes}{$LanguageCode2}, $LanguageID);
    }

    foreach my $CustomerGroupItem (@{$hPriceList->{'CustomerGroups'}}) {
        # add Prefix "CG_epages6_" if CustomerGroup Alias is NewCustomer and it is not IsNonMemberDefault
        if ($CustomerGroupItem->{'Alias'} eq 'NewCustomer' && $Shop->get('IsMigrationPrepShop')) {
            if ($Shop->folder('Groups')->existsChild($CustomerGroupItem->{'Alias'})) {
                my $PrepShopCustomerGroup = $Shop->folder('Groups')->child($CustomerGroupItem->{'Alias'});
                if (!$PrepShopCustomerGroup->get('IsNonMemberDefault')) {
                    $CustomerGroupItem->{'Alias'} = "CG_epages6_".$CustomerGroupItem->{'Alias'};
                }
            }else{
                $CustomerGroupItem->{'Alias'} = "CG_epages6_".$CustomerGroupItem->{'Alias'};
            }
        }
        my $CustomerGroup = $Shop->folder('Groups')->child($CustomerGroupItem->{'Alias'});
        $NewPriceList->addCustomerGroup($CustomerGroup);
    }
    return;
}

#========================================================================================
# function     _getPriceListCustomerGroupData
# state        private
#----------------------------------------------------------------------------------------
# syntax       _getPriceListCustomerGroupData($PriceList);
#----------------------------------------------------------------------------------------
# description  Returns the customer groups for a specified price list.
#----------------------------------------------------------------------------------------
# input        $PriceList         | price list from a shop | object
#----------------------------------------------------------------------------------------
# return       $hCustomerGroups | hash with following fields
#               <ul>
#                   <li>Alias           - string</li>
#               </ul>
#========================================================================================
sub _getPriceListCustomerGroupData {
    my ($PriceList) = @_;

    my $aCustomerGroups = $PriceList->get('CustomerGroups');

    return [ map {
        {
            'Alias'   => $_->get('Alias'),
        }
    } sort {$a->get('Alias') <=> $b->get('Alias')} @$aCustomerGroups ];
}

#========================================================================================
# function     DeletePriceListsPrepShop
# state        public
#----------------------------------------------------------------------------------------
# syntax       DeletePriceListsPrepShop($Shop);
#----------------------------------------------------------------------------------------
# description  deletes pricelists from prepshop
#----------------------------------------------------------------------------------------
# input        $Shop | prep shop - shop object | object
#========================================================================================
sub DeletePriceListsPrepShop {
    my ($Shop) = @_;

    if ($Shop->get('IsMigrationPrepShop')) {
        my $aPriceLists = $Shop->get('PriceLists');
        foreach my $PriceList (@{$aPriceLists}) {
            $PriceList->delete;
        }
    }
    return;
}

#========================================================================================
# function     CreateAdditionalPriceListsPrepShop
# state        public
#----------------------------------------------------------------------------------------
# syntax       CreateAdditionalPriceListsPrepShop($Shop)
#----------------------------------------------------------------------------------------
# description  Clone Pricelist in Prepshop and assign CustomerGroup to it, if Pricelist
#               has multiple CustomerGroups assigned. Import ScalePrices, QunatityDiscounts
#               and GlobalValueDiscounts for each new cloned PriceList.
#----------------------------------------------------------------------------------------
# input        $Shop | prep shop - shop object | object
#========================================================================================
sub CreateAdditionalPriceListsPrepShop {
    my ($Shop) = @_;

    my $PriceLists = $Shop->get('PriceLists');

    foreach my $PriceList (@$PriceLists) {
        my $aPriceListCustomerGroups = $PriceList->get('CustomerGroups');

        if (scalar(@{$aPriceListCustomerGroups}) > 1) {
            my $Changed;

            foreach my $CustomerGroup (@$aPriceListCustomerGroups) {
                if ($CustomerGroup->alias eq 'NewCustomer') {
                    $Changed = 1;
                    next;
                }

                # Rename PriceList to match first CustomerGroup Alias
                if (!defined $Changed && $PriceList->alias ne 'PriceListCustomer_' . $Shop->get('CurrencyID')) {
                    my $PriceListAlias = $PriceList->alias;
                    $PriceList->set({ 'Alias' => $CustomerGroup->alias });
                    LogDebug(sprintf("Rename PriceList '%s' to match first CustomerGroup alias '%s'!\n",
                        $PriceListAlias, $CustomerGroup->alias), 1);
                    $Changed = 1;
                    next;
                }

                if ($PriceList->alias eq 'PriceListCustomer_' . $Shop->get('CurrencyID') && $CustomerGroup->alias ne "NewCustomer") {
                    # Remove CustomerGroup from current PriceList, because we can add only one CustomerGroup to a active Pricelist
                    $PriceList->removeCustomerGroup($CustomerGroup);
                    LogDebug(sprintf("Remove CustomerGroup '%s' from '%s' PriceList'!\n",
                        $CustomerGroup->alias, $PriceList->alias), 1);

                    # Clone current PriceList and add removed CustomerGroup to it
                    my $ClonedPriceList = $PriceList->clone({ 'Alias'          => $CustomerGroup->alias,
                                                              'CustomerGroups' => [ $CustomerGroup ] }, 1);
                    LogDebug(sprintf("Clone PriceList '%s' as new '%s' PriceList and add CustomerGroup '%s'!\n",
                        $PriceList->alias, $ClonedPriceList->alias, $CustomerGroup->alias), 1);

                    unless (defined $Changed) {
                        $Changed = 1;
                    }
                    next;
                }

                # Remove CustomerGroup from current PriceList, because we can add only one CustomerGroup to a active Pricelist
                $PriceList->removeCustomerGroup($CustomerGroup);
                LogDebug(sprintf("Remove CustomerGroup '%s' from '%s' PriceList'!\n",
                    $CustomerGroup->alias, $PriceList->alias), 1);

                # Clone current PriceList and add removed CustomerGroup to it
                my $ClonedPriceList = $PriceList->clone({ 'Alias'          => $CustomerGroup->alias,
                                                          'CustomerGroups' => [ $CustomerGroup ] }, 1);
                LogDebug(sprintf("Clone PriceList '%s' as new '%s' PriceList and add CustomerGroup '%s'!\n",
                    $PriceList->alias, $ClonedPriceList->alias, $CustomerGroup->alias), 1);
            }
        }
    }
    return;
}

1;
