#========================================================================================
# package      DE_EPAGES::ShopMigration::t::PriceListGlobalValueDiscountExportImport
# base         DE_EPAGES::Test::API::UnitTest::Test
#----------------------------------------------------------------------------------------
# description  unit test for DE_EPAGES::ShopMigration::API::JSONLines::PriceListGlobalValueDiscountsExportImport
#========================================================================================
package DE_EPAGES::ShopMigration::t::PriceListGlobalValueDiscountExportImport;
use base DE_EPAGES::Test::API::StoreUnitTest;

use strict;
use utf8;
use warnings;
use File::Spec::Functions;
use File::Temp qw(tempdir);
use Test::Differences qw(eq_or_diff);
use Test::More;

use DE_EPAGES::Core::API::File qw(GetFileContent GetFileContentLines ExistsFile DeleteDirectory DeleteFile Text2Unix GetFileNames);
use DE_EPAGES::Object::API::Factory qw(LoadRootObject LoadClassByAlias);
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw(ImportObjects);
use DE_EPAGES::PriceList::API::Table::GlobalValueDiscount qw ( GetGlobalValueDiscountsByPriceListID );
use DE_EPAGES::ShopMigration::API::JSON qw(DecodeJSONArray);
use DE_EPAGES::ShopMigration::API::JSONLines::PriceListGlobalValueDiscountsExportImport qw(ExportPriceListGlobalValueDiscountsJSON ImportPriceListGlobalValueDiscountsJSON);
use DE_EPAGES::ShopMigration::API::MigrationShop qw(CreatePrepShop ExistsPrepShop SetShopFeatureValues);
use DE_EPAGES::ShopMigration::API::PriceListsCommon qw (CreateAdditionalPriceListsPrepShop);

my $TmpDir = tempdir();

if ($ENV{SKIP_IN_PR_TESTS}) {
    plan skip_all => 'Skipping it because SKIP_IN_PR_TESTS == 1';
}

sub test {
    testExportWithPriceListGlobalValueDiscountsExportImportBase2Prep();
    testImportWithPriceListGlobalValueDiscountsExportImportBase2Prep();
    testExportWithPriceListGlobalValueDiscountsExportImportPrep2Base();
    testImportWithPriceListGlobalValueDiscountsExportImportPrep2Base();
    cleanup();
    done_testing();
}

my %TestData = (
    TestShop      => undef,
    TestShopAlias => 'TestShop',
    ImportPath    => '/Shops/TestShop',
);
use constant FILENAME => 'pricelistglobalvaluediscounts.json';
use constant FILE_DIR => [ qw(ShopData.d export TestShop) ];
use constant FILE_LINES => 3;
use constant FILE_LINES_PREP => 19;

# This is a bit hacky, the json exporter *can* export a numeric value either as 1 or as '1' (string) depending on how it was used before
# I got random results, so we just check both (as it just concerns the 'IsActive' key)
sub testExportWithPriceListGlobalValueDiscountsExportImportBase2Prep {
    my $Shop = $TestData{TestShop};

    note('test PriceListGlobalValueDiscountsExportDriver (Base2Prep)');
    isnt($Shop, undef, 'Testshop exists');
    is(scalar @{$Shop->get('Languages')}, 1, 'Testshop has now one Language');
    is(scalar @{$Shop->get('SiteLocales')}, 1, 'Testshop has one SiteLocale');
    my $Exporter = ExportPriceListGlobalValueDiscountsJSON($Shop, $TmpDir);
    ok(ExistsFile($Exporter->fileName), 'export file exists: ' . $Exporter->fileName);
    my $LinesCount = scalar @{GetFileContentLines($Exporter->fileName, { encoding => 'utf8' })};
    is($LinesCount, FILE_LINES, "export file lines count is ${\FILE_LINES}");
    ok($Exporter->fileName =~ /${\FILENAME}/, "export file name matches ${\FILENAME}");
    my $sFileContent = GetFileContent($Exporter->fileName, { encoding => 'utf8' });
    my $sExpected = GetFileContent(catfile(@{FILE_DIR()}, 'base2prep_' . FILENAME), { encoding => 'utf8' });
    compareGlobalValueDiscountHashes($sFileContent, $sExpected);
    DeleteFile($Exporter->fileName());
    ok(!ExistsFile($Exporter->fileName()), 'export file cleaned up');

    $Exporter = ExportPriceListGlobalValueDiscountsJSON($Shop, $TmpDir);
    ok(ExistsFile($Exporter->fileName), 'export file exists: ' . $Exporter->fileName);
    is(scalar @{GetFileContentLines($Exporter->fileName, { encoding => 'utf8' })}, 3, 'Export file has 3 lines');
    DeleteFile($Exporter->fileName());
    ok(!ExistsFile($Exporter->fileName()), 'export file cleaned up');
}

sub testExportWithPriceListGlobalValueDiscountsExportImportPrep2Base {
    my $PrepShop = $TestData{PrepShop};

    note('test PriceListGlobalValueDiscountsExportDriver (Prep2Base)');
    isnt($PrepShop, undef, 'Prepshop exists');
    is(scalar @{$PrepShop->get('Languages')}, 1, 'Prepshop has now one Language');
    is(scalar @{$PrepShop->get('SiteLocales')}, 1, 'Prepshop has one SiteLocale');
    my $Exporter = ExportPriceListGlobalValueDiscountsJSON($PrepShop, $TmpDir);
    ok(ExistsFile($Exporter->fileName), 'export file exists: ' . $Exporter->fileName);
    my $LinesCount = scalar @{GetFileContentLines($Exporter->fileName, { encoding => 'utf8' })};
    is($LinesCount, FILE_LINES, "export file lines count is ${\FILE_LINES}");
    ok($Exporter->fileName =~ /${\FILENAME}/, "export file name matches ${\FILENAME}");
    my $sFileContent = GetFileContent($Exporter->fileName, { encoding => 'utf8' });
    my $sExpected = GetFileContent(catfile(@{FILE_DIR()}, 'prep2base_' . FILENAME), { encoding => 'utf8' });
    compareGlobalValueDiscountHashes($sFileContent, $sExpected);
    DeleteFile($Exporter->fileName());
    ok(!ExistsFile($Exporter->fileName()), 'export file cleaned up');

    $Exporter = ExportPriceListGlobalValueDiscountsJSON($PrepShop, $TmpDir);
    ok(ExistsFile($Exporter->fileName), 'export file exists: ' . $Exporter->fileName);
    is(scalar @{GetFileContentLines($Exporter->fileName, { encoding => 'utf8' })}, 3, 'Export file has 3 lines');
    DeleteFile($Exporter->fileName());
    ok(!ExistsFile($Exporter->fileName()), 'export file cleaned up');
}

sub compareGlobalValueDiscountHashes {
    my ($CurrentContent, $ExpectedContent) = @_;

    my $aCurrent = DecodeJSONArray($$CurrentContent);
    my $aExpected = DecodeJSONArray($$ExpectedContent);
    eq_or_diff($aCurrent, $aExpected, 'file content matches');
}

sub testImportWithPriceListGlobalValueDiscountsExportImportBase2Prep {

    my $ExpectedValues = [
        { 'Value' => '10.0000', 'Discount' => '10' },
        { 'DiscountAbsolute' => '20.0000', 'Value' => '20.0000' },
        { 'DiscountAbsolute' => '30.0000', 'Value' => '30.0000', 'Discount' => '30' }
        ];

    my $PrepShop = $TestData{PrepShop};
    my $DirName = catdir('ShopData.d', 'import');
    note('test PriceListGlobalValueDiscountsImportHandler (Base2Prep)');
    isnt($PrepShop, undef, 'Prepshop exists');
    my $Importer = ImportPriceListGlobalValueDiscountsJSON($PrepShop, $DirName);
    CreateAdditionalPriceListsPrepShop($PrepShop);
    ok($Importer->fileName() =~ /TestShop.base2prep_pricelistglobalvaluediscounts[.]json/, 'ImportFileName was constructed properly');
    my $PriceList = $PrepShop->folder('PriceLists')->child('PL_epages6_RegularCustomer_GBP');

    my $hPriceListGlobalValueDiscounts = GetGlobalValueDiscountsByPriceListID($PrepShop->id,$PriceList->id);
    eq_or_diff($ExpectedValues, $hPriceListGlobalValueDiscounts, 'Global value discounts match (Base2Prep)');
}

sub testImportWithPriceListGlobalValueDiscountsExportImportPrep2Base {

    my $ExpectedValues = [
        { 'Value' => '10.0000', 'Discount' => '10' },
        { 'DiscountAbsolute' => '20.0000', 'Value' => '20.0000' },
        { 'DiscountAbsolute' => '30.0000', 'Value' => '30.0000', 'Discount' => '30' }
        ];

    my $TestShop = $TestData{TestShop};
    my $DirName = catdir('ShopData.d', 'import');
    note('test PriceListGlobalValueDiscountsImportHandler (Prep2Base)');
    isnt($TestShop, undef, 'Prepshop exists');
    #delete existing PriceLists in TestShop
    $TestShop->folder('PriceLists')->deleteChildren;
    #Import pricelists from Now into Base shop.
    ImportObjects('ImportFiles/PriceListsNow.xml', $TestShop->pathString());
    my $Importer = ImportPriceListGlobalValueDiscountsJSON($TestShop, $DirName);
    ok($Importer->fileName() =~ /TestShop.prep2base_pricelistglobalvaluediscounts[.]json/, 'ImportFileName was constructed properly');
    my $PriceList = $TestShop->folder('PriceLists')->child('RegularCustomer');

    my $hPriceListGlobalValueDiscounts = GetGlobalValueDiscountsByPriceListID($TestShop->id,$PriceList->id);
    eq_or_diff($ExpectedValues, $hPriceListGlobalValueDiscounts, 'Global value discounts match (Prep2Base)');
}


# We need to create a fake ShopTypeRef UnityDE + ECom5000 in case they don't exists
# otherwise the Mapper won't work
sub init {
    my $self = shift;

    $self->SUPER::init;
    my $ShopAlias = $TestData{TestShopAlias};
    my $System = LoadRootObject();
    my $Shop;
    my $TestFolder = $System->folder('Shops');
    my $ShopTypeRefFolder = $System->folder('ShopTypeRefs');
    my $Class = LoadClassByAlias('ShopTypeRef');
    $Class->insertObject({ Alias => 'UnityDE', 'Parent' => $ShopTypeRefFolder })
        unless ($ShopTypeRefFolder->existsChild('UnityDE'));
    $Class->insertObject({ Alias => 'ECom5000', 'Parent' => $ShopTypeRefFolder })
        unless ($ShopTypeRefFolder->existsChild('ECom5000'));
    if ($TestFolder->existsChild($ShopAlias)) {
        $Shop = $TestFolder->child($ShopAlias);
    } else {
        $Shop = $TestFolder->insertChild('Shop', {
            Alias                   => $ShopAlias,
            ShopTypeRef             => $ShopTypeRefFolder->child('ECom5000'),
            WebServerScriptNamePart => $ShopAlias,
        });
        SetShopFeatureValues($Shop, [
            { Alias => 'Regions', MaxValue => '2' },
            { Alias => 'Languages', MaxValue => '15' },
            { Alias => 'ProductTypes', MaxValue => '5'},
            { Alias => 'PriceLists', MaxValue => '10'},
            { Alias => 'Products', MaxValue => '100'},
            { Alias => 'Variations', MaxValue => '5'},
            { Alias => 'ProductTypeAttributes', MaxValue => '10'},
            { Alias => 'CustomerGroups', MaxValue => '5'},
            { Alias => 'BasketDiscount', MaxValue => '1'}
        ]);
    }
    $TestData{TestShop} = $Shop;

    $TestData{PrepShop} = ExistsPrepShop($Shop) ? $Shop->get('MigrationPrepShop') : CreatePrepShop($Shop);
    SetShopFeatureValues($TestData{PrepShop}, [
        { Alias => 'Languages', MaxValue => '15' },
        { Alias => 'PriceLists', MaxValue => '10'},
        { Alias => 'ProductTypes', MaxValue => '5'},
        { Alias => 'Products', MaxValue => '100'},
        { Alias => 'Variations', MaxValue => '5'},
        { Alias => 'ProductTypeAttributes', MaxValue => '10'},
        { Alias => 'ProductMaxSubProducts', MaxValue => '50' },
        { Alias => 'CustomerGroups', MaxValue => '5'},
        { Alias => 'BasketDiscount', MaxValue => '1'}
    ]);
    foreach my $JSONFile (@{GetFileNames('ShopData.d', { 'recursive' => 1 })}) {
        Text2Unix($JSONFile) if ($^O eq 'linux');
    }
    ImportObjects('ImportFiles/PriceListGlobalValueDiscountsBase.xml', $TestData{TestShop}->pathString());
    ImportObjects('ImportFiles/PriceListGlobalValueDiscountsNow.xml', $TestData{PrepShop}->pathString());
    return;
}

sub cleanup {
    my $self = shift;

    my $ShopAlias = $TestData{TestShopAlias};
    my $TestFolder = LoadRootObject()->folder('Shops');
    if ($TestFolder->existsChild($ShopAlias)) {
        my $Shop = $TestFolder->child($ShopAlias);
        my $PrepShop = $Shop->get('MigrationPrepShop');
        $PrepShop->delete() if (defined $PrepShop);
        $Shop->delete();
    }
    DeleteDirectory($TmpDir, { 'recursive' => 1 });
    return;
}

__PACKAGE__->runScript();
1;
