#========================================================================================
# package      DE_EPAGES::Zettle::API::Object::ZettleEnvironment
# state        public
#----------------------------------------------------------------------------------------
# description  object for calling functions on the ZettleEnvironment
#========================================================================================
package DE_EPAGES::Zettle::API::Object::ZettleEnvironment;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;

use Readonly;

# Names of the attributes used to store token information in ZettleEnvironment object
Readonly my $ACCESS_TOKEN_KEY => 'AccessToken';
Readonly my $REFRESH_TOKEN_KEY => 'RefreshToken';
Readonly my $ACCESS_TOKEN_EXPIRATION_KEY => 'AccessTokenExpiryDate';

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( LogError );
use DE_EPAGES::Core::API::PerlTools qw ( AddDuration GetCurrentDateTime );
use DE_EPAGES::Core::API::TestParameter qw ( TestObject );
use DE_EPAGES::Object::API::Factory qw ( LoadClassByAlias );
use DE_EPAGES::Zettle::DAL::Zettle qw ( DAL_GetDeletedObjects );

#ERRORDEFINITION
sub ErrorDefinition () {
    return {
        'CURRENCY_MISMATCH' =>
          'Shop currency (#ShopCurrency) does not match Zettle currrency (#ZettleCurrency)',
    };
}
#ERRORDEFINITION END

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::Zettle::API::Object::ZettleEnvironment->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;

    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'ReplaceZettleLibrary'          => 0,
        'ExportScheduleInterval'        => 3,
        'ProductIsExporting'            => 0,
        'ExportProductPrices'           => 0,
        'PriceIsExporting'              => 0,
        'InitialExportStatus'           => 0,
        'MarkedForDeletion'             => undef,
    };
}

#========================================================================================
# function     accessToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->accessToken();
# syntax       $self->accessToken($AccessToken);
#----------------------------------------------------------------------------------------
# description  Get or set access token.
#----------------------------------------------------------------------------------------
# input        $AccessToken | optional. Access token retrieved from Zettle | string
#----------------------------------------------------------------------------------------
# return       $Result | Stored access token, if $AccessToken argument is not present | string
#========================================================================================
sub accessToken {
    my $self = shift;
    my ($AccessToken) = @_;
    
    return $self->get($ACCESS_TOKEN_KEY) unless defined $AccessToken;

    $self->set({$ACCESS_TOKEN_KEY => $AccessToken});
}

#========================================================================================
# function     refreshToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->refreshToken();
# syntax       $self->refreshToken($RefreshToken);
#----------------------------------------------------------------------------------------
# description  Get or set refresh token.
#----------------------------------------------------------------------------------------
# input        $RefreshToken | optional. Refresh token retrieved from Zettle | string
#----------------------------------------------------------------------------------------
# return       $Result | Stored refresh token, if $RefreshToken argument is not present | string
#========================================================================================
sub refreshToken {
    my $self = shift;
    my ($RefreshToken) = @_;
    
    return $self->get($REFRESH_TOKEN_KEY) unless defined $RefreshToken;

    $self->set({$REFRESH_TOKEN_KEY => $RefreshToken});
}

#========================================================================================
# function     accessTokenExpiration
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->accessTokenExpiration();
# syntax       $self->accessTokenExpiration($AccessTokenExpiration);
#----------------------------------------------------------------------------------------
# description  Get or set access token expiration date.
#----------------------------------------------------------------------------------------
# input        $AccessTokenExpiration | optional. Access token expiration date retrieved from Zettle | Datetime
#----------------------------------------------------------------------------------------
# return       $Result | Stored access token expiration date, if $AccessTokenExpiration argument is not present | Datetime
#========================================================================================
sub accessTokenExpiration {
    my $self = shift;
    my ($AccessTokenExpiration) = @_;
    
    return $self->get($ACCESS_TOKEN_EXPIRATION_KEY) unless defined $AccessTokenExpiration;

    $self->set({$ACCESS_TOKEN_EXPIRATION_KEY => $AccessTokenExpiration});
}

#========================================================================================
# function     updateTokenData
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->updateTokenData($AccessToken, $RefreshToken, $ExpirationDate);
#----------------------------------------------------------------------------------------
# description  Update token-related data in Zettle environment.
#----------------------------------------------------------------------------------------
# input        $AccessToken    | Access token | string
# input        $RefreshToken   | Refresh token | string
# input        $ExpirationDate | Access token expiration date | Datetime
#========================================================================================
sub updateTokenData {
    my $self = shift;
    my ($AccessToken, $RefreshToken, $ExpirationDate) = @_;

    $self->set({
        $ACCESS_TOKEN_KEY     => $AccessToken,
        $REFRESH_TOKEN_KEY    => $RefreshToken,
        $ACCESS_TOKEN_EXPIRATION_KEY  => $ExpirationDate,
    });
}

#========================================================================================
# function     setupInitialExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupInitialExport($InitialExportUUID, $ExportPrices);
#----------------------------------------------------------------------------------------
# description  Update initial export attributes:
#               <ul>
#                 <li> Assign $InitialExportUUID (this enables us to query export status afterwards)
#                 <li> Set 'LastProductExport' to current datetime.
#                 <li> Set 'LastPriceExport' to current datetime if $ExportPrices is TRUE.
#               </ul>
#----------------------------------------------------------------------------------------
# input        $InitialExportUUID | UUID of initial export | string
# input        $ExportPrices | TRUE if prices are to be exported with products, false otherwise | boolean
#========================================================================================
sub setupInitialExport {
    my $self = shift;
    my ($InitialExportUUID, $ExportPrices) = @_;

    if($ExportPrices){
        $self->set({
            'InitialExportUUID' => $InitialExportUUID,
            'LastProductExport' => GetCurrentDateTime(),
            'LastPriceExport'   => GetCurrentDateTime(),
        });
    }else{
        $self->set({
            'InitialExportUUID' => $InitialExportUUID,
            'LastProductExport' => GetCurrentDateTime(),
        });
    }
}

sub lastPriceExport {
    my $self = shift;
    my ($LastPriceExport) = @_;
    
    return $self->get('LastPriceExport') unless defined $LastPriceExport;

    $self->set({'LastPriceExport' => $LastPriceExport});
}

sub lastProductExport {
    my $self = shift;
    my ($LastProductExport) = @_;
    
    return $self->get('LastProductExport') unless defined $LastProductExport;

    $self->set({'LastProductExport' => $LastProductExport});
}

sub exportScheduleInterval {
    my $self = shift;
    my ($ExportScheduleInterval) = @_;
    
    return $self->get('ExportScheduleInterval') unless defined $ExportScheduleInterval;

    $self->set({'ExportScheduleInterval' => $ExportScheduleInterval});
}

#========================================================================================
# function     initialExportStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $InitialExportStatus = $self->initialExportStatus();
# syntax       $self->initialExportStatus($Status);
#----------------------------------------------------------------------------------------
# description  Get or update initial export status.
#----------------------------------------------------------------------------------------
# input        $Status | optional. Status value. See DE_EPAGES::Zettle::API::Constants::%INITIAL_EXPORT_STATUS | integer
#----------------------------------------------------------------------------------------
# return       $Result | Stored initial export status, if $Status argument is not present | integer
#========================================================================================
sub initialExportStatus {
    my $self = shift;
    my ($Status) = @_;
    
    return $self->get('InitialExportStatus') unless defined $Status;

    $self->set({'InitialExportStatus' => $Status});
}

#========================================================================================
# function     lastStockSync
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LastStockSync = $self->lastStockSync();
# syntax       $self->lastStockSync($LastStockSync);
#----------------------------------------------------------------------------------------
# description  Get or update last local stock update datetime.
#----------------------------------------------------------------------------------------
# input        $LastStockSync | optional. Last stock sync datetime | Datetime
#----------------------------------------------------------------------------------------
# return       $Result | Stored last stock sync datetime, if $LastStockSync argument is not present | Datetime
#========================================================================================
sub lastStockSync {
    my $self = shift;
    my ($LastStockSync) = @_;
    
    return $self->get('ProductStockLevelUpdateTime') unless defined $LastStockSync;

    $self->set({'ProductStockLevelUpdateTime' => $LastStockSync});
}

#========================================================================================
# function     lastStockSyncZettle
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LastStockSync = $self->lastStockSyncZettle();
# syntax       $self->lastStockSyncZettle($LastStockSync);
#----------------------------------------------------------------------------------------
# description  Get or update last Zettle stock update datetime.
#----------------------------------------------------------------------------------------
# input        $LastStockSync | optional. Last stock sync datetime | Datetime
#----------------------------------------------------------------------------------------
# return       $Result | Stored last stock sync datetime, if $LastStockSync argument is not present | Datetime
#========================================================================================
sub lastStockSyncZettle {
    my $self = shift;
    my ($LastStockSync) = @_;
    
    return $self->get('ProductStockLevelUpdateTimeZettle') unless defined $LastStockSync;

    $self->set({'ProductStockLevelUpdateTimeZettle' => $LastStockSync});
}

#========================================================================================
# function     initialExportUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ExportUUID = $self->initialExportUUID();
# syntax       $self->initialExportUUID($ExportUUID);
#----------------------------------------------------------------------------------------
# description  Get or update initial export UUID.
#----------------------------------------------------------------------------------------
# input        $ExportUUID | optional. Initial export UUIDv1 | string
#----------------------------------------------------------------------------------------
# return       $ExportUUID | Stored initial export UUID if $ExportUUID parameter is not present | string
#========================================================================================
sub initialExportUUID {
    my $self = shift;
    my ($ExportUUID) = @_;

    return $self->get('InitialExportUUID') unless defined $ExportUUID;

    $self->set({'InitialExportUUID' => $ExportUUID});
}

#========================================================================================
# function     setMerchantInfo
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setMerchantInfo($hOptions);
#----------------------------------------------------------------------------------------
# description  Configure merchant account information in Zettle environment.
#----------------------------------------------------------------------------------------
# input        $hOptions | options setup info with keys:
#                          <ul>
#                              <li>uuid - organization UUID - string
#                              <li>email - zettle login email - string
#                              <li>country - ISO3166-1 code (i.e: GB) - char(2)
#                              <li>currency - ISO4217 code (i.e: GBP) - char(3)
#                              <li>language - ISO639-1 code (i.e: en) - char(2)
#                              <li>taxationType - zettle vat model (NONE, VAT, SALES_TAX) - string
#                              <li>taxationMode - zettle price structure (INCLUSIVE, EXCLUSIVE) - string
#                          </ul> | ref.hash
#========================================================================================
sub setMerchantInfo {
    my $self = shift;
    my %hOptions = @_;

    my $UUID = $hOptions{'uuid'};
    my $Email = $hOptions{'email'};
    my $CountryCode = $hOptions{'country'};
    my $LanguageCode = $hOptions{'language'};
    my $CurrencyCode = $hOptions{'currency'};
    my $TaxationType = $hOptions{'taxationType'};
    my $TaxationMode = $hOptions{'taxationMode'};
    
    # Check shop currency matches zettle currency. If currencies don't match we'll throw an error and integration won't be setup
    my $Shop = $self->getSite();
    my $CurrencyID = $Shop->get('CurrencyID');
    Error('CURRENCY_MISMATCH', {
        'ShopCurrency' => $CurrencyID,
        'ZettleCurrency' => $CurrencyCode
    }) if $CurrencyID ne $CurrencyCode;

    $self->set({
        'ZettleUUID'         => $UUID,
        'ZettleEmail'        => $Email,
        'ZettleCountryCode'  => $CountryCode,
        'ZettleCurrencyCode' => $CurrencyCode,
        'ZettleLanguageCode' => $LanguageCode,
        'ZettleTaxationType' => $TaxationType,
        'ZettleTaxationMode' => $TaxationMode,
    });
}

#========================================================================================
# function     getMerchantUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $OrganizationUUID = $self->getMerchantUUID();
#----------------------------------------------------------------------------------------
# description  Retrieve organization UUID.
#----------------------------------------------------------------------------------------
# return       $OrganizationUUID | Organization UUID associated to merchant in Zettle site  | string
#========================================================================================
sub getMerchantUUID {
    return shift->get('ZettleUUID');
}

#========================================================================================
# function     getMerchantEmail
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Email = $self->getMerchantEmail();
#----------------------------------------------------------------------------------------
# description  Get Zettle merchant email.
#----------------------------------------------------------------------------------------
# return       $Email | Email merchant uses to login to Zettle site  | string
#========================================================================================
sub getMerchantEmail {
    return shift->get('ZettleEmail');
}

#========================================================================================
# function     getMerchantCountry
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CountryCode = $self->getMerchantCountry();
#----------------------------------------------------------------------------------------
# description  Retrieve merchant country.
#----------------------------------------------------------------------------------------
# return       $CountryCode | ISO3166-1 code for country setup by merchant in Zettle site (i.e: GB)  | char(2)
#========================================================================================
sub getMerchantCountry {
    return shift->get('ZettleCountryCode');
}

#========================================================================================
# function     getMerchantCurrency
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CurrencyCode = $self->getMerchantCurrency();
#----------------------------------------------------------------------------------------
# description  Retrieve merchant currency.
#----------------------------------------------------------------------------------------
# return       $CurrencyCode | ISO4217 code for currency setup by merchant in Zettle site (i.e: GBP)  | char(3)
#========================================================================================
sub getMerchantCurrency {
    return shift->get('ZettleCurrencyCode');
}

#========================================================================================
# function     getMerchantLanguage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LanguageCode = $self->getMerchantLanguage();
#----------------------------------------------------------------------------------------
# description  Retrieve merchant language.
#----------------------------------------------------------------------------------------
# return       $LanguageCode | ISO639-1 code for language setup by merchant in Zettle site (i.e: en)  | char(2)
#========================================================================================
sub getMerchantLanguage {
    return shift->get('ZettleLanguageCode');
}

#========================================================================================
# function     markForDeletion
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->markForDeletion($Delete);
#----------------------------------------------------------------------------------------
# description  Mark current environment for deletion.
#----------------------------------------------------------------------------------------
# input       $Delete | TRUE to mark environment to be deleted by batch job, false otherwise  | boolean
#========================================================================================
sub markForDeletion {
    my $self = shift;
    my ($Delete) = @_;

    if($Delete){
        $self->set({'MarkedForDeletion', GetCurrentDateTime()});
    }else{
        $self->set({'MarkedForDeletion', undef});        
    }
}

#========================================================================================
# function     setupInventorySubscription
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupInventorySubscription($InventorySubscriptionUUID, $InventorySubscriptionSigningKey);
#----------------------------------------------------------------------------------------
# description  Configure inventory subscription with data received from Zettle after creating event subscription.
#----------------------------------------------------------------------------------------
# input        $InventorySubscriptionUUID | Inventory subscription UUIDv1 | string
# input        $InventorySubscriptionSigningKey | Inventory subscription signing key | string
#========================================================================================
sub setupInventorySubscription {
    my $self = shift;
    my ($InventorySubscriptionUUID, $InventorySubscriptionSigningKey) = @_;

    $self->set({
        'InventorySubscriptionUUID' => $InventorySubscriptionUUID,
        'InventorySubscriptionSigningKey' => $InventorySubscriptionSigningKey,
    });
}

#========================================================================================
# function     getInventorySubscriptionUUID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $InventorySubscriptionUUID = $self->getInventorySubscriptionUUID();
#----------------------------------------------------------------------------------------
# description  Retrieve inventory subscription UUID.
#----------------------------------------------------------------------------------------
# return       $InventorySubscriptionUUID | Subscription UUID | string
#========================================================================================
sub getInventorySubscriptionUUID {
    return shift->get('InventorySubscriptionUUID');
}

#========================================================================================
# function     getInventorySubscriptionSigningKey
# state        public
#----------------------------------------------------------------------------------------
# syntax       $InventorySubscriptionSigningKey = $self->getInventorySubscriptionSigningKey();
#----------------------------------------------------------------------------------------
# description  Retrieve inventory subscription signing key.
#----------------------------------------------------------------------------------------
# return       $InventorySubscriptionSigningKey | Subscription signing key | string
#========================================================================================
sub getInventorySubscriptionSigningKey {
    return shift->get('InventorySubscriptionSigningKey');
}

#========================================================================================
# function     locationBinUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationBinUUID = $self->locationBinUuid();
# syntax       $self->locationBinUuid($LocationBinUUID);
#----------------------------------------------------------------------------------------
# description  Get or set BIN location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationBinUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationBinUUID | UUID for BIN location in Zettle site, if $LocationBinUUID argument is not present  | string
#========================================================================================
sub locationBinUuid {
    my $self = shift;
    my ($LocationBinUUID) = @_;
    
    return $self->get('LocationBinUUID') unless defined $LocationBinUUID;

    $self->set({'LocationBinUUID' => $LocationBinUUID});
}

#========================================================================================
# function     locationSoldUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationSoldUUID = $self->locationSoldUuid();
# syntax       $self->locationSoldUuid($LocationSoldUUID);
#----------------------------------------------------------------------------------------
# description  Get or set SOLD location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationSoldUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationSoldUUID | UUID for SOLD location in Zettle site, if $LocationSoldUUID argument is not present  | string
#========================================================================================
sub locationSoldUuid {
    my $self = shift;
    my ($LocationSoldUUID) = @_;
    
    return $self->get('LocationSoldUUID') unless defined $LocationSoldUUID;

    $self->set({'LocationSoldUUID' => $LocationSoldUUID});
}

#========================================================================================
# function     locationStoreUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationStoreUUID = $self->locationStoreUuid();
# syntax       $self->locationStoreUuid($LocationStoreUUID);
#----------------------------------------------------------------------------------------
# description  Get or set STORE location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationStoreUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationStoreUUID | UUID for STORE location in Zettle site, if $LocationStoreUUID argument is not present  | string
#========================================================================================
sub locationStoreUuid {
    my $self = shift;
    my ($LocationStoreUUID) = @_;
    
    return $self->get('LocationStoreUUID') unless defined $LocationStoreUUID;

    $self->set({'LocationStoreUUID' => $LocationStoreUUID});
}

#========================================================================================
# function     locationSupplierUuid
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocationSupplierUUID = $self->locationSupplierUuid();
# syntax       $self->locationSupplierUuid($LocationSupplierUUID);
#----------------------------------------------------------------------------------------
# description  Get or set SUPPLIER location UUID.
#----------------------------------------------------------------------------------------
# input        $LocationSupplierUUID | optional. Location UUID | string
#----------------------------------------------------------------------------------------
# return       $LocationSupplierUUID | UUID for SUPPLIER location in Zettle site, if $LocationSupplierUUID argument is not present  | string
#========================================================================================
sub locationSupplierUuid {
    my $self = shift;
    my ($LocationSupplierUUID) = @_;
    
    return $self->get('LocationSupplierUUID') unless defined $LocationSupplierUUID;

    $self->set({'LocationSupplierUUID' => $LocationSupplierUUID});
}

#========================================================================================
# function     taxationType
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TaxationType = $self->taxationType();
# syntax       $self->taxationType($TaxationType);
#----------------------------------------------------------------------------------------
# description  Get or set Zettle merchant taxation type: NONE, VAT, SALES_TAX.
#               See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.4.1.1. Taxation Types.
#----------------------------------------------------------------------------------------
# input        $TaxationType | optional. Tax model used by merchant in Zettle | string
#----------------------------------------------------------------------------------------
# return       $TaxationType | Tax model for merchant in Zettle site, if $TaxationType argument is not present  | string
#========================================================================================
sub taxationType {
    my $self = shift;
    my ($TaxationType) = @_;
    
    return $self->get('ZettleTaxationType') unless defined $TaxationType;

    $self->set({'ZettleTaxationType' => $TaxationType});
}

#========================================================================================
# function     taxationMode
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TaxationMode = $self->taxationMode();
# syntax       $self->taxationMode($TaxationMode);
#----------------------------------------------------------------------------------------
# description  Get or set Zettle merchant taxation type: NONE, VAT, SALES_TAX.
#               See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.4.1.2. Price Types.
#----------------------------------------------------------------------------------------
# input        $TaxationMode | optional. Tax model used by merchant in Zettle | string
#----------------------------------------------------------------------------------------
# return       $TaxationMode | Tax model for merchant in Zettle site, if $TaxationMode argument is not present  | string
#========================================================================================
sub taxationMode {
    my $self = shift;
    my ($TaxationMode) = @_;
    
    return $self->get('ZettleTaxationMode') unless defined $TaxationMode;

    $self->set({'ZettleTaxationMode' => $TaxationMode});
}

#========================================================================================
# function     exportProductPrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ExportProductPrices = $self->exportProductPrices();
# syntax       $self->exportProductPrices($ExportProductPrices);
#----------------------------------------------------------------------------------------
# description  Get or set ExportProductPrices.
#----------------------------------------------------------------------------------------
# input        $ExportProductPrices | optional. TRUE if prices are to be exported during scheduled export | boolean
#----------------------------------------------------------------------------------------
# return       $ExportProductPrices | Export product prices value , if $ExportProductPrices argument is not present  | boolean
#========================================================================================
sub exportProductPrices {
    my $self = shift;
    my ($ExportProductPrices) = @_;
    
    return $self->get('ExportProductPrices') unless defined $ExportProductPrices;

    $self->set({'ExportProductPrices' => $ExportProductPrices});
}

#========================================================================================
# function     getDeletedObjects
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ahDeletedObjects = $self->getDeletedObjects($DeletedAfter);
# syntax       $ahDeletedObjects = $self->getDeletedObjects($DeletedAfter, $ClassName);
#----------------------------------------------------------------------------------------
# description  Returns information about objects of a given type in the site this environment belongs to
#               that have been deleted after a given date.
#               For example: products deleted from the shop during the last week.
#----------------------------------------------------------------------------------------
# input        $DeletedAfter | date and time | DateTime
# input        $ClassName | class name (optional - default: 'Product') | string
#----------------------------------------------------------------------------------------
# return       $Iterator | iterator over hash references with following keys:
#               <ul>
#                   <li>GUID          - GUID of the deleted object (primary key)- varchar(36)
#                   <li>PathFromSite  - Object Path relative to the object site - varchar(1000)
#                   <li>ZettleUUID    - Zettle UUID of the deleted object - varchar(36)
#                   <li>DeletedOn     - Date and time of the deletion on the object - datetime
#               </ul> | DE_EPAGES::Core::API::Iterator
#========================================================================================
sub getDeletedObjects {
    my $self = shift;
    my ($DeletedAfter, $ClassName) = @_;

    $ClassName //= 'Product';
    
    TestObject('DeletedAfter', $DeletedAfter, 'DateTime');

    my $SiteID = $self->getSite()->id;
    my $ClassID = LoadClassByAlias($ClassName)->id;
    $DeletedAfter = AddDuration($DeletedAfter, 'seconds', -1); # subtract 1 second
    return DAL_GetDeletedObjects($SiteID, $ClassID, $DeletedAfter);
}

#========================================================================================
# function     requiresCleanup
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DoCleanup = $self->requiresCleanup();
#----------------------------------------------------------------------------------------
# description  Determines whether current environment needs to be cleaned up.
#               When merchant disconnects from Zettle the environment datetime attribute 'MarkedForDeletion' is set.
#               This flag is used to identify which environments need cleanup.
#----------------------------------------------------------------------------------------
# return       $DoCleanup | TRUE if environment needs to be cleaned up, FALSE otherwise | boolean
#========================================================================================
sub requiresCleanup {
    my $self = shift;

    return defined $self->get('MarkedForDeletion');
}

#========================================================================================
# function     isScheduledExportPending
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isScheduledExportPending();
#----------------------------------------------------------------------------------------
# description  Determine whether scheduled export shall be executed for this environment.
#               Export process shall be run when:
#                   date(LasProductExport)+ScheduledExportInterval >= date(NOW)
#----------------------------------------------------------------------------------------
# return       $IsPending | TRUE if scheduled export is pending, FALSE otherwise | boolean
#========================================================================================
sub isScheduledExportPending {
    my $self = shift;

    # we only take into account date part ignoring time part to avoid errors when handling hours/minutes/seconds
    # time is not relevant for required calculations
    my $Now = GetCurrentDateTime();
    my $LastProductExport = $self->lastProductExport();
    my $ExportScheduleInterval = $self->exportScheduleInterval();

    return 0 unless defined $LastProductExport;
    
    # Calculate absolute days difference (discarding time difference) between current datetime and LastProductExport
    # See https://metacpan.org/pod/DateTime#$dt-%3Edelta_days($datetime)
    my $DeltaDaysDuration = $Now->delta_days($LastProductExport);
    my $DeltaDaysAbsolute = $DeltaDaysDuration->in_units('days');

    return $DeltaDaysAbsolute >= $ExportScheduleInterval;
}

#========================================================================================
# function     exportProductUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportProductUpdates();
# syntax       $self->exportProductUpdates($hOptions);
#----------------------------------------------------------------------------------------
# description  Export product updates to Zettle for products whose attributes of interest
#               to Zettle were modified after previous product export.
#----------------------------------------------------------------------------------------
# input        $hOptions | optional. Hash with the following keys
#                           <ul>
#                               <li>IncludePrices - TRUE to include prices in export (default: ExportProductPrices value) - boolean
#                           </ul> | ref.hash
#========================================================================================
sub exportProductUpdates {
    my $self = shift;
    my ($hOptions) = @_;

    $self->set({'ProductIsExporting' => 1});

    $hOptions //= {};
    my $IncludePrices = $hOptions->{'IncludePrices'}
                            // $self->get('ExportProductPrices');

    eval{
        my $ZettleService = DE_EPAGES::Zettle::API::ZettleService->new($self);
        $ZettleService->exportProductUpdates($IncludePrices);
    };
    if(ExistsError()){
        my $Error = GetError();
        LogError($Error);
    }
    
    $self->set({'ProductIsExporting' => 0});
}

#========================================================================================
# function     exportPrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportPrices();
#               $self->exportPrices($ExportAllPrices);
#----------------------------------------------------------------------------------------
# description  Export product prices to Zettle:
#               <ul>
#                   <li>if $ExportAllPrices is TRUE, export prices for all exportable products (see <function GetExportableProducts>).
#                   <li>if $ExportAllPrices is FALSE or undef, only export prices modified after last price export.
#               </ul>
#----------------------------------------------------------------------------------------
# input        $ExportAllPrices   | Determines whether to export prices for all products or only modified prices | boolean
#========================================================================================
sub exportPrices {
    my $self = shift;
    my ($ExportAllPrices) = @_;

    $self->set({'PriceIsExporting' => 1});

    eval{
        my $ZettleService = DE_EPAGES::Zettle::API::ZettleService->new($self);
        $ZettleService->exportProductPrices($ExportAllPrices);
    };
    if(ExistsError()){
        my $Error = GetError();
        LogError($Error);
    }
    
    $self->set({'PriceIsExporting' => 0});
}

1;
