#========================================================================================
# package      DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory
# state        public
#----------------------------------------------------------------------------------------
# description  Factory for creating different DTO objects.
#========================================================================================
package DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory;

use strict;

use DE_EPAGES::Core::API::Error qw( Error );
use DE_EPAGES::Core::API::TestParameter qw( TestObject );
use DE_EPAGES::Object::API::Table::Country qw ( GetPKeyCountryByCode2 );
use DE_EPAGES::Shop::API::Constants qw( TAX_MODEL_GROSS TAX_MODEL_NET );
use DE_EPAGES::Zettle::API::Constants qw (
    ZETTLE_ENVIRONMENT_CLASS
    ZETTLE_TAXATION_TYPE_NONE
    ZETTLE_TAXATION_MODE_INCLUSIVE
    ZETTLE_TAXATION_MODE_EXCLUSIVE
);
use DE_EPAGES::Zettle::API::Tools qw ( GenerateUUID );
use DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO;
use DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO;
use DE_EPAGES::Zettle::API::WebService::VariationOptionHelper qw ( BuildOptionDefinitions BuildVariantOptions );

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'NoProductUuidError' => 'ZettleUUID missing for product ID=#ProductId',
        'NoProductPriceError' => 'No list price available for product ID=#ProductId',
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory->new($Shop);
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# input        $Shop | shop object | DE_EPAGES::Shop::API::Object::Shop
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::WebService::DTO::DTOFactory
#========================================================================================
sub new {
    my $class = shift;
    my ($Shop) = @_;
    TestObject('Shop', $Shop, 'DE_EPAGES::Shop::API::Object::Shop');

    my $self = bless {
        'Shop' => $Shop,
        'Environment' => $Shop->get(ZETTLE_ENVIRONMENT_CLASS),
    }, $class;
    
    return $self;
}

#========================================================================================
# function     buildProductDTO
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->buildProductDTO($Product, $IncludePrice, $hImageUploadResults);
#----------------------------------------------------------------------------------------
# description  build product DTO required for Zettle updates based on the given product
#----------------------------------------------------------------------------------------
# input        $Product | a product object | DE_EPAGES::Product::API::Object::Product
# input        $IncludePrice | determines whether price information will be included (optional) | boolean
# input        $hImageUploadResults | a map of ImageUploadResult objects indexed by product GUID
#                                      Image lookupKeys for products to be exported will be searched here.
#                                      See <DE_EPAGES::Zettle::API::Object::ImageUploadResult> | ref.hash.object
#========================================================================================
sub buildProductDTO {
    my $self = shift;
    my ($Product, $IncludePrice, $hImageUploadResults) = @_;
    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');

    my $Shop = $self->_getSite();
    my $LanguageID = $Shop->get('LanguageID');
    my $Environment = $self->_getEnvironment();
    my $ZettleCurrencyCode = $Environment->getMerchantCurrency();
    my $ZettleCountryCode = $Environment->getMerchantCountry();
    
    my $ProductDTO = DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO->new();

    # Construct ProductDTO fields from superproduct data if this is a subproduct
    my $SuperProduct = (defined $Product->superProduct()) ? $Product->superProduct() : $Product;

    my $Uuid = $self->_getProductUUID($SuperProduct);
    $ProductDTO->uuid($Uuid);

    my $ProductName = $SuperProduct->get('Name', $LanguageID);
    $ProductDTO->name($ProductName);

    my $ProductDescription = $SuperProduct->get('Description', $LanguageID);
    $ProductDTO->description($ProductDescription);

    my $ProductGuid = $SuperProduct->guid();

    # Only include imageLookupKeys info if image was updated (meaning modified or removed)
    if(exists $hImageUploadResults->{$ProductGuid}){
        my $aImageUploadResult = $hImageUploadResults->{$ProductGuid};
        my @ProductImageKeys = ();
        foreach my $ImageUploadResult ( @{$aImageUploadResult} ){
            my $ProductImageLookupKey = $ImageUploadResult->lookupKey();
            push(@ProductImageKeys, $ProductImageLookupKey);
        }
        # if $ProductImageKeys is an empty array, image will be removed from product in zettle (if previous image present)
        $ProductDTO->imageLookupKeys(\@ProductImageKeys);
    }

    my $ProductVariations = $self->_buildProductVariationDTOs($Product, $IncludePrice, $LanguageID, $ZettleCurrencyCode);
    $ProductDTO->variations($ProductVariations);

    $ProductDTO->externalReference($ProductGuid);

    my $OrderUnit = $Product->get('OrderUnit');
    my $Unit;
    $Unit = $OrderUnit->alias if defined $OrderUnit && $OrderUnit->alias ne 'piece';
    $ProductDTO->unitName($Unit) if defined $Unit;

    # Only include VAT percentage if Zettle merchant is setup to use taxes
    my $VatPercentage = $self->_getVatPercentage($SuperProduct, $ZettleCountryCode);
    $ProductDTO->vatPercentage($VatPercentage) if defined $VatPercentage;

    my $VariantOptionDefinitions = BuildOptionDefinitions($SuperProduct);
    $ProductDTO->variantOptionDefinitions($VariantOptionDefinitions);

    # NOTE: for unused fields see DE_EPAGES/Zettle/Documents/API/ZettleAPI-Extracts.md, section 1.2.3.3. Unused ProductDTO fields

    return $ProductDTO;
}

#========================================================================================
# function     buildProductPriceDTO
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductDTO = $self->buildProductPriceDTO($Product);
#----------------------------------------------------------------------------------------
# description  Build product DTO required for Zettle updates based on the given product/subproduct.
#               Generated DTO will only include the necessary information to export product price.
#               For a master product it will create a fake Zettle variation since prices are handled at variation level.
#----------------------------------------------------------------------------------------
# input        $Product | a (sub)product | DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $ProductDTO | a ProductDTO object which only includes price information | DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO
#========================================================================================
sub buildProductPriceDTO {
    my $self = shift;
    my ($Product) = @_;
    TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');

    my $Shop = $self->_getSite();
    my $LanguageID = $Shop->get('LanguageID');
    my $Environment = $self->_getEnvironment();
    my $ZettleCurrencyCode = $Environment->getMerchantCurrency();
    my $ZettleCountryCode = $Environment->getMerchantCountry();
    
    my $ProductDTO = DE_EPAGES::Zettle::API::WebService::DTO::ProductDTO->new();

    # Construct ProductDTO fields from superproduct data if this is a subproduct
    my $SuperProduct = (defined $Product->superProduct()) ? $Product->superProduct() : $Product;
    
    my $Uuid = $self->_getProductUUID($SuperProduct);
    $ProductDTO->uuid($Uuid);

    my $ProductName = $SuperProduct->get('Name', $LanguageID);
    $ProductDTO->name($ProductName);

    my $aVariationDTO = $self->_buildProductPriceVariationDTOs($Product, $ZettleCurrencyCode);
    $ProductDTO->variations($aVariationDTO);
    
    # Only include VAT percentage if Zettle merchant is setup to use taxes
    my $VatPercentage = $self->_getVatPercentage($SuperProduct, $ZettleCountryCode);
    $ProductDTO->vatPercentage($VatPercentage) if defined $VatPercentage;

    return $ProductDTO;
}

sub buildPatchVariationDTO {
    my $self = shift;
    my ($SubProduct) = @_;

    my $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

    my $Uuid = $self->_getProductUUID($SubProduct);
    Error('NoProductUuidError', {'ProductId' => $SubProduct->id()}) unless defined $Uuid;
    $VariationDTO->uuid($Uuid);

    return $VariationDTO;
}


#========================================================================================


#========================================================================================
# function     _buildProductVariationDTOs
#----------------------------------------------------------------------------------------
# syntax       $aZettleVariations = $self->_buildProductVariationDTOs($Product, $IncludePrice, $LanguageID, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Create a VariationDTO object from the given product.
#----------------------------------------------------------------------------------------
# input        $Product  | a superproduct (either normal or master product) | DE_EPAGES::Product::API::Product
# input        $IncludePrice | determines whether price information will be included (optional) | boolean
# input        $LanguageID  | language ID (used to retrieve localized attributes)  | int
# input        $CurrencyID  | currency ID (ISO 4217: https://publications.europa.eu/code/en/en-5000700.htm) | char(3)
#----------------------------------------------------------------------------------------
# return       $aZettleVariations | a list of VariationDTO objects | ref.DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO
#========================================================================================
sub _buildProductVariationDTOs {
    my $self = shift;
    my ($Product, $IncludePrice, $LanguageID, $CurrencyID) = @_;

    my @aZettleVariations = ();

    my $aSubProducts = $Product->get('SubProducts');
    if(scalar @{$aSubProducts} > 0){
        foreach my $SubProduct (@$aSubProducts) {
            my $VariationDTO = $self->_buildVariationDTO($SubProduct, $IncludePrice, $LanguageID, $CurrencyID);
            push(@aZettleVariations, $VariationDTO);
        }
    }else{
        # Create a fake variation for product without variations (Zettle handles stock at variation level at the moment)
        my $FakeVariationDTO = $self->_buildVariationDTO($Product, $IncludePrice, $LanguageID, $CurrencyID);
        push(@aZettleVariations, $FakeVariationDTO);
    }

    return \@aZettleVariations;
}

#========================================================================================
# function     _buildVariationDTO
#----------------------------------------------------------------------------------------
# syntax       $VariationDTO = $self->_buildVariationDTO($Product, $IncludePrice, $LanguageID, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Create a VariationDTO object from the given product.
#----------------------------------------------------------------------------------------
# input        $Product  | a subproduct (or master product) | DE_EPAGES::Product::API::Product
# input        $IncludePrice | determines whether price information will be included (optional) | boolean
# input        $LanguageID  | language ID (used to retrieve localized attributes)  | int
# input        $CurrencyID  | currency ID (ISO 4217: https://publications.europa.eu/code/en/en-5000700.htm) | char(3)
#----------------------------------------------------------------------------------------
# note         $Product can be a master product in which case fake variation options will be used.
#               This is due to the fact that Zettle only handles stock levels in variations at the moment.
#----------------------------------------------------------------------------------------
# return       $VariationDTO | a VariationDTO object | DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO
#========================================================================================
sub _buildVariationDTO {
    my $self = shift;
    my ($Product, $IncludePrice, $LanguageID, $CurrencyID) = @_;

    my $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

    my $hLocAttributes = $Product->get(['Name', 'Description'], $LanguageID);
    my $hAttributes = $Product->get(['ManufacturerSKU', 'UPCEAN']);

    my $Uuid = $self->_getProductUUID($Product);
    $VariationDTO->uuid($Uuid);

    # We fill the name field for the sake of completeness but when Zettle processes variation it assigns a new name based on variation options values (i.e: blue, M, Cotton)
    my $Name = $hLocAttributes->{'Name'};
    $VariationDTO->name($Name) if $Name;

    my $Description = $hLocAttributes->{'Description'};
    $VariationDTO->description($Description) if $Description;

    my $Sku = $hAttributes->{'ManufacturerSKU'};
    $VariationDTO->sku($Sku) if $Sku;

    my $Barcode = $hAttributes->{'UPCEAN'};
    $VariationDTO->barcode($Barcode) if $Barcode;

    # NOTE: although imageLookupKeys can be added to variations, they are not displayed in Zettle, and they advised not to upload variation images

    if($IncludePrice){
        my $Price = $self->_getProductPrice($Product, $CurrencyID);
        # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
        $VariationDTO->price($Price*100) if defined $Price;
        $VariationDTO->currencyCode($CurrencyID);
    }

    my $Options = BuildVariantOptions($Product);
    $VariationDTO->options($Options);

    return $VariationDTO;
}

#========================================================================================
# function     _buildProductPriceVariationDTOs
#----------------------------------------------------------------------------------------
# syntax       $aZettleVariations = $self->_buildProductPriceVariationDTOs($Product, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Create a list of VariationDTOs that will include only UUID and price information for $Product.
#----------------------------------------------------------------------------------------
# input        $Product  | a subproduct (or master product) | DE_EPAGES::Product::API::Product
# input        $CurrencyID  | currency ID (ISO 4217: https://publications.europa.eu/code/en/en-5000700.htm) | char(3)
#----------------------------------------------------------------------------------------
# return       $aZettleVariations | a list of VariationDTO object | ref.DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO
#----------------------------------------------------------------------------------------
# throws NoProductUuidError
# throws NoProductPriceError
#========================================================================================
sub _buildProductPriceVariationDTOs {
    my $self = shift;
    my ($Product, $CurrencyID) = @_;

    my @aZettleVariations = ();

    my ($VariationDTO, $Uuid, $Price);
    
    my $aSubProducts = $Product->get('SubProducts');
    if(scalar @{$aSubProducts} > 0){
        foreach my $SubProduct (@$aSubProducts) {
            $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

            $Uuid = $self->_getProductUUID($SubProduct);
            $VariationDTO->uuid($Uuid);
            
            $Price = $self->_getProductPrice($SubProduct, $CurrencyID);
            next unless defined $Price;
            # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
            $VariationDTO->price($Price*100);
            $VariationDTO->currencyCode($CurrencyID);

            my $SubProductOptions = BuildVariantOptions($SubProduct);
            $VariationDTO->options($SubProductOptions);
            
            push(@aZettleVariations, $VariationDTO);
        }
    }else{
        $VariationDTO = DE_EPAGES::Zettle::API::WebService::DTO::VariationDTO->new();

        $Uuid = $self->_getProductUUID($Product);
        Error('NoProductUuidError', {'ProductId' => $Product->id()}) unless defined $Uuid;
        $VariationDTO->uuid($Uuid);
        
        $Price = $self->_getProductPrice($Product, $CurrencyID);
        Error('NoProductPriceError', {'ProductId' => $Product->id()}) unless defined $Price;

        if(defined $Price){
            # NOTE: Price in cents. See DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.5.1.4. Product prices
            $VariationDTO->price($Price*100);
            $VariationDTO->currencyCode($CurrencyID);
            
            push(@aZettleVariations, $VariationDTO);
        }

        my $ProductOptions = BuildVariantOptions($Product);
        $VariationDTO->options($ProductOptions);
    }

    return \@aZettleVariations;
}

#========================================================================================
# function     _getProductPrice
#----------------------------------------------------------------------------------------
# syntax       $ProductPrice = $self->_getProductPrice($Product, $CurrencyID);
#----------------------------------------------------------------------------------------
# description  Returns the price for $Product in currency given in $CurrencyID.
#               Price will include taxes if configured in Environment, that is:
#               TaxationType=INCLUSIVE and TaxationMode=VAT
#----------------------------------------------------------------------------------------
# input        $Product     | a product | DE_EPAGES::Product::API::Product
# input        $CurrencyID  | currency ID (ISO 4217: https://publications.europa.eu/code/en/en-5000700.htm) | char(3)
#----------------------------------------------------------------------------------------
# return       $ProductPrice | Price of the product with(out) taxes | float
#========================================================================================

sub _getProductPrice {
    my $self = shift;
    my ($Product, $CurrencyID) = @_;

    my $Environment = $self->_getEnvironment();
    my $TaxModel;
    if($Environment->taxationMode eq ZETTLE_TAXATION_MODE_INCLUSIVE){
        # For TAXATION_TYPE=NONE we don't include taxes
        # For VAT or SALES_TAX we include taxes
        $TaxModel = ($Environment->taxationType eq ZETTLE_TAXATION_TYPE_NONE) ? TAX_MODEL_NET : TAX_MODEL_GROSS;
    }else{
        # For TAXATION_MODE=EXCLUSIVE we don't include taxes
        $TaxModel = TAX_MODEL_NET;
    }

    my $DisplayPrice = $Product->displayPrice({
        CurrencyID => $CurrencyID,
        TaxModel => $TaxModel,
    });

    return $DisplayPrice;
}

#========================================================================================
# function     _getVatPercentage
#----------------------------------------------------------------------------------------
# syntax       $VatPercentage = $self->_getVatPercentage($Product, $Country);
#----------------------------------------------------------------------------------------
# description  Returns VAT percentage for $Product applied in $Country.
#----------------------------------------------------------------------------------------
# input        $Product  | a product | DE_EPAGES::Product::API::Product
# input        $Country  | country iso code alpha 2 | char(2)
#----------------------------------------------------------------------------------------
# return       $VatPercentage | VAT applied to $Product in $Country (min: 0, max: 100).
#                                return undef if Environment is configured to exclude taxes | int
#========================================================================================
sub _getVatPercentage {
    my $self = shift;
    my ($Product, $Country) = @_;

    my $Environment = $self->_getEnvironment();
    return undef if $Environment->taxationMode eq ZETTLE_TAXATION_MODE_EXCLUSIVE;

    my $CountryID = GetPKeyCountryByCode2($Country);
    my $Shop = $Product->getSite();
    my $TaxMatrix = $Shop->get('TaxMatrix');
    my $TaxClass = $Product->get('TaxClass')
                    // $TaxMatrix->defaultTaxClass;
    my $TaxArea = $TaxMatrix->taxAreaByCountry($CountryID)
                    // $TaxMatrix->defaultTaxArea;
    my $TaxRate = $TaxMatrix->taxRate($TaxArea,  $TaxClass);
    # Minimum 0, maximum 100
    my $VatPercentage = $TaxRate*100;

    return $VatPercentage;
}

#========================================================================================
# function     _getProductUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Uuid = $self->_getProductUUID($Product);
#----------------------------------------------------------------------------------------
# description  Retrieve Zettle UUID for $Product. If UUID is empty it will be generated and updated.
#----------------------------------------------------------------------------------------
# input        $Product | a (sub) product | DE_EPAGES::Product::API::Object::Product
#----------------------------------------------------------------------------------------
# return       $Uuid | Product UUID used for Zettle | string
#========================================================================================
sub _getProductUUID {
    my $self = shift;
    my ($Product) = @_;

    my $Uuid = $Product->get('ZettleUUID');

    # If product has not been assigned a UUID, generate it and update product attribute
    if(not defined $Uuid){
        $Uuid = GenerateUUID();
        $Product->set({'ZettleUUID' => $Uuid});
    }

    return $Uuid;
}

sub _getSite {
    return shift->{'Shop'};
}

sub _getEnvironment {
    return shift->{'Environment'};
}

1;
