#========================================================================================
# package      DE_EPAGES::Zettle::API::ZettleService
# state        public
#----------------------------------------------------------------------------------------
# description  object for calling functions on the ZettleService which serves as a Zettle facade
#========================================================================================
package DE_EPAGES::Zettle::API::ZettleService;

use strict;

use Data::Compare;
use JSON qw ( decode_json );

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime unique );
use DE_EPAGES::Core::API::TestParameter  qw( TestObject );
use DE_EPAGES::Object::API::Factory  qw ( ExistsObjectByGUID ExistsObjectByPath LoadObjectByGUID LoadObjectByPath );
use DE_EPAGES::Zettle::API::Constants qw ( %INITIAL_EXPORT_STATUS WAIT_INTERVAL_AFTER_DELETE );
use DE_EPAGES::Zettle::API::ImageHelper qw ( GetGuidImageMap GetImageUploadResults );
use DE_EPAGES::Zettle::API::ProductChangeHelper qw (
    GetModifiedImageProducts
    GetModifiedBasicDataProducts
    GetModifiedPriceProducts
);
use DE_EPAGES::Zettle::API::TokenHelper;
use DE_EPAGES::Zettle::API::Tools qw ( GetExportableProducts LoadProductByZettleUUID );
use DE_EPAGES::Zettle::API::WebService::AuthService;
use DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO;
use DE_EPAGES::Zettle::API::WebService::DTO::WebhookEventDTO;
use DE_EPAGES::Zettle::API::WebService::ImageService;
use DE_EPAGES::Zettle::API::WebService::InventoryService;
use DE_EPAGES::Zettle::API::WebService::LocationService;
use DE_EPAGES::Zettle::API::WebService::MerchantService;
use DE_EPAGES::Zettle::API::WebService::PusherService;
use DE_EPAGES::Zettle::API::WebService::ProductService;
use DE_EPAGES::Zettle::API::WebService::VariationOptionHelper qw ( BuildVariantOptions );


#ERRORDEFINITION
sub ErrorDefinition { return {}
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Zettle::API::ZettleService->new($Environment)
#----------------------------------------------------------------------------------------
# description  creates a new ZettleService object
#----------------------------------------------------------------------------------------
# input        $Environment | the Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::Zettle::API::ZettleService
#========================================================================================
sub new {
    my $class = shift;
    my ($Environment)  = @_;
    TestObject('ZettleEnvironment', $Environment, 'DE_EPAGES::Zettle::API::Object::ZettleEnvironment');
    
    my $self  = bless {
        'Environment' => $Environment
    }, $class;

    return $self;
}

#========================================================================================
# function     getEnvironment
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Environment = $self->getEnvironment()
#----------------------------------------------------------------------------------------
# description  Returns the Zettle Environment object.
#----------------------------------------------------------------------------------------
# return       $Environment | Zettle Environment | Object
#========================================================================================
sub getEnvironment {
    return shift->{'Environment'};
}

#========================================================================================
# function     connect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->connect($AuthCode)
#----------------------------------------------------------------------------------------
# description  Use given $AuthCode to retrieve tokens from Zettle OAuth API and create/update zettle environment storing token data.
#               See also <function DE_EPAGES::Zettle::API::WebService::AuthService::getTokensOnAuthCode>.
#========================================================================================
sub connect {
    my $self = shift;
    my ($AuthCode) = @_;

    my $Environment = $self->getEnvironment();

    # Use OAuth service to exchange AuthCode for Access and Refresh tokens
    my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
    my $hTokens = $AuthService->getTokensOnAuthCode($AuthCode);
    
    my $TokenHelper = $self->_createTokenHelper($Environment);
    $TokenHelper->updateTokenData($hTokens);

    my $MerchantService = DE_EPAGES::Zettle::API::WebService::MerchantService->new($Environment);
    # See example in: DE_EPAGES/Zettle/Documents/API/ZettlePartnerDocumentation-Extracts.md, section 4.4.2. Retrieve a Merchant's Account Information)
    my $AccountInfo = $MerchantService->getAccountInfo();
    
    my $ContactEmail = $AccountInfo->{'contactEmail'};
    my $ReceiptEmail = $AccountInfo->{'receiptEmail'};
    my %MerchantInfo = (
        'uuid'     => $AccountInfo->{'uuid'},
        'email'    => $ContactEmail // $ReceiptEmail,
        'country'  => $AccountInfo->{'country'},
        'language' => $AccountInfo->{'language'},
        'currency' => $AccountInfo->{'currency'},
        'taxationType' => $AccountInfo->{'taxationType'},
        'taxationMode' => $AccountInfo->{'taxationMode'},
    );

    eval {
        $Environment->setMerchantInfo(%MerchantInfo);
    };
    # Abort setup if merchant info could not be updated:
    # - disconnect from Zettle (revoke tokens)
    # - delete ZettleEnvironment object
    if( ExistsError() ){
        my $Error = GetError();
        $self->disconnect();
        $Environment->delete();
        Error($Error);
    };
}

#========================================================================================
# function     disconnect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->disconnect($MarkForDeletion)
#----------------------------------------------------------------------------------------
# input        $MarkForDeletion | Optional. Determines whether ZettleEnvironment will be marked to be deleted by batch job (default: TRUE) | boolean
#----------------------------------------------------------------------------------------
# description  Disconnect from Zettle API and remove token data from zettle environment
#========================================================================================
sub disconnect {
    my $self = shift;
    my ($MarkForDeletion) = @_;

    $MarkForDeletion //= 1;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();

    # Delete webhooks
    my $InventorySubscriptionUUID = $Environment->get('InventorySubscriptionUUID');
    my $PusherService = DE_EPAGES::Zettle::API::WebService::PusherService->new($Environment);
    eval {
        $PusherService->deleteSubscription($InventorySubscriptionUUID);
    };
    if(ExistsError()){ LogError(GetError()); }
                    
    # Disable inventory tracking for all tracked shop products
    my @aTrackedProducts = grep { $_->get('ZettleTrackingEnabled') } @{ $Shop->get('Products') };
    my $InventoryService = DE_EPAGES::Zettle::API::WebService::InventoryService->new($Environment);
    eval {
        $self->_disableInventoryTracking($InventoryService, \@aTrackedProducts);
    };
    if(ExistsError()){ LogError(GetError()); }

    # Delete connection
    my $AuthService = DE_EPAGES::Zettle::API::WebService::AuthService->new($Environment);
    eval {
        $AuthService->disconnect();
    };
    if(ExistsError()){ LogError(GetError()); }

    $Environment->initialExportStatus($INITIAL_EXPORT_STATUS{PENDING});

    my $TokenHelper = $self->_getTokenHelper();
    $TokenHelper->deleteTokenData();

    # Configure ZettleEnvironment so that batch job proceeds with deletion afterwards if MarkForDeletion is true
    $Environment->markForDeletion($MarkForDeletion);
}

#========================================================================================
# function     isTokenAcquired
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isTokenAcquired()
#----------------------------------------------------------------------------------------
# description  Check if tokens were retrieved from Zettle
#               See <function DE_EPAGES::Zettle::API::TokenHelper::isTokenAcquired>
#----------------------------------------------------------------------------------------
# return        TRUE if tokens present in ZettleEnvironment, false otherwise
#========================================================================================
sub isTokenAcquired {
    my $self = shift;    
    
    my $TokenHelper = $self->_getTokenHelper();
    return $TokenHelper->isTokenAcquired();
}

#========================================================================================
# function     initialExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->initialExport()
#----------------------------------------------------------------------------------------
# description  Run an initial product export which will send to Zettle all products that match criteria, and update InitialExportStatus in environment.
#               See <function DE_EPAGES::Zettle::API::WebService::ProductService::exportProducts>.
#               See <function DE_EPAGES::Zettle::API::WebService::DTO::Requirements::IsProductExportable>.
#----------------------------------------------------------------------------------------
# input       $ReplaceZettleLibrary | Determines whether all products will be removed from Zettle before export | boolean
# input       $IncludePrices | Determines if prices will be included in product export | boolean
#----------------------------------------------------------------------------------------
# return      Number of items to be processed by Zettle.
#========================================================================================
sub initialExport {
    my $self = shift;
    my ($ReplaceZettleLibrary, $IncludePrices) = @_;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    
    my $MethodName = (caller(0))[3];
    GetLog->debug("$MethodName: Shop=".$Shop->alias.", ReplaceZettleLibrary=$ReplaceZettleLibrary, IncludePrices=$IncludePrices");

    if ($ReplaceZettleLibrary){
        $self->_deleteLibrary();
        # PATCH: we must wait a few seconds before starting initial export process so that Zettle correctly shows all products in dashboard
        sleep(WAIT_INTERVAL_AFTER_DELETE);
    }else{
        # [STEP] ZettleUUIDs: use product info retrieved from Zettle to fill in product ZettleUUID attribute for previously exported products/variations
        my $aProductResponseDTO = $ProductService->getAllProducts();
        $self->_updateShopProductsUUID($aProductResponseDTO);
    }

    # [STEP] Retrieve all exportable products
    my $aProducts = GetExportableProducts($Shop);

    # [STEP] ProductImages: export all product images and retrieve imageUploadResults to be used in product export.
    my $hImageUploadResults = $self->_exportImages($aProducts, 'ImageLarge');

    # [STEP] ProductData: export fields of interest (which will include image lookupKeys from $hImageUploadResults)
    my $hResponse = $ProductService->exportProducts($aProducts, $IncludePrices, $hImageUploadResults);
    my $InitialExportUUID = $hResponse->{uuid};
    $Environment->setupInitialExport($InitialExportUUID, $IncludePrices);
    
    Error('InitialExportError', $hResponse) if $hResponse->{state} eq 'FINISHED_FAILED';
    
    my $State = $hResponse->{'state'};
    my $Status = $self->_getStatusFromState($State);
    $Environment->initialExportStatus($Status);

    # [STEP] Retrieve location UUIDs from Zettle
    $self->_updateAllLocationUUID($Environment);
    
    my $InventoryService = DE_EPAGES::Zettle::API::WebService::InventoryService->new($Environment);

    # [STEP] Enable InventoryTracking for exported products to ensure newly created products are tracked
    my $hProductBalanceBulk = $self->_enableInventoryTracking($InventoryService, $aProducts);
    
    # [STEP] Export stock levels
    if(scalar @{$aProducts}){
        $self->_exportStockLevels($InventoryService, $aProducts, $hProductBalanceBulk);
    }

    return $hResponse->{items};
}

#========================================================================================
# function     isInitialExportComplete
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->isInitialExportComplete()
#----------------------------------------------------------------------------------------
# description  Check if initial product export has been completed
#----------------------------------------------------------------------------------------
# return        TRUE if completed, false otherwise
#========================================================================================
sub isInitialExportComplete {
    my $self = shift;
    my $InitialExportStatus = $self->getInitialExportStatus();
    return ($InitialExportStatus == $INITIAL_EXPORT_STATUS{COMPLETED}) ? 1 : 0;
}

#========================================================================================
# function     getInitialExportStatus
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->getInitialExportStatus()
#----------------------------------------------------------------------------------------
# description  Retrieve updated initial export status querying Zettle if necessary (if export was RUNNING), updating environment export status is status changed.
#----------------------------------------------------------------------------------------
# return       $ExportStatus | See %INITIAL_EXPORT_STATUS values in <DE_EPAGES::Zettle::API::Constants> | integer
#========================================================================================
sub getInitialExportStatus {
    my $self = shift;
    my $Environment = $self->getEnvironment();
    my $InitialExportStatus = $Environment->initialExportStatus();

    # If last known status was 'running' we query Zettle for updated status
    if($InitialExportStatus == $INITIAL_EXPORT_STATUS{RUNNING}){
        my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
        my $ExportUUID = $Environment->initialExportUUID();
        my $State;
        eval {
            $State = $ProductService->getInitialExportState($ExportUUID);
        };
        if(ExistsError()){
            my $Error = GetError();
            LogError($Error);
            $State = $Error->vars()->{'state'};
        }
        $InitialExportStatus = $self->_getStatusFromState($State);
        # Update export status in environment if status changed
        $Environment->initialExportStatus($InitialExportStatus) unless $InitialExportStatus == $INITIAL_EXPORT_STATUS{RUNNING};
    }

    return $InitialExportStatus;
}

#========================================================================================
# function     exportProductUpdates
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportProductUpdates($IncludePrices);
#----------------------------------------------------------------------------------------
# description  Export product basic information to Zettle for products modified after previous export.
#               Also export products created and products deleted after LastProductExport.
# notes        Previous product export datetime is stored in ZettleEnvironment 'LastProductExport' field.
#========================================================================================
sub exportProductUpdates {
    my $self = shift;
    my ($IncludePrices) = @_;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    my $LastProductExport = $Environment->lastProductExport();
    
    # [STEP] Retrieve shop products created, deleted and updated
    # Register current datetime: we store datetime before retrieving products as to not lose any update
    my $NewLastUpdate = GetCurrentDateTime();
    # Retrieve products deleted since last export
    my $ZettleDeleteIterator = $Environment->getDeletedObjects($LastProductExport);
    my @ahZettleDeletes = ();
    while(my $hZettleDelete = <$ZettleDeleteIterator>){
        push(@ahZettleDeletes, $hZettleDelete);
    }
    # Send ZettleUUIDs to delete
    $self->_exportProductDeletions(\@ahZettleDeletes) if scalar(@ahZettleDeletes);

    # Retrieve products whose image was updated
    my $aProductsModifiedImage = GetModifiedImageProducts($Shop, $LastProductExport);
    # Retrieve products whose basic Zettle-related fields were updated (also includes products cloned or created)
    my $aProductsBasicDataUpdated = GetModifiedBasicDataProducts($Shop, $LastProductExport);
    
    # Basic data shall be exported for newly created products, for products whose basic Zettle-related fields were modified and for products whose image was modified
    my @aDataProducts = unique(@$aProductsModifiedImage, @{$aProductsBasicDataUpdated});

    # [STEP] InitialExport must export all product images whereas regular export should only export modified images.
    my $hImageUploadResults = $self->_exportImages($aProductsModifiedImage, 'ImageLarge');

    # [STEP] Export product basic information
    my $hResponse = {};
    if(scalar @aDataProducts > 0){
        my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
        $hResponse = $ProductService->exportProducts(\@aDataProducts, $IncludePrices, $hImageUploadResults);

        # Retrieve products created since last export to:
        # - Enable InventoryTracking
        # - Export stock levels
        my @aNewProducts = grep { DateTime->compare($LastProductExport, $_->get('CreationDate')) < 0 } @aDataProducts;
        if(scalar @aNewProducts){
            my $InventoryService = DE_EPAGES::Zettle::API::WebService::InventoryService->new($Environment);

            my $hProductBalanceBulk = $self->_enableInventoryTracking($InventoryService, \@aNewProducts);
            
            $self->_exportStockLevels($InventoryService, \@aNewProducts, $hProductBalanceBulk);
        }
    }
            
    # Update LastProductExport timestamp: if there were errors we'll update it anyway so as to not get stuck forever
    $Environment->lastProductExport($NewLastUpdate);

    # Update LastPriceExport timestamp: if there were errors we'll update it anyway so as to not get stuck forever
    $Environment->lastPriceExport($NewLastUpdate) if $IncludePrices;

    return $hResponse;
}

#========================================================================================
# function     exportProductPrices
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportProductPrices();
#               $self->exportProductPrices($ExportAllPrices);
#----------------------------------------------------------------------------------------
# description  Export product prices to Zettle:
#               <ul>
#                   <li>if $ExportAllPrices is TRUE, export prices for all exportable products (see <function GetExportableProducts>).
#                   <li>if $ExportAllPrices is FALSE or undef, only export prices modified after last price export.
#               </ul>
#----------------------------------------------------------------------------------------
# input        $ExportAllPrices   | Determines whether to export prices for all products or only modified prices | boolean
#----------------------------------------------------------------------------------------
# notes        Last price export datetime is stored in ZettleEnvironment 'LastPriceExport' field.
#========================================================================================
sub exportProductPrices {
    my $self = shift;
    my ($ExportAllPrices) = @_;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();
    
    my $aProducts;
    if($ExportAllPrices == 1){
        $aProducts = GetExportableProducts($Shop);
    }else{
        my $LastPriceExport = $Environment->lastPriceExport();
        # Export prices for all products if no previous price export was executed, otherwise export only modified prices
        $aProducts = defined $LastPriceExport
                            ? GetModifiedPriceProducts($Shop, $LastPriceExport)
                            : GetExportableProducts($Shop);
    }
    my $NewLastPriceExport = GetCurrentDateTime();

    my $hResponse = {};
    if(scalar @{$aProducts} > 0){
        my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
        $hResponse = $ProductService->exportPrices($aProducts);
    }

    # Update LastPriceExport timestamp: if there were errors we'll update it anyway so as to not get stuck forever
    $Environment->lastPriceExport($NewLastPriceExport);

    return $hResponse;
}

#========================================================================================
# function     setupStockWebhook
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->setupStockWebhook($Destination, $ContactEmail);
#----------------------------------------------------------------------------------------
# description  Subscribe to inventory events to receive stock level updates via Zettle webhooks.
#               See also: <function DE_EPAGES::Zettle::API::ZettleService::incomingStockUpdate>.
# notes        After subscription is created a POST request is sent to $Destination when subscribed event triggers.
#               If Zettle request fails (is not correctly processed by ePages) an email is sent to merchant email.
#========================================================================================
sub setupStockWebhook {
    my $self = shift;
    my ($Destination, $ContactEmail) = @_;

    my $Environment = $self->getEnvironment();

    # Inventory-related events
    my @aEventNames = qw (
        InventoryBalanceChanged
        InventoryTrackingStarted
        InventoryTrackingStopped
    );
    
    my $MethodName = (caller(0))[3];
    GetLog->debug("$MethodName: registering stock webhooks - ".join(',', @aEventNames));
    
    my $PusherService = DE_EPAGES::Zettle::API::WebService::PusherService->new($Environment);
    my $SubscriptionResponseDTO = $PusherService->createSubscription($Destination, $ContactEmail, \@aEventNames);

    my $Uuid = $SubscriptionResponseDTO->uuid();
    my $SigningKey = $SubscriptionResponseDTO->signingKey();
    $Environment->setupInventorySubscription($Uuid, $SigningKey);
}

#========================================================================================
# function     incomingStockUpdate
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->incomingStockUpdate($hEventContent);
#----------------------------------------------------------------------------------------
# description  Process $hEventContent which contains inventory-related events.
#               Incoming messages for the following events:
#               <ul>
#                   <li>InventoryBalanceChanged - includes stock level changes for a list of products
#                   <li>InventoryTrackingStarted - informs tracking was enabled for a product
#                   <li>InventoryTrackingStopped - informs tracking was disabled for a product
#               </ul>
#               Events will be parsed and the shop will be updated accordingly:
#               update stock levels or (un)set ZettleTrackingEnabled product attributes.
#----------------------------------------------------------------------------------------
# input        $hEventContent | contents of an incoming event message with the following keys:
#                                <ul>
#                                    <li>organizationUuid - string
#                                    <li>messageUuid - UUID v1 based on timestamp and messageId - string
#                                    <li>eventName - https://github.com/iZettle/api-documentation/blob/master/pusher-api/api-reference.md#supported-events - string
#                                    <li>messageId - unique UUID of the message - string
#                                    <li>payload - specific payload depending on the type of event.
#                                        see https://github.com/iZettle/api-documentation/blob/master/pusher-api/user-guides/subscriptions.md#payloads - hash
#                                    <li>timestamp - event timestamp in ISO-8601 format - string
#                                </ul>
#========================================================================================
sub incomingStockUpdate {
    my $self = shift;
    my ($hEventContent) = @_;
    my $MethodName = (caller(0))[3];

    my $Payload = $hEventContent->{'payload'};
    # Decode event payload (which is also encoded)
    $hEventContent->{'payload'} = decode_json($Payload) if defined $Payload;

    my $Environment = $self->getEnvironment();
    my $Shop = $Environment->getSite();

    my $WebhookEventDTO = DE_EPAGES::Zettle::API::WebService::DTO::WebhookEventDTO->fromHash($hEventContent);

    #TODO: skip duplicate events comparing messageId field with stored values

    my $OrganizationUuid = $WebhookEventDTO->organizationUuid();
    my $EventName = $WebhookEventDTO->eventName();
    my $MerchantUuid = $Environment->getMerchantUUID();
    my $MessageId = $WebhookEventDTO->messageId();
    Error("Invalid organizationUuid received in $EventName event: $OrganizationUuid (MessageId: $MessageId)") unless $MerchantUuid eq $OrganizationUuid;

    my $hPayload = $WebhookEventDTO->payload;

    if($EventName eq 'InventoryBalanceChanged'){
        my $StockUpdateDTO = DE_EPAGES::Zettle::API::WebService::DTO::StockUpdateDTO->fromHash($hPayload);

        # Skip incoming events triggered by updates sent to Zettle from the shop
        my $externalUuid = $StockUpdateDTO->externalUuid();
        return if defined $externalUuid && ($externalUuid eq $MerchantUuid);

        my $aBalanceBeforeDTO = $StockUpdateDTO->balanceBefore();
        my $aBalanceAfterDTO = $StockUpdateDTO->balanceAfter();

        # Loop balance changes and update stock levels
        for(my $idx=0; $idx < scalar @{$aBalanceBeforeDTO}; $idx++){
            eval{
                my $BalanceBeforeDTO = @{$aBalanceBeforeDTO}[$idx];
                my $BalanceBefore = $BalanceBeforeDTO->balance();
                my $VariationUuid = $BalanceBeforeDTO->variationUuid();

                my $BalanceAfterDTO = @{$aBalanceAfterDTO}[$idx];
                my $BalanceAfter = $BalanceAfterDTO->balance();

                my $BalanceDifference = $BalanceAfter-$BalanceBefore;

                my $Product = LoadProductByZettleUUID($Shop, $VariationUuid);

                my $CurrentStockLevel = $Product->get('StockLevel');
                # Only update if incoming stock level is different from current stock level (this prevents unnecessary updates)
                if($BalanceAfter != $CurrentStockLevel){
                    $Product->set({
                        'StockLevel' => $CurrentStockLevel + $BalanceDifference,
                    },
                    undef,
                    {
                        'IgnoreZettleUpdate' => 1,  # Add a additionalOptions used in hook OnBeforeUpdateProduct to prevent sending Zettle updates back to Zettle
                    });
                }
            };
            if(ExistsError()){
                my $Error = GetError();
                $Error->addVars({
                    'IncomingEvent' => $EventName,
                    'MessageId' => $MessageId
                });
                LogError($Error);
            }else{
                $Environment->lastStockSyncZettle(GetCurrentDateTime());
            }
        }
    }elsif($EventName eq 'InventoryTrackingStarted'){
        my ($TrackedProductId, $TrackedProductZettleUuid);
        eval{
            $TrackedProductZettleUuid = $hPayload->{'productUuid'};
            my $TrackedProduct = LoadProductByZettleUUID($Shop, $TrackedProductZettleUuid);
            if(defined $TrackedProduct){
                $TrackedProductId = $TrackedProduct->id();
                $TrackedProduct->set({'ZettleTrackingEnabled' => 1});
            }
        };
        if(ExistsError()){
            my $Error = GetError();
            $Error->addVars({
                'IncomingEvent' => $EventName,
                'MessageId' => $MessageId,
                'ZettleProductUuid' => $TrackedProductZettleUuid,
                'ProductID' => $TrackedProductId,
            });
            LogError($Error);
        }
    }elsif($EventName eq 'InventoryTrackingStopped'){
        my ($UntrackedProductId, $UntrackedProductZettleUuid);
        eval{
            $UntrackedProductZettleUuid = $hPayload->{'productUuid'};
            my $UntrackedProduct = LoadProductByZettleUUID($Shop, $UntrackedProductZettleUuid);
            if(defined $UntrackedProduct){
                $UntrackedProductId = $UntrackedProduct->id();
                $UntrackedProduct->set({'ZettleTrackingEnabled' => 0});
            }
        };
        if(ExistsError()){
            my $Error = GetError();
            # When a product is deleted from the shop inventory tracking is disabled and this generates an incoming request
            # This request refers the ZettleUUID of the deleted product which is not available anymore so we just ignore it
            if($Error->code ne 'NoResults_ProductByZettleUUID'){
                $Error->addVars({
                    'IncomingEvent' => $EventName,
                    'MessageId' => $MessageId,
                    'ZettleProductUuid' => $UntrackedProductZettleUuid,
                    'ProductID' => $UntrackedProductId,
                });
                LogError($Error);
            }
        }
    }else{
        GetLog->debug("$MethodName - ignoring unexpected event of type ".$EventName);
    }

    return;
}

#========================================================================================
# function     deleteProducts
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->deleteProducts($aProducts)
#----------------------------------------------------------------------------------------
# description  Delete given products from Zettle library.
#----------------------------------------------------------------------------------------
# syntax       $aProducts | list of products to delete | ref.array.object
#========================================================================================
sub deleteProducts {
    my $self = shift;
    my ($aProducts) = @_;

    my $Environment = $self->getEnvironment();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    
    my @aProductUUIDs = ();
    foreach my $Product (@{$aProducts}){
        my $ProductUUID = $Product->get('ZettleUUID');
        push(@aProductUUIDs, $ProductUUID) if defined $ProductUUID;
    }
    
    $ProductService->deleteProducts(\@aProductUUIDs) if scalar(@aProductUUIDs);
}


#========================================================================================


#========================================================================================
# function     _exportProductDeletions
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hResponse = $self->_exportProductDeletions($ahZettleDeletes)
#----------------------------------------------------------------------------------------
# description  Export to Zettle images for given products in the specified format.
#----------------------------------------------------------------------------------------
# input        $ahZettleDeletes | List of ZettleDelete hashes which hold information of objects
#                                  deleted from the shop and their ZettleUUIDs (required to delete from Zettle)
#                                  A ZettleDelete hash contains the following keys:
#                                  <ul>
#                                      <li>GUID - object GUID - string
#                                      <li>SiteID - shop ID - int
#                                      <li>ClassID - class ID - int
#                                      <li>PathFromSite - path of object relative to site - string
#                                      <li>ZettleUUID - Zettle UUID of the object - string
#                                      <li>DeletedOn - date of deletion - datetime
#                                  </ul> | ref.array.hash
#----------------------------------------------------------------------------------------
# return       $hResponse | an HTTP response | HTTPResponse
#========================================================================================
sub _exportProductDeletions {
    my $self = shift;
    my ($ahZettleDeletes) = @_;

    my $hResponse;
    eval{
        my $Environment = $self->getEnvironment();
        my $ShopPath = $Environment->getSite()->get('Path');
        my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);

        # [STEP] Delete superproducts (with or without variations)
        my %deletedParents = map { $_->{'PathFromSite'} => $_ } grep { $_->{'PathFromSite'} !~ /\/SubProducts\// } @{$ahZettleDeletes};
        if(scalar values %deletedParents){
            my @aDeleteZettleUUIDs = map { $_->{'ZettleUUID'} } values %deletedParents;
            $hResponse = $ProductService->deleteProducts(\@aDeleteZettleUUIDs);
        }

        # [STEP] Compile deleted variations whose superproducts were not deleted too
        my %deletedVariations = ();
        foreach my $DeletedVariation ( @{$ahZettleDeletes} ){
            my $Path = $DeletedVariation->{'PathFromSite'};
            # Skip superproducts
            next if $Path !~ /SubProducts/;
            
            $Path =~ s/\/SubProducts\/.*//;
            # Skip subproducts if its superproduct is included in %deletedParents (subproducts will be cascade deleted in Zettle)
            next if exists $deletedParents{$Path};

            # Create a new array hash entry indexed by superproduct path
            $deletedVariations{$Path} //= [];
            push(@{$deletedVariations{$Path}}, $DeletedVariation);
        }

        # If all variations were deleted for a product we have to create a fake variation
        my $aProductsForFakeVariations = ();
        # If only some variations of a superproduct were deleted, we have to "delete" them using an HTTP PUT request
        my $aProductsForPutUpdate = ();
        foreach my $ParentPath (keys %deletedVariations){
            my $ProductFullPath = "$ShopPath/$ParentPath";
            unless(ExistsObjectByPath($ProductFullPath)){
                print STDERR "Product not found: $ProductFullPath\n";
                next;
            }

            my $Product = LoadObjectByPath($ProductFullPath);
            my $SubProductCount = scalar $Product->get('SubProducts');
            if($SubProductCount > 0){
                push(@{$aProductsForPutUpdate}, $Product);
            }else{
                push(@{$aProductsForFakeVariations}, $Product);
            }
        }

        my $ZettleUUID;
        my $HttpResponse;
        my $ETag;

        # [STEP] Use updateProduct (HTTP PUT) to delete some, but not all, variations of a product
        foreach my $ProductForUpdate (@{$aProductsForPutUpdate}){
            $ZettleUUID = $ProductForUpdate->get('ZettleUUID');
            $HttpResponse = $ProductService->getProduct($ZettleUUID);
            $ETag = $HttpResponse->header('ETag');
            $ProductService->updateProduct($ProductForUpdate, $ETag);
        }

        # [STEP] Use updateProduct (HTTP PUT) to create a fake variation for a product without subproducts
        foreach my $ProductForFake (@{$aProductsForFakeVariations}){
            $ZettleUUID = $ProductForFake->get('ZettleUUID');
            $HttpResponse = $ProductService->getProduct($ZettleUUID);
            $ETag = $HttpResponse->header('ETag');
            $ProductService->updateProduct($ProductForFake, $ETag);
        }
    };
    if(ExistsError()){
        LogError(GetError());
    }
}

#========================================================================================
# function     _exportImages
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hImageUploadResults = $self->_exportImages($aProducts, $ImageFormat)
#----------------------------------------------------------------------------------------
# description  Export to Zettle images for given products in the specified format.
#----------------------------------------------------------------------------------------
# input        $aProducts | list of products objects | ref.array.object
# input        $ImageFormat | optional. name of the image format to be processed,
#                              any of: ImageLarge, ImageMedium, ImageSmall, ImageThumbnail
#                              If argument not present, all product images will be exported | string
#----------------------------------------------------------------------------------------
# return       $hImageUploadResults | a map of ImageUploadResult objects indexed by product GUID.
#                                      See <DE_EPAGES::Zettle::API::Object::ImageUploadResult> | ref.hash.array.object
#========================================================================================
sub _exportImages {
    my $self = shift;
    my ($aProducts, $ImageFormat) = @_;
    
    my $Environment = $self->getEnvironment();
    my $ImageService = DE_EPAGES::Zettle::API::WebService::ImageService->new($Environment);
    my $hImageUrls = GetGuidImageMap($aProducts, $ImageFormat);

    # Merge the list of image URLs of each product
    my @aImageUrls;
    foreach my $aImageUrlSubList (values %{$hImageUrls}) {
        push(@aImageUrls, @{$aImageUrlSubList});
    }

    return {} if scalar(@aImageUrls) == 0;

    my $BulkImageUploadResponseDTO = $ImageService->bulkUploadImages(\@aImageUrls);

    my $hImageUploadResults = GetImageUploadResults($BulkImageUploadResponseDTO, $hImageUrls);

    return $hImageUploadResults;
}

#========================================================================================
# function     _updateShopProductsUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateShopProductsUUID($aProductResponseDTO)
#----------------------------------------------------------------------------------------
# input        $aProductResponseDTO | Array of ProductResponseDTO (see DE_EPAGES::Zettle::API::WebService::DTO::ProductResponseDTO) | ref.array.object
#----------------------------------------------------------------------------------------
# description  Use product info retrievd from Zettle to fill in product ZettleUUID attribute for products previously exported to Zettle.
#               The goal is to avoid generating product duplicates in Zettle given that UUID information is the unique identifier used in Zettle API requests.
#               This let's us reuse previously generated UUIDs instead of generating new ones, which would end up with duplicates in Zettle.
#----------------------------------------------------------------------------------------
# return       $hUpdatedProducts | hash of updated products where key is product/variation ID and value is updated UUID | ref.hash
#========================================================================================
sub _updateShopProductsUUID {
    my $self = shift;
    my ($aProductResponseDTO) = @_;

    # Hash of product UUID updates
    my $hUpdatedProducts = {};
    foreach my $ProductResponseDTO (@{$aProductResponseDTO}){
        my $ZettleProductUuid = $ProductResponseDTO->uuid();
        # If product was exported from the shop, externalReference will contain product GUID
        my $ZettleExternalReference = $ProductResponseDTO->externalReference();
        if(defined $ZettleExternalReference && ExistsObjectByGUID($ZettleExternalReference)){
            eval{
                my $Product = LoadObjectByGUID($ZettleExternalReference);
                TestObject('Product', $Product, 'DE_EPAGES::Product::API::Object::Product');
                $Product->set({'ZettleUUID' => $ZettleProductUuid});
                $hUpdatedProducts->{$Product->id} = $ZettleProductUuid;
                
                # Also update UUID for product variations
                my $ahRawVariations = $ProductResponseDTO->variations();
                foreach my $hRawVariation (@{$ahRawVariations}){
                    my $ZettleVariationUuid = $hRawVariation->{'uuid'};
                    my $ZettleVariationOptions = $hRawVariation->{'options'};
                    
                    # NOTE: Zettle variations don't have externalReference field so the best we can do is match by variation attribute options.
                    my $SubProductMatch = $self->_matchSubProductByOptions($Product, $ZettleVariationOptions);
                    if(defined $SubProductMatch){
                        $SubProductMatch->set({'ZettleUUID' => $ZettleVariationUuid});
                        $hUpdatedProducts->{$SubProductMatch->id} = $ZettleVariationUuid;
                    }
                }
            };
            if(ExistsError()){
                my $Error = GetError();
                LogError($Error);
            }
        }
    }

    return $hUpdatedProducts;
}

#========================================================================================
# function     _matchSubProductByOptions
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SubProduct = $self->_matchSubProductByOptions($Product, $ZettleVariationOptions)
#----------------------------------------------------------------------------------------
# input        $Product | a super product | DE_EPAGES::Product::API::Object::Product
# input        $ZettleVariationOptions | attribute values for a zettle variation | ref.array.hash
#               [
#                   { 'name' => 'Colour', 'value' => 'Blue' },
#                   { 'name' => 'Size', 'value' => 'M' },
#                   { 'name' => 'Material', 'value' => 'Cotton' },
#               ]
#----------------------------------------------------------------------------------------
# description  Search SubProducts of given $Product and return the one whose variation attribute options match $ZettleVariationOptions.
#----------------------------------------------------------------------------------------
# return       $SubProduct | a SubProduct that matches given options or undef | DE_EPAGES::Product::API::Object::Product
#========================================================================================
sub _matchSubProductByOptions {
    my $self = shift;
    my ($Product, $ZettleVariationOptions) = @_;

    foreach my $SubProduct (@{$Product->get('SubProducts')}){
        my $SubProductOptions = BuildVariantOptions($SubProduct);
        # We need to specify full Compare package to avoid perl critic complaining about PerlTools not 'use'd
        return $SubProduct if Data::Compare::Compare($SubProductOptions, $ZettleVariationOptions);
    }

    return;
}

#========================================================================================
# function     _deleteLibrary
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_deleteLibrary()
#----------------------------------------------------------------------------------------
# description  Delete all products from Zettle library.
#========================================================================================
sub _deleteLibrary {
    my $self = shift;

    my $Environment = $self->getEnvironment();
    my $ProductService = DE_EPAGES::Zettle::API::WebService::ProductService->new($Environment);
    
    # Retrieve existing products
    my $aProductResponseDTO = $ProductService->getAllProducts();
    my @aProductUUIDs;
    foreach my $ProductResponseDTO (@{$aProductResponseDTO}){
        my $ProductUUID = $ProductResponseDTO->uuid;
        push(@aProductUUIDs, $ProductUUID);
    }
    # Delete products retrieved
    $ProductService->deleteProducts(\@aProductUUIDs) if scalar(@aProductUUIDs);
}

#========================================================================================
# function     _enableInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_enableInventoryTracking($InventoryService, $aProducts)
#----------------------------------------------------------------------------------------
# description  Enable InventoryTracking for given products.
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
# input        $aProducts | List of products | ref.array.object
#========================================================================================
sub _enableInventoryTracking {
    my $self = shift;
    my ($InventoryService, $aProducts) = @_;

    my $TrackingStatus = 'START_TRACKING';
    return $self->_updateInventoryTracking($InventoryService, $aProducts, $TrackingStatus);
}

#========================================================================================
# function     _disableInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_disableInventoryTracking($InventoryService, $aProducts)
#----------------------------------------------------------------------------------------
# description  Disable InventoryTracking for given products.
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
# input        $aProducts | List of products | ref.array.object
#========================================================================================
sub _disableInventoryTracking {
    my $self = shift;
    my ($InventoryService, $aProducts) = @_;

    my $TrackingStatus = 'STOP_TRACKING';
    return $self->_updateInventoryTracking($InventoryService, $aProducts, $TrackingStatus);
}

#========================================================================================
# function     _updateInventoryTracking
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateInventoryTracking($InventoryService, $aProducts, $TrackingStatus)
#----------------------------------------------------------------------------------------
# description  Disable InventoryTracking for given products.
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
# input        $aProducts | List of products | ref.array.object
# input        $TrackingStatus | Any of: START_TRACKING, STOP_TRACKING | string
#========================================================================================
sub _updateInventoryTracking {
    my $self = shift;
    my ($InventoryService, $aProducts, $TrackingStatus) = @_;

    return unless ($TrackingStatus eq 'START_TRACKING') || ($TrackingStatus eq 'STOP_TRACKING');

    my @aProductUuid = map { $_->get('ZettleUUID') } @{$aProducts};

    my $hProductBalanceBulk = $InventoryService->bulkUpdateProductTracking(\@aProductUuid, $TrackingStatus);

    # If $TrackingStatus eq 'START_TRACKING' we subscribe to InventoryTrackingStarted event and it will generate an incoming callback for each of the products tracked.
    # Even though, we force here TrackingEnabled in all products given that asynchronous messages might take a while to be received
    # This would cause stock levels not to be exported for some products
    my $TrackingEnabled = ($TrackingStatus eq 'START_TRACKING') ? 1 : 0;
    $_->set({'ZettleTrackingEnabled' => $TrackingEnabled}) foreach @{$aProducts};

    return $hProductBalanceBulk;
}

#========================================================================================
# function     _updateAllLocationUUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_updateAllLocationUUID($Environment)
#----------------------------------------------------------------------------------------
# description  Retrieve all location UUIDs from Zettle and store them in ZettleEnvironment.
#----------------------------------------------------------------------------------------
# input        $Environment | a Zettle environment object | DE_EPAGES::Zettle::API::Object:ZettleEnvironment
#----------------------------------------------------------------------------------------
# return       $aLocationDTO | List of locations.
#                               See DE_EPAGES::Zettle::API::WebService::DTO::LocationDTO | ref.array.object
#========================================================================================
sub _updateAllLocationUUID {
    my $self = shift;
    my ($Environment) = @_;

    my $LocationService = DE_EPAGES::Zettle::API::WebService::LocationService->new($Environment);
    my $aLocationDTO = $LocationService->getAllLocations();
    foreach my $LocationDTO (@{$aLocationDTO}){
        my $LocationUuid = $LocationDTO->uuid();
        if($LocationDTO->name() eq 'BIN'){
            $Environment->locationBinUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'SOLD'){
            $Environment->locationSoldUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'STORE'){
            $Environment->locationStoreUuid($LocationUuid);
        }elsif($LocationDTO->name() eq 'SUPPLIER'){
            $Environment->locationSupplierUuid($LocationUuid);
        }
    }

    return $aLocationDTO;
}

#========================================================================================
# function     _exportStockLevels
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_exportStockLevels($InventoryService, $aProducts, $hProductBalanceBulk)
#----------------------------------------------------------------------------------------
# description  Export stock levels for given products considering $hProductBalanceBulk (current stock levels in Zettle).
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
# input        $aProducts | List of products | ref.array.object
# input        $hProductBalanceBulk | Current Zettle stock levels. List of variation balances with the following keys:
#                                      <ul>
#                                          <li>locationUuid - Identifies the location - string
#                                          <li>variants - List of balances - array.hash
#                                              <ul>
#                                                  <li>productUuid - Identifies the product - string
#                                                  <li>variantUuid - Identifies the variant - string
#                                                  <li>balance - Balance of variation in location - string
#                                              </ul>
#                                      </ul> | ref.hash.*
#----------------------------------------------------------------------------------------
# return       $hExportStockBulk | List of updated variation balances.
#                                   Same structure as $hProductBalanceBulk parameter | ref.hash.*
#========================================================================================
sub _exportStockLevels {
    my $self = shift;
    my ($InventoryService, $aProducts, $hProductBalanceBulk) = @_;

    my $hBalanceUpdate = {};
    # Use $hProductBalanceBulk as a reference to calculate the stock difference to send to Zettle for each variation
    # If variation is not present in hash we'll then send current variation stock level
    my $ahVariationBalances = $hProductBalanceBulk->{'variants'};
    
    # If variant balances is empty we need to make sure we use the right balances retrieving them from Zettle
    $ahVariationBalances = $self->_getZettleSTOREVariationBalances($InventoryService) unless defined $ahVariationBalances && scalar @{$ahVariationBalances};

    # Only export stock levels for tracked products or we'll get an error
    my @aTrackedProducts = grep { $_->get('ZettleTrackingEnabled') == 1 } @{$aProducts};
    foreach my $Product (@aTrackedProducts){
        my $ProductUuid = $Product->get('ZettleUUID');
        my $aSubProducts = $Product->get('SubProducts');
        my $hVariationUpdates = {};
        if(scalar @{$aSubProducts}){
            # Calculate stock update for each subproduct: difference between shop stock and zettle stock
            foreach my $SubProduct (@{$aSubProducts}){
                my $VariantUuid = $SubProduct->get('ZettleUUID');
                my $StockDelta = $self->_getProductStockDelta($SubProduct, $ahVariationBalances);
                # Stock changes sent to Zettle must be greater than zero
                $hVariationUpdates->{$VariantUuid} = $StockDelta if $StockDelta > 0;
            }
        }else{
            # If no SubProducts, search for a fake zettle variant with same ZettleUuid as $Product
            my $VariantUuid = $Product->get('ZettleUUID');
            my $StockDelta = $self->_getProductStockDelta($Product, $ahVariationBalances);
            $hVariationUpdates->{$VariantUuid} = $StockDelta if $StockDelta > 0;
        }
        $hBalanceUpdate->{$ProductUuid} = $hVariationUpdates;
    }

    my $hExportStockBulk = $InventoryService->bulkUpdateProductBalance($hBalanceUpdate);
    
    my $Environment = $self->getEnvironment();
    $Environment->lastStockSync(GetCurrentDateTime());

    return $hExportStockBulk;
}

#========================================================================================
# function     _getZettleSTOREVariationBalances
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ahVariationBalances = $self->_getZettleSTOREVariationBalances($InventoryService);
#----------------------------------------------------------------------------------------
# description  Use InventoryService to retrieve all variation balances from Zettle STORE location.
#----------------------------------------------------------------------------------------
# input        $InventoryService | Inventory service | DE_EPAGES::Zettle::API::WebService::InventoryService
#----------------------------------------------------------------------------------------
# return       $ahVariationBalances | List of variation balances with the following keys:
#                                      <ul>
#                                          <li>productUuid - Identifies the product - string
#                                          <li>variantUuid - Identifies the variant - string
#                                          <li>balance - Balance of variation in location - string
#                                      </ul> | ref.array.hash
#========================================================================================
sub _getZettleSTOREVariationBalances {
    my $self = shift;
    my ($InventoryService) = @_;

    my @ahVariationBalances = ();
    # Retrieve balances from Zettle STORE location
    my $LocationInventoryDTO = $InventoryService->getLocationInventory();
    foreach my $VariationBalanceDTO (@{$LocationInventoryDTO->variationBalances()}){
        my $VariationBalance = {
            'productUuid' => $VariationBalanceDTO->productUuid(),
            'variantUuid' => $VariationBalanceDTO->variationUuid(),
            'balance' => $VariationBalanceDTO->balance(),
        };
        push(@ahVariationBalances, $VariationBalance);
    }

    return \@ahVariationBalances;
}

#========================================================================================
# function     _getProductStockDelta
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_getProductStockDelta($Product, $ahVariationBalances)
#----------------------------------------------------------------------------------------
# description  Calculate difference between shop stock level and zettle stock level.
#               Search $Product ZettleUUID in $ahVariationBalances in order to calculate difference.
#               If ZettleUUID is not present in hash return $Product StockLevel.
#               If it's present, return ShopStock - ZettleStock.
#----------------------------------------------------------------------------------------
# input        $Product | a Product object | DE_EPAGES::Product::API::Object::Product
# input        $ahVariationBalances | List of variation balances with the following keys:
#                                      <ul>
#                                          <li>productUuid - Identifies the product - string
#                                          <li>variantUuid - Identifies the variant - string
#                                          <li>balance - Balance of variation in location - string
#                                      </ul> | ref.array.hash
#----------------------------------------------------------------------------------------
# return       $StockDelta | Difference between Zettle stock and shop stock | integer
#========================================================================================
sub _getProductStockDelta {
    my $self = shift;
    my ($Product, $ahVariationBalances) = @_;

    my $ZettleUuid = $Product->get('ZettleUUID');
    my $ShopStock = $Product->get('StockLevel') // 0;
    
    # Search $ahVariationBalances for a variant that matches $ZettleUuid
    my $ZettleStock;
    foreach my $hVariantBalance (@{$ahVariationBalances}){
        if($hVariantBalance->{'variantUuid'} eq $ZettleUuid){
            $ZettleStock = $hVariantBalance->{'balance'};
            last;
        }
    }

    my $StockDelta;
    if(defined $ZettleStock){
        $StockDelta = $ShopStock - $ZettleStock;
    }else{
        $StockDelta = $ShopStock;
    }

    return $StockDelta;
}

#========================================================================================
sub _getTokenHelper {
    my $self = shift;
    my ($Environment) = $self->getEnvironment();

    $self->{'TokenHelper'} //= $self->_createTokenHelper($Environment);
    return $self->{'TokenHelper'};
}

#========================================================================================
sub _createTokenHelper {
    my $self = shift;
    my ($Environment) = @_;
    $self->{'TokenHelper'} //= DE_EPAGES::Zettle::API::TokenHelper->new($Environment);
    return $self->{'TokenHelper'};
}

#========================================================================================
sub _getStatusFromState {
    my $self = shift;
    my ($State) = @_;

    my $ExportStatus;
    if($State eq 'IMPORTING'){
        $ExportStatus = $INITIAL_EXPORT_STATUS{RUNNING};
    }elsif($State eq 'FINISHED_SUCCESS'){
        $ExportStatus = $INITIAL_EXPORT_STATUS{COMPLETED};
    }elsif($State eq 'FINISHED_FAILED'){
        $ExportStatus = $INITIAL_EXPORT_STATUS{FAILED};
    }

    return $ExportStatus;
}

1;
