use strict;

use Getopt::Long;
use Time::HiRes qw (time);

use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Core::API::Warning qw ( Warning BlockWarnings );
use DE_EPAGES::Core::API::Log qw ( GetLog LogError );
use DE_EPAGES::Database::API::Connection qw ( RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( ExistsClassByAlias LoadClassByAlias LoadObjectByPath );
use DE_EPAGES::Shop::API::Feature qw( LoadFeatureByAlias );
use DE_EPAGES::WebInterface::API::MessageCenter qw ( SynchronizeCache );
use DE_EPAGES::Zettle::API::Constants qw ( ZETTLE_ENVIRONMENT_CLASS ZETTLE_OBJECT_ALIAS );
use DE_EPAGES::Zettle::API::ZettleService;

# Store to be used if 'storename' argument is not present
use constant DEFAULT_STORE => 'Store';

my $ShopAlias;

# Export product prices for all ZettleEnvironments whose merchants already ran initial export and also selected to include prices in product export.
sub Main {

    local $| = 1;

    my ($Help, $NoOutput, $Password, $StoreName, $Update, $ExportAll, $Force);

    GetOptions(
        'help'        => \$Help,
        'nooutput'    => \$NoOutput,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shop=s'      => \$ShopAlias,
        'update'      => \$Update,
        'exportall'   => \$ExportAll,
        'force'       => \$Force,
    );

    $StoreName //= DEFAULT_STORE;

    usage() if $Help;

    # Either 'update' or 'exportall' flag must be present
    usage() unless $Update || $ExportAll;
    usage() if $Update && $ExportAll;

    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub {
            # make sure that the cartridge is installed
            unless(ExistsCartridgeByPackage('DE_EPAGES::Zettle')) {
                printf STDERR ("[%s] WARN: Cartridge Zettle is not installed.\n", $StoreName) unless $NoOutput;
                return;
            }

            my @EnvironmentsForExport = ();
            # Run script on a specific shop if 'shop' argument is present (and different from 'all')
            if(defined $ShopAlias && $ShopAlias ne 'all'){
                my $Shop;
                eval {
                    $Shop = LoadObjectByPath('/Shops')->child( $ShopAlias );
                };
                if(ExistsError()){
                    printf STDERR ("[%s] WARN: Shop '%s' does not exist.\n", $StoreName, $ShopAlias);
                    return;
                }

                return unless $Shop->existsChild(ZETTLE_OBJECT_ALIAS);

                my $ZettleEnvironment = $Shop->child(ZETTLE_OBJECT_ALIAS);

                push (@EnvironmentsForExport, $ZettleEnvironment) if _MatchPriceExportCriteria($ZettleEnvironment) || $Force;
            } else {
                unless(ExistsClassByAlias(ZETTLE_ENVIRONMENT_CLASS)){
                    printf STDERR ("[%s] WARN: Class '%s' does not exist.\n", $StoreName, ZETTLE_ENVIRONMENT_CLASS);
                    return;
                }

                # Retrieve all Zettle instances from store
                my $aZettleEnvironments = LoadClassByAlias(ZETTLE_ENVIRONMENT_CLASS)->objects;
                
                foreach my $ZettleEnvironment ( @{ $aZettleEnvironments }){
                    push (@EnvironmentsForExport, $ZettleEnvironment) if _MatchPriceExportCriteria($ZettleEnvironment) || $Force;
                }
            }

            # Export product prices to Zettle
            my $ExportAllPrices = (defined $ExportAll ? 1 : 0);
            RunExportPricesJob(\@EnvironmentsForExport, $StoreName, $NoOutput, $ExportAllPrices);
        },
    );

    my $time2   = time;
    my $seconds = $time2 - $time1;
    my $timemsg = sprintf( 'need %.3f seconds', $seconds );
    print $timemsg. "\n" unless $NoOutput;
    GetLog()->debug("Zettle price export on Store '$StoreName': $timemsg");

    eval { SynchronizeCache(); };
    Warning( GetError() ) if ExistsError();
}

# Evaluate whether given environment matches the criteria to export product prices
sub _MatchPriceExportCriteria {
    my ($Environment) = @_;

    return 0 unless $Environment->exportProductPrices();

    my $ZettleService = DE_EPAGES::Zettle::API::ZettleService->new($Environment);
    return 0 unless $ZettleService->isInitialExportComplete();

    return 1;
}

# Export product prices
sub RunExportPricesJob {    ## no critic (ProhibitFunctionSize)
    my ($Environments, $StoreName, $NoOutput, $ExportAllPrices) = @_;
    
    return unless scalar @{$Environments};

    my $Feature = LoadFeatureByAlias('Zettle');
    foreach my $Environment ( @{$Environments} ) {
        my $EnvPath = $Environment->pathString;
        my $Shop = $Environment->getSite();

        # skip export if Zettle feature is disabled for this shop
        my $FeatureValue = $Shop->featureMaxValue($Feature);
        next if ( not $FeatureValue );

        # skip if shop is ClosedByProvider
        next if ($Shop->get('ClosedByProvider'));

        my $aWarnings = BlockWarnings(
            sub {
                eval{
                    # Send product prices to Zettle
                    printf STDOUT ("[%s] Exporting product prices to Zettle for shop: %s\n",
                                    $StoreName, $Shop->alias) unless $NoOutput;
                    $Environment->exportPrices($ExportAllPrices);
                };
                if(ExistsError()){
                    my $Error = GetError();
                    $Error->addVars({ 'Path' => $EnvPath });
                    Warning($Error);
                }
            }
        );

        foreach my $Warning ( @{$aWarnings} ) {
            LogError($Warning);

            printf STDERR ("[%s] WARN: %s: %s\n",
                            $StoreName,
                            $Warning->vars->{Path}//'path unknown' ,
                            substr( $Warning->shortMessage, 0, 4096 ) ); ## no critic (ProhibitMagicNumbers)
        }
    }

    return;
}

sub usage {
    print <<END_USAGE;
Description:
    Export product prices to Zettle.
    Either -update or -exportall flag must be present.
    Export will only run for shops where InitialExport is completed and ExportProductPrices is enabled, unless -force flag is used.

Usage:
    perl $0 [options] [flags]

options:
    -passwd       database user password              ()
    -storename    name of store                       (Default: Store)
    -shop         shop alias             (optional)   (Default: all)
flags:
    -nooutput   do not print any output
    -help       show the command line options
    -update     export only product prices modified since last price export
    -exportall  export all product prices
    -force      export product prices (independently of InitialExport state or ExportProductPrices configuration)

Example:
    perl $0 -shop DemoShop -update
    perl $0 -storename Store -exportall
    perl $0 -storename Store -exportall -force
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );
